/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "animation_storer_dialog.h"

#include <algorithm>

#include <qcheckbox.h>
#include <qgroupbox.h>
#include <qlabel.h>
#include <qvbox.h>

#include <kapp.h>
#include <kconfig.h>
#include <klocale.h>
#include <knuminput.h>

#include "map.h"
#include "theme.h"




AnimationStorerDialog::AnimationStorerDialog(Map const & map, Theme const * theme,
                                             QWidget * parent, char const * name) :
    AdvancedOptionsDialog(parent, name, true, i18n("Animation setup"), Ok | Help | Cancel, Ok, true),
    m_theme(theme),
    m_map_width(map.width()),
    m_map_height(map.height())
{
    QVBox * page = makeVBoxMainWidget();

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    createSizeGroup(page, config);
    createBackgroundGroup(page, config);
    createAdvancedOptionsButton(page);
    createDelayGroup(page, config);
    createCycleGroup(page, config);
    createQualityGroup(page, config);

    setHelp("export-animation-dialog", QString::null);
}


AnimationStorerDialog::~AnimationStorerDialog()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Animation piece size", m_piece_size->value());
    config->writeEntry("Animation transparent background", m_transparent_background->isChecked());
    config->writeEntry("Animation start delay", m_start_delay->value());
    config->writeEntry("Animation delay", m_delay->value());
    config->writeEntry("Animation cylce", m_cycle->isChecked());
    config->writeEntry("Animation end delay", m_end_delay->value());
    config->writeEntry("Animation low quality images", m_low_quality->isChecked());
}


bool AnimationStorerDialog::transparentBackground() const
{
    return m_transparent_background->isChecked();
}


int AnimationStorerDialog::pieceSize() const
{
    return m_piece_size->value();
}


int AnimationStorerDialog::startDelay() const
{
    return m_start_delay->value();
}


int AnimationStorerDialog::delay() const
{
    return m_delay->value();
}


bool AnimationStorerDialog::cylcle() const
{
    return m_cycle->isChecked();
}


int AnimationStorerDialog::endDelay() const
{
    return m_end_delay->value();
}


bool AnimationStorerDialog::lowQuality() const
{
    return m_low_quality->isChecked();
}


void AnimationStorerDialog::pieceSizeChanged(int size)
{
    int const width = m_map_width * size + static_cast<int> (m_theme->leftBorder() * size) +
                      static_cast<int> (m_theme->rightBorder() * size);
    int const height = m_map_height * size + static_cast<int> (m_theme->upperBorder() * size) +
                       static_cast<int> (m_theme->lowerBorder() * size);

    m_whole_size->setText(i18n("The whole size will be %1x%2.").arg(width).arg(height));
}


void AnimationStorerDialog::cycleChanged(bool cycle)
{
    m_end_delay->setEnabled(cycle);
}


void AnimationStorerDialog::createSizeGroup(QWidget * parent, KConfig * config)
{
    QGroupBox * box = new QGroupBox(2, Vertical, i18n("Set the size of the animation"), parent);

    int piece_size = config->readNumEntry("Animation piece size", 32);
    piece_size = std::min(std::max(piece_size, 4), 256);

    m_piece_size = new KIntNumInput(piece_size, box);
    m_piece_size->setRange(4, 256);
    m_piece_size->setSuffix(" " + i18n("pixel"));
    m_piece_size->setLabel(i18n("Size of a field"));
    connect(m_piece_size, SIGNAL(valueChanged(int)), this, SLOT(pieceSizeChanged(int)));

    m_whole_size = new QLabel(box);
    pieceSizeChanged(m_piece_size->value());
}


void AnimationStorerDialog::createBackgroundGroup(QWidget * parent, KConfig * config)
{
    QGroupBox * box = new QGroupBox(1, Vertical, i18n("Make background transparent"), parent);

    m_transparent_background = new QCheckBox(i18n("Make background transparent"), box);
    m_transparent_background->setChecked(config->readNumEntry("Animation transparent background", true));
}


void AnimationStorerDialog::createDelayGroup(QWidget * parent, KConfig * config)
{
    QGroupBox * box = new QGroupBox(2, Vertical, i18n("Delays between two animation frames"), parent);
    addAdvancedWidget(box);

    int start_delay = config->readNumEntry("Animation start delay", 5000);
    start_delay = std::min(std::max(start_delay, 1), 100000);

    m_start_delay = new KIntNumInput(start_delay, box);
    m_start_delay->setRange(1, 100000);
    m_start_delay->setSuffix(" " + i18n("miliseconds"));
    m_start_delay->setLabel(i18n("Delay for the first frame"));

    int delay = config->readNumEntry("Animation delay", 100);
    delay = std::min(std::max(delay, 1), 10000);

    m_delay = new KIntNumInput(delay, box);
    m_delay->setRange(1, 10000);
    m_delay->setSuffix(" " + i18n("miliseconds"));
    m_delay->setLabel(i18n("For the other frames"));
}


void AnimationStorerDialog::createCycleGroup(QWidget * parent, KConfig * config)
{
    QGroupBox * box = new QGroupBox(2, Vertical, i18n("Cylce the animation"), parent);
    addAdvancedWidget(box);

    m_cycle = new QCheckBox(i18n("Repeat animation from start if it ended"), box);
    m_cycle->setChecked(config->readNumEntry("Animation cylce", true));
    connect(m_cycle, SIGNAL(toggled(bool)), this, SLOT(cycleChanged(bool)));

    int end_delay = config->readNumEntry("Animation end delay", 5000);
    end_delay = std::min(std::max(end_delay, 1), 100000);

    m_end_delay = new KIntNumInput(end_delay, box);
    m_end_delay->setRange(1, 100000);
    m_end_delay->setSuffix(" " + i18n("miliseconds"));
    m_end_delay->setLabel(i18n("Delay between last frame and first frame"));
    cycleChanged(m_cycle->isChecked());
}


void AnimationStorerDialog::createQualityGroup(QWidget * parent, KConfig * config)
{
    QGroupBox * box = new QGroupBox(3, Vertical, i18n("Set quality of the animation"), parent);
    addAdvancedWidget(box);

    m_low_quality = new QCheckBox(i18n("Use low quality images"), box);
    m_low_quality->setChecked(config->readNumEntry("Animation low quality images", false));
}
