/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "bookmarks.h"

#include <cassert>
#include <utility>

#include <qdatastream.h>
#include <qfile.h>
#include <qpoint.h>
#include <qregexp.h>
#include <qtextstream.h>

#include <kglobal.h>
#include <klocale.h>
#include <kstddirs.h>

#include "atomic_move.h"
#include "collection.h"




bool Bookmarks::s_is_initialized = false;
bool Bookmarks::s_modified = false;
int Bookmarks::s_number_of_bookmarks;
std::map<int, int> Bookmarks::s_index_to_index_map;
QStringList Bookmarks::s_annotations;
QStringList Bookmarks::s_collection_names;
std::vector<int> Bookmarks::s_levels;
std::vector<CompressedMap> Bookmarks::s_maps;
std::vector<Movements> Bookmarks::s_moves;
std::vector<QDateTime> Bookmarks::s_dates;


void Bookmarks::load()
{
    assert(s_is_initialized == false);

    s_is_initialized = true;

    s_number_of_bookmarks = 0;

    QString filename = KGlobal::dirs()->findResource("data", "easysok/bookmarks.dat");

    if (filename.isEmpty())
    {
        return;
    }

    QFile file(filename);

    if (!file.open(IO_ReadOnly))
    {
        return;
    }

    QDataStream stream(&file);

    Q_INT32 version;
    stream >> version;

    Q_INT32 number_of_bookmarks;
    stream >> number_of_bookmarks;
    s_number_of_bookmarks = number_of_bookmarks;

    for (int i = 0; i < s_number_of_bookmarks; ++i)
    {
        Q_INT32 index;
        stream >> index;

        QString collection;
        stream >> collection;

        Q_INT32 level;
        stream >> level;

        CompressedMap map(stream);

        Movements moves(stream);

        QDateTime date;
        stream >> date;

        QString annotation;
        stream >> annotation;

        s_index_to_index_map.insert(std::make_pair(static_cast<int> (index), i));
        s_collection_names.append(collection);
        s_levels.push_back(level);
        s_maps.push_back(map);
        s_moves.push_back(moves);
        s_dates.push_back(date);
        s_annotations.append(annotation);
    }
}


void Bookmarks::save()
{
    assert(s_is_initialized);

    if (!s_modified)
    {
        return;
    }

    QString filename = KGlobal::dirs()->saveLocation("data", "easysok/", true) + "bookmarks.dat";

    if (filename.isEmpty())
    {
        return;
    }

    QFile file(filename);

    if (!file.open(IO_WriteOnly | IO_Truncate))
    {
        return;
    }

    QDataStream stream(&file);

    Q_INT32 const version = 0;
    stream << version;

    Q_INT32 const number_of_bookmarks = s_number_of_bookmarks;
    stream << number_of_bookmarks;

    std::map<int, int>::const_iterator it = s_index_to_index_map.begin();
    std::map<int, int>::const_iterator const end_it = s_index_to_index_map.end();

    for (; it != end_it; ++it)
    {
        Q_INT32 bookmark_index = it->first;
        stream << bookmark_index;

        int const index = it->second;

        stream << s_collection_names[index];

        Q_INT32 const level = s_levels[index];
        stream << level;

        s_maps[index].writeToStream(stream);

        s_moves[index].writeToStream(stream);

        stream << s_dates[index];

        stream << s_annotations[index];
    }
}


bool Bookmarks::hasBookmark(int index)
{
    assert(s_is_initialized);

    return indexToIndex(index) != -1;
}


QString Bookmarks::collectionName(int index)
{
    assert(s_is_initialized);
    assert(hasBookmark(index));

    return s_collection_names[indexToIndex(index)];
}


int Bookmarks::level(int index)
{
    assert(s_is_initialized);
    assert(hasBookmark(index));

    return s_levels[indexToIndex(index)];
}


CompressedMap Bookmarks::map(int index)
{
    assert(s_is_initialized);
    assert(hasBookmark(index));

    return s_maps[indexToIndex(index)];
}


Movements Bookmarks::moves(int index)
{
    assert(s_is_initialized);
    assert(hasBookmark(index));

    return s_moves[indexToIndex(index)];
}


QDateTime Bookmarks::date(int index)
{
    assert(s_is_initialized);
    assert(hasBookmark(index));

    return s_dates[indexToIndex(index)];
}


QString Bookmarks::annotation(int index)
{
    assert(s_is_initialized);
    assert(hasBookmark(index));

    return s_annotations[indexToIndex(index)];
}


void Bookmarks::replaceBookmark(int index, QString const & annotation, QString const & collection_name,
                                int level, CompressedMap const & map, Movements const & moves)
{
    assert(s_is_initialized);

    s_modified = true;
    int const old_index = indexToIndex(index);

    if (old_index == -1)
    {
        int const new_index = s_number_of_bookmarks;
        ++s_number_of_bookmarks;

        s_index_to_index_map.insert(std::make_pair(index, new_index));
        s_collection_names.append(collection_name);
        s_levels.push_back(level);
        s_maps.push_back(map);
        s_moves.push_back(moves);
        s_moves.back().truncateToCurrent();
        s_dates.push_back(QDateTime::currentDateTime());
        s_annotations.append(annotation);
    }
    else
    {
        s_collection_names[old_index] = collection_name;
        s_levels[old_index] = level;
        s_maps[old_index] = map;
        s_moves[old_index] = moves;
        s_moves[old_index].truncateToCurrent();
        s_dates[old_index] = QDateTime::currentDateTime();
        s_annotations[old_index] = annotation;
    }
}


bool Bookmarks::hasKSokobanBookmark(int ksokoban_index)
{
    assert(s_is_initialized);
    assert(ksokoban_index >= 1);
    assert(ksokoban_index <= 10);

    QString const bookmark_name = "ksokoban/bookmark" + QString::number(ksokoban_index);
    QString const bookmark_file = KGlobal::dirs()->findResource("data", bookmark_name);

    if (bookmark_file.isEmpty())
    {
        return false;
    }

    return QFile::exists(bookmark_file);
}


QString Bookmarks::kSokobanBookmarkCollectionAndLevel(int ksokoban_index, int & level)
{
    assert(s_is_initialized);
    assert(ksokoban_index >= 1);
    assert(ksokoban_index <= 10);
    assert(index >= 0);

    QString result;

    QString const raw_bookmark_name = "ksokoban/bookmark" + QString::number(ksokoban_index);
    QString const bookmark_name = KGlobal::dirs()->findResource("data", raw_bookmark_name);

    QFile bookmark_file(bookmark_name);

    if (!bookmark_file.open(IO_ReadOnly))
    {
        return result;
    }

    QTextStream stream(&bookmark_file);

    int collection_index;

    stream >> collection_index;

    QString collection_file = collectionFileForKSokobanCollection(collection_index);

    if (collection_file.isEmpty())
    {
        return result;
    }

    Collection collection(collection_file);

    stream >> level;

    if ( (level < 0) && (level >= collection.numberOfLevels()) )
    {
        return result;
    }

    result = collection.name();

    return result;
}


bool Bookmarks::importKSokobanBookmark(int ksokoban_index, int index)
{
    assert(s_is_initialized);
    assert(ksokoban_index >= 1);
    assert(ksokoban_index <= 10);
    assert(index >= 0);

    QString const raw_bookmark_name = "ksokoban/bookmark" + QString::number(ksokoban_index);
    QString const bookmark_name = KGlobal::dirs()->findResource("data", raw_bookmark_name);

    QFile bookmark_file(bookmark_name);

    if (!bookmark_file.open(IO_ReadOnly))
    {
        return false;
    }

    QTextStream stream(&bookmark_file);

    int collection_index;

    stream >> collection_index;

    QString collection_file = collectionFileForKSokobanCollection(collection_index);

    if (collection_file.isEmpty())
    {
        return false;
    }

    Collection collection(collection_file);

    int level_index;

    stream >> level_index;

    if ( (level_index < 0) && (level_index >= collection.numberOfLevels()) )
    {
        return false;
    }

    int pushes;

    stream >> pushes;

    QString raw_moves;

    stream >> raw_moves;

    if (raw_moves.find('-', 0) == -1)
    {
        return false;
    }

    int pos = 0;

    QRegExp digit("[0-9]");

    Movements moves;

    QPoint keeper_pos = collection.level(level_index).map().keeper();

    while (raw_moves[pos] != '-')
    {
        int repeat = 1;

        if (digit.search(raw_moves, pos) == pos)
        {
            repeat = raw_moves[pos].latin1() - '0';
            ++pos;

            if (digit.search(raw_moves, pos) == pos)
            {
                repeat *= 10;
                repeat = raw_moves[pos].latin1() - '0';
                ++pos;
            }
        }

        bool push = false;
        AtomicMove::AtomicMoveType atomic_move;

        switch(raw_moves[pos].latin1())
        {
        case 'L':
        case 'W':
            push = true;

        case 'l':
        case 'w':
            atomic_move = AtomicMove::LEFT;
            break;

        case 'R':
        case 'E':
            push = true;

        case 'r':
        case 'e':
            atomic_move = AtomicMove::RIGHT;
            break;

        case 'U':
        case 'N':
            push = true;

        case 'u':
        case 'n':
            atomic_move = AtomicMove::UP;
            break;

        case 'D':
        case 'S':
            push = true;

        case 'd':
        case 's':
            atomic_move = AtomicMove::DOWN;
            break;

        default:
            return false;
        }

        for (int i = 0; i < repeat; ++i)
        {
            Move const move(keeper_pos, atomic_move, push);
            moves.addMove(move);
            keeper_pos = move.to();
        }

        ++pos;
    }

    if (!collection.level(level_index).map().areValidMoves(moves))
    {
        return false;
    }

    replaceBookmark(index, i18n("KSokoban bookmark #%1").arg(ksokoban_index), collection.name(),
                    level_index, collection.level(level_index).compressedMap(), moves);

    return true;
}


int Bookmarks::indexToIndex(int index)
{
    assert(s_is_initialized);

    std::map<int, int>::const_iterator const it = s_index_to_index_map.find(index);

    if (it != s_index_to_index_map.end())
    {
        return it->second;
    }
    else
    {
        return -1;
    }
}


QString Bookmarks::collectionFileForKSokobanCollection(int collection_index)
{
    assert(s_is_initialized);

    QString collection_file;

    switch (collection_index)
    {
    case 0:
        collection_file = "xsokoban.xsb";
        break;

    case 1:
        collection_file = "xsokobanextra.xsb";
        break;

    case 4:
        collection_file = "autogenerated.xsb";
        break;

    case 6:
        collection_file = "dimitryyorick.xsb";
        break;

    case 10:
        collection_file = "sasquatch.xsb";
        break;

    case 11:
        collection_file = "massasquatch.xsb";
        break;

    case 12:
        collection_file = "sasquatchiii.xsb";
        break;

    case 13:
        collection_file = "microban.xsb";
        break;

    default:
        return collection_file;
    }

    return KGlobal::dirs()->findResource("data", "easysok/levels/" + collection_file);
}
