/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_BOOKMARKS_INC_GUARD_H
#define EASYSOK_BOOKMARKS_INC_GUARD_H


#include <map>
#include <vector>

#include <qdatetime.h>
#include <qstring.h>
#include <qstringlist.h>

#include "compressed_map.h"
#include "movements.h"




/**
 * Used to store and retrieve bookmarks.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class Bookmarks
{

public:

    /**
     * Loads the bookmarks from disk.
     *
     * This should be the first function called (and called only once).
     */

    static void load();


    /**
     * Writes the bookmarks to disk.
     */

    static void save();


    /**
     * Returns true, if a bookmark with the given index exists.
     *
     * @param index The index of the bookmark.
     */

    static bool hasBookmark(int index);


    /**
     * Returns the collection name of the bookmark.
     *
     * @param index The index of the bookmark.
     */

    static QString collectionName(int index);


    /**
     * The index of the level in the collection.
     *
     * @param index The index of the bookmark.
     */

    static int level(int index);


    /**
     * Returns the start map of the bookmark.
     *
     * @param index The index of the bookmark.
     */

    static CompressedMap map(int index);


    /**
     * Returns the moves of the bookmark.
     *
     * @param index The index of the bookmark.
     */

    static Movements moves(int index);


    /**
     * Returns the date, when the bookmark was created.
     *
     * @param index The index of the bookmark.
     */

    static QDateTime date(int index);


    /**
     * Returns the annotation of the bookmark.
     *
     * @param index The index of the bookmark.
     */

    static QString annotation(int index);


    /**
     * Replaces or adds the given bookmark.
     *
     * @param index The index of the bookmark.
     * @param annotation The annotation of the bookmark.
     * @param collection_name The name of the collection.
     * @param level The level in the collection.
     * @param map The start map of the level.
     * @param moves The moves made.
     */

    static void replaceBookmark(int index, QString const & annotation, QString const & collection_name,
                                int level, CompressedMap const & map, Movements const & moves);


    /**
     * Returns true, if a ksokoban bookmark with the given index exists.
     *
     * @param ksokoban_index The index of the KSokoban bookmar (1 - 10).
     */

    static bool hasKSokobanBookmark(int ksokoban_index);


    /**
     * Returns the collection name of the KSokoban bookmark with the given index and
     * stores the level.
     *
     * @param ksokoban_index The index of the KSokoban bookmar (1 - 10).
     * @param lebel Here we store the level.
     */

    static QString kSokobanBookmarkCollectionAndLevel(int ksokoban_index, int & level);


    /**
     * Tries to import a ksokoban bookmark.
     *
     * @param ksokoban_index The index of the KSokoban bookmark (1 - 10).
     * @param index The index in which to store the ksokoban bookmark.
     * @return True, if the import operation was successful.
     */

    static bool importKSokobanBookmark(int ksokoban_index, int index);


private:

    /**
     * Returns the internal index for the given bookmark index (or -1 if we don't have the
     * bookmark).
     *
     * @param index The index of the bookmark.
     */

    static int indexToIndex(int index);


    /**
     * Returns the filename of the collection for a ksokoban collection index.
     *
     * @param collection_index The index from the ksokoban bookmark.
     * @return The filename of the collection or an empty string if the collection could
     *         not be found.
     */

    static QString collectionFileForKSokobanCollection(int collection_index);


    /**
     * If true, we have called init already.
     */

    static bool s_is_initialized;


    /**
     * If true, the bookmarks were modified.
     */

    static bool s_modified;


    /**
     * The number of bookmarks stored.
     */

    static int s_number_of_bookmarks;


    /**
     * Maps bookmark indices to internal indices.
     */

    static std::map<int, int> s_index_to_index_map;


    /**
     * Here we store the annotations.
     */

    static QStringList s_annotations;


    /**
     * Here we store the collection names.
     */

    static QStringList s_collection_names;


    /**
     * Here we store the level numbers.
     */

    static std::vector<int> s_levels;


    /**
     * Here we store the compressed maps.
     */

    static std::vector<CompressedMap> s_maps;


    /**
     * Here we store the moves.
     */

    static std::vector<Movements> s_moves;


    /**
     * Here we store the dates.
     */

    static std::vector<QDateTime> s_dates;
};


#endif
