/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "collection.h"

#include <cassert>

#include <qdatastream.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qstringlist.h>
#include <qtextstream.h>

#include <klocale.h>




Collection::Collection(QString const & name, QStringList const & authors, QStringList const & emails,
                       QString const & homepage, QString const & copyright,
                       QString const & info, int difficulty) :
    m_authors(authors),
    m_emails(emails),
    m_homepage(homepage),
    m_copyright(copyright),
    m_name(name),
    m_info(info)
{
    setDifficulty(difficulty);
}



Collection::Collection(QString const & collection)
{
    QFile file(collection);

    if (!file.open(IO_ReadOnly))
    {
        return;
    }

    QTextStream stream(&file);
    QStringList lines;

    while (!stream.atEnd())
    {
        lines.append(stream.readLine());
    }

    Level::getInfo(lines, m_authors, m_emails, m_homepage, m_copyright, m_name, m_info, m_difficulty);

    while (!lines.isEmpty())
    {
        Level level(lines, m_authors, m_emails, m_homepage, m_copyright, m_info, m_difficulty);

        if (level.map().isValid())
        {
            m_levels.push_back(level);
        }
    }
}


Collection::Collection(QDataStream & stream, int version)
{
    Q_INT32 levels;

    stream >> levels;
    stream >> m_authors;
    stream >> m_emails;
    stream >> m_homepage;
    stream >> m_copyright;
    stream >> m_name;
    stream >> m_info;

    if (version >= 1)
    {
        Q_INT8 difficulty;
        stream >> difficulty;

        setDifficulty(difficulty);
    }
    else
    {
        m_difficulty = -1;
    }

    for (Q_INT32 i = 0; i < levels; ++i)
    {
        Level level(stream, version);

        if (level.map().isValid())
        {
            m_levels.push_back(level);
        }
    }
}


void Collection::writeToStream(QDataStream & stream) const
{
    Q_INT32 levels = static_cast<Q_INT32> (m_levels.size());

    stream << levels;
    stream << m_authors;
    stream << m_emails;
    stream << m_homepage;
    stream << m_copyright;
    stream << m_name;
    stream << m_info;

    Q_INT8 difficulty = static_cast<Q_INT8> (m_difficulty);
    stream << difficulty;

    for (Q_INT32 i = 0; i < levels; ++i)
    {
        m_levels[i].writeToStream(stream);
    }
}


int Collection::numberOfLevels() const
{
    return m_levels.size();
}


QStringList const & Collection::authors() const
{
    return m_authors;
}


QString Collection::authorEmailLine() const
{
    return Level::createAuthorEmailLine(m_authors, m_emails);
}


void Collection::setAuthors(QStringList const & authors)
{
    m_authors = authors;
}


void Collection::setAuthorEmailLine(QString author_email_line)
{
    Level::parseAuthorEmailLine(author_email_line, m_authors, m_emails);
}


QStringList const & Collection::emails() const
{
    return m_emails;
}


void Collection::setEmails(QStringList const & emails)
{
    m_emails = emails;
}


QString const & Collection::homepage() const
{
    return m_homepage;
}


void Collection::setHomepage(QString const & homepage)
{
    m_homepage = homepage;
}


QString const & Collection::copyright() const
{
    return m_copyright;
}


void Collection::setCopyright(QString const & copyright)
{
    m_copyright = copyright;
}


QString const & Collection::name() const
{
    return m_name;
}


void Collection::setName(QString const & name)
{
    m_name = name;
}


QString const & Collection::info() const
{
    return m_info;
}


void Collection::setInfo(QString const & info)
{
    m_info = info;
}


int Collection::difficulty() const
{
    return m_difficulty;
}


void Collection::setDifficulty(int difficulty)
{
    if ( (difficulty >= 0) && (difficulty <= 10) )
    {
        m_difficulty = difficulty;
    }
    else
    {
        m_difficulty = -1;
    }
}


Level const & Collection::level(int index) const
{
    assert(index >= 0);
    assert(index < numberOfLevels());

    return m_levels[index];
}


void Collection::addLevel(Level const & level)
{
    m_levels.push_back(level);
}


void Collection::insertLevel(Level const & level, int index)
{
    assert(index >= 0);
    assert(index <= numberOfLevels());

    m_levels.insert(m_levels.begin() + index, level);
}


void Collection::replaceLevel(Level const & level, int index)
{
    assert(index >= 0);
    assert(index < numberOfLevels());

    m_levels[index] = level;
}


void Collection::removeLevel(int index)
{
    assert(index >= 0);
    assert(index < numberOfLevels());

    m_levels.erase(m_levels.begin() + index);
}


QString Collection::toText() const
{
    QString result;

    result += "Name: " + m_name + '\n';
    result += "Author: " + authorEmailLine() + '\n';
    result += "Hompage: " + m_homepage + '\n';
    result += "Copyright: " + m_copyright + '\n';
    result += QStringList::split('\n', m_info, true).join("Info: ") + "\n\n";

    if (m_difficulty != -1)
    {
        result += "Difficulty: " + QString::number(m_difficulty) + '\n';
    }

    int const number_of_levels = static_cast<int> (m_levels.size());

    for (int i = 0; i < number_of_levels; ++i)
    {
        result += m_levels[i].toText(m_authors, m_emails, m_homepage, m_copyright, m_info, m_difficulty);
    }

    return result;
}


QString Collection::levelNameAndCollection(Level const & level, int level_nr, Collection const & collection)
{
    QString result = "Name: ";

    if (level.name().isEmpty())
    {
        result += i18n("Level %1 of %2").arg(level_nr + 1).arg(collection.numberOfLevels());
    }
    else
    {
        result += level.name();
    }

    result += " " + i18n("from collection %1").arg(collection.name()) + "\n";

    return result;
}
