/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_COLLECTION_INC_GUARD_H
#define EASYSOK_COLLECTION_INC_GUARD_H


#include <vector>

#include <qstring.h>
#include <qstringlist.h>

#include "level.h"


class QDataStream;




/**
 * This class represents a collection of sokoban levels.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class Collection
{

public:

    /**
     * Creates an empty collection.
     *
     * @param name The name of the collection.
     * @param authors The authors of the collection.
     * @param emails The emails addresses of the authors of the collection.
     * @param homepage The homepage of the collection.
     * @param copyright A copyright string.
     * @param info Additional info.
     * @param difficulty The difficulty of the level in the range[-1:10], with 0 being trivially easy.
     *                   -1 means an unknown difficulty.
     */

    Collection(QString const & name, QStringList const & authors, QStringList const & emails,
               QString const & homepage, QString const & copyright, QString const & info, int difficulty);


    /**
     * Loads the level collection from a file in xsokoban format.
     *
     * @param collection The filename of the collection to load.
     */

    Collection(QString const & collection);


    /**
     * Reads the level collection from a data stream.
     *
     * @param stream The stream to use.
     * @param version The version to use.
     */

    Collection(QDataStream & stream, int version);


    /**
     * Writes the level collection to a data stream.
     *
     * @param stream The stream to use.
     */

    void writeToStream(QDataStream & stream) const;


    /**
     * Returns the number of levels.
     */

    int numberOfLevels() const;


    /**
     * Returns the name of the authors.
     */

    QStringList const & authors() const;


    /**
     * Returns the author-email-line (the authors with appended email, separated by commas).
     */

    QString authorEmailLine() const;


    /**
     * Sets the name of the authors.
     *
     * @param authors The new names of the authors.
     */

    void setAuthors(QStringList const & authors);


    /**
     * Sets the authors and emails from the given line.
     *
     * @param author_email_line The line containing (separated by ',') the authors and emails.
     */

    void setAuthorEmailLine(QString author_email_line);


    /**
     * Returns the email adresses of the authors.
     */

    QStringList const & emails() const;


    /**
     * Sets the email adddresses of the authors.
     *
     * @param emails The email adddresses of the authors.
     */

    void setEmails(QStringList const & emails);


    /**
     * Returns the home of the collection/author.
     */

    QString const & homepage() const;


    /**
     * Sets a new homepage.
     *
     * @param homepage The new homepage.
     */

    void setHomepage(QString const & homepage);


    /**
     * Returns the copyright of the collection.
     */

    QString const & copyright() const;


    /**
     * Sets a new copyright.
     *
     * @param copyright The new copyright.
     */

    void setCopyright(QString const & copyright);


    /**
     * Returns the name of the level collection.
     */

    QString const & name() const;


    /**
     * Sets a new name.
     *
     * @param name The new name.
     */

    void setName(QString const & name);


    /**
     * Returns additional info of the level collection.
     */

    QString const & info() const;


    /**
     * Sets a new info.
     *
     * @param info The new info.
     */

    void setInfo(QString const & info);


    /**
     * Returns the difficulty of the collection.
     */

    int difficulty() const;


    /**
     * Sets the difficulty of the collection.
     *
     * @param difficulty The new difficulty.
     */

    void setDifficulty(int difficulty);


    /**
     * Returns a pointer to the level with the given index.
     *
     * @param index The index of the level.
     */

    Level const & level(int index) const;


    /**
     * Adds a new level to the end of the collection.
     *
     * @param level The level to add.
     */

    void addLevel(Level const & level);


    /**
     * Insert a new level at the given index (see std::vector<>::insert for where it is inserted).
     *
     * @param level The level to insert.
     * @param index The index to insert.
     */

    void insertLevel(Level const & level, int index);


    /**
     * Replaces the level at the given index.
     *
     * @param level The new level.
     * @param index The index of the level to replace.
     */

    void replaceLevel(Level const & level, int index);


    /**
     * Removes the level at the given index.
     *
     * @param index The position of the level.
     */

    void removeLevel(int index);


    /**
     * Returns the maps of the level collection in xsb format.
     */

    QString toText() const;


    /**
     * Returns a string containing the name (if the it has a name and the number otherwise) of
     * a level plus the collection name of a given level in a collection.
     *
     * @param level The level.
     * @param level_nr The number of the level in the collection.
     * @param collection The collection of the level.
     */

    static QString levelNameAndCollection(Level const & level, int level_nr, Collection const & collection);


private:

    /**
     * The levels of the collection.
     */

    std::vector<Level> m_levels;


    /**
     * The authors.
     */

    QStringList m_authors;


    /**
     * The email adddresses of the authors.
     */

    QStringList m_emails;


    /**
     * The homepage.
     */

    QString m_homepage;


    /**
     * The copyright.
     */

    QString m_copyright;


    /**
     * The name of the collection.
     */

    QString m_name;


    /**
     * Additional info of the collection.
     */

    QString m_info;


    /**
     * The difficulty.
     */

    int m_difficulty;
};


#endif
