/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "collection_comparer.h"

#include "collection.h"
#include "compressed_map.h"
#include "level.h"
#include "map.h"




CollectionComparer::CollectionComparer(Collection const & collection1, Collection const & collection2) :
    m_nr_reordered(0),
    m_nr_removed(0),
    m_nr_added(0),
    m_changes_properties(false),
    m_not_trivial_changes(false)
{
    bool collection_name_changed = false;

    if (collection1.authorEmailLine() != collection2.authorEmailLine())
    {
        m_changes_properties = true;
    }
    else if (collection1.homepage() != collection2.homepage())
    {
        m_changes_properties = true;
    }
    else if (collection1.copyright() != collection2.copyright())
    {
        m_changes_properties = true;
    }
    else if (collection1.name() != collection2.name())
    {
        collection_name_changed = true;
        m_not_trivial_changes = true;
    }
    else if (collection1.info() != collection2.info())
    {
        m_changes_properties = true;
    }
    else if (collection1.difficulty() != collection2.difficulty())
    {
	m_changes_properties = true;
    }

    int const nr_level_1 = collection1.numberOfLevels();
    int const nr_level_2 = collection2.numberOfLevels();

    for (int i = 0; i < nr_level_1; ++i)
    {
        Level const & level1 = collection1.level(i);
        CompressedMap const & map1 = level1.compressedMap();
        bool found = false;

        for (int j = 0; j < nr_level_2; ++j)
        {
            Level const & level2 = collection2.level(j);
            CompressedMap const & map2 = level2.compressedMap();

            if (map1 == map2)
            {
                found = true;

                if (i != j)
                {
                    ++m_nr_reordered;
                }

                if (!m_changes_properties)
                {
                    if (level1.authorEmailLine() != level2.authorEmailLine())
                    {
                        m_changes_properties = true;
                    }
                    else if (level1.homepage() != level2.homepage())
                    {
                        m_changes_properties = true;
                    }
                    else if (level1.copyright() != level2.copyright())
                    {
                        m_changes_properties = true;
                    }
                    else if (level1.name() != level2.name())
                    {
                        m_changes_properties = true;
                    }
                    else if (level1.info() != level2.info())
                    {
                        m_changes_properties = true;
                    }
		    else if (level1.difficulty() != level2.difficulty())
		    {
			m_changes_properties = true;
		    }
                }

                break;
            }
        }

        if (!found)
        {
            ++m_nr_added;
        }
    }

    m_reorders_levels = (m_nr_reordered > 0);
    m_adds_levels = (m_nr_added > 0);
    m_nr_removed = nr_level_2 - (nr_level_1 - m_nr_added);
    m_removes_levels = (m_nr_removed > 0);
    m_are_equal = (!m_changes_properties && !m_reorders_levels && !m_adds_levels && !m_removes_levels);
    m_are_unrelated = (m_nr_added == nr_level_1);
    m_only_collection_name_changed = m_are_equal && collection_name_changed;
    m_are_equal = m_are_equal && !collection_name_changed;
    m_changes_properties = m_changes_properties || m_only_collection_name_changed;
    m_not_trivial_changes = m_not_trivial_changes || m_reorders_levels || m_adds_levels || m_removes_levels;
}


bool CollectionComparer::areEqual() const
{
    return m_are_equal;
}


bool CollectionComparer::areUnrelated() const
{
    return m_are_unrelated;
}


bool CollectionComparer::onlyTrivialChanges() const
{
    return !m_are_equal && !m_not_trivial_changes;
}


bool CollectionComparer::onlyCollectionNameChanged() const
{
    return m_only_collection_name_changed;
}


bool CollectionComparer::addsLevels() const
{
    return m_adds_levels;
}


bool CollectionComparer::removesLevels() const
{
    return m_removes_levels;
}


bool CollectionComparer::changesProperties() const
{
    return m_changes_properties;
}


bool CollectionComparer::reordersLevels() const
{
    return m_reorders_levels;
}


bool CollectionComparer::operator < (CollectionComparer const & other) const
{
    if (m_are_unrelated)
    {
        return true;
    }

    if (m_are_equal)
    {
        return false;
    }

    if (m_nr_removed > other.m_nr_removed)
    {
        return true;
    }
    else if (m_nr_removed < other.m_nr_removed)
    {
        return false;
    }

    if (m_nr_added > other.m_nr_added)
    {
        return true;
    }
    else if (m_nr_added < other.m_nr_added)
    {
        return false;
    }

    if (m_nr_reordered > other.m_nr_reordered)
    {
        return true;
    }
    else if (m_nr_reordered < other.m_nr_reordered)
    {
        return false;
    }

    return m_changes_properties;
}
