/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "collection_holder.h"

#include <cassert>

#include <qdatetime.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qstring.h>
#include <qstringlist.h>

#include <kapp.h>
#include <kconfig.h>
#include <kglobal.h>
#include <klineeditdlg.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kstddirs.h>

#include "collection.h"
#include "collection_comparer.h"
#include "compressed_map.h"
#include "level.h"




std::vector<Collection *> CollectionHolder::s_collections;
std::vector<int> CollectionHolder::s_temporary;
bool CollectionHolder::s_modified = false;
bool CollectionHolder::s_initialized = false;


void CollectionHolder::load()
{
    assert(s_initialized == false);

    s_initialized = true;

    QString compiled_file = KGlobal::dirs()->findResource("data", "easysok/levels.dat");

    if (!QFile::exists(compiled_file))
    {
        QStringList files = KGlobal::dirs()->findAllResources("data", "easysok/levels/*.xsb");
        getCollections(files);
    }
    else
    {
        getCollections(compiled_file);

        KConfig * config = KApplication::kApplication()->config();
        config->setGroup("");

        QDateTime last_update;
        last_update.setTime_t(0);
        last_update = config->readDateTimeEntry("Last collection update", &last_update);

        if (updateCollections(last_update, false))
        {
            setModified();
            save();
        }
    }

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    if (config->readNumEntry("Times started", 0) == 5)
    {
        s_modified = true;
        save();
    }
}


void CollectionHolder::save()
{
    assert(s_initialized == true);

    if (!s_modified)
    {
        return;
    }

    QString levels_file = KGlobal::dirs()->saveLocation("data", "easysok/", true);
    levels_file += "levels.dat";

    QFile file(levels_file);

    if (file.open(IO_WriteOnly | IO_Truncate))
    {
        QDataStream stream(&file);

        Q_INT32 version = 1;
        stream << version;

        int number_of_collections = numberOfCollections();
        Q_INT32 number_of_not_temporary_collections = 0;

        for (int i = 0; i < number_of_collections; ++i)
        {
            if (!isTemporary(i))
            {
                ++ number_of_not_temporary_collections;
            }
        }

        stream << number_of_not_temporary_collections;

        for (Q_INT32 i = 0; i < number_of_collections; ++i)
        {
            if (!isTemporary(i))
            {
                s_collections[i]->writeToStream(stream);
            }
        }
    }
}


int CollectionHolder::numberOfCollections()
{
    assert(s_initialized == true);

    return static_cast<int> (s_collections.size());
}


int CollectionHolder::numberOfPermanentCollections()
{
    assert(s_initialized == true);

    return numberOfCollections() - numberOfTemporaryCollections();
}


int CollectionHolder::numberOfTemporaryCollections()
{
    assert(s_initialized == true);

    int const number_of_collections = numberOfCollections();
    int temporary_collections = 0;

    for (int i = 0; i < number_of_collections; ++i)
    {
        if (s_temporary[i])
        {
            ++temporary_collections;
        }
    }

    return temporary_collections;
}


int CollectionHolder::indexFromName(QString const & name)
{
    assert(s_initialized == true);

    int const number_of_collections = numberOfCollections();

    for (int i = 0; i < number_of_collections; ++i)
    {
        if (collection(i)->name() == name)
        {
            return i;
        }
    }

    return -1;
}


Collection * CollectionHolder::collection(int index)
{
    assert(s_initialized == true);
    assert(index >= 0);
    assert(index < numberOfCollections());

    return s_collections[index];
}


bool CollectionHolder::isTemporary(int index)
{
    assert(s_initialized == true);
    assert(index >= 0);
    assert(index < numberOfCollections());

    return s_temporary[index];
}


void CollectionHolder::setTemporary(int index, bool temporary)
{
    assert(s_initialized == true);
    assert(index >= 0);
    assert(index < numberOfCollections());

    if (isTemporary(index) != temporary)
    {
        s_temporary[index] = temporary;
        setModified();
    }
}


int CollectionHolder::addCollection(Collection * collection, bool temporary)
{
    assert(s_initialized == true);
    assert(collection != 0);

    int const old_index = indexFromName(collection->name());

    if (old_index == -1)
    {
        s_collections.push_back(collection);
        s_temporary.push_back(temporary);

        s_modified = !temporary;

        return numberOfCollections() - 1;
    }
    else
    {
        delete s_collections[old_index];
        s_collections[old_index] = collection;
        s_temporary[old_index] = temporary;

        s_modified = true;

        return old_index;
    }
}


void CollectionHolder::removeCollection(int index)
{
    assert(s_initialized == true);
    assert(index >= 0);
    assert(index < numberOfCollections());

    s_modified = true;

    delete s_collections[index];
    s_collections.erase(s_collections.begin() + index);
    s_temporary.erase(s_temporary.begin() + index);
}


void CollectionHolder::setModified()
{
    assert(s_initialized == true);

    s_modified = true;
}


void CollectionHolder::findCompressedMap(CompressedMap const & compressed_map,
                                         int & collection_nr, int & level_nr)
{
    assert(s_initialized == true);

    int const number_of_collections = numberOfCollections();

    for (int i = 0; i < number_of_collections; ++i)
    {
        Collection * act_collection = collection(i);
        int const number_of_levels = act_collection->numberOfLevels();

        for (int j = 0; j < number_of_levels; ++j)
        {
            if (compressed_map == act_collection->level(j).compressedMap())
            {
                collection_nr = i;
                level_nr = j;

                return;
            }
        }
    }

    collection_nr = -1;
}


bool CollectionHolder::updateCollections(QDateTime const & min_date, bool show_trival_changes)
{
    QStringList all_files = KGlobal::dirs()->findAllResources("data", "easysok/levels/*.xsb");
    QStringList new_files;

    int const nr_old_files = all_files.count();

    for (int i = 0; i < nr_old_files; ++i)
    {
        QFileInfo file_info(all_files[i]);

        if (file_info.lastModified() >= min_date)
        {
            new_files.append(all_files[i]);
        }
    }

    if (new_files.isEmpty())
    {
        return false;
    }

    bool found_updated = false;
    int const nr_files = new_files.count();
    int const nr_collections = numberOfCollections();

    for (int i = 0; i < nr_files; ++i)
    {
        Collection act_collection(new_files[i]);

        if (act_collection.numberOfLevels() == 0)
        {
            continue;
        }

        CollectionComparer comparer(act_collection, *collection(0));
        int best_fit = 0;

        for (int j = 1; j < nr_collections; ++j)
        {
            CollectionComparer new_comparer(act_collection, *collection(j));

            if (comparer < new_comparer)
            {
                comparer = new_comparer;
                best_fit = j;
            }
        }

        if (comparer.areEqual())
        {
            continue;
        }

        bool add_collection = false;
        bool replace_collection = false;
        bool is_trivial = false;

        if (comparer.areUnrelated())
        {
            QString msg = i18n("Do you want to install the new collection\n'%1'?").arg(act_collection.name());

            if (KMessageBox::questionYesNo(0, msg) == KMessageBox::Yes)
            {
                add_collection = true;
                found_updated = true;
            }
        }
        else
        {
            QString msg;

            if (comparer.onlyCollectionNameChanged())
            {
                msg = i18n("The new collection\n'%1' changes the name of the collection\n'%2'.");
            }
            else if (comparer.removesLevels())
            {
                if (comparer.addsLevels())
                {
                    msg = i18n("The new collection\n'%1' removes and adds levels to the collection\n'%2'.");
                }
                else
                {
                    msg = i18n("The new collection\n'%1' removes levels of the collection\n'%2'.");
                }
            }
            else if (comparer.addsLevels())
            {
                if (comparer.reordersLevels())
                {
                    msg = i18n("The new collection\n'%1' adds and reorders levels of the collection\n'%2'.");
                }
                else
                {
                    msg = i18n("The new collection\n'%1' adds levels to the collection\n'%2'.");
                }
            }
            else if (comparer.reordersLevels())
            {
                if (comparer.changesProperties())
                {
                    msg = i18n("The new collection\n'%1' reorders levels and changes properties "
                               "of the collection\n'%2'.");
                }
                else
                {
                    msg = i18n("The new collection\n'%1' reorders levels of the collection\n'%2'.");
                }
            }
            else if (comparer.changesProperties())
            {
                msg = i18n("The new collection\n'%1' changes some properties of the collection\n'%2'.");
                is_trivial = comparer.onlyTrivialChanges();
            }
            else
            {
                continue;
            }

            if (!is_trivial || show_trival_changes)
            {
                msg = msg.arg(act_collection.name()).arg(collection(best_fit)->name());
                msg += "\n\n" + i18n("Do you want to replace the old collection or add the new collection?");

                int const result = KMessageBox::warningYesNoCancel(0, msg, "", i18n("Replace"), i18n("Add"));

                found_updated = true;

                if (result == KMessageBox::Yes)
                {
                    replace_collection = true;
                }
                else if (result == KMessageBox::No)
                {
                    add_collection = true;
                }
            }
            else
            {
                replace_collection = true;
            }
        }

        if (add_collection)
        {
            while (indexFromName(act_collection.name()) != -1)
            {
                QString const msg = i18n("A collection with the same name already exists.\n"
                                         "Please enter another name.");

                KLineEditDlg dialog(msg, act_collection.name(), 0);

                if (!dialog.exec())
                {
                    add_collection = false;

                    break;
                }

                act_collection.setName(dialog.text());
            }

            if (add_collection)
            {
                addCollection(new Collection(act_collection), false);
            }
        }
        else if (replace_collection)
        {
            while ( (indexFromName(act_collection.name()) != -1) &&
                    (act_collection.name() != collection(best_fit)->name()) )
            {
                QString const msg = i18n("A collection with the same name already exists.\n"
                                         "Please enter another name.");

                KLineEditDlg dialog(msg, act_collection.name(), 0);

                if (!dialog.exec())
                {
                    replace_collection = false;

                    break;
                }

                act_collection.setName(dialog.text());
            }

            if (replace_collection)
            {
                removeCollection(indexFromName(collection(best_fit)->name()));
                addCollection(new Collection(act_collection), false);
            }
        }
    }

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Last collection update", QDateTime::currentDateTime());

    return found_updated;
}


void CollectionHolder::getCollections(QString const & filename)
{
    assert(s_initialized == true);

    QFile file(filename);

    if (!file.open(IO_ReadOnly))
    {
        return;
    }

    QDataStream stream(&file);

    Q_INT32 version;
    stream >> version;

    // Updated to version 1
    if (version < 1)
    {
        setModified();
    }

    Q_INT32 number_of_collections;
    stream >> number_of_collections;

    for (Q_INT32 i = 0; i < number_of_collections; ++i)
    {
        Collection * collection = new Collection(stream, version);
        s_collections.push_back(collection);
        s_temporary.push_back(false);
    }
}


void CollectionHolder::getCollections(QStringList const & files)
{
    assert(s_initialized == true);

    QStringList::ConstIterator it = files.begin();
    QStringList::ConstIterator const end_it = files.end();

    for (; it != end_it; ++it)
    {
        Collection * collection = new Collection(*it);
        s_collections.push_back(collection);
        s_temporary.push_back(false);
    }
}
