/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_COLLECTION_HOLDER_INC_GUARD_H
#define EASYSOK_COLLECTION_HOLDER_INC_GUARD_H


#include <vector>


class Collection;
class CompressedMap;
class QDateTime;
class QString;
class QStringList;




/**
 * This class holds the collections.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class CollectionHolder
{

public:

    /**
     * Loads the collections. Call this first and only once!
     */

    static void load();


    /**
     * Saves the collections, if they were modified.
     */

    static void save();


    /**
     * Returns the number of collections.
     */

    static int numberOfCollections();


    /**
     * Returns the number of non temporary collections.
     */

    static int numberOfPermanentCollections();


    /**
     * Returns the number of temporary collections.
     */

    static int numberOfTemporaryCollections();


    /**
     * Returns the index of the collection with the given name or -1 if it does not exists.
     *
     * @param name The name of the collection.
     */

    static int indexFromName(QString const & name);


    /**
     * Returns a pointer to the given collection.
     *
     * @param index The index of the collection.
     */

    static Collection * collection(int index);


    /**
     * Returns true, if the given collection is only temporary.
     *
     * @param index The index of the collection.
     */

    static bool isTemporary(int index);


    /**
     * Makes the current collection temporary or not.
     *
     * @param index The index of the collection.
     * @param temporary If true the collection will be temporary.
     */

    static void setTemporary(int index, bool temporary);


    /**
     * Adds a new collection.
     *
     * If a collection with the same name exists, it will be deleted!
     *
     * This object takes over ownership of the collection.
     *
     * @param collection The new collection.
     * @param temporary If true, the collection is only temporary and will not be saved.
     * @return The index of the collection.
     */

    static int addCollection(Collection * collection, bool temporary);


    /**
     * Removes the collection with the given index.
     *
     * The collection must be modifyable.
     *
     * @param index The index of the collection to remove.
     */

    static void removeCollection(int index);


    /**
     * Marks a collection as modified.
     */

    static void setModified();


    /**
     * Gets the index of the collection and level of a compressed map.
     *
     * If the compressed map is not found, collection_nr will be set to -1.
     *
     * @param compressed_map The compressed to search.
     * @param collection_nr Here we store the colleciton number.
     * @param level_nr Here we store the level number.
     */

    static void findCompressedMap(CompressedMap const & compressed_map, int & collection_nr, int & level_nr);


    /**
     * Searches for updated or new collections.
     *
     * @param min_date The minimum date of collections to consider.
     * @param show_trival_changes If false, trivial changes are not shown but simply applied.
     * @return True, if at least one changed or new collection was found.
     */

    static bool updateCollections(QDateTime const & min_date, bool show_trival_changes);


private:

    /**
     * Loads the collections from the given compiled file.
     *
     * @param filename The file to use.
     */

    static void getCollections(QString const & filename);


    /**
     * Loads the collections from the given xsb-files.
     *
     * @param files The files from which to load the collections.
     */

    static void getCollections(QStringList const & files);


    /**
     * The collections.
     */

    static std::vector<Collection *> s_collections;


    /**
     * The flags, if the level are only temporary.
     */

    static std::vector<int> s_temporary;


    /**
     * If true, the collection was modified.
     */

    static bool s_modified;


    /**
     * If true, load was called.
     */

    static bool s_initialized;
};


#endif
