/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "compressed_map.h"

#include <cassert>

#include <qdatastream.h>

#include "map.h"




CompressedMap::CompressedMap(Map const & map)
{
    assert(map.width() < 128);
    assert(map.height() < 128);

    m_width = static_cast<Q_UINT8> (map.width());
    m_height = static_cast<Q_UINT8> (map.height());

    if (map.isValid())
    {
        m_empty_goals = static_cast<Q_UINT16> (map.numberOfEmptyGoals());
        m_keeper_index = static_cast<Q_UINT16> (map.getIndex(map.keeper()));
    }

    int const size = m_width * m_height;

    int const codes_length = codesLength();
    m_codes.resize(codes_length);

    int code_index = 0;
    int fields_in_code = 0;
    Q_UINT32 act_code = 0;

    for (int i = 0; i < size; ++i)
    {
        act_code >>= 3;
        act_code += map.getPiece(i) * 0x08000000;
        ++fields_in_code;

        if (fields_in_code == 10)
        {
            m_codes[code_index] = act_code;
            ++code_index;
            act_code = 0;
            fields_in_code = 0;
        }
    }

    if (fields_in_code != 0)
    {
        for (int i = fields_in_code; i < 10; ++i)
        {
            act_code >>= 3;
        }

        m_codes[code_index] = act_code;
    }
}


CompressedMap::CompressedMap(QDataStream & stream)
{
    stream >> m_width;
    stream >> m_height;
    stream >> m_empty_goals;
    stream >> m_keeper_index;

    int const codes_length = codesLength();
    m_codes.resize(codes_length);
    stream.readRawBytes(reinterpret_cast<char *> (&(m_codes[0])), codes_length * 4);
}


int CompressedMap::width() const
{
    return m_width;
}


int CompressedMap::height() const
{
    return m_height;
}


int CompressedMap::numberOfEmptyGoals() const
{
    return m_empty_goals;
}


int CompressedMap::keeperIndex() const
{
    return m_keeper_index;
}


void CompressedMap::setPieces(std::vector<int> & pieces) const
{
    int const size = static_cast<int> (m_width) * m_height;
    pieces.resize(size);

    int code_index = 0;
    int fields_in_code = 0;
    Q_UINT32 act_code = m_codes[0];

    for (int i = 0; i < size; ++i)
    {
        if (fields_in_code == 10)
        {
            ++code_index;
            act_code = m_codes[code_index];
            fields_in_code = 0;
        }

        pieces[i] = act_code & 7;
        act_code >>= 3;

        ++fields_in_code;
    }
}


bool CompressedMap::operator == (CompressedMap const & other_map) const
{
    if (m_keeper_index != other_map.m_keeper_index)
    {
        return false;
    }

    if (m_empty_goals != other_map.m_empty_goals)
    {
        return false;
    }

    if (m_width != other_map.m_width)
    {
        return false;
    }

    if (m_height != other_map.m_height)
    {
        return false;
    }

    return m_codes == other_map.m_codes;
}


bool CompressedMap::operator != (CompressedMap const & other_map) const
{
    return !(*this == other_map);
}


bool CompressedMap::operator < (CompressedMap const & other_map) const
{
    if (m_keeper_index < other_map.m_keeper_index)
    {
        return true;
    }
    else if (m_keeper_index > other_map.m_keeper_index)
    {
        return false;
    }

    if (m_empty_goals < other_map.m_empty_goals)
    {
        return true;
    }
    else if (m_empty_goals > other_map.m_empty_goals)
    {
        return false;
    }

    if (m_width < other_map.m_width)
    {
        return true;
    }
    else if (m_width > other_map.m_width)
    {
        return false;
    }

    if (m_height < other_map.m_height)
    {
        return true;
    }
    else if (m_height > other_map.m_height)
    {
        return false;
    }

    return m_codes < other_map.m_codes;
}


void CompressedMap::writeToStream(QDataStream & stream) const
{
    stream << m_width;
    stream << m_height;
    stream << m_empty_goals;
    stream << m_keeper_index;

    stream.writeRawBytes(reinterpret_cast<char const *> (&(m_codes[0])), codesLength() * 4);
}


int CompressedMap::codesLength() const
{
    return (m_width * m_height - 1) / 10 + 1;
}
