/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_COMPRESSED_MAP_INC_GUARD_H
#define EASYSOK_COMPRESSED_MAP_INC_GUARD_H


#include <vector>

#include <qglobal.h>


class Map;
class QDataStream;




/**
 * This class represents a compressed version of a Map.
 *
 * This class is used for fast loading and storing and to check if to maps are equal.
 * It also provides all operators to use it as a key in std::map<>.
 *
 * Note, that Map itself uses a CompressedMap for loading and saving.
 *
 * The format of a compressed map is easy. We simply store the pieces (which are in the
 * range [0:8[) in 32 bit integers. Because a piece takes 3 bits, we can store 10 pieces
 * per int.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class CompressedMap
{

public:

    /**
     * Constructs the compressed map from a normal map.
     *
     * @param map The map to use.
     */

    CompressedMap(Map const & map);


    /**
     * Constructs the compressed map from a data stream.
     *
     * @param stream The stream to use.
     */

    CompressedMap(QDataStream & stream);


    /**
     * Returns the width of the map.
     */

    int width() const;


    /**
     * Returns the height of the map.
     */

    int height() const;


    /**
     * Returns the number of empty goals (without a gem).
     */

    int numberOfEmptyGoals() const;


    /**
     * Returns the index of the keeper.
     */

    int keeperIndex() const;


    /**
     * Fills the given vector with the piece codes suitable to create a Map.
     *
     * @param pieces The vector to fill.
     */

    void setPieces(std::vector<int> & pieces) const;


    /**
     * Compares to maps for equality.
     *
     * @param other_map The other map.
     */

    bool operator == (CompressedMap const & other_map) const;


    /**
     * Compares to maps for inequality.
     *
     * @param other_map The other map.
     */

    bool operator != (CompressedMap const & other_map) const;


    /**
     * This is needed for associative containers like std::map.
     *
     * @param other_map The other map.
     */

    bool operator < (CompressedMap const & other_map) const;


    /**
     * Writes the compressed map to a data stream.
     *
     * @param stream The stream to use.
     */

    void writeToStream(QDataStream & stream) const;


private:

    /**
     * Returns the length of the code array.
     */

    int codesLength() const;


    /**
     * The width of the map.
     */

    Q_UINT8 m_width;


    /**
     * The height of the map.
     */

    Q_UINT8 m_height;


    /**
     * The number of empty goals in the map.
     */

    Q_UINT16 m_empty_goals;


    /**
     * The index of the keeper.
     */

    Q_UINT16 m_keeper_index;


    /**
     * Here we store the compressed moves.
     */

    std::vector<Q_UINT32> m_codes;
};


#endif
