/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "compressed_movements.h"

#include <algorithm>
#include <cassert>

#include <qdatastream.h>


#include "movements.h"




CompressedMovements::CompressedMovements(Movements const & movements)
{
    m_number_of_moves = movements.numberOfMoves();

    if (m_number_of_moves < 1)
    {
        return;
    }

    std::vector<Q_UINT32> compressed_moves;

    QPoint act_from = movements.move(0).from();

    Q_UINT32 raw_data = (act_from.x() << 18) + (act_from.y() << 25);
    int bits_used = 14;

    for (int i = 0; i < m_number_of_moves; ++i)
    {
        Move const move = movements.move(i);
        int const stone_pushed = move.stonePushed();

        if (move.from() != act_from)
        {
            m_number_of_moves = 0;

            return;
        }

        assert(stone_pushed >= 0);
        assert(stone_pushed <= 1);

        if (move.isAtomicMove())
        {
            int const atomic_move = static_cast<int> (move.atomicMove().type());

            assert(atomic_move >= 0);
            assert(atomic_move < 4);

            if (bits_used + 4 > 32)
            {
                raw_data >>= (32 - bits_used);
                compressed_moves.push_back(raw_data);
                bits_used = 0;
            }

            raw_data >>= 4;
            raw_data += (1 << 28) + (stone_pushed << 29) + (atomic_move << 30);
            bits_used += 4;
        }
        else
        {
            int const to_x = move.to().x();
            int const to_y = move.to().y();

            assert(to_x >= 0);
            assert(to_x < 128);
            assert(to_y >= 0);
            assert(to_y < 128);

            if (bits_used + 16 > 32)
            {
                raw_data >>= (32 - bits_used);
                compressed_moves.push_back(raw_data);
                bits_used = 0;
            }

            raw_data >>= 16;
            raw_data += (stone_pushed << 17) + (to_x << 18) + (to_y << 25);
            bits_used += 16;
        }

        act_from = move.to();
    }

    if (bits_used != 0)
    {
        raw_data >>= (32 - bits_used);
        compressed_moves.push_back(raw_data);
    }

    m_moves = compressed_moves;
}


CompressedMovements::CompressedMovements(QDataStream & stream)
{
    Q_INT32 number_of_moves;
    stream >> number_of_moves;

    m_number_of_moves = static_cast<int> (number_of_moves);

    Q_INT32 number_of_data;
    stream >> number_of_data;

    m_moves.resize(number_of_data);

    for (int i = 0; i < number_of_data; ++i)
    {
        stream >> m_moves[i];
    }
}


int CompressedMovements::numberOfMoves() const
{
    return m_number_of_moves;
}


void CompressedMovements::setMoves(std::vector<Move> & moves) const
{
    moves.resize(m_number_of_moves);

    if (m_number_of_moves < 1)
    {
        return;
    }

    QPoint last_from(m_moves[0] & 127, (m_moves[0] & 16256) >> 7);
    Q_UINT32 raw_data = m_moves[0];
    raw_data >>= 14;
    int bits_used = 14;
    int raw_moves_pointer = 0;

    for (int i = 0; i < m_number_of_moves; ++i)
    {
        Move move;

        if (raw_data & 1)
        {
            if (bits_used + 4 > 32)
            {
                ++raw_moves_pointer;
                raw_data = m_moves[raw_moves_pointer];
                bits_used = 0;
            }
        }
        else
        {
            if (bits_used + 16 > 32)
            {
                ++raw_moves_pointer;
                raw_data = m_moves[raw_moves_pointer];
                bits_used = 0;
            }
        }

        if (raw_data & 1)
        {
            int const stone_pushed = raw_data & 2;
            int const type = ((raw_data & 12) >> 2);
            raw_data >>= 4;
            bits_used += 4;

            move = Move(last_from, static_cast<AtomicMove::AtomicMoveType> (type), stone_pushed);
        }
        else
        {
            int const stone_pushed = raw_data & 2;
            int const to_x = (raw_data & 508) >> 2;
            int const to_y = (raw_data & 65024) >> 9;
            raw_data >>= 16;
            bits_used += 16;

            move = Move(last_from, QPoint(to_x, to_y), stone_pushed);
        }

        moves[i] = move;
        last_from = move.to();
    }
}


void CompressedMovements::writeToStream(QDataStream & stream) const
{
    Q_INT32 number_of_moves = static_cast<Q_INT32> (m_number_of_moves);

    stream << number_of_moves;

    Q_INT32 number_of_data = static_cast<Q_INT32> (m_moves.size());

    stream << number_of_data;

    for (int i = 0; i < number_of_data; ++i)
    {
        stream << m_moves[i];
    }
}
