/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "configuration_dialog.h"

#include <algorithm>
#include <cassert>

#include <qbuttongroup.h>
#include <qcheckbox.h>
#include <qframe.h>
#include <qgroupbox.h>
#include <qlayout.h>
#include <qradiobutton.h>

#include <kapp.h>
#include <kconfig.h>
#include <kiconloader.h>
#include <klocale.h>
#include <knuminput.h>




ConfigurationDialog::ConfigurationDialog(QWidget * parent, char const * name ) :
    KDialogBase(IconList, i18n("Options"), Ok | Cancel | Default | Apply | Help, Ok, parent, name, true, true)
{
    setupCorePage();
    setupMousePage();
    setupSolverPage();
    setupBookmarksPage();
    setupAnimationSpeedPage();
    setupScalingPage();
    setupConfirmationPage();

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    if (config->hasKey("Configuration window geometry"))
    {
        setGeometry(config->readRectEntry("Configuration window geometry", 0));
    }

    setHelp("configuration-dialog");
}


ConfigurationDialog::~ConfigurationDialog()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Configuration window geometry", geometry());
}


void ConfigurationDialog::slotOk()
{
    applyCoreSettings();
    applyMouseSettings();
    applySolverSettings();
    applyBookmarksSettings();
    applyAnimationSpeedSettings();
    applyScalingSettings();
    applyConfirmationSettings();

    emit configurationChanged();
    accept();
}


void ConfigurationDialog::slotApply()
{
    switch (activePageIndex())
    {
    case CORE:
        applyCoreSettings();
        break;

    case MOUSE:
        applyMouseSettings();
        break;

    case SOLVER:
        applySolverSettings();
        break;

    case BOOKMARKS:
        applyBookmarksSettings();
        break;

    case ANIMATION_SPEEDS:
        applyAnimationSpeedSettings();
        break;

    case SCALING:
        applyScalingSettings();
        break;

    case CONFIRMATIONS:
        applyConfirmationSettings();
        break;

    default:
        assert(false);
    }

    emit configurationChanged();
}


void ConfigurationDialog::slotDefault()
{
    switch (activePageIndex())
    {
    case CORE:
        setupCoreDefaults();
        break;

    case MOUSE:
        setupMouseDefaults();
        break;

    case SOLVER:
        setupSolverDefaults();
        break;

    case BOOKMARKS:
        setupBookmarksDefaults();
        break;

    case ANIMATION_SPEEDS:
        setupAnimationSpeedDefaults();
        break;

    case SCALING:
        setupScalingDefaults();
        break;

    case CONFIRMATIONS:
        setupConfirmationDefaults();
        break;

    default:
        assert(false);
    }
}


void ConfigurationDialog::setupCorePage()
{
    QFrame * page = addPage(i18n("Core"), i18n("Here you can adjust some core settings"),
                            BarIcon("easysok", KIcon::SizeMedium));

    QVBoxLayout * layout = new QVBoxLayout(page, marginHint(), spacingHint());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_next_after_solved = new QCheckBox(i18n("Jump to next level when solved a level"), page);
    bool const next_level = config->readBoolEntry("Jump to next level when solved", true);
    m_next_after_solved->setChecked(next_level);
    layout->addWidget(m_next_after_solved);

    m_illegal_as_virtual = new QCheckBox(i18n("Treat illegal moves as virtual moves"), page);
    m_illegal_as_virtual->setChecked(config->readBoolEntry("Illegal moves as virtual", false));
    layout->addWidget(m_illegal_as_virtual);

    m_wrap_virtual_keeper = new QCheckBox(i18n("Wrap the virtual keeper"), page);
    m_wrap_virtual_keeper->setChecked(config->readBoolEntry("Wrap virtual keeper", true));
    layout->addWidget(m_wrap_virtual_keeper);

    m_goto_any_level = new QCheckBox(i18n("Allow goto to any level"), page);
    m_goto_any_level->setChecked(config->readBoolEntry("Goto any level", true));
    layout->addWidget(m_goto_any_level);

    m_always_show_advanced_options = new QCheckBox(i18n("Always show advanced options"), page);
    m_always_show_advanced_options->setChecked(config->readBoolEntry("Always show advanced options", false));
    layout->addWidget(m_always_show_advanced_options);

    QGroupBox * auto_save_box = new QGroupBox(2, Vertical, i18n("Automatic saving of collections, "
                                                                "solutions and bookmarks"), page);
    layout->addWidget(auto_save_box);

    int const auto_save_time = config->readNumEntry("Auto save time", 15);

    m_auto_save_enabled = new QCheckBox(i18n("Enable auto save"), auto_save_box);
    m_auto_save_enabled->setChecked(auto_save_time > 0);

    m_auto_save_time = new KIntNumInput(auto_save_time, auto_save_box);
    m_auto_save_time->setLabel(i18n("Auto save time"));
    m_auto_save_time->setSuffix(" " + i18n("minutes"));
    m_auto_save_time->setRange(1, 60);
    connect(m_auto_save_enabled, SIGNAL(toggled(bool)), m_auto_save_time, SLOT(setEnabled(bool)));

    QGroupBox * recent_collections_box = new QGroupBox(1, Vertical, i18n("Number of recent collections to "
                                                                         "display"), page);
    layout->addWidget(recent_collections_box);

    int nr_of_recent_collections = config->readNumEntry("Number of recent collections", 5);

    m_recent_collections = new KIntNumInput(nr_of_recent_collections, recent_collections_box);
    m_recent_collections->setRange(1, 10);

    QGroupBox * status_bar_box = new QGroupBox(3, Vertical,
                                               i18n("Information displayed in the status bar"), page);
    layout->addWidget(status_bar_box);

    bool const show_difficulty = config->readBoolEntry("Show difficulty", true);
    m_show_difficulty = new QCheckBox(i18n("Show difficulty"), status_bar_box);
    m_show_difficulty->setChecked(show_difficulty);

    bool const show_linear_pushes = config->readBoolEntry("Show linear pushes", false);
    m_show_linear_pushes = new QCheckBox(i18n("Show number of linear pushes"), status_bar_box);
    m_show_linear_pushes->setChecked(show_linear_pushes);

    bool const show_gem_changes = config->readBoolEntry("Show gem changes", false);
    m_show_gem_changes = new QCheckBox(i18n("Show number of gem changes"), status_bar_box);
    m_show_gem_changes->setChecked(show_gem_changes);

    bool const show_gems_left = config->readBoolEntry("Show gems left", true);
    m_show_gems_left = new QCheckBox(i18n("Show gems left"), status_bar_box);
    m_show_gems_left->setChecked(show_gems_left);

    bool const show_author = config->readBoolEntry("Show author", true);
    m_show_author = new QCheckBox(i18n("Show author of level"), status_bar_box);
    m_show_author->setChecked(show_author);

    bool const show_user = config->readBoolEntry("Show user", true);
    m_show_user = new QCheckBox(i18n("Show current user"), status_bar_box);
    m_show_user->setChecked(show_user);

    layout->addStretch(1);
}


void ConfigurationDialog::setupCoreDefaults()
{
    m_next_after_solved->setChecked(true);
    m_illegal_as_virtual->setChecked(false);
    m_wrap_virtual_keeper->setChecked(true);
    m_goto_any_level->setChecked(true);
    m_always_show_advanced_options->setChecked(false);
    m_show_difficulty->setChecked(true);
    m_show_linear_pushes->setChecked(false);
    m_show_gem_changes->setChecked(false);
    m_show_gems_left->setChecked(true);
    m_show_author->setChecked(true);
    m_show_user->setChecked(true);
    m_auto_save_enabled->setChecked(true);
    m_auto_save_time->setValue(15);
    m_recent_collections->setValue(5);
}


void ConfigurationDialog::applyCoreSettings()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Jump to next level when solved", m_next_after_solved->isChecked());
    config->writeEntry("Illegal moves as virtual", m_illegal_as_virtual->isChecked());
    config->writeEntry("Wrap virtual keeper", m_wrap_virtual_keeper->isChecked());
    config->writeEntry("Goto any level", m_goto_any_level->isChecked());
    config->writeEntry("Always show advanced options", m_always_show_advanced_options->isChecked());
    config->writeEntry("Show difficulty", m_show_difficulty->isChecked());
    config->writeEntry("Show linear pushes", m_show_linear_pushes->isChecked());
    config->writeEntry("Show gem changes", m_show_gem_changes->isChecked());
    config->writeEntry("Show gems left", m_show_gems_left->isChecked());
    config->writeEntry("Show author", m_show_author->isChecked());
    config->writeEntry("Show user", m_show_user->isChecked());
    config->writeEntry("Auto save time", m_auto_save_enabled ? m_auto_save_time->value() : 0);
    config->writeEntry("Number of recent collections", m_recent_collections->value());
}


void ConfigurationDialog::setupMousePage()
{
    QFrame * page = addPage(i18n("Mouse"), i18n("Here you can adjust the mouse settings"),
                            BarIcon("mouse", KIcon::SizeMedium));

    QVBoxLayout * layout = new QVBoxLayout(page, marginHint(), spacingHint());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QGroupBox * hide_box = new QGroupBox(2, Vertical, i18n("Auto hiding of mouse cursor"), page);
    layout->addWidget(hide_box);

    m_hide_cursor = new QCheckBox(i18n("Enable automatically hiding of unused mouse cursor"), hide_box);
    m_hide_cursor->setChecked(config->readBoolEntry("Cursor hide", true));

    int delay = config->readNumEntry("Cursor hide delay", 5);
    delay = std::max(std::min(delay, 60), 1);
    m_hide_cursor_delay = new KIntNumInput(delay, hide_box);
    m_hide_cursor_delay->setLabel(i18n("Hiding delay"));
    m_hide_cursor_delay->setSuffix(" " + i18n("seconds"));
    m_hide_cursor_delay->setEnabled(m_hide_cursor->isChecked());
    m_hide_cursor_delay->setRange(1, 60);
    connect(m_hide_cursor, SIGNAL(toggled(bool)), m_hide_cursor_delay, SLOT(setEnabled(bool)));

    QGroupBox * repeat_box = new QGroupBox(2, Vertical, i18n("Mouse repeat"), page);
    layout->addWidget(repeat_box);

    int repeat_delay = config->readNumEntry("Mouse repeat delay", 1000);
    m_mouse_repeat_delay = new KIntNumInput(repeat_delay, repeat_box);
    m_mouse_repeat_delay->setLabel(i18n("Mouse repeat delay"));
    m_mouse_repeat_delay->setSuffix(" " + i18n("miliseconds"));
    m_mouse_repeat_delay->setRange(1, 2000);

    int repeat_rate = config->readNumEntry("Mouse repeat rate");
    m_mouse_repeat_rate = new KIntNumInput(repeat_rate, repeat_box);
    m_mouse_repeat_rate->setLabel(i18n("Mouse repeat rate"));
    m_mouse_repeat_rate->setSuffix(" " + i18n("repeat(s) per second"));
    m_mouse_repeat_rate->setRange(1, 100);

    layout->addStretch(1);
}


void ConfigurationDialog::setupMouseDefaults()
{
    m_hide_cursor->setChecked(true);
    m_hide_cursor_delay->setValue(5);
    m_mouse_repeat_delay->setValue(1000);
    m_mouse_repeat_rate->setValue(4);
}


void ConfigurationDialog::applyMouseSettings()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Cursor hide delay", m_hide_cursor_delay->value());
    config->writeEntry("Cursor hide", m_hide_cursor->isChecked());
    config->writeEntry("Mouse repeat delay", m_mouse_repeat_delay->value());
    config->writeEntry("Mouse repeat rate", m_mouse_repeat_rate->value());
}


void ConfigurationDialog::setupSolverPage()
{
    QFrame * page = addPage(i18n("Solver"), i18n("Here you can adjust the solver"),
                            BarIcon("idea", KIcon::SizeMedium));

    QVBoxLayout * layout = new QVBoxLayout(page, marginHint(), spacingHint());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_solver_steps_per_call = new KIntNumInput(page);
    m_solver_steps_per_call->setRange(1, 100000);
    m_solver_steps_per_call->setValue(config->readNumEntry("Solver steps per call", 1000));
    m_solver_steps_per_call->setLabel("Number of solver steps in a row");
    layout->addWidget(m_solver_steps_per_call);

    m_solver_cache_size = new KIntNumInput(page);
    m_solver_cache_size->setRange(1000, 10000000);
    m_solver_cache_size->setValue(config->readNumEntry("Solver cache size", 10000));
    m_solver_cache_size->setLabel("Number of entries in the cache");
    layout->addWidget(m_solver_cache_size);

    layout->addStretch(1);
}


void ConfigurationDialog::setupSolverDefaults()
{
    m_solver_steps_per_call->setValue(1000);
    m_solver_cache_size->setValue(10000);
}


void ConfigurationDialog::applySolverSettings()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Solver steps per call", m_solver_steps_per_call->value());
    config->writeEntry("Solver cache size", m_solver_cache_size->value());
}


void ConfigurationDialog::setupBookmarksPage()
{
    QFrame * page = addPage(i18n("Bookmarks"), i18n("Here you can adjust some properties of bookmarking"),
                            BarIcon("bookmark", KIcon::SizeMedium));

    QVBoxLayout * layout = new QVBoxLayout(page, marginHint(), spacingHint());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_bookmarks = new KIntNumInput(page);
    m_bookmarks->setRange(1, 200);
    m_bookmarks->setValue(config->readNumEntry("Number of bookmarks", 10));
    m_bookmarks->setLabel(i18n("Number of bookmarks"));
    layout->addWidget(m_bookmarks);

    QString const group_text = i18n("Information displayed in bookmarks");
    QButtonGroup * group = new QButtonGroup(3, Vertical, group_text, page);
    layout->addWidget(group);

    int const bookmark_info = config->readNumEntry("Bookmark format", 16);

    QString const subgroup1_text = i18n("Date and Time information");
    QButtonGroup * subgroup1 = new QButtonGroup(3, Vertical, subgroup1_text, group);
    subgroup1->setRadioButtonExclusive(true);

    m_no_bookmark_date = new QRadioButton(i18n("Show no date and time"), subgroup1);
    m_no_bookmark_date->setChecked((bookmark_info & 3) == 0);

    m_bookmark_date = new QRadioButton(i18n("Show only date"), subgroup1);
    m_bookmark_date->setChecked((bookmark_info & 3) == 1);

    m_bookmark_time = new QRadioButton(i18n("Show date and time"), subgroup1);
    m_bookmark_time->setChecked((bookmark_info & 3) == 2);


    QString const subgroup2_text = i18n("Collection and level information");
    QButtonGroup * subgroup2 = new QButtonGroup(3, Vertical, subgroup2_text, group);
    subgroup2->setRadioButtonExclusive(true);

    m_no_bookmark_collection = new QRadioButton(i18n("Show no level/collection information"), subgroup2);
    m_no_bookmark_collection->setChecked((bookmark_info & 12) == 0);

    m_bookmark_collection = new QRadioButton(i18n("Show collection name"), subgroup2);
    m_bookmark_collection->setChecked((bookmark_info & 12) == 4);

    m_bookmark_level = new QRadioButton(i18n("Show collection name and level number"), subgroup2);
    m_bookmark_level->setChecked((bookmark_info & 12) == 8);

    QButtonGroup * subgroup3 = new QButtonGroup(2, Vertical, i18n("Annotations"), group);
    subgroup3->setRadioButtonExclusive(true);

    m_no_bookmark_annotation = new QRadioButton(i18n("Show no annotations"), subgroup3);
    m_no_bookmark_annotation->setChecked((bookmark_info & 16) == 0);

    m_bookmark_annotation = new QRadioButton(i18n("Show annotations"), subgroup3);
    m_bookmark_annotation->setChecked((bookmark_info & 16) == 16);

    layout->addStretch(1);
}


void ConfigurationDialog::setupBookmarksDefaults()
{
    m_bookmarks->setValue(10);
    m_no_bookmark_date->setChecked(true);
    m_no_bookmark_collection->setChecked(true);
    m_bookmark_annotation->setChecked(true);
}


void ConfigurationDialog::applyBookmarksSettings()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Number of bookmarks", m_bookmarks->value());

    int bookmark_info = 0;

    if (m_bookmark_date->isChecked())
    {
        bookmark_info = 1;
    }
    else if (m_bookmark_time->isChecked())
    {
        bookmark_info = 2;
    }

    if (m_bookmark_collection->isChecked())
    {
        bookmark_info |= 4;
    }
    else if (m_bookmark_level->isChecked())
    {
        bookmark_info |= 8;
    }

    if (m_bookmark_annotation->isChecked())
    {
        bookmark_info |= 16;
    }

    config->writeEntry("Bookmark format", bookmark_info);
}


void ConfigurationDialog::setupAnimationSpeedPage()
{
    QFrame * page = addPage(i18n("Animation"), i18n("Here you can fine tune the three animation speeds"),
                            BarIcon("goto", KIcon::SizeMedium));

    QVBoxLayout * layout = new QVBoxLayout(page, marginHint(), spacingHint());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_fast_animation_speed = new KIntNumInput(page);
    m_fast_animation_speed->setRange(0, 1000);
    m_fast_animation_speed->setValue(config->readNumEntry("Fast animation time", 5));
    m_fast_animation_speed->setSuffix(" " + i18n("miliseconds"));
    m_fast_animation_speed->setLabel(i18n("Time between two moves in fast animation speed"));
    layout->addWidget(m_fast_animation_speed);

    m_normal_animation_speed = new KIntNumInput(page);
    m_normal_animation_speed->setRange(0, 1000);
    m_normal_animation_speed->setValue(config->readNumEntry("Normal animation time", 50));
    m_normal_animation_speed->setSuffix(" " + i18n("miliseconds"));
    m_normal_animation_speed->setLabel(i18n("Time between two moves in normal animation speed"));
    layout->addWidget(m_normal_animation_speed);

    m_slow_animation_speed = new KIntNumInput(page);
    m_slow_animation_speed->setRange(0, 1000);
    m_slow_animation_speed->setValue(config->readNumEntry("Slow animation time", 500));
    m_slow_animation_speed->setSuffix(" " + i18n("miliseconds"));
    m_slow_animation_speed->setLabel(i18n("Time between two moves in slow animation speed"));
    layout->addWidget(m_slow_animation_speed);

    layout->addStretch(1);
}


void ConfigurationDialog::setupAnimationSpeedDefaults()
{
    m_fast_animation_speed->setValue(5);
    m_normal_animation_speed->setValue(50);
    m_slow_animation_speed->setValue(500);
}


void ConfigurationDialog::applyAnimationSpeedSettings()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Fast animation time", m_fast_animation_speed->value());
    config->writeEntry("Normal animation time", m_normal_animation_speed->value());
    config->writeEntry("Slow animation time", m_slow_animation_speed->value());
}


void ConfigurationDialog::setupScalingPage()
{
    QFrame * page = addPage(i18n("Scaling"), i18n("Here you can choose the parameters of image scaling"),
                            BarIcon("viewmag", KIcon::SizeMedium));

    QVBoxLayout * layout = new QVBoxLayout(page, marginHint(), spacingHint());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_min_square_size = new KIntNumInput(page);
    m_min_square_size->setRange(2, 100);
    m_min_square_size->setValue(config->readNumEntry("Minimum field size", 10));
    m_min_square_size->setLabel(i18n("The minimum size of a field"));
    layout->addWidget(m_min_square_size);

    int const scaling_mode = config->readNumEntry("Scaling mode", 2);

    QButtonGroup * group = new QButtonGroup(3, Vertical, i18n("Scaling quality"), page);
    group->setRadioButtonExclusive(true);
    layout->addWidget(group);

    m_bad_scaling = new QRadioButton(i18n("Use bad, but fast scaling"), group);
    m_bad_scaling->setChecked(scaling_mode == 0);

    m_good_scaling = new QRadioButton(i18n("Use good, but slow scaling"), group);
    m_good_scaling->setChecked(scaling_mode == 1);

    m_timed_scaling = new QRadioButton(i18n("Switch from bad to good scaling"), group);
    m_timed_scaling->setChecked(scaling_mode == 2);

    m_scaling_switch_time = new KIntNumInput(page);
    m_scaling_switch_time->setRange(0, 1000);
    m_scaling_switch_time->setValue(config->readNumEntry("Scaling switch time", 500));
    m_scaling_switch_time->setSuffix(" " + i18n("miliseconds"));
    m_scaling_switch_time->setLabel(i18n("Time between switching from bad to good scaling"));
    layout->addWidget(m_scaling_switch_time);

    layout->addStretch(1);
}


void ConfigurationDialog::setupScalingDefaults()
{
    m_min_square_size->setValue(10);
    m_timed_scaling->setChecked(true);
    m_scaling_switch_time->setValue(500);
}


void ConfigurationDialog::applyScalingSettings()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Minimum field size", m_min_square_size->value());

    if (m_bad_scaling->isChecked())
    {
        config->writeEntry("Scaling mode", 0);
    }
    else if (m_good_scaling->isChecked())
    {
        config->writeEntry("Scaling mode", 1);
    }
    else
    {
        config->writeEntry("Scaling mode", 2);
    }

    config->writeEntry("Scaling switch time", m_scaling_switch_time->value());
}


void ConfigurationDialog::setupConfirmationPage()
{
    QFrame * page = addPage(i18n("Confirmations"), i18n("Change, what to you want to have to confirm"),
                            BarIcon("ok", KIcon::SizeMedium));

    QVBoxLayout * layout = new QVBoxLayout(page, marginHint(), spacingHint());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("Notification Messages");

    QString const bookmark_text = i18n("Confirmations for bookmarks");
    QButtonGroup * bookmark_group = new QButtonGroup(1, Vertical, bookmark_text, page);
    layout->addWidget(bookmark_group);

    m_overwrite_bookmarks = new QCheckBox(i18n("Ask when overwriting bookmarks?"), bookmark_group);
    m_overwrite_bookmarks->setChecked(config->readBoolEntry("Bookmark overwrite", true));


    QString const export_text = i18n("Confirmations for loading/saving");
    QButtonGroup * export_group = new QButtonGroup(6, Vertical, export_text, page);
    layout->addWidget(export_group);

    m_overwrite_game = new QCheckBox(i18n("Ask when overwriting a game in export?"), export_group);
    m_overwrite_game->setChecked(config->readBoolEntry("Overwrite game", true));

    m_overwrite_collection = new QCheckBox(i18n("Ask when overwriting a collection in export?"), export_group);
    m_overwrite_collection->setChecked(config->readBoolEntry("Overwrite collection", true));

    m_overwrite_solutions = new QCheckBox(i18n("Ask when overwriting solutions?"), export_group);
    m_overwrite_solutions->setChecked(config->readBoolEntry("Overwrite solution", true));

    m_overwrite_image = new QCheckBox(i18n("Ask when overwriting images?"), export_group);
    m_overwrite_image->setChecked(config->readBoolEntry("Overwrite image", true));

    m_overwrite_animation = new QCheckBox(i18n("Ask when overwriting animations?"), export_group);
    m_overwrite_animation->setChecked(config->readBoolEntry("Overwrite animation", true));

    m_warn_big_files = new QCheckBox(i18n("Warn when loading very big files?"), export_group);
    m_warn_big_files->setChecked(config->readBoolEntry("Warn because of big files", true));

    QString const solver_text = i18n("Confirmations for solver");
    QButtonGroup * solver_group = new QButtonGroup(1, Vertical, solver_text, page);
    layout->addWidget(solver_group);

    m_warn_many_gems = new QCheckBox(i18n("Warn when solving maps with many gems?"), solver_group);
    m_warn_many_gems->setChecked(config->readBoolEntry("Warn when solving map with many gems", true));

    layout->addStretch(1);

    config->setGroup("");
}


void ConfigurationDialog::setupConfirmationDefaults()
{
    m_overwrite_bookmarks->setChecked(true);
    m_overwrite_game->setChecked(true);
    m_overwrite_collection->setChecked(true);
    m_overwrite_solutions->setChecked(true);
    m_overwrite_image->setChecked(true);
    m_overwrite_animation->setChecked(true);
    m_warn_big_files->setChecked(true);
    m_warn_many_gems->setChecked(true);
}


void ConfigurationDialog::applyConfirmationSettings()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("Notification Messages");

    config->writeEntry("Bookmark overwrite", m_overwrite_bookmarks->isChecked());
    config->writeEntry("Overwrite game", m_overwrite_game->isChecked());
    config->writeEntry("Overwrite collection", m_overwrite_collection->isChecked());
    config->writeEntry("Overwrite solution", m_overwrite_solutions->isChecked());
    config->writeEntry("Overwrite image", m_overwrite_image->isChecked());
    config->writeEntry("Overwrite animation", m_overwrite_animation->isChecked());
    config->writeEntry("Warn because of big files", m_warn_big_files->isChecked());
    config->writeEntry("Warn when solving map with many gems", m_warn_many_gems->isChecked());

    config->setGroup("");
}
