/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_CONFIGURATION_DIALOG_INC_GUARD_H
#define EASYSOK_CONFIGURATION_DIALOG_INC_GUARD_H


#include <kdialogbase.h>


class KIntNumInput;
class QCheckBox;
class QRadioButton;




/**
 * The configuration dialog.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class ConfigurationDialog : public KDialogBase
{

    Q_OBJECT

public:

    /**
     * Constructs the dialog.
     *
     * @param parent The parent widget.
     * @param name The internal name.
     */

    ConfigurationDialog(QWidget * parent = 0, char const * name = 0);


    /**
     * Stores the window geometry.
     */

    virtual ~ConfigurationDialog();


signals:

    /**
     * Called when the configuration has changed.
     */

    void configurationChanged();


protected slots:

    /**
     * Apply the options and quit.
     */

    virtual void slotOk();


    /**
     * Apply the options.
     */

    virtual void slotApply();


    /**
     * Set defaults.
     */

    virtual void slotDefault();


private:

    /**
     * Setups the core page.
     */

    void setupCorePage();


    /**
     * Setups the core defaults.
     */

    void setupCoreDefaults();


    /**
     * Applies the core settings.
     */

    void applyCoreSettings();


    /**
     * Sets up the mouse page.
     */

    void setupMousePage();


    /**
     * Sets up the mouse defaults.
     */

    void setupMouseDefaults();


    /**
     * Applies the mouse settings.
     */

    void applyMouseSettings();


    /**
     * Sets up the solver page.
     */

    void setupSolverPage();


    /**
     * Sets up the solver defaults.
     */

    void setupSolverDefaults();


    /**
     * Applies the solver settings.
     */

    void applySolverSettings();


    /**
     * Setup the bookmarks page.
     */

    void setupBookmarksPage();


    /**
     * Setup bookmarks defaults.
     */

    void setupBookmarksDefaults();


    /**
     * Applies the bookmarks settings.
     */

    void applyBookmarksSettings();


    /**
     * Setup the animation speed page.
     */

    void setupAnimationSpeedPage();


    /**
     * Setup animation speed defaults.
     */

    void setupAnimationSpeedDefaults();


    /**
     * Applies the animation speed settings.
     */

    void applyAnimationSpeedSettings();


    /**
     * Setup the scaling page.
     */

    void setupScalingPage();


    /**
     * Setup scaling defaults.
     */

    void setupScalingDefaults();


    /**
     * Applies the scaling settings.
     */

    void applyScalingSettings();


    /**
     * Setup the confirmation page.
     */

    void setupConfirmationPage();


    /**
     * Setup confirmation defaults.
     */

    void setupConfirmationDefaults();


    /**
     * Applies the confirmation settings.
     */

    void applyConfirmationSettings();


    /**
     * Used to number the pages.
     */

    enum Page
    {
        CORE,
        MOUSE,
        SOLVER,
        BOOKMARKS,
        ANIMATION_SPEEDS,
        SCALING,
        CONFIRMATIONS
    };


    /**
     * If set, we can jump to the next level after solved.
     */

    QCheckBox * m_next_after_solved;


    /**
     * If set, illegal moves are treated as virtual keeper moves.
     */

    QCheckBox * m_illegal_as_virtual;


    /**
     * If set, we can wrap the virtual keeper.
     */

    QCheckBox * m_wrap_virtual_keeper;


    /**
     * If set, we can goto any level.
     */

    QCheckBox * m_goto_any_level;


    /**
     * If set, the advanced options button are not shown.
     */

    QCheckBox * m_always_show_advanced_options;


    /**
     * The number of recent collections.
     */

    KIntNumInput * m_recent_collections;


    /**
     * If set, we automatically hide the cursor after a time unused.
     */

    QCheckBox * m_hide_cursor;


    /**
     * The time in seconds after which to hide the cursor.
     */

    KIntNumInput * m_hide_cursor_delay;


    /**
     * The mouse repeat delay.
     */

    KIntNumInput * m_mouse_repeat_delay;


    /**
     * The mouse repeat rate.
     */

    KIntNumInput * m_mouse_repeat_rate;


    /**
     * Auto save enabled.
     */

    QCheckBox * m_auto_save_enabled;

    /**
     * The auto save time.
     */

    KIntNumInput * m_auto_save_time;


    /**
     * Show the difficulty of the level.
     */

    QCheckBox * m_show_difficulty;


    /**
     * Show linear pushes.
     */

    QCheckBox * m_show_linear_pushes;


    /**
     * Show gem changes.
     */

    QCheckBox * m_show_gem_changes;


    /**
     * Show the gems left.
     */

    QCheckBox * m_show_gems_left;


    /**
     * Show the author.
     */

    QCheckBox * m_show_author;


    /**
     * Shows the current user.
     */

    QCheckBox * m_show_user;


    /**
     * The number of solver steps per call widget.
     */

    KIntNumInput * m_solver_steps_per_call;


    /**
     * The size of the solver cache widget.
     */

    KIntNumInput * m_solver_cache_size;


    /**
     * The bookmarks widget.
     */

    KIntNumInput * m_bookmarks;


    /**
     * If set, we don't display dates and times in bookmarks.
     */

    QRadioButton * m_no_bookmark_date;


    /**
     * If set, we display the dates in bookmarks.
     */

    QRadioButton * m_bookmark_date;


    /**
     * If set, we display dates and times in bookmarks.
     */

    QRadioButton * m_bookmark_time;


    /**
     * If set, we display no collection and level name.
     */

    QRadioButton * m_no_bookmark_collection;


    /**
     * If set, we display the collection name.
     */

    QRadioButton * m_bookmark_collection;


    /**
     * If set, we display the collection name and level number.
     */

    QRadioButton * m_bookmark_level;


    /**
     * If set, we display no bookmark annotations.
     */

    QRadioButton * m_no_bookmark_annotation;


    /**
     * If set, we display bookmark annotations.
     */

    QRadioButton * m_bookmark_annotation;


    /**
     * The speed for fast animation.
     */

    KIntNumInput * m_fast_animation_speed;


    /**
     * The speed for normal animation.
     */

    KIntNumInput * m_normal_animation_speed;


    /**
     * The speed for slow animation.
     */

    KIntNumInput * m_slow_animation_speed;


    /**
     * The minimum size of a field.
     */

    KIntNumInput * m_min_square_size;


    /**
     * If set, we use bad pixmap scaling.
     */

    QRadioButton * m_bad_scaling;


    /**
     * If set, we use good pixmap scaling.
     */

    QRadioButton * m_good_scaling;


    /**
     * If set, we switch from bad to good scaling after a some time.
     */

    QRadioButton * m_timed_scaling;


    /**
     * The time, after which to switch from good to bad scaling.
     */

    KIntNumInput * m_scaling_switch_time;


    /**
     * Should we ask when we overwrite a bookmark in an export operation.
     */

    QCheckBox * m_overwrite_bookmarks;


    /**
     * Should we ask when we overwrite a game in an export operation.
     */

    QCheckBox * m_overwrite_game;


    /**
     * Should we ask when we overwrite a collection in an export operation.
     */

    QCheckBox * m_overwrite_collection;


    /**
     *  Should we ask when we overwrite solutions in an export operation.
     */

    QCheckBox * m_overwrite_solutions;


    /**
     * Should we ask when we overwrite images in an export operation.
     */

    QCheckBox * m_overwrite_image;


    /**
     * Should we ask when we overwrite animations in an export operation.
     */

    QCheckBox * m_overwrite_animation;


    /**
     * Should be ask when we load very big files?
     */

    QCheckBox * m_warn_big_files;


    /**
     * Should be ask when solving really big levels with many gems.
     */

    QCheckBox * m_warn_many_gems;
};


#endif
