/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "create_solutions_dialog.h"

#include <cassert>

#include <qtimer.h>

#include <klocale.h>

#include "collection.h"
#include "collection_holder.h"
#include "compressed_map.h"
#include "level.h"
#include "map.h"
#include "solution_holder.h"




CreateSolutionsDialog::CreateSolutionsDialog(std::vector<int> const & collections,
                                             std::vector<int> const & levels, QString const regexp,
                                             bool only_best, QWidget * parent, char const * name) :
    QMessageBox(i18n("Creating solutions"), i18n("Creating solutions (%1 percent finished) ...").arg(0),
                Information, Cancel | Default, NoButton, NoButton, parent, name, true),
    m_collections(collections),
    m_levels(levels),
    m_level(0),
    m_use_regexp(!regexp.isEmpty()),
    m_regexp(regexp),
    m_only_best(only_best)
{
    assert(m_collections.size() == m_levels.size());

    m_timer = new QTimer(this);
    connect(m_timer, SIGNAL(timeout()), this, SLOT(createSolutions()));

    m_timer->start(100, true);
}


QString CreateSolutionsDialog::solutions() const
{
    return m_solutions;
}


bool CreateSolutionsDialog::finished() const
{
    return m_level >= static_cast<int> (m_levels.size());
}


void CreateSolutionsDialog::createSolutions()
{
    if (finished())
    {
        accept();

        return;
    }

    setText(i18n("Creating solutions (%1 percent finished) ...").
            arg(100 * m_level / static_cast<int> (m_levels.size())));
    adjustSize();

    Collection * collection = CollectionHolder::collection(m_collections[m_level]);
    Level const & level = collection->level(m_levels[m_level]);
    CompressedMap const & compressed_map = level.compressedMap();
    Map const & map = level.map();

    int number_of_solutions = 1;

    if (!m_only_best)
    {
        number_of_solutions = SolutionHolder::numberOfSolutions(compressed_map);
    }

    for (int i = 0; i < number_of_solutions; ++i)
    {
        QString info = SolutionHolder::infoOfSolution(compressed_map, i);

        if (m_use_regexp)
        {
            if (m_regexp.search(info, 0) == -1)
            {
                continue;
            }
        }

        Movements moves = map.expandMoves(SolutionHolder::movements(compressed_map, i), false);

        QString result = map.toText();
        result += Collection::levelNameAndCollection(level, m_levels[m_level], *collection);
        result += "Author: " + level.authorEmailLine() + '\n';
        result += "Info: " + info + '\n';
        result += moves.toText();

        m_solutions += result;
    }

    ++m_level;

    m_timer->start(1, true);
}
