/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "export_solutions_dialog.h"

#include <algorithm>
#include <cassert>

#include <qbuttongroup.h>
#include <qhbox.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qvbox.h>

#include <kapp.h>
#include <kconfig.h>
#include <klineedit.h>
#include <klocale.h>
#include <kmessagebox.h>

#include <kparts/componentfactory.h>
#include <kregexpeditorinterface.h>




ExportSolutionsDialog::ExportSolutionsDialog(bool export_current_level,
                                             QWidget * parent, char const * name) :
    KDialogBase(parent, name, true, i18n("Export solutions"), Ok | Cancel | Help)
{
    QVBox * page = makeVBoxMainWidget();

    QButtonGroup * group1 = new QButtonGroup(3, Vertical, i18n("Solutions for which level"), page);
    group1->setRadioButtonExclusive(true);

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    int options = config->readNumEntry("Export solutions", 2);
    options = std::min(std::max(options, 0), 2);

    if (!export_current_level && (options == 2))
    {
        options = 1;
    }

    m_export_all = new QRadioButton(i18n("Solutions of all collections"), group1);
    m_export_all->setChecked(options == 0);

    m_export_collection = new QRadioButton(i18n("Solutions of current collection"), group1);
    m_export_collection->setChecked(options == 1);

    m_export_level = new QRadioButton(i18n("Solution of current level"), group1);
    m_export_level->setChecked(options == 2);
    m_export_level->setEnabled(export_current_level);


    QButtonGroup * group2 = new QButtonGroup(4, Vertical, i18n("Which solutions of a level to export"), page);
    group2->setRadioButtonExclusive(true);

    int which_solutions = config->readNumEntry("Which solutions to export", 0);

    m_export_best = new QRadioButton(i18n("Export only the best solution of a level"), group2);
    m_export_best->setChecked(which_solutions == 0);

    m_export_all_solutions = new QRadioButton(i18n("Export all solutions of a level"), group2);
    m_export_all_solutions->setChecked(which_solutions == 1);

    QString const by_name_text = i18n("Export solutions which match regexp:");
    m_export_solutions_by_name = new QRadioButton(by_name_text, group2);
    m_export_solutions_by_name->setChecked(which_solutions == 2);
    connect(m_export_solutions_by_name, SIGNAL(toggled(bool)), this, SLOT(regexpButtonToggled(bool)));

    QHBox * group2a = new QHBox(group2);
    group2a->setSpacing(10);

    QString const solutions_regexp = config->readEntry("Export solutions regexp", "");
    m_solutions_regexp = new KLineEdit(solutions_regexp, group2a);
    m_solutions_regexp->setEnabled(which_solutions == 2);

    QPushButton * button = new QPushButton(i18n("Start RegExp Editor"), group2a);
    connect(button, SIGNAL(clicked()), this, SLOT(showRegExpEditor()));

    setHelp("export-solutions-dialog", QString::null);
}


ExportSolutionsDialog::~ExportSolutionsDialog()
{
    int options1 = 0;

    if (exportCurrentLevel())
    {
        options1 = 2;
    }
    else if (exportCollection())
    {
        options1 = 1;
    }

    int options2 = 0;

    if (exportAll())
    {
        options2 = 1;
    }
    else if (exportRegexpMatches())
    {
        options2 = 2;
    }

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Export solutions", options1);
    config->writeEntry("Which solutions to export", options2);
    config->writeEntry("Export solutions regexp", m_solutions_regexp->text());
}


bool ExportSolutionsDialog::exportCurrentLevel() const
{
    return m_export_level->isChecked();
}


bool ExportSolutionsDialog::exportCollection() const
{
    return m_export_collection->isChecked();
}


bool ExportSolutionsDialog::exportOnlyBest() const
{
    return m_export_best->isChecked();
}


bool ExportSolutionsDialog::exportAll() const
{
    return m_export_all_solutions->isChecked();
}


bool ExportSolutionsDialog::exportRegexpMatches() const
{
    return m_export_solutions_by_name->isChecked();
}


void ExportSolutionsDialog::showRegExpEditor()
{
    QDialog * dialog;
    dialog = KParts::ComponentFactory::createInstanceFromQuery<QDialog>("KRegExpEditor/KRegExpEditor");

    if (dialog)
    {
        if (KMessageBox::questionYesNo(this, i18n("The editor may crash EasySok!\n"
                                                  "Do you want to continue?"),
                                       i18n("Procceed?")) == KMessageBox::No)
        {
            return;
        }

        KRegExpEditorInterface * editor = dynamic_cast<KRegExpEditorInterface *>(editor);
        assert(editor != 0);

        QString regexp = m_solutions_regexp->text();
        editor->setRegExp(regexp);

        if (dialog->exec())
        {
            m_solutions_regexp->setText(editor->regExp());
        }
    }
    else
    {
        KMessageBox::error(this, i18n("You must have the kdeutils package installed "
                                      "to get the regular expression editor!"));
    }
}


QString ExportSolutionsDialog::solutionRegexp() const
{
    return m_solutions_regexp->text();
}


void ExportSolutionsDialog::regexpButtonToggled(bool state)
{
    m_solutions_regexp->setEnabled(state);
}
