/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "game.h"

#include <cassert>

#include <qtimer.h>

#include <kapp.h>
#include <kconfig.h>




int Game::s_fast_time = -1;
int Game::s_normal_time = -1;
int Game::s_slow_time = -1;


Game::Game(Map * map, Movements const & moves) :
    m_no_update(false),
    m_show_arrows(false),
    m_honor_deadlocks(false),
    m_keeper_direction(AtomicMove::UP),
    m_timer(new QTimer(this)),
    m_animation(0),
    m_retro_mode(false),
    m_backup_map(*map),
    m_prevent_update(false),
    m_in_play(false)
{
    connect(m_timer, SIGNAL(timeout()), this, SLOT(processMoveQueue()));

    if ( (s_fast_time == -1) || (s_normal_time == -1) || (s_slow_time == -1) )
    {
        configChanged();
    }

    setMapAndMoves(map, moves);
}


void Game::setMapAndMoves(Map * map, Movements const & moves)
{
    m_timer->stop();

    m_retro_mode = false;
    m_is_solved = false;
    m_number_of_pushes = 0;
    m_number_of_linear_pushes = 0;
    m_number_of_gem_changes = 0;
    m_number_of_moves = 0;
    m_all_number_of_linear_pushes.resize(0);
    m_all_number_of_gem_changes.resize(0);
    m_all_last_pushed_gems.resize(0);
    m_all_last_push_directions.resize(0);
    m_last_pushed_gem = QPoint(-1, -1);
    m_last_push_direction = QPoint(0, 0);
    m_width = map->width();
    m_height = map->height();
    m_size = m_width * m_height;
    m_virtual_keeper = QPoint(-1, -1);
    m_keeper_direction = AtomicMove::UP;
    m_no_update = false;
    m_map = map;
    m_backup_map = *map;
    m_moves = Movements();

    m_xy_offsets[0] = -1;
    m_xy_offsets[1] = 1;
    m_xy_offsets[2] = -m_width;
    m_xy_offsets[3] = m_width;

    if (m_honor_deadlocks)
    {
        m_map->calcDeadlocks();
        m_map->crossDeadlocks();
    }

    m_map->calcReachable();

    if (m_show_arrows)
    {
        calcArrows();
    }

    m_move_queue.clear();

    setMoves(moves);
}


bool Game::retroMode() const
{
    return m_retro_mode;
}


bool Game::setRetroMode(bool retro_mode)
{
    emptyMoveQueue();

    if (retro_mode == m_retro_mode)
    {
        return true;
    }

    m_retro_mode = retro_mode;

    if (m_retro_mode)
    {
        assert(!m_map->containsGoal(m_map->keeper()));

        m_backup_moves = m_moves;
        m_backup_moves.truncateToCurrent();
        m_moves = Movements();
        makeEndMap();
        m_number_of_moves = 0;
        m_number_of_pushes = 0;
        m_is_solved = false;

        if (m_show_arrows)
        {
            calcArrows();
        }

        if (m_honor_deadlocks)
        {
            m_map->calcDeadlocks();
            m_map->crossDeadlocks();
        }

        forceUpdate();
    }
    else
    {
        *m_map = m_backup_map;
        m_is_solved = m_map->isSolved();
        m_moves = Movements();

        setMapAndMoves(m_map, m_backup_moves);
    }

    return true;
}


bool Game::tryMove(AtomicMove const & move)
{
    emptyMoveQueue();

    if (m_is_solved)
    {
        forceUpdate();

        return false;
    }

    QPoint const diff = move.diff();
    QPoint const from = m_map->keeper();
    QPoint const to = from + diff;

    if (m_map->containsGem(to))
    {
        if (m_retro_mode)
        {
            return false;
        }

        if (!isDeadlockField(to + diff))
        {
            return tryMove(Move(from, to, true));
        }
        else
        {
            forceUpdate();

            return false;
        }
    }
    else
    {
        return tryMove(Move(from, to, false));
    }
}


bool Game::tryMove(Move const & move)
{
    emptyMoveQueue();

    if (m_is_solved)
    {
        forceUpdate();

        return false;
    }

    m_handmade_moves = true;

    bool const valid_move = m_map->isValidMove(move, m_retro_mode);

    if (valid_move)
    {
        doMove(move);
    }
    else
    {
        forceUpdate();
    }

    return valid_move;
}


bool Game::tryPullMove(AtomicMove const & move)
{
    emptyMoveQueue();

    if (m_is_solved)
    {
        forceUpdate();

        return false;
    }

    QPoint const diff = move.diff();
    QPoint const from = m_map->keeper();
    QPoint const gem_pos = from - diff;
    QPoint const to = from + diff;

    if (m_retro_mode)
    {
        if (m_map->containsGem(gem_pos) && m_map->canDropGem(from) && m_map->canDropKeeper(to))
        {
            doMove(Move(from, to, true));

            return true;
        }
        else if (m_map->canDropKeeper(to))
        {
            doMove(Move(from, to, false));

            return true;
        }

        return false;
    }
    else
    {
        if (m_map->containsGem(gem_pos) && !isDeadlockField(from))
        {
            m_map->calcReachable();

            QPoint const keeper_push_pos = gem_pos - diff;

            if (m_map->isReachable(keeper_push_pos))
            {
                m_no_update = true;

                doMove(Move(from, keeper_push_pos, false));
                doMove(Move(keeper_push_pos,gem_pos, true));
                tryMoveKeeper(gem_pos, to);
                emptyMoveQueue();
                m_no_update = false;

                forceUpdate();

                return true;
            }
        }

        return tryMove(move);
    }
}


bool Game::tryFarMove(AtomicMove const & move)
{
    emptyMoveQueue();

    if (m_is_solved)
    {
        forceUpdate();

        return false;
    }

    QPoint const keeper = m_map->keeper();
    QPoint const diff = move.diff();
    QPoint to = keeper + diff;

    if (m_map->containsGem(to))
    {
        bool test_for_goal = ((m_map->numberOfEmptyGoals() == 1) && !m_map->containsGoal(to));

        to += diff;

        while (!isDeadlockField(to) && m_map->canDropGem(to))
        {
            if (test_for_goal && m_map->containsGoal(to))
            {
                to += diff;

                break;
            }

            to += diff;
        }

        to -= 2 * diff;

        if (to != keeper)
        {
            return tryMove(Move(keeper, to, true));
        }
        else
        {
            forceUpdate();

            return false;
        }
    }
    else
    {
        while (m_map->canDropKeeper(to))
        {
            to += diff;
        }

        to -= diff;

        if (to != keeper)
        {
            return tryMove(Move(keeper, to, false));
        }
        else
        {
            forceUpdate();

            return false;
        }
    }
}


bool Game::moveVirtualKeeper(AtomicMove const & move)
{
    if (m_is_solved)
    {
        return false;
    }

    if (m_virtual_keeper == QPoint(-1, -1))
    {
        m_virtual_keeper = m_map->keeper();
    }

    m_virtual_keeper += move.diff();

    if (m_virtual_keeper.x() < 0)
    {
        if (m_wrap_virtual_keeper)
        {
            m_virtual_keeper.setX(m_width - 1);
        }
        else
        {
            m_virtual_keeper.setX(0);
        }
    }
    else if (m_virtual_keeper.x() >= m_width)
    {
        if (m_wrap_virtual_keeper)
        {
            m_virtual_keeper.setX(0);
        }
        else
        {
            m_virtual_keeper.setX(m_width - 1);
        }
    }
    else if (m_virtual_keeper.y() < 0)
    {
        if (m_wrap_virtual_keeper)
        {
            m_virtual_keeper.setY(m_height - 1);
        }
        else
        {
            m_virtual_keeper.setY(0);
        }
    }
    else if (m_virtual_keeper.y() >= m_height)
    {
        if (m_wrap_virtual_keeper)
        {
            m_virtual_keeper.setY(0);
        }
        else
        {
            m_virtual_keeper.setY(m_height - 1);
        }
    }

    m_map->calcReachable();

    if ( (m_map->isReachable(m_virtual_keeper)) && (m_map->keeper() != m_virtual_keeper) )
    {
        tryMove(Move(m_map->keeper(), m_virtual_keeper, false));

        return true;
    }
    else
    {
        if (!m_prevent_update)
        {
            emit virtualKeeperChanged(m_virtual_keeper);
            emit mapChanged();
        }

        return false;
    }
}


bool Game::canUndo() const
{
    return m_moves.hasPrevMove();
}


bool Game::canRedo() const
{
    return m_moves.hasNextMove();
}


int Game::numberOfPushes() const
{
    return m_number_of_pushes;
}


int Game::numberOfLinearPushes() const
{
    return m_number_of_linear_pushes;
}


int Game::numberOfGemChanges() const
{
    return m_number_of_gem_changes;
}


int Game::numberOfMoves() const
{
    return m_number_of_moves;
}


int Game::numberOfEmptyGoals() const
{
    return m_map->numberOfEmptyGoals();
}


Movements const & Game::moves() const
{
    return m_moves;
}


void Game::setMoves(Movements const & moves)
{
    emptyMoveQueue();

    if (m_is_solved)
    {
        emit wasUnsolved();
    }

    m_no_update = true;

    while (canUndo())
    {
        undo();
    }

    emptyMoveQueue();

    m_moves = moves;

    int const move_index = m_moves.movePointer();
    m_moves.setToFirstPosition();

    for (int i = 0; i < move_index; ++i)
    {
        redo();
    }

    m_no_update = false;
    emptyMoveQueue();

    forceUpdate();
}


bool Game::isSolved() const
{
    return m_map->isSolved();
}


bool Game::showArrows() const
{
    return m_show_arrows;
}


bool Game::honorDeadlocks() const
{
    return m_honor_deadlocks;
}


int Game::animation() const
{
    return m_animation;
}


AtomicMove::AtomicMoveType Game::keeperDirection() const
{
    return m_keeper_direction;
}


void Game::configChanged()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    s_fast_time = config->readNumEntry("Fast animation time", 5);
    s_normal_time = config->readNumEntry("Normal animation time", 50);
    s_slow_time = config->readNumEntry("Slow animation time", 500);
    m_wrap_virtual_keeper = config->readBoolEntry("Wrap virtual keeper", true);
}


void Game::setShowArrows(bool show)
{
    m_show_arrows = show;

    if (m_show_arrows)
    {
        calcArrows();
    }
    else
    {
        if (!m_prevent_update)
        {
            emit arrowsChanged(std::vector<Move> ());
        }
    }
}


void Game::setHonorDeadlock(bool honor_deadlocks)
{
    if (m_honor_deadlocks == honor_deadlocks)
    {
        return;
    }

    m_honor_deadlocks = honor_deadlocks;

    if (m_honor_deadlocks == false)
    {
        m_map->uncrossAll();
    }
    else
    {
        m_map->calcDeadlocks();
        m_map->crossDeadlocks();
    }

    if (m_show_arrows)
    {
        calcArrows();
    }
}


void Game::setAnimation(int animation)
{
    m_animation = animation;
}


void Game::emptyMoveQueue()
{
    bool old_update = m_no_update;
    m_no_update = true;
    m_in_play = false;

    while (!m_move_queue.empty())
    {
        processMove();
    }

    if (m_show_arrows && !old_update)
    {
        calcArrows();
    }

    m_no_update = old_update;
}


bool Game::tryMoveKeeper(QPoint const & from, QPoint const & to)
{
    emptyMoveQueue();

    if (m_is_solved)
    {
        forceUpdate();

        return false;
    }

    if (from == to)
    {
        forceUpdate();

        return false;
    }

    if (m_map->keeper() != from)
    {
        forceUpdate();

        return false;
    }

    m_map->calcReachable();

    if (!m_map->isReachable(to))
    {
        forceUpdate();

        return false;
    }

    m_handmade_moves = true;
    doMove(Move(from, to, false));

    return true;
}


bool Game::tryMoveGem(QPoint const & from, QPoint const & to)
{
    emptyMoveQueue();

    if (m_is_solved)
    {
        forceUpdate();

        return false;
    }

    if (from == to)
    {
        forceUpdate();

        return false;
    }

    if (!m_map->containsGem(from))
    {
        forceUpdate();

        return false;
    }

    if (!m_map->canDropGem(to) && !isDeadlockField(to))
    {
        forceUpdate();

        return false;
    }

    Movements moves = m_map->getShortestPathForGem(from, to, m_retro_mode);

    if (moves.isEmpty())
    {
        forceUpdate();

        return false;
    }

    m_handmade_moves = true;

    if (!m_retro_mode)
    {
        moves.setToFirstPosition();

        Map tmp_map = *m_map;
        Movements safe_moves;

        while (moves.hasNextMove())
        {
            Move move = moves.nextMove();
            tmp_map.doMove(move, m_retro_mode);
            safe_moves.addMove(move);

            if (tmp_map.isSolved())
            {
                break;
            }
        }

        doMoves(safe_moves);
    }
    else
    {
        doMoves(moves);
    }

    return false;
}


void Game::jumpToStart()
{
    m_no_update = true;

    emptyMoveQueue();

    while (canUndo())
    {
        undo();
    }

    emptyMoveQueue();

    m_no_update = false;

    forceUpdate();
}


void Game::jumpToEnd()
{
    m_no_update = true;
    m_handmade_moves = false;

    emptyMoveQueue();

    while (canRedo())
    {
        redo();
    }

    emptyMoveQueue();

    m_no_update = false;

    forceUpdate();
}


bool Game::tryUndo()
{
    if (canUndo())
    {
        undo();

        return true;
    }
    else
    {
        return false;
    }
}


bool Game::tryRedo()
{
    if (canRedo())
    {
        redo();

        return true;
    }
    else
    {
        return false;
    }
}


void Game::playMoves(bool from_start)
{
    m_handmade_moves = false;

    if (from_start)
    {
        jumpToStart();
    }

    Movements play_moves;
    int number_of_moves = m_moves.numberOfMoves();

    for (int i = m_moves.movePointer(); i < number_of_moves; ++i)
    {
        play_moves.addMove(m_moves.move(i));
    }

    doMoves(play_moves);
    m_in_play = true;
}


void Game::processMoveQueue()
{
    int animation = m_animation;

    if (m_in_play && (animation == 0))
    {
        animation = 1;
    }

    while (true)
    {
        if (m_move_queue.empty())
        {
            if (!m_no_update && (m_animation == 0))
            {
                forceUpdate();
            }

            m_in_play = false;

            return;
        }

        processMove();

        if (m_is_solved)
        {
            return;
        }

        if (animation > 0)
        {
            switch (animation)
            {
            case 1:
                m_timer->start(s_fast_time, true);
                break;

            case 2:
                m_timer->start(s_normal_time, true);
                break;

            case 3:
                m_timer->start(s_slow_time, true);
                break;

            default:
                assert(false);
                break;
            }

            return;
        }
    }
}


void Game::removeVirtualKeeper()
{
    if (m_virtual_keeper != QPoint(-1, -1))
    {
        m_virtual_keeper = QPoint(-1, -1);

        if (!m_prevent_update)
        {
            emit virtualKeeperChanged(m_virtual_keeper);
            emit mapChanged();
        }
    }
}


void Game::undo()
{
    assert(canUndo());

    m_handmade_moves = false;
    doUndoMove(m_moves.prevMove());
}


void Game::redo()
{
    assert(canRedo());

    m_handmade_moves = false;
    doMove(m_moves.peekNextMove());
}


void Game::doMove(Move const & move)
{
    emptyMoveQueue();

    m_moves.addMove(move);

    doAtomicMoves(simplifyMove(move, m_map));
}


void Game::doAtomicMoves(Movements const & moves)
{
    removeVirtualKeeper();

    int const number_of_moves = moves.numberOfMoves();
    m_undo_moves = false;

    for (int i = 0; i < number_of_moves; ++i)
    {
        addToMoveQueue(moves.move(i));
    }
}


void Game::doMoves(Movements const & moves)
{
    assert(!m_map->expandMoves(moves, m_retro_mode).isEmpty());

    emptyMoveQueue();

    int const number_of_moves = moves.numberOfMoves();

    for (int i = 0; i < number_of_moves; ++i)
    {
        m_moves.addMove(moves.move(i));
    }

    doAtomicMoves(m_map->expandMoves(moves, m_retro_mode));
}


void Game::doUndoMove(Move const & move)
{
    removeVirtualKeeper();
    emptyMoveQueue();

    m_undo_moves = true;

    if (move.isAtomicMove())
    {
        addToMoveQueue(move);
    }
    else
    {
        Movements simple_moves = simplifyUndoMove(move, m_map);
        int const number_of_moves = simple_moves.numberOfMoves();

        for (int i = 0; i < number_of_moves; ++i)
        {
            addToMoveQueue(simple_moves.move(i));
        }
    }
}


void Game::addToMoveQueue(Move const & move)
{
    assert(move.isAtomicMove());

    if (m_move_queue.empty())
    {
        m_timer->stop();
        m_timer->start(0, true);
    }

    m_move_queue.push_back(move);
}


void Game::processMove()
{
    assert(!m_move_queue.empty());

    Move move = m_move_queue.front();
    m_move_queue.pop_front();

    bool is_solved = m_is_solved;
    bool const was_last_move = m_move_queue.empty();
    bool const do_update = !m_no_update && ((m_animation != 0) || was_last_move || m_in_play);

    AtomicMove::AtomicMoveType new_direction = move.atomicMove().type();

    if (m_undo_moves)
    {
        m_map->doUndoMove(move, m_retro_mode);

        --m_number_of_moves;

        if (move.stonePushed())
        {
            --m_number_of_pushes;

            m_number_of_linear_pushes = m_all_number_of_linear_pushes.back();
            m_all_number_of_linear_pushes.pop_back();

            m_number_of_gem_changes = m_all_number_of_gem_changes.back();
            m_all_number_of_gem_changes.pop_back();

            m_last_pushed_gem = m_all_last_pushed_gems.back();
            m_all_last_pushed_gems.pop_back();

            m_last_push_direction = m_all_last_push_directions.back();
            m_all_last_push_directions.pop_back();

            if (is_solved)
            {
                emit wasUnsolved();
                is_solved = false;
            }
        }
        else
        {
            m_last_push_direction = QPoint(0, 0);
        }
    }
    else
    {
        m_map->doMove(move, m_retro_mode);

        ++m_number_of_moves;

        if (move.stonePushed())
        {
            ++m_number_of_pushes;

            m_all_number_of_linear_pushes.push_back(m_number_of_linear_pushes);
            m_all_number_of_gem_changes.push_back(m_number_of_gem_changes);
            m_all_last_pushed_gems.push_back(m_last_pushed_gem);
            m_all_last_push_directions.push_back(m_last_push_direction);

            if (move.to() != m_last_pushed_gem)
            {
                ++m_number_of_gem_changes;
            }

            QPoint const push_direction = AtomicMove(new_direction).diff();

            if (push_direction != m_last_push_direction)
            {
                ++m_number_of_linear_pushes;
            }

            m_last_push_direction = push_direction;
            m_last_pushed_gem = move.to() + push_direction;

            is_solved = m_map->isSolved();

            if (is_solved && !m_is_solved)
            {
                emit wasSolved();
            }
            else if (!is_solved && m_is_solved)
            {
                emit wasUnsolved();
            }
        }
        else
        {
            m_last_push_direction = QPoint(0, 0);
        }
    }

    if (do_update)
    {
        if (m_keeper_direction != new_direction)
        {
            m_keeper_direction = new_direction;

            if (!m_prevent_update)
            {
                emit keeperDirectionChanged(m_keeper_direction);
            }
        }

        if (m_show_arrows && move.stonePushed())
        {
            calcArrows();
        }

        if (!m_prevent_update)
        {
            emit mapChanged();
            emit movesChanged(was_last_move);
        }
    }
    else
    {
        m_keeper_direction = new_direction;
    }

    if (!m_retro_mode)
    {
        m_is_solved = is_solved;

        if (is_solved && m_handmade_moves)
        {
            emit wasSolvedByHand();
        }
    }
}


void Game::forceUpdate()
{
    if (m_show_arrows)
    {
        calcArrows();
    }

    if (!m_prevent_update)
    {
        emit keeperDirectionChanged(m_keeper_direction);
        emit mapChanged();
        emit movesChanged(true);
    }
}


void Game::preventUpdate(bool prevent_update)
{
    m_prevent_update = prevent_update;
}


void Game::calcArrows()
{
    std::vector<Move> arrows;

    m_map->calcReachable();

    int const from = m_width + 1;
    int const to = m_size - m_width - 1;

    if (m_retro_mode)
    {
        for (int i = from; i < to; ++i)
        {
            if (isDeadlockField(i) || !m_map->containsGem(i))
            {
                continue;
            }

            for (int j = 0; j < 4; ++j)
            {
                int const gem_nb_index = i - m_xy_offsets[j];
                int const keeper_nb_index = gem_nb_index - m_xy_offsets[j];

                if (m_map->canDropGem(gem_nb_index) &&
                    m_map->canDropKeeper(keeper_nb_index) &&
                    m_map->isReachable(keeper_nb_index))
                {
                    arrows.push_back(Move(m_map->getPoint(i), m_map->getPoint(gem_nb_index), false));
                }
            }
        }
    }
    else
    {
        for (int i = from; i < to; ++i)
        {
            if (isDeadlockField(i) || !m_map->containsGem(i))
            {
                continue;
            }

            for (int j = 0; j < 4; ++j)
            {
                int const gem_nb_index = i + m_xy_offsets[j];
                int const keeper_nb_index = i - m_xy_offsets[j];
                int const can_drop_gem = (m_map->canDropGem(gem_nb_index) && !isDeadlockField(gem_nb_index));

                if (can_drop_gem && m_map->canDropKeeper(keeper_nb_index) &&
                    m_map->isReachable(keeper_nb_index))
                {
                    arrows.push_back(Move(m_map->getPoint(i), m_map->getPoint(gem_nb_index), false));
                }
            }
        }
    }

    if (!m_prevent_update)
    {
        emit arrowsChanged(arrows);
    }
}


Movements Game::simplifyMove(Move const & move, Map const * map) const
{
    return map->expandMove(move, m_retro_mode);
}


Movements Game::simplifyUndoMove(Move const & move, Map const * map) const
{
    return map->expandUndoMove(move);
}


bool Game::isDeadlockField(QPoint const & position) const
{
    return m_honor_deadlocks && m_map->isDeadlock(position);
}


bool Game::isDeadlockField(int index) const
{
    return m_honor_deadlocks && m_map->isDeadlock(index);
}


void Game::makeEndMap()
{
    std::vector<int> goal_indices;
    std::vector<int> gem_indices;

    for (int i = 0; i < m_size; ++i)
    {
        if (m_map->containsGem(i) && !m_map->containsGoal(i))
        {
            gem_indices.push_back(i);
        }
        else if (m_map->containsGoal(i) && !m_map->containsGem(i))
        {
            goal_indices.push_back(i);
        }
    }

    assert(goal_indices.size() == gem_indices.size());

    int const gems_to_move = static_cast<int> (gem_indices.size());

    for (int i = 0; i< gems_to_move; ++i)
    {
        assert(!m_map->containsKeeper(goal_indices[i]));

        m_map->moveGem(gem_indices[i], goal_indices[i]);
    }
}
