/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_GAME_INC_GUARD_H
#define EASYSOK_GAME_INC_GUARD_H


#include <deque>
#include <vector>

#include <qobject.h>
#include <qpoint.h>

#include "map.h"
#include "map_widget.h"
#include "move.h"
#include "movements.h"

class AtomicMove;
class QTimer;




/**
 * This class represents a game for a map.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class Game : public QObject
{
    Q_OBJECT

public:

    /**
     * Constructs the game.
     *
     * @param map The start map.
     * @param moves The initial moves.
     */

    Game(Map * map, Movements const & moves);


    /**
     * Sets a new map with movements.
     *
     * @param map The start map.
     * @param moves The initial moves.
     */

    void setMapAndMoves(Map * map, Movements const & moves);


    /**
     * Returns true, if we are in retro mode.
     *
     * Note that setMapAndMoves() will unset the retro mode.
     */

    bool retroMode() const;


    /**
     * Sets the retro mode.
     *
     * If you switch to retro mode and the keeper sits on a goal, we don't switch
     * to retro mode.
     *
     * @param retro_mode If true, we switch to retro mode.
     * @return True, if the switching was successful.
     */

    bool setRetroMode(bool retro_mode);


    /**
     * Tries to make the given atomic move.
     *
     * @param move The atomic move to try.
     * @return True if the move was made.
     */

    bool tryMove(AtomicMove const & move);


    /**
     * Tries to make the given move.
     *
     * @param move The move to try.
     * @return True if the move was made.
     */

    bool tryMove(Move const & move);


    /**
     * Tries to move the keeper and pull a gem.
     *
     * @param move The atomic move to try.
     * @return True, if the move was made (with or without pulling a gem).
     */

    bool tryPullMove(AtomicMove const & move);


    /**
     * Tries to move the keeper as far in a direction as possible.
     *
     * If the first move would push a stone, we move the stone as far as
     * possible and in the other case we move as far, as we reach a wall or a stone.
     *
     * @param move The direction of the move.
     */

    bool tryFarMove(AtomicMove const & move);


    /**
     * Moves the virtual keeper.
     *
     * @param move The atomic move to to withe the virtual keeper.
     * @return True, if the virtual keeper reached a reachable field.
     */

    bool moveVirtualKeeper(AtomicMove const & move);


    /**
     * Returns true, if we can undo a move.
     */

    bool canUndo() const;


    /**
     * Returns true, if we can redo a move.
     */

    bool canRedo() const;


    /**
     * Returns the number of pushes currently made.
     */

    int numberOfPushes() const;


    /**
     * Returns the number of linear pushes currently made.
     */

    int numberOfLinearPushes() const;


    /**
     * Returns the number of gem changes currently made.
     */

    int numberOfGemChanges() const;


    /**
     * Returns the number of moves currently made.
     */

    int numberOfMoves() const;


    /**
     * Returns the number of empty goals (goals without a gem).
     */

    int numberOfEmptyGoals() const;


    /**
     * Returns the moves.
     *
     * Note that unless you call emptyMoveQueue(), the map may lay behind the moves.
     */

    Movements const & moves() const;


    /**
     * Sets new moves.
     *
     * Note that we automatically jump to the actual move.
     *
     * @param moves The moves to set.
     */

    void setMoves(Movements const & moves);


    /**
     * Returns true, if the map is currently solved.
     */

    bool isSolved() const;


    /**
     * Returns true, if we show arrows.
     */

    bool showArrows() const;


    /**
     * Returns true, if we honor deadlock fields.
     */

    bool honorDeadlocks() const;


    /**
     * Returns the current animation speed.
     *
     * 0: no animation
     * 1: fast animation
     * 2: normal animation
     * 3: slow animation
     */

    int animation() const;


    /**
     * Returns the current direction of the keeper.
     */

    AtomicMove::AtomicMoveType keeperDirection() const;


    /**
     * Call this, when the config file has changed.
     */

    void configChanged();


    /**
     * Sets if we should display arrows.
     *
     * @param show If true, we display arrows for valid moves.
     */

    void setShowArrows(bool show);


    /**
     * Sets if the should honor deadlock fields.
     *
     * @param honor_deadlocks If true, we honor deadlock fields.
     */

    void setHonorDeadlock(bool honor_deadlocks);


    /**
     * Sets the animation speed.
     *
     * @param animation The animation speed.
     */

    void setAnimation(int animation);


    /**
     * Empties the move queue.
     *
     * After this call, the map and the moves are corresponding.
     */

    void emptyMoveQueue();


    /**
     * Tries to move the keeper.
     *
     * @param from The start position of the keeper.
     * @param to The desired end position of the keeper.
     * @param True, if the keeper was moved.
     */

    bool tryMoveKeeper(QPoint const & from, QPoint const & to);


    /**
     * Tries to move a gem from one point to the other.
     *
     * @param from The start position of the gem.
     * @param end The end position of the gem.
     * @return True, if the gem was moved.
     */

    bool tryMoveGem(QPoint const & from, QPoint const & to);


    /**
     * Jumps to the start position.
     */

    void jumpToStart();


    /**
     * Jumps to the end position of the moves made.
     */

    void jumpToEnd();


    /**
     * Tries to undo a move.
     *
     * @return True, if an undo move existed.
     */

    bool tryUndo();


    /**
     * Tries to redo a move.
     *
     * @return True, if a redo move existed.
     */

    bool tryRedo();


    /**
     * Plays the moves.
     *
     * @param from_start If true, we als jump to the start first.
     */

    void playMoves(bool from_start);


    /**
     * Updates the map.
     */

    void forceUpdate();


    /**
     * Enables or disables all updates.
     *
     * @param prevent_update If true, all updates to the map widget are prevented.
     */

    void preventUpdate(bool prevent_update);


signals:

    /**
     * This is emitted, whenever the map is solved.
     */

    void wasSolved();


    /**
     * Emitted of we leave a solution.
     */

    void wasUnsolved();


    /**
     * This is emitted, whenever the map was solved by a try***() function
     * (so not by undo, redo, play ...).
     */

    void wasSolvedByHand();


    /**
     * This is emitted, if the map was changed.
     *
     * @param force If true, we force an update of the status bar.
     */

    void mapChanged();


    /**
     * This is emitted, if the moves changed changed.
     *
     * @param force If true, we force an update of the status bar.
     */

    void movesChanged(bool force);


    /**
     * This is emitted, if the arrows we changed.
     *
     * It's guaranteed, that a mapChanged() signal will be emitted later.
     *
     * @param arrows The new arrows.
     */

    void arrowsChanged(std::vector<Move> const &arrows);


    /**
     * This is emitted, if the keeper direction was changed.
     *
     * It's guaranteed, that a mapChanged() signal will be emitted later.
     *
     * @param direction The new keeper direction.
     */

    void keeperDirectionChanged(AtomicMove::AtomicMoveType direction);


    /**
     * This is emitted, if the virtual keeper was moved.
     *
     * @param position The new position of the virtual keeper.
     */

    void virtualKeeperChanged(const QPoint & position);


private slots:

    /**
     * Processes them move queue.
     */

    void processMoveQueue();


private:

    /**
     * Removes the virtual keeper.
     */

    void removeVirtualKeeper();


    /**
     * Undoes a move.
     */

    void undo();


    /**
     * Redoes a move.
     */

    void redo();


    /**
     * Puts the given move in the move queue.
     *
     * Note that the move queue will be emptied before.
     *
     * The move will be added to m_moves.
     *
     * @param move The move to put (splitted into atomic moves) in the move queue.
     */

    void doMove(Move const & move);


    /**
     * Puts the given atomic moves in the move queue.
     *
     * Note that the move queue will <e>not</e> be emptied and the moves are <e>not</e>
     * stored in m_moves.
     *
     * @param moves The atomic moves to put in the queue.
     */

    void doAtomicMoves(Movements const & moves);


    /**
     * Puts the given moves in the move queue.
     *
     * Note that the move queue will be emptied before.
     *
     * The move will be added to m_moves.
     *
     * @param moves The moves to put (splitted into atomic moves) in the move queue.
     */

    void doMoves(Movements const & moves);


    /**
     * Puts the given undo move in the move queue.
     *
     * Note that the move queue will be emptied before.
     *
     * m_moves will not be changed.
     *
     * @param move The undo move to put (splitted into atomic moves) in the move queue.
     */

    void doUndoMove(Move const & move);


    /**
     * Adds an atomic move to the move queue.
     *
     * @param move The move to add.
     */

    void addToMoveQueue(Move const & move);


    /**
     * Processes a single move in the move queue.
     */

    void processMove();


    /**
     * Calculates the possible moves and shows the arrows.
     */

    void calcArrows();


    /**
     * Returns a simplified version of a move.
     *
     * @param move The move to simplify.
     * @param map The map to use.
     */

    Movements simplifyMove(Move const & move, Map const * map) const;


    /**
     * Returns a simplified version of a undo move.
     *
     * @param move The undo move to simplify.
     * @param map The map to use.
     */

    Movements simplifyUndoMove(Move const & move, Map const * map) const;


    /**
     * Returns true, if the given field is a deadlock field.
     *
     * If m_honor_deadlocks is false, this alsways returns false.
     *
     * @param position The position of the field to test.
     */

    bool isDeadlockField(QPoint const & position) const;


    /**
     * @overload
     */

    bool isDeadlockField(int index) const;


    /**
     * Moves all gems to the goals and sets the keeper to a reasonable position.
     *
     * Note that the keeper should not be located on a goal field.
     */

    void makeEndMap();


    /**
     * The current map.
     */

    Map * m_map;


    /**
     * The width of the map.
     */

    int m_width;


    /**
     * The height of the map.
     */

    int m_height;


    /**
     * m_width * m_height
     */

    int m_size;


    /**
     * The offsets for nb calculation.
     */

    int m_xy_offsets[4];


    /**
     * The current moves.
     */

    Movements m_moves;


    /**
     * The moves in the move queue.
     */

    std::deque<Move> m_move_queue;


    /**
     * If true, the moves in the queue are undo moves.
     */

    bool m_undo_moves;


    /**
     * If true, the moves in the queue are handmade moves.
     */

    bool m_handmade_moves;


    /**
     * If true, we don't call mapChanged() in processMoveQueue().
     */

    bool m_no_update;


    /**
     * If true, the level is solved.
     */

    bool m_is_solved;


    /**
     * The number of moves made.
     */

    int m_number_of_moves;


    /**
     * The number of pushes made.
     */

    int m_number_of_pushes;


    /**
     * The number of linear pushes.
     */

    int m_number_of_linear_pushes;


    /**
     * The number of gem changes.
     */

    int m_number_of_gem_changes;


    /**
     * Here we store the number of linear pushes for each push (used for undo).
     */

    std::vector<int> m_all_number_of_linear_pushes;


    /**
     * Here we store the number of gem changes for each push (used for undo).
     */

    std::vector<int> m_all_number_of_gem_changes;


    /**
     * Here we store the last gem positions for each push (used for undo).
     */

    std::vector<QPoint> m_all_last_pushed_gems;


    /**
     * Here we store the last gem directions for each push (used for undo).
     */

    std::vector<QPoint> m_all_last_push_directions;


    /**
     * The position of the last gem pushed.
     */

    QPoint m_last_pushed_gem;


    /**
     * The last push direction (or (0, 0) if last move was no push).
     */

    QPoint m_last_push_direction;


    /**
     * If true, we also calculate valid pushes and the corresponding arrows.
     */

    bool m_show_arrows;


    /**
     * If true, we honor deadlock fields.
     */

    bool m_honor_deadlocks;


    /**
     * The keeper direction.
     */

    AtomicMove::AtomicMoveType m_keeper_direction;


    /**
     * The position of the virtual keeper.
     */

    QPoint m_virtual_keeper;


    /**
     * If true, we wrap the virtual keeper around the sides of the map.
     */

    bool m_wrap_virtual_keeper;


    /**
     * The animation timer.
     */

    QTimer * m_timer;


    /**
     * The animation speed.
     */

    int m_animation;


    /**
     * If true, we are in retro mode.
     */

    bool m_retro_mode;


    /**
     * The backup map.
     */

    Map m_backup_map;


    /**
     * The backup moves.
     */

    Movements m_backup_moves;


    /**
     * If true, we prevent updates.
     */

    bool m_prevent_update;


    /**
     * If true, we are playing the moves.
     */

    bool m_in_play;


    /**
     * The time for fast animation.
     */

    static int s_fast_time;


    /**
     * The time for normal animation.
     */

    static int s_normal_time;


    /**
     * The time for slow animation.
     */

    static int s_slow_time;
};


#endif
