/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "hash.h"

#include <algorithm>
#include <cassert>

#include "map.h"




Hash::Hash(Map const & map)
{
    static std::vector<int> positions;
    positions.resize(0);

    int const max_size = map.width() * map.height();

    map.calcDeadlocks();

    for (int i = map.width() + 1; i < max_size; ++i)
    {
        int const piece = map.getPiece(i);

        if ( (piece != Map::OUTSIDE) && (piece != Map::WALL) && !map.isDeadlock(i) )
        {
            positions.push_back(i);
        }
    }

    calcHash(map, positions);
}


Hash::Hash(Map const & map, std::vector<int> const & positions)
{
    calcHash(map, positions);
}


Hash::Hash(Hash const & other_hash) :
    m_data0(other_hash.m_data0)
{
    if (onHeap())
    {
        Q_UINT32 const size = other_hash.rest.on_heap.m_size;
        rest.on_heap.m_size = size;
        rest.on_heap.m_data = new Q_UINT32[size];
        Q_UINT32 * my_data = rest.on_heap.m_data;
        Q_UINT32 const * other_data = other_hash.rest.on_heap.m_data;

        std::copy(other_data, other_data + size, my_data);
    }
    else
    {
        rest.on_stack.m_data1 = other_hash.rest.on_stack.m_data1;
        rest.on_stack.m_data2 = other_hash.rest.on_stack.m_data2;
    }
}


Hash::~Hash()
{
    if (onHeap())
    {
        delete[] rest.on_heap.m_data;
    }
}


Hash & Hash::operator = (Hash const & other_hash)
{
    assert(onHeap() == other_hash.onHeap());
    assert(!onHeap() || (rest.on_heap.m_size == other_hash.rest.on_heap.m_size));

    m_data0 = other_hash.m_data0;

    if (onHeap())
    {
        if (&other_hash != this)
        {
            Q_UINT32 const size = rest.on_heap.m_size;
            Q_UINT32 * my_data = rest.on_heap.m_data;
            Q_UINT32 const * other_data = other_hash.rest.on_heap.m_data;

            std::copy(other_data, other_data + size, my_data);
        }
    }
    else
    {
        rest.on_stack.m_data1 = other_hash.rest.on_stack.m_data1;
        rest.on_stack.m_data2 = other_hash.rest.on_stack.m_data2;
    }

    return *this;
}


bool Hash::operator == (Hash const & other_hash) const
{
    assert(onHeap() == other_hash.onHeap());
    assert(!onHeap() || (rest.on_heap.m_size == other_hash.rest.on_heap.m_size));

    if (m_data0 != other_hash.m_data0)
    {
        return false;
    }

    if (onHeap())
    {
        Q_UINT32 const size = rest.on_heap.m_size;
        Q_UINT32 const * my_data = rest.on_heap.m_data;
        Q_UINT32 const * other_data = other_hash.rest.on_heap.m_data;

        for (Q_UINT32 i = 0; i < size; ++i)
        {
            if (my_data[i] != other_data[i])
            {
                return false;
            }
        }

        return true;
    }
    else
    {
        return ( (rest.on_stack.m_data1 == other_hash.rest.on_stack.m_data1) &&
                 (rest.on_stack.m_data2 == other_hash.rest.on_stack.m_data2) );
    }
}


bool Hash::operator != (Hash const & other_hash) const
{
    assert(onHeap() == other_hash.onHeap());
    assert(!onHeap() || (rest.on_heap.m_size == other_hash.rest.on_heap.m_size));

    if (m_data0 != other_hash.m_data0)
    {
        return true;
    }

    if (onHeap())
    {
        Q_UINT32 const size = rest.on_heap.m_size;
        Q_UINT32 const * my_data = rest.on_heap.m_data;
        Q_UINT32 const * other_data = other_hash.rest.on_heap.m_data;

        for (Q_UINT32 i = 0; i < size; ++i)
        {
            if (my_data[i] != other_data[i])
            {
                return true;
            }
        }

        return false;
    }
    else
    {
        return !( (rest.on_stack.m_data1 == other_hash.rest.on_stack.m_data1) &&
                  (rest.on_stack.m_data2 == other_hash.rest.on_stack.m_data2) );
    }
}


bool Hash::operator < (Hash const & other_hash) const
{
    assert(onHeap() == other_hash.onHeap());
    assert(!onHeap() || (rest.on_heap.m_size == other_hash.rest.on_heap.m_size));

    if (m_data0 < other_hash.m_data0)
    {
        return true;
    }
    else if (m_data0 > other_hash.m_data0)
    {
        return false;
    }

    if (onHeap())
    {
        Q_UINT32 const size = rest.on_heap.m_size;
        Q_UINT32 const * my_data = rest.on_heap.m_data;
        Q_UINT32 const * other_data = other_hash.rest.on_heap.m_data;

        for (Q_UINT32 i = 0; i < size; ++i)
        {
            if (my_data[i] < other_data[i])
            {
                return true;
            }
            else if (my_data[i] > other_data[i])
            {
                return false;
            }
        }

        return false;
    }
    else
    {
        if (rest.on_stack.m_data1 < other_hash.rest.on_stack.m_data1)
        {
            return true;
        }
        else if (rest.on_stack.m_data1 > other_hash.rest.on_stack.m_data1)
        {
            return false;
        }
        else
        {
            return rest.on_stack.m_data2 < other_hash.rest.on_stack.m_data2;
        }
    }
}


void Hash::calcHash(Map const & map, std::vector<int> const & positions)
{
    static std::vector<Q_UINT32> act_hash;
    act_hash.resize(0);

    Q_UINT32 act_val;

    map.calcReachable();

    int const max_size = map.width() * map.height();

    for (int i = map.width() + 1; i < max_size; ++i)
    {
        if (map.isReachable(i))
        {
            act_val = static_cast<Q_UINT32> (i);

            break;
        }
    }

    int bits_in_val = 15;
    size_t const position_size = positions.size();

    for (size_t i = 0; i < position_size; ++i)
    {
        act_val <<= 1;
        ++bits_in_val;

        if (map.containsGem(positions[i]))
        {
            ++act_val;
        }

        if (bits_in_val == 32)
        {
            act_hash.push_back(act_val);
            bits_in_val = 0;
            act_val = 0;
        }
    }

    if (bits_in_val != 0)
    {
        act_hash.push_back(act_val);
    }

    act_hash.front() <<= 1;
    m_data0 = act_hash.front();
    int const size = static_cast<Q_UINT32> (act_hash.size()) - 1;

    if (size < 3)
    {
        if (size != 0)
        {
            rest.on_stack.m_data1 = act_hash[1];

            if (size == 2)
            {
                rest.on_stack.m_data2 = act_hash[2];
            }
            else
            {
                rest.on_stack.m_data2 = 0;
            }
        }
        else
        {
            rest.on_stack.m_data1 = 0;
            rest.on_stack.m_data2 = 0;
        }
    }
    else
    {
        m_data0 |= 1;
        rest.on_heap.m_size = size;
        rest.on_heap.m_data = new Q_UINT32[size];
        Q_UINT32 * my_data = rest.on_heap.m_data;
        Q_UINT32 const * other_data = &(act_hash[1]);

        std::copy(other_data, other_data + size, my_data);
    }
}


bool Hash::onHeap() const
{
    return m_data0 & 1;
}
