/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_HASH_INC_GUARD_H
#define EASYSOK_HASH_INC_GUARD_H


#include <vector>

#include <qglobal.h>


class Map;




/**
 * This class represents a hash value of a map.
 *
 * Note that unlike a real hash, it always allows you to differentiate
 * between different states of the same map.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 * @see Map
 */

class Hash
{

public:

    /**
     * Constructs the hash.
     *
     * @param map The map for the hash.
     */

    Hash(Map const & map);


    /**
     * Constructs the hash.
     * The positions in the given vector are used in the hash calculation.
     * Note that the vector holds the indices of all nonwall, nonoutside and nondeadlock fields,
     * the result is equivalent to that of the normal hash function, if the vector is sorted.
     *
     * @param map The map for the hash.
     * @param positions A vector with the indices of the relevant positions.
     */

    Hash(Map const & map, std::vector<int> const & positions);


    /**
     * The copy constructor.
     *
     * @param other_hash The hash to copy.
     */

    Hash(Hash const & other_hash);


    /**
     * Releasy the resources.
     */

    ~Hash();


    /**
     * Copies a hash.
     *
     * @param other_hash The hash to copie. Must be of the same map as this hash.
     */

    Hash & operator = (Hash const & other_hash);


    /**
     * True if the two hashs are the same.
     *
     * @param other_hash The hash to compare. Must be of the same map as this hash.
     */

    bool operator == (Hash const & other_hash) const;


    /**
     * True if the two hashs are not the same.
     *
     * @param other_hash The hash to compare. Must be of the same map as this hash.
     */

    bool operator != (Hash const & other_hash) const;


    /**
     * Used for associative containers.
     *
     * @param other_hash The hash to compare. Must be of the same map as this hash.
     */

    bool operator < (Hash const & other_hash) const;


private:

    /**
     * Calculates the actual hash value.
     *
     * @param map The map for the hash.
     * @param positions A vector with the indices of the relevant positions.
     */

    void calcHash(Map const & map, std::vector<int> const & positions);


    /**
     * Returns true, if we have allocated on the heap.
     */

    bool onHeap() const;


    /**
     * If the LSB is set, we use the freestore.
     */

    Q_UINT32 m_data0;


    /**
     * The union used to store the pieces.
     */

    union
    {
        struct OnStack
        {
            Q_UINT32 m_data1;
            Q_UINT32 m_data2;
        } on_stack;

        struct
        {
            Q_UINT32 m_size;
            Q_UINT32 * m_data;
        } on_heap;
    } rest;

};



#endif
