/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_IMAGE_EFFECT_INC_GUARD_H
#define EASYSOK_IMAGE_EFFECT_INC_GUARD_H


#include <vector>

#include <qcolor.h>


class QImage;




/**
 * This class adds functionality not present in KImageEffect.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class ImageEffect
{
public:

    /**
     * Blends the upper image on the lower.
     *
     * In contrast to KImageEffect::blendOnLower() the alpha channel of the resulting
     * image is calculated correctly. The implementation is taken mostly from
     * KImageEffect::blendOnLower(), with only small adjustments.
     *
     * Note that both images must be 32bit images.
     *
     * @param x_offset The x-offset of the upper image.
     * @param y_offset The y-offset of the upper image.
     * @param upper The upper image.
     * @param lower The lower image, which also holds the result.
     */

    static void blendOnLower(int x_offset, int y_offset, QImage const & upper, QImage & lower);


    /**
     * Blend the image to the given color.
     *
     * @param image The image to fade.
     * @param The blend ammount (in the range [0:1]).
     * @param color The color to blend to.
     * @param fade_rgb If true, we blend the rgb channels.
     * @param fade_alpha If true, we blend the alpha channel.
     */

    static void blend(QImage & image, float val, QRgb color, bool fade_rgb, bool fade_alpha);


    /**
     * @overload
     */

    static void blend(QImage & image, int val, QRgb color, bool fade_rgb, bool fade_alpha);


    /**
     * Colorizes the image to the given color.
     *
     * @param image The image to fade.
     * @param The colorize ammount (in the range [0:1]).
     * @param color The color to fade to.
     * @param fade_rgb If true, we fade the rgb channels.
     * @param fade_alpha If true, we fade the alpha channel.
     */

    static void colorize(QImage & image, float val, QRgb color);


    /**
     * @overload
     */

    static void colorize(QImage & image, int val, QRgb color);


    /**
     * Scales the image.
     *
     * @param image The image to use.
     * @param new_width The new width.
     * @param new_height The new height.
     */

    static void scale(QImage & image, int new_width, int new_height);


private:

    /**
     * Creates the scale factor vectors.
     *
     * @param from_pixels The original number of pixels.
     * @param to_pixels The new number of pixels.
     * @param sum The sum of the factors.
     * @param lengths Here we store the number pixels for the new pixels.
     * @param starts Here we store of indices of the start pixels.
     * @param factor_offsets Here we store the offsets in the factors array.
     * @param factors Here we store the factos.
     */

    static void calcScaleFactors(int from_pixels, int to_pixels, int sum, std::vector<int> & lengths,
                                 std::vector<int> & starts, std::vector<int> & factor_offsets,
                                 std::vector<Q_UINT32> & factors);
};


#endif
