/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "image_storer.h"

#include <cassert>
#include <vector>

#include <qiodevice.h>
#include <qpixmap.h>
#include <qstring.h>

#include <kglobal.h>
#include <kimageeffect.h>
#include <kstddirs.h>

#include "image_effect.h"
#include "map.h"
#include "pixmap_provider.h"
#include "theme.h"




bool ImageStorer::store(QIODevice * device, char const * format, Map const & map, int piece_size,
                        Theme const & theme, int direction, bool use_background, bool low_quality)
{
    assert(format != 0);
    assert(device != 0);

    QImageIO image_io(device, format);
    image_io.setImage(createImage(map, piece_size, theme, direction, use_background, low_quality));

    return image_io.write();
}


QImage ImageStorer::createImage(Map const & map, int piece_size, Theme const & theme,
                                int direction, bool use_background, bool low_quality)
{
    assert(piece_size > 0);
    assert(direction >= 0);
    assert(direction < 4);

    int const width = map.width();
    int const height = map.height();
    int const x_offset = static_cast<int> (theme.leftBorder() * piece_size);
    int const y_offset = static_cast<int> (theme.upperBorder() * piece_size);
    int const image_width = width * piece_size + x_offset +
                            static_cast<int> (theme.rightBorder() * piece_size);
    int const image_height = height * piece_size + y_offset +
                             static_cast<int> (theme.lowerBorder() * piece_size);

    QImage image(image_width, image_height, 32);
    image.fill(qRgba(0, 0, 0, 0));
    image.setAlphaBuffer(true);

    if (use_background)
    {
        bool use_image = false;

        if (!theme.backgroundImage().isEmpty())
        {
            QString const filename = KGlobal::dirs()->findResource("wallpaper", theme.backgroundImage());
            QImage background_image;

            if (background_image.load(filename))
            {
                use_image = true;
                int const background_width = background_image.width();
                int const background_height = background_image.height();

                int const x_tiles = (image_width - 1) / background_width + 1;
                int const y_tiles = (image_height - 1) / background_height + 1;

                background_image = background_image.convertDepth(32);
                background_image.setAlphaBuffer(true);

                for (int y = 0; y < y_tiles; ++y)
                {
                    for (int x = 0; x < x_tiles; ++x)
                    {
                        ImageEffect::blendOnLower(x * background_width, y * background_height,
                                                  background_image, image);
                    }
                }
            }
        }

        if (!use_image)
        {
            image.setAlphaBuffer(false);
            image.fill(theme.backgroundColor().rgb());
            image.setAlphaBuffer(true);
        }
    }

    PixmapProvider pixmap_provider(theme);

    int const nr_of_images = theme.nrOfPieceImages();
    std::vector<QImage> images(nr_of_images);

    for (int i = 0; i < nr_of_images; ++i)
    {
        QPixmap pixmap = pixmap_provider.createPixmap(i, piece_size);
        images[i] = pixmap.convertToImage();
    }

    for (int y = 0; y < height; ++y)
    {
        for (int x = 0; x < width; ++x)
        {
            std::vector<int> image_indices = theme.imageIndicesFromPosition(direction, QPoint(x, y), map);
            int const nr_of_indices = static_cast<int> (image_indices.size());

            for (int i = 0; i < nr_of_indices; ++i)
            {
                QPoint const offset = pixmap_provider.offset(image_indices[i], piece_size);

                ImageEffect::blendOnLower(x * piece_size + offset.x() + x_offset,
                                          y * piece_size + offset.y() + y_offset,
                                          images[image_indices[i]], image);
            }
        }
    }

    if (low_quality)
    {
        image = image.convertDepth(8, Qt::DiffuseAlphaDither | Qt::DiffuseDither |
                                   Qt::AutoColor | Qt::AvoidDither);
    }

    return image;
}
