/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "info_dialog.h"

#include <cassert>

#include <qframe.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qstring.h>
#include <qstringlist.h>

#include <kapp.h>
#include <klocale.h>
#include <kurl.h>
#include <kurllabel.h>

#include "collection.h"
#include "difficulty.h"
#include "level.h"




InfoDialog::InfoDialog(Level const & level, bool temporary, QWidget * parent, char const * name) :
    KDialogBase(parent, name, true, i18n("Level Information"), Close, Close, true)
{
    setupDialog(level.authors(), level.emails(), level.homepage(),
                level.copyright(), level.name(), level.info(), level.difficulty(),
                temporary ? i18n("The level is in a temporary collection!") : QString::null);
}


InfoDialog::InfoDialog(Collection const & collection, bool temporary, QWidget * parent, char const * name) :
    KDialogBase(parent, name, true, i18n("Collection Information"), Close, Close, true)
{
    setupDialog(collection.authors(), collection.emails(), collection.homepage(),
                collection.copyright(), collection.name(), collection.info(), collection.difficulty(),
                temporary ? i18n("The collection is temporary!") : QString::null );
}


void InfoDialog::emailClicked(const QString & url)
{
    KApplication::kApplication()->invokeMailer(KURL(url));
}


void InfoDialog::homepageClicked(const QString & url)
{
    KApplication::kApplication()->invokeBrowser(url);
}


void InfoDialog::setupDialog(QStringList const & authors, QStringList const & emails, QString const & homepage,
                             QString const & copyright, QString const & name, QString const & info,
                             int difficulty, QString const & temporary)
{
    assert(authors.count() == emails.count());

    int const number_of_authors = authors.count();
    int lines = std::max(1, number_of_authors) + 5;

    if (!temporary.isEmpty())
    {
        ++lines;
    }

    QFrame * frame = makeMainWidget();
    QGridLayout * layout = new QGridLayout(frame, lines, 2, marginHint(), spacingHint());

    QLabel * label = new QLabel(i18n("Author(s):") + " ", frame);
    layout->addWidget(label, 0, 0, AlignLeft | AlignVCenter);

    int line = 1;

    if (number_of_authors > 0)
    {
        for (int i = 0; i < number_of_authors; ++i)
        {
            KURLLabel * author_label = new KURLLabel(frame);
            author_label->setText(authors[i]);

            if (!emails[i].isEmpty())
            {
                author_label->setURL("mailto:" + emails[i]);
                connect(author_label, SIGNAL(leftClickedURL(const QString &)),
                        this,  SLOT(emailClicked(const QString &)));
            }

            layout->addWidget(author_label, i, 1, AlignLeft | AlignVCenter);
        }

        line = number_of_authors;
    }
    else
    {
        label = new QLabel(i18n("unkown"), frame);
        layout->addWidget(label, 0, 0, AlignLeft | AlignVCenter);
    }

    if (!homepage.isEmpty())
    {
        label = new QLabel(i18n("Homepage:") + " ", frame);
        layout->addWidget(label, line, 0, AlignLeft | AlignVCenter);

        KURLLabel * homepage_label = new KURLLabel(homepage, QString::null, frame);
        layout->addWidget(homepage_label, line, 1, AlignLeft | AlignVCenter);
        connect(homepage_label, SIGNAL(leftClickedURL(const QString &)),
                this, SLOT(homepageClicked(const QString &)));

        ++line;
    }

    if (!copyright.isEmpty())
    {
        label = new QLabel(i18n("Copyright:") + " ", frame);
        layout->addWidget(label, line, 0, AlignLeft | AlignVCenter);

        label = new QLabel(copyright, frame);
        layout->addWidget(label, line, 1, AlignLeft | AlignVCenter);

        ++line;
    }

    label = new QLabel(i18n("Difficulty:") + " ", frame);
    layout->addWidget(label, line, 0, AlignLeft | AlignVCenter);

    label = new QLabel(Difficulty::text(difficulty), frame);
    layout->addWidget(label, line, 1, AlignLeft | AlignVCenter);

    ++line;

    if (!name.isEmpty())
    {
        label = new QLabel(i18n("Name:") + " ", frame);
        layout->addWidget(label, line, 0, AlignLeft | AlignVCenter);

        label = new QLabel(name, frame);
        layout->addWidget(label, line, 1, AlignLeft | AlignVCenter);

        ++line;
    }

    if (!info.isEmpty())
    {
        label = new QLabel(i18n("Info:") + " ", frame);
        layout->addWidget(label, line, 0, AlignLeft | AlignTop);

        label = new QLabel(info, frame);
        layout->addWidget(label, line, 1, AlignLeft | AlignTop);

        ++line;
    }

    if (!temporary.isEmpty())
    {
        label = new QLabel(i18n("Warning:") + " ", frame);
        layout->addWidget(label, line, 0, AlignLeft | AlignTop);

        label = new QLabel(temporary, frame);
        layout->addWidget(label, line, 1, AlignLeft | AlignTop);
    }
}
