/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "level.h"

#include <algorithm>
#include <cassert>

#include <qdatastream.h>
#include <qregexp.h>
#include <qstringlist.h>




Level::Level(Map const & map, QStringList const & authors, QStringList const & emails,
             QString const & homepage, QString const & copyright,
             QString const & name,  QString const & info, int difficulty) :
    m_compressed_map(map),
    m_map(map),
    m_authors(authors),
    m_emails(emails),
    m_homepage(homepage),
    m_copyright(copyright),
    m_name(name),
    m_info(info)
{
    assert(m_map.isValid());
    assert(authors.count() == emails.count());

    setDifficulty(difficulty);
}


Level::Level(QStringList & lines, QStringList const & authors, QStringList const & emails,
             QString const & homepage, QString const & copyright, QString const & info, int difficulty) :
    m_compressed_map(Map(1, 1, std::vector<int> (1, 0))),
    m_map(lines),
    m_authors(authors),
    m_emails(emails),
    m_homepage(homepage),
    m_copyright(copyright),
    m_info(info)
{
    assert(authors.count() == emails.count());

    m_compressed_map = CompressedMap(m_map);
    getInfo(lines, m_authors, m_emails, m_homepage, m_copyright, m_name, m_info, m_difficulty);

    if (m_difficulty == -1)
    {
        setDifficulty(difficulty);
    }
}


Level::Level(QDataStream & stream, int version) :
    m_compressed_map(stream),
    m_map(m_compressed_map)
{
    assert(m_map.isValid());

    stream >> m_authors;
    stream >> m_emails;
    stream >> m_homepage;
    stream >> m_copyright;
    stream >> m_name;
    stream >> m_info;

    if (version >= 1)
    {
        Q_INT8 difficulty;
        stream >> difficulty;

        setDifficulty(difficulty);
    }
    else
    {
        m_difficulty = -1;
    }

    assert(m_authors.count() == m_emails.count());
}


Map const & Level::map() const
{
    return m_map;
}


void Level::setMap(Map const & map)
{
    m_map = map;
    m_compressed_map = CompressedMap(map);
}


CompressedMap const & Level::compressedMap() const
{
    return m_compressed_map;
}


QStringList const & Level::authors() const
{
    return m_authors;
}


QString Level::authorLine() const
{
    QString result;

    int const number_of_authors = m_authors.count();

    for (int i = 0; i < number_of_authors; ++i)
    {
        result += m_authors[i];

        if (i != number_of_authors - 1)
        {
            result += ", ";
        }
    }

    return result;
}


QString Level::authorEmailLine() const
{
    return createAuthorEmailLine(m_authors, m_emails);
}


void Level::setAuthors(QStringList const & authors)
{
    m_authors = authors;
}


void Level::setAuthorEmailLine(QString author_email_line)
{
    parseAuthorEmailLine(author_email_line, m_authors, m_emails);
}


QStringList const & Level::emails() const
{
    return m_emails;
}


void Level::setEmails(QStringList const & emails)
{
    m_emails = emails;
}


QString const & Level::homepage() const
{
    return m_homepage;
}


void Level::setHomepage(QString const & homepage)
{
    m_homepage = homepage;
}


QString const & Level::copyright() const
{
    return m_copyright;
}


void Level::setCopyright(QString const & copyright)
{
    m_copyright = copyright;
}


QString const & Level::name() const
{
    return m_name;
}


void Level::setName(QString const & name)
{
    m_name = name;
}


QString const & Level::info() const
{
    return m_info;
}


void Level::setInfo(QString const & info)
{
    m_info = info;
}


int Level::difficulty() const
{
    return m_difficulty;
}


void Level::setDifficulty(int difficulty)
{
    if ( (difficulty >= 0) && (difficulty <= 10) )
    {
        m_difficulty = difficulty;
    }
    else
    {
        m_difficulty = -1;
    }
}


void Level::writeToStream(QDataStream & stream) const
{
    m_compressed_map.writeToStream(stream);

    stream << m_authors;
    stream << m_emails;
    stream << m_homepage;
    stream << m_copyright;
    stream << m_name;
    stream << m_info;

    Q_INT8 difficulty = static_cast<Q_INT8> (m_difficulty);
    stream << difficulty;
}


QString Level::toText(QStringList const & authors, QStringList const & emails, QString const & homepage,
                      QString const & copyright, QString const & info, int difficulty) const
{
    assert(authors.count() == emails.count());

    QString result = m_map.toText();

    if ( ((m_authors != authors) && !m_authors.isEmpty()) || ((m_emails != emails) && !m_emails.isEmpty()) )
    {
        result += "Author: " + authorEmailLine() + '\n';
    }

    if ((m_homepage != homepage) && !m_homepage.isEmpty())
    {
        result += "Homepage: " + m_homepage + '\n';
    }

    if ((m_copyright != copyright) && !m_copyright.isEmpty())
    {
        result += "Copyright: " + m_copyright + '\n';
    }

    if (!m_name.isEmpty())
    {
        result += "Name: " + m_name + '\n';
    }

    if ((m_info != info) && !m_info.isEmpty())
    {
        result += QStringList::split('\n', m_info, true).join("Info: ") + '\n';
    }

    if (m_difficulty != difficulty)
    {
        result += "Difficulty: " + QString::number(m_difficulty) + '\n';
    }

    return result;
}


void Level::getInfo(QStringList & lines, QStringList & authors, QStringList & emails,
                    QString & homepage, QString & copyright, QString & name,
                    QString & info, int & difficulty)
{
    assert(authors.count() == emails.count());

    difficulty = -1;

    bool had_info = !info.isEmpty();

    while (!lines.isEmpty() && !Map::isMapLine(lines.first()) && (lines.first() != "+-+-"))
    {
        QString const act_line = lines.first();
        lines.remove(lines.begin());

        if (act_line.find("author:", 0, false) == 0)
        {
            parseAuthorEmailLine(act_line.right(act_line.length() - 7), authors, emails);
        }
        else if (act_line.find("homepage:", 0, false) == 0)
        {
            homepage = act_line.right(act_line.length() - 9).stripWhiteSpace();
        }
        else if (act_line.find("copyright:", 0, false) == 0)
        {
            copyright = act_line.right(act_line.length() - 10).stripWhiteSpace();
        }
        else if (act_line.find("name:", 0, false) == 0)
        {
            name = act_line.right(act_line.length() - 5).stripWhiteSpace();
        }
        else if (act_line.find("title:", 0, false) == 0)
        {
            name = act_line.right(act_line.length() - 6).stripWhiteSpace();
        }
        else if ( (act_line.find("info:", 0, false) == 0) || (act_line.find("comment:", 0, false) == 0) )
        {
	    if (had_info)
	    {
	        had_info = false;
		info = "";
	    }

            info += act_line.right(act_line.length() - 5).stripWhiteSpace() + '\n';
        }
        else if (act_line.find("difficulty:", 0, false) == 0)
        {
            difficulty = act_line.right(act_line.length() - 11).stripWhiteSpace().toInt();

            if ( (difficulty < 0) || (difficulty > 10) )
            {
                difficulty = -1;
            }
        }
    }
}


void Level::parseAuthorEmailLine(QString const & author_email_line,
                                 QStringList & authors, QStringList & emails)
{
    static QRegExp email_regexp("[<>()]");

    authors.clear();
    emails.clear();

    QStringList raw_authors = QStringList::split(',', author_email_line);
    int const number_of_authors = raw_authors.count();

    for (int i = 0; i < number_of_authors; ++i)
    {
        QStringList author_email = QStringList::split(email_regexp, raw_authors[i]);

        authors.append(author_email[0].stripWhiteSpace().simplifyWhiteSpace());

        if (author_email.count() > 1)
        {
            emails.append(author_email[1].stripWhiteSpace().simplifyWhiteSpace());
        }
        else
        {
            emails.append("");
        }
    }
}


QString Level::createAuthorEmailLine(QStringList const & authors, QStringList const & emails)
{
    assert(authors.count() == emails.count());

    QString result;

    int const number_of_authors = authors.count();

    for (int i = 0; i < number_of_authors; ++i)
    {
        result += authors[i];

        if (!emails[i].isEmpty())
        {
            result += " <" + emails[i] + ">";
        }

        if (i != number_of_authors - 1)
        {
            result += ", ";
        }
    }

    return result;
}
