/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_LEVEL_INC_GUARD_H
#define EASYSOK_LEVEL_INC_GUARD_H


#include <qregexp.h>
#include <qstring.h>
#include <qstringlist.h>

#include "compressed_map.h"
#include "map.h"


class Collection;
class QDataStream;
class QStringList;




/**
 * This class represents a sokoban level (which is a map plus some other properties).
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class Level
{

public:

    /**
     * Creates the level.
     *
     * @param map The map to use.
     * @param authors The author(s) the of the level.
     * @param email The email addresses of the authors (must be of the same size as authors).
     * @param homepage The homepage of the author.
     * @param copyright The copyright of the level.
     * @param name The name of the level.
     * @param info Additional info for this level.
     * @param difficulty The difficulty of the level in the range[-1:10], with 0 being trivially easy.
     *                   -1 means an unknown difficulty.
     */

    Level(Map const & map, QStringList const & authors, QStringList const & emails, QString const & homepage,
          QString const & copyright, QString const & name, QString const & info, int difficulty);


    /**
     * Constructs the map from a list of lines in xsb format.
     *
     * The constructor removes all lines in the string list, which were before the
     * actual map and the lines of the actual map itself and all lines up to another
     * map (in these lines we also look for additional info like author etc.).
     * Be sure to call map.isValid() afterwards.
     *
     * Note that you have to supply default values for all author etc, which will only
     * be overwritten, if there exist other information for the level.
     *
     * @param lines The list with the lines.
     * @param authors The author(s) the of the level.
     * @param emails The email addresses of the authors.
     * @param homepage The homepage of the author.
     * @param copyright The copyright of the level.
     * @param info The additional information of the level.
     * @param difficulty The difficulty of the level in the range[-1:10], with 0 being trivially easy.
     *                   -1 means an unknown difficulty.
     */

    Level(QStringList & lines, QStringList const & authors, QStringList const & emails,
          QString const & homepage, QString const & copyright, QString const & info, int difficulty);


    /**
     * Creates a level from a QDataStream.
     *
     * @param stream The stream to use.
     * @param version The version to use.
     */

    Level(QDataStream & stream, int version);


    /**
     * Returns the map of the level.
     */

    Map const & map() const;


    /**
     * Sets the map of the level.
     *
     * @param map The new map.
     */

    void setMap(Map const & map);


    /**
     * Returns the compressed map of the level.
     */

    CompressedMap const & compressedMap() const;


    /**
     * Returns the authors of the level.
     */

    QStringList const & authors() const;


    /**
     * Returns the authors in one line, separated by ", ".
     */

    QString authorLine() const;


    /**
     * Returns the authors and emails in one line, separated by ", ".
     */

    QString authorEmailLine() const;


    /**
     * Sets the authors of the level.
     *
     * @param authors The new authosr.
     */

    void setAuthors(QStringList const & authors);


    /**
     * Sets the authors and emails created from the line.
     *
     * @param author_email_line The line containing (separated by ',') the authors and emails.
     */

    void setAuthorEmailLine(QString author_email_line);


    /**
     * Returns the email adresses of the authors.
     */

    QStringList const & emails() const;


    /**
     * Sets the email adddresses of the authors.
     *
     * @param emails The email adddresses of the authors.
     */

    void setEmails(QStringList const & emails);


    /**
     * Returns the homepage of the level.
     */

    QString const & homepage() const;


    /**
     * Sets the homepage of the level.
     *
     * @param homepage The new homepage.
     */

    void setHomepage(QString const & homepage);


    /**
     * Returns the copyright the the level.
     */

    QString const & copyright() const;


    /**
     * Sets the copyright of the level.
     *
     * @param copyright The new copyright.
     */

    void setCopyright(QString const & copyright);


    /**
     * Returns the name of the level.
     */

    QString const & name() const;


    /**
     * Sets the name of the level.
     *
     * @param name The new name.
     */

    void setName(QString const & name);


    /**
     * Returns the additional info for this level.
     */

    QString const & info() const;


    /**
     * Sets the info of the level.
     *
     * @param info The new info.
     */

    void setInfo(QString const & info);


    /**
     * Returns the difficulty of the level.
     */

    int difficulty() const;


    /**
     * Sets the difficulty of the level.
     *
     * @param difficulty The new difficulty.
     */

    void setDifficulty(int difficulty);


    /**
     * Writes the level to a QDataStream.
     *
     * @param stream The stream to use.
     */

    void writeToStream(QDataStream & stream) const;


    /**
     * Returns the map of the level collection plus additional info in xsb format.
     *
     * Note that only these informations are added, which differ from the ones in
     * the collection.
     *
     * @param authors The authors of the collection.
     * @param emails The email addresses of the authors of the collection.
     * @param homepage The homepage of the collection.
     * @param copyright The copyright of the collection.
     * @param info The info of the collection.
     * @param difficulty The difficulty of the collection.
     */

    QString toText(QStringList const & authors, QStringList const & emails, QString const & homepage,
                   QString const & copyright, QString const & info, int difficulty) const;


private:

    /**
     * To let collection use the information routines (maybe this should be refactored).
     */

    friend class Collection;


    /**
     * Read author etc. information of a list of lines,  unitl a sokoban map is found.
     *
     * The constructor removes all lines in the string list, which were before the
     * actual map.
     *
     * Note that you have to supply default values for all author etc, which will only
     * be overwritten, if there exist other information.
     *
     * @param lines The list with the lines.
     * @param authors The authors the of the level.
     * @param emails The the email adresses of the authors the of the level.
     * @param homepage The homepage of the author.
     * @param copyright The copyright of the level.
     * @param name The name of the level.
     * @param difficulty The difficulty of the level.
     */

    static void getInfo(QStringList & lines, QStringList & authors, QStringList & emails,
                        QString & homepage, QString & copyright, QString & name,
                        QString & info, int & difficulty);


    /**
     * Creates authors and emails from an author-email-line.
     *
     * @param author_email_line The author-email-line.
     * @param authors Here we store the authors.
     * @param email Here we store the email addresses.
     */

    static void parseAuthorEmailLine(QString const & author_email_line,
                                     QStringList & authors, QStringList & emails);


    /**
     * Creates the author-email-line.
     *
     * @param authors The authors.
     * @param emails The emails.
     */

    static QString createAuthorEmailLine(QStringList const & authors, QStringList const & emails);


    /**
     * The compressed map.
     */

    CompressedMap m_compressed_map;


    /**
     * The map.
     */

    Map m_map;


    /**
     * The authors.
     */

    QStringList m_authors;


    /**
     * The email adresses of the authors.
     */

    QStringList m_emails;


    /**
     * The homepage.
     */

    QString m_homepage;


    /**
     * The copyright.
     */

    QString m_copyright;


    /**
     * The name.
     */

    QString m_name;


    /**
     * The additional level info.
     */

    QString m_info;


    /**
     * The difficulty.
     */

    int m_difficulty;
};


#endif
