/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "level_editor.h"

#include <algorithm>
#include <cassert>

#include <qvbox.h>

#include <kaccel.h>
#include <kaction.h>
#include <kapp.h>
#include <kconfig.h>
#include <kedittoolbar.h>
#include <kkeydialog.h>
#include <knuminput.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kstdaction.h>

#include "image_storer.h"
#include "level_generator.h"
#include "level_properties_dialog.h"
#include "map_size_dialog.h"
#include "map_widget.h"
#include "movements.h"
#include "pixmap_provider.h"
#include "solver.h"
#include "solver_dialog.h"




int LevelEditor::s_pieces[65] =
{
 Map::WALL, Map::WALL, Map::WALL, Map::WALL, Map::WALL,
 Map::WALL, Map::EMPTY, Map::EMPTY, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::KEEPER, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::EMPTY, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::GOAL, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::EMPTY, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::GEM, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::EMPTY, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::EMPTY, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::EMPTY, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::WALL, Map::EMPTY, Map::WALL,
 Map::WALL, Map::EMPTY, Map::EMPTY, Map::EMPTY, Map::WALL,
 Map::WALL, Map::WALL, Map::WALL, Map::WALL, Map::WALL,
};


LevelEditor::LevelEditor(Level const & level, int collection_nr, int level_nr, Theme * theme,
                         QWidget * parent, const char * name) :
    KMainWindow(parent, name),
    m_level(level),
    m_original_level(level),
    m_collection_nr(collection_nr),
    m_level_nr(level_nr),
    m_map(level.map()),
    m_theme(*theme),
    m_active_element(EMPTY),
    m_undo_pointer(0),
    m_was_saved(true),
    m_exited_signaled(false)
{
    resize(640, 480);
    assert(theme != 0);

    setCaption(i18n("Editor"));

    m_map.clearDeadlocks();
    m_map.uncrossAll();

    m_maps.push_back(level.compressedMap());

    m_theme.setHideGems(false);
    m_theme.setHideGoals(false);
    m_theme.setOutsideAsWall(false);

    m_map_widget = new MapWidget(&m_map, new PixmapProvider(m_theme), &m_theme, true, this);

    connect(m_map_widget, SIGNAL(fieldClicked(QPoint)), this, SLOT(fieldClicked(QPoint)));
    connect(m_map_widget, SIGNAL(gemMoved(QPoint, QPoint)), this, SLOT(gemMoved(QPoint, QPoint)));
    connect(m_map_widget, SIGNAL(keeperMoved(QPoint, QPoint)), this, SLOT(keeperMoved(QPoint, QPoint)));
    connect(m_map_widget, SIGNAL(undo()), this, SLOT(undo()));
    connect(m_map_widget, SIGNAL(redo()), this, SLOT(redo()));
    connect(m_map_widget, SIGNAL(mouseDragStarted()), this, SLOT(mouseDragStarted()));
    connect(m_map_widget, SIGNAL(mouseDragged(QPoint, QPoint)), this, SLOT(mouseDragged(QPoint, QPoint)));
    connect(m_map_widget, SIGNAL(mouseDragEnded()), this, SLOT(mouseDragEnded()));

    setCentralWidget(m_map_widget);

    createAccels();
    createActions();
    createGUI("easysokeditorui.rc");
}


LevelEditor::~LevelEditor()
{
    if (!m_was_saved && !m_exited_signaled)
    {
        saveUnsavedChanges(false);
    }

    if (!m_exited_signaled)
    {
        emit exited(this);
        m_exited_signaled = true;
    }
}


void LevelEditor::setOriginalLevel(Level const & level, int collection_nr, int level_nr)
{
    assert(collection_nr >= 0);
    assert(level_nr >= 0);

    m_original_level = level;
    m_collection_nr = collection_nr;
    m_level_nr = level_nr;
}


Level const & LevelEditor::level() const
{
    m_map.clearDeadlocks();
    m_map.clearReachable();

    m_level.setMap(m_map);

    return m_level;
}


Level const & LevelEditor::originalLevel() const
{
    return m_original_level;
}


int LevelEditor::collectionNr() const
{
    return m_collection_nr;
}


int LevelEditor::levelNr() const
{
    return m_level_nr;
}


void LevelEditor::showEditorHelp()
{
    KApplication::kApplication()->invokeHelp("edit-level-dialog");
}


void LevelEditor::saveAndExit()
{
    if (save())
    {
        exitWithoutSave();
    }
}


bool LevelEditor::save()
{
   Map::MapValidity validity = m_map.validity();

    if (validity == Map::IS_VALID)
    {
        emit saveLevel(this);
        m_was_saved = true;

        return true;
    }

    showValidityMessage(validity);

    return false;
}


void LevelEditor::exitWithoutSave()
{
    saveUnsavedChanges(true);

    if (m_exited_signaled)
    {
        close();
    }
}


void LevelEditor::undo()
{
    if (m_undo_pointer > 0)
    {
        --m_undo_pointer;
        m_map = m_maps[m_undo_pointer];
        m_map_widget->setMap(&m_map);
        m_map_widget->updateDisplay();

        updateUndoRedoState();
    }
}


void LevelEditor::redo()
{
    if (m_undo_pointer + 1 < static_cast<int> (m_maps.size()))
    {
        ++m_undo_pointer;
        m_map = m_maps[m_undo_pointer];
        m_map_widget->setMap(&m_map);
        m_map_widget->updateDisplay();

        updateUndoRedoState();
    }
}


void LevelEditor::fieldClicked(QPoint position)
{
    if (m_map.containsKeeper(position))
    {
        return;
    }

    int piece = m_map.getPiece(position);

    if (piece == Map::OUTSIDE)
    {
        piece = Map::EMPTY;
        m_map.setPiece(position, piece);
    }

    bool const was_wall = (piece == Map::WALL);

    if (m_active_element == KEEPER)
    {
        if (was_wall)
        {
            return;
        }

        m_map.setKeeper(position);
    }
    else if (m_active_element == EMPTY)
    {
        if (m_map.getPiece(position) == Map::EMPTY)
        {
            return;
        }

        m_map.setPiece(position, Map::EMPTY);
    }
    else if (m_active_element == GOAL)
    {
        if (was_wall)
        {
            return;
        }

        if (m_map.containsGoal(position))
        {
            return;
        }

        m_map.setPiece(position, piece + 1);
    }
    else if (m_active_element == GEM)
    {
        if (was_wall)
        {
            return;
        }

        if (m_map.containsGem(position))
        {
            return;
        }

        m_map.setPiece(position, piece - 2);
    }
    else if (m_active_element == WALL)
    {
        if (was_wall)
        {
            return;
        }

        m_map.setPiece(position, Map::WALL);
    }
    else
    {
        return;
    }

    updateOutside();
    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::gemMoved(QPoint from, QPoint to)
{
    if (from == to)
    {
        return;
    }

    if (!m_map.containsGem(from))
    {
        return;
    }

    if (m_map.containsKeeper(to))
    {
        return;
    }

    int to_piece = m_map.getPiece(to);

    if (to_piece == Map::WALL)
    {
        return;
    }

    if (to_piece == Map::OUTSIDE)
    {
        to_piece = Map::EMPTY;
        m_map.setPiece(to, to_piece);
    }

    if ( (to_piece == Map::KEEPER) || (to_piece == Map::GEM) )
    {
        m_map.setPiece(to, Map::EMPTY);
    }
    else if ( (to_piece == Map::KEEPER_ON_GOAL) || (to_piece == Map::GEM_ON_GOAL) )
    {
        m_map.setPiece(to, Map::GOAL);
    }

    m_map.moveGem(from, to);
    updateOutside();
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::keeperMoved(QPoint from, QPoint to)
{
    if (from == to)
    {
        return;
    }

    if (!m_map.containsKeeper(from))
    {
        return;
    }

    int to_piece = m_map.getPiece(to);

    if (to_piece == Map::WALL)
    {
        return;
    }

    if (to_piece == Map::OUTSIDE)
    {
        to_piece = Map::EMPTY;
        m_map.setPiece(to, to_piece);
    }

    if (to_piece == Map::GEM)
    {
        m_map.setPiece(to, Map::EMPTY);
    }
    else if (to_piece == Map::GEM_ON_GOAL)
    {
        m_map.setPiece(to, Map::GOAL);
    }

    m_map.setKeeper(to);
    updateOutside();
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::rotateLeft()
{
    m_map.rotateLeft();

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::rotateRight()
{
    m_map.rotateRight();

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::mirrorHorizontally()
{
    m_map.mirrorHorizontally();

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::mirrorVertically()
{
    m_map.mirrorVertically();

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::simplify()
{
    m_map = m_map.simplify();

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::fillEdges()
{
    m_map = m_map.fillEdges();

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::fillOutside()
{
    int const size = m_map.width() * m_map.height();

    for (int i = 0; i < size; ++i)
    {
        if (m_map.getPiece(i) == Map::OUTSIDE)
        {
            m_map.setPiece(i, Map::WALL);
        }
    }

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::changeProperties()
{
    LevelPropertiesDialog dialog(m_level, this);

    if (dialog.exec())
    {
        m_level.setAuthorEmailLine(dialog.author());
        m_level.setHomepage(dialog.homepage());
        m_level.setCopyright(dialog.copyright());
        m_level.setName(dialog.name());
        m_level.setInfo(dialog.info());
        m_level.setDifficulty(dialog.difficulty());
    }

    m_was_saved = false;
}


void LevelEditor::changeSize()
{
    MapSizeDialog dialog(m_map.width(), m_map.height(), this);

    if (dialog.exec())
    {
        int const old_width = m_map.width();
        int const old_height = m_map.height();
        int const new_width = dialog.width();
        int const new_height = dialog.height();
        int const copy_width = std::min(old_width, new_width);
        int const copy_height = std::min(old_height, new_height);
        std::vector<int> new_pieces(new_width * new_height, Map::EMPTY);

        if ( (m_map.keeper().x() >= new_width) || (m_map.keeper().y() >= new_height) )
        {
            int dist = 0;
            QPoint new_keeper(0, 0);

            for (int y = 0; y < copy_height; ++y)
            {
                for (int x = 0; x < copy_width; ++x)
                {
                    if (m_map.canDropKeeper(QPoint(x, y)))
                    {
                        if (x + y > dist)
                        {
                            new_keeper  = QPoint(x, y);
                            dist = x + y;
                        }
                    }
                }
            }

            m_map.setPiece(new_keeper, Map::KEEPER);
        }

        for (int y = 0; y < copy_height; ++y)
        {
            for (int x = 0; x < copy_width; ++x)
            {
                new_pieces[x + new_width * y] = m_map.getPiece(QPoint(x, y));
            }
        }

        m_map = Map(new_width, new_height, new_pieces);
        m_map_widget->setMap(&m_map);
        m_map_widget->updateDisplay();
        insertMap();
    }
}


void LevelEditor::adjustSize()
{
    m_map = m_map.adjustSize();
    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::createEmptyMap()
{
    KDialogBase dialog(this, 0, true, QString::null, KDialogBase::Ok | KDialogBase::Cancel);

    KConfig * config = KApplication::kApplication()->config();
    int fill_ammount = config->readNumEntry("Wall fill ammount", 50);
    fill_ammount = std::max(std::min(fill_ammount, 100), 0);

    KIntNumInput * fill = new KIntNumInput(fill_ammount, dialog.makeVBoxMainWidget());
    fill->setSuffix(i18n(" percent filled"));
    fill->setRange(0, 100);
    fill->setLabel(i18n("Set fill ammount"));

    if (dialog.exec())
    {
        m_map = LevelGenerator::createEmptyMap(m_map.width(), m_map.height(), fill->value() / 100.0);

        m_map_widget->setMap(&m_map);
        m_map_widget->updateDisplay();
        insertMap();
    }

    config->setGroup("");
    config->writeEntry("Wall fill ammount", fill->value());
}


void LevelEditor::createGoals()
{
    KDialogBase dialog(this, 0, true, QString::null, KDialogBase::Ok | KDialogBase::Cancel);

    KConfig * config = KApplication::kApplication()->config();
    int const nr_of_goals = std::max(config->readNumEntry("Number of goals", 5), 1);

    KIntNumInput * goals = new KIntNumInput(nr_of_goals, dialog.makeVBoxMainWidget());
    goals->setRange(1, 200);
    goals->setLabel(i18n("Set number of goals"));

    if (dialog.exec())
    {
        m_map = LevelGenerator::createGoals(m_map, goals->value());

        m_map_widget->setMap(&m_map);
        m_map_widget->updateDisplay();
        insertMap();
    }

    config->setGroup("");
    config->writeEntry("Number of goals", goals->value());
}


void LevelEditor::createGems()
{
    m_map = LevelGenerator::createGems(m_map);

    m_map_widget->setMap(&m_map);
    m_map_widget->updateDisplay();
    insertMap();
}


void LevelEditor::createSolvableMap()
{
    KDialogBase dialog(this, 0, true, QString::null, KDialogBase::Ok | KDialogBase::Cancel);

    KConfig * config = KApplication::kApplication()->config();
    int const nr_of_pushes = std::max(config->readNumEntry("Number of pushes in autocreate", 50), 1);

    KIntNumInput * pushes = new KIntNumInput(nr_of_pushes, dialog.makeVBoxMainWidget());
    pushes->setRange(1, 1000);
    pushes->setLabel(i18n("Set number of pushes for solution"));

    if (dialog.exec())
    {
        Map map = m_map;
        bool found_one = false;

        for (int tries = 0; tries < 1; ++tries)
        {
            int const nr_of_pushes = pushes->value();
            int max_effort = nr_of_pushes * 1000;

            Movements moves = LevelGenerator::createSolution(m_map, nr_of_pushes, max_effort, map);

            if (!moves.isEmpty())
            {
                m_map = map;
                found_one = true;
                break;
            }
        }

        if (found_one)
        {
            m_map_widget->setMap(&m_map);
            m_map_widget->updateDisplay();
            insertMap();
        }
        else
        {
            KMessageBox::error(this, i18n("Unable to create a solution of the given length!"));
        }
    }

    config->setGroup("");
    config->writeEntry("Number of pushes in autocreate", pushes->value());
}


void LevelEditor::checkValidity()
{
    Map::MapValidity validity = m_map.validity();

    showValidityMessage(validity);
}


void LevelEditor::isSolvable()
{
    Map::MapValidity validity = m_map.validity();

    if (validity != Map::IS_VALID)
    {
        KMessageBox::error(this, i18n("The map is invalid!"));

        return;
    }

    if (m_map.isSolved())
    {
        KMessageBox::information(this, i18n("The map is already solved!"));

        return;
    }

    if (m_map.numberOfGems() > 10)
    {
        if (KMessageBox::warningContinueCancel(this, i18n("The current map contains more than 10 gems.\n"
                                                          "Experience shows, that the solver is\n"
                                                          "almost always too dumb to solve maps of that\n"
                                                          "size. Do you really want to continue?"),
                                               QString::null, i18n("Continue"),
                                               "Warn when solving map with many gems",
                                               false) == KMessageBox::Cancel)
        {
            return;
        }
    }

    m_map.clearDeadlocks();
    m_map.clearReachable();
    SolverDialog dialog(m_map, this);

    if (dialog.exec() != QMessageBox::Cancel)
    {
        if (dialog.solver()->moves().isEmpty())
        {
            KMessageBox::information(this, i18n("The map is unsolvable."));
        }
        else
        {
            QString const text = i18n("The map is solvable with %1 pushes.").arg(dialog.solver()->depth());
            KMessageBox::information(this, text);
        }
    }
}


void LevelEditor::activeElementChanged()
{
    if (m_empty_field_action->isChecked())
    {
        m_active_element = EMPTY;
    }
    else if (m_wall_field_action->isChecked())
    {
        m_active_element = WALL;
    }
    else if (m_goal_field_action->isChecked())
    {
        m_active_element = GOAL;
    }
    else if (m_keeper_field_action->isChecked())
    {
        m_active_element = KEEPER;
    }
    else if (m_gem_field_action->isChecked())
    {
        m_active_element = GEM;
    }
}


void LevelEditor::showKeyDialog()
{
    KKeyDialog::configureKeys(actionCollection(), xmlFile());
}


void LevelEditor::showToolbarDialog()
{
    KEditToolbar dialog(actionCollection(), "easysokeditorui.rc");
    connect(&dialog, SIGNAL(newToolbarConfig()), this, SLOT(updateToolbar()));

    if (dialog.exec())
    {
        createGUI("easysokeditorui.rc");
    }
}


void LevelEditor::updateToolbar()
{
    createGUI("easysokeditorui.rc");
}


void LevelEditor::mouseDragStarted()
{

    if (m_active_element != KEEPER)
    {
        m_in_drag = true;
    }
}


void LevelEditor::mouseDragged(QPoint from, QPoint to)
{
    if (m_in_drag)
    {
        if ((to - from).manhattanLength() <= 1)
        {
            if ((to.x() >= 0) && (to.y() >= 0) && (to.x() < m_map.width()) && (to.y() < m_map.height()))
            {
                fieldClicked(to);
            }
        }
        else
        {
            QPoint act_point = from;

            int const dx = to.x() - from.x();
            int const abs_dx = std::abs(dx);
            int const s_dx = dx > 0 ? -1 : 1;
            double r_dx = 0;
            double act_rx = 1.0;

            if (to.x() != from.x())
            {
                r_dx = 1.0 / abs_dx;
                act_rx -= r_dx / 2.0;
            }
            else
            {
                act_rx = 0.0;
            }

            int const dy = to.y() - from.y();
            int const abs_dy = std::abs(dy);
            int const s_dy = dy > 0 ? -1 : 1;
            double r_dy = 0;
            double act_ry = 1.0;

            if (to.y() != from.y())
            {
                r_dy = 1.0 / abs_dy;
                act_ry -= r_dy / 2.0;
            }
            else
            {
                act_ry = 0.0;
            }

            int const steps = abs_dx + abs_dy;

            for (int i = 0; i < steps; ++i)
            {
                if (act_ry > act_rx)
                {
                    act_point += QPoint(0, s_dy);
                    act_ry -= r_dy;
                }
                else
                {
                    act_point += QPoint(0, s_dx);
                    act_rx -= r_dx;
                }

                fieldClicked(act_point);
            }
        }
    }
}


void LevelEditor::mouseDragEnded()
{
    m_in_drag = false;
}


void LevelEditor::saveUnsavedChanges(bool cancel_possible)
{
    while (!m_was_saved)
    {
        QString msg = i18n("The level is not saved!\nDo you want to save it?");
        int result;

        if (cancel_possible)
        {
            result = KMessageBox::warningYesNoCancel(this, msg, QString::null, QString::null, QString::null,
                                                     "exitwithoutsave");
        }
        else
        {
            result = KMessageBox::warningContinueCancel(this, msg, QString::null, i18n("Yes"),
                                                     "exitwithoutsave");

            if (result == KMessageBox::Continue)
            {
                result = KMessageBox::Yes;
            }
            else
            {
                result = KMessageBox::No;
            }
        }

        if (result == KMessageBox::Yes)
        {
            save();
        }
        else if (result == KMessageBox::Cancel)
        {
            return;
        }
        else
        {
            break;
        }
    }

    emit exited(this);
    m_exited_signaled = true;
}


void LevelEditor::createAccels()
{
    KAccel * acc = accel();

    acc->insertItem("Save level and exit", "SaveAndQuit", SHIFT | Key_Q);
    acc->insertItem("Undo", "Undo", Key_U);
    acc->insertItem("Redo", "Redo", Key_R);
    acc->insertItem("Change properties of level", "ChangeProperties", Key_P);
    acc->insertItem("Change size", "ChangeSize", Key_S);
    acc->insertItem("Auto adjust size", "AutoAdjustSize", Key_A);
    acc->insertItem("Empty field", "Empty", Key_E);
    acc->insertItem("Wall",  "Wall",  Key_W);
    acc->insertItem("Keeper", "Keeper", Key_K);
    acc->insertItem("Goal", "Goal", Key_O);
    acc->insertItem("Gem", "Gem", Key_G);
    acc->insertItem("Rotate left", "RotateLeft", SHIFT | Key_L);
    acc->insertItem("Rotate right", "RotateRight", SHIFT | Key_R);
    acc->insertItem("Mirror horizontally", "MirrorHorizontally", SHIFT | Key_H);
    acc->insertItem("Mirror vertically", "MirrorVertically", SHIFT | Key_V);
    acc->insertItem("Simplify map", "Simplify", Key_I);
    acc->insertItem("Fill edges", "FillEdges", CTRL | Key_E);
    acc->insertItem("Fill outside", "FillOutside", CTRL | Key_O);
    acc->insertItem("Create map without goals", "CreateMap", SHIFT | Key_C);
    acc->insertItem("Create goals", "CreateGoals", SHIFT | Key_O);
    acc->insertItem("Create gems", "CreateGems", SHIFT | Key_G);
    acc->insertItem("Make solvable map", "CreateSolvable", SHIFT | Key_M);
    acc->insertItem("Is map valid ?", "CheckValidity", CTRL | Key_V);
    acc->insertItem("Is solvable ?", "IsSolvable", SHIFT | Key_S);
}


void LevelEditor::createActions()
{
    int const piece_size = 32;
    int x_offset = static_cast<int> (m_theme.leftBorder() * piece_size);
    int y_offset = static_cast<int> (m_theme.upperBorder() * piece_size);

    std::vector<int> pieces(s_pieces, s_pieces + 65);
    Map image_map(5, 13, pieces);
    QImage image = ImageStorer::createImage(image_map, piece_size, m_theme, 0, true, false);

    QPixmap wall_pixmap;
    wall_pixmap.convertFromImage(image.copy(2 * piece_size + x_offset, 10 * piece_size + y_offset,
                                            piece_size, piece_size));

    QPixmap empty_pixmap;
    empty_pixmap.convertFromImage(image.copy(2 * piece_size + x_offset, 8 * piece_size + y_offset,
                                             piece_size, piece_size));

    QPixmap keeper_pixmap;
    keeper_pixmap.convertFromImage(image.copy(2 * piece_size + x_offset, 2 * piece_size + y_offset,
                                              piece_size, piece_size));

    QPixmap gem_pixmap;
    gem_pixmap.convertFromImage(image.copy(2 * piece_size + x_offset, 6 * piece_size + y_offset,
                                           piece_size, piece_size));

    QPixmap goal_pixmap;
    goal_pixmap.convertFromImage(image.copy(2 * piece_size + x_offset, 4 * piece_size + y_offset,
                                            piece_size, piece_size));


    KActionCollection * action_col = actionCollection();
    KAccel * acc = accel();

    new KAction(i18n("&Save Level"), "filesave", KStdAccel::key(KStdAccel::Save),
                this, SLOT(save()), action_col, "Save");
    new KAction(i18n("S&ave Level And Exit"), acc->currentKey("SaveAndQuit"),
                this, SLOT(saveAndExit()), action_col, "SaveAndQuit");
    new KAction(i18n("&Exit"), "exit", KStdAccel::key(KStdAccel::Quit),
                this, SLOT(exitWithoutSave()), action_col, "Quit");

    m_undo_action = new KAction(i18n("&Undo"), "undo", acc->currentKey("Undo"),
                                this, SLOT(undo()), action_col, "Undo");
    m_redo_action = new KAction(i18n("&Redo"), "redo", acc->currentKey("Redo"),
                                this, SLOT(redo()), action_col, "Redo");
    new KAction(i18n("Change &Properties"), acc->currentKey("ChangeProperties"),
                this, SLOT(changeProperties()), action_col, "ChangeProperties");
    new KAction(i18n("Change &Size"), acc->currentKey("ChangeSize"),
                this, SLOT(changeSize()), action_col, "ChangeSize");
    new KAction(i18n("&Auto Adjust Size"), acc->currentKey("AutoAdjustSize"),
                this, SLOT(adjustSize()), action_col, "AutoAdjustSize");


    m_empty_field_action = new KRadioAction(i18n("&Empty Field"), empty_pixmap, acc->currentKey("Empty"),
                                            this, SLOT(activeElementChanged()), action_col, "Empty");
    m_empty_field_action->setExclusiveGroup("activeelement");
    m_empty_field_action->setChecked(m_active_element == EMPTY);

    m_wall_field_action = new KRadioAction(i18n("&Wall"), wall_pixmap, acc->currentKey("Wall"),
                                           this, SLOT(activeElementChanged()), action_col, "Wall");
    m_wall_field_action->setExclusiveGroup("activeelement");
    m_wall_field_action->setChecked(m_active_element == WALL);

    m_keeper_field_action = new KRadioAction(i18n("&Keeper"), keeper_pixmap, acc->currentKey("Keeper"),
                                             this, SLOT(activeElementChanged()), action_col, "Keeper");
    m_keeper_field_action->setExclusiveGroup("activeelement");
    m_keeper_field_action->setChecked(m_active_element == KEEPER);

    m_goal_field_action = new KRadioAction(i18n("G&oal"), goal_pixmap, acc->currentKey("Goal"),
                                           this, SLOT(activeElementChanged()), action_col, "Goal");
    m_goal_field_action->setExclusiveGroup("activeelement");
    m_goal_field_action->setChecked(m_active_element == GOAL);

    m_gem_field_action = new KRadioAction(i18n("&Gem"), gem_pixmap, acc->currentKey("Gem"),
                                          this, SLOT(activeElementChanged()), action_col, "Gem");
    m_gem_field_action->setExclusiveGroup("activeelement");
    m_gem_field_action->setChecked(m_active_element == GEM);

    new KAction(i18n("Rotate &Left"), "rotate_ccw", acc->currentKey("RotateLeft"),
                this, SLOT(rotateLeft()), action_col, "RotateLeft");
    new KAction(i18n("Rotate &Right"), "rotate_cw", acc->currentKey("RotateRight"),
                this, SLOT(rotateRight()), action_col, "RotateRight");
    new KAction(i18n("Mirror &Horizontally"), "vmirror", acc->currentKey("MirrorHorizontally"),
                this, SLOT(mirrorHorizontally()), action_col, "MirrorHorizontally");
    new KAction(i18n("Mirror &Vertically"), "hmirror", acc->currentKey("MirrorVertically"),
                this, SLOT(mirrorVertically()), action_col, "MirrorVertically");
    new KAction(i18n("&Simplify Map"), acc->currentKey("Simplify"),
                this, SLOT(simplify()), action_col, "Simplify");
    new KAction(i18n("Fill &Edges"), acc->currentKey("FillEdges"),
                this, SLOT(fillEdges()), action_col, "FillEdges");
    new KAction(i18n("Fill &Outside"), acc->currentKey("FillOutside"),
                this, SLOT(fillOutside()), action_col, "FillOutside");

    new KAction(i18n("Create &Map Without Goals"), acc->currentKey("CreateMap"),
                this, SLOT(createEmptyMap()), action_col, "CreateMap");
    new KAction(i18n("Create G&oals"), acc->currentKey("CreateGoals"),
                this, SLOT(createGoals()), action_col, "CreateGoals");
    new KAction(i18n("Create &Gems"), acc->currentKey("CreateGems"),
                this, SLOT(createGems()), action_col, "CreateGems");
    new KAction(i18n("Make &Solvable Map"), acc->currentKey("CreateSolvable"),
                this, SLOT(createSolvableMap()), action_col, "CreateSolvable");

    KStdAction::keyBindings(this, SLOT(showKeyDialog()), action_col);
    KStdAction::configureToolbars(this, SLOT(showToolbarDialog()), action_col);

    KStdAction::helpContents(this, SLOT(showEditorHelp()), action_col, "EditHelp");
    KStdAction::whatsThis(this, SLOT(whatsThis()), action_col, "WhatsThis");
    new KAction(i18n("Is Map &Valid?"), acc->currentKey("CheckValidity"),
                this, SLOT(checkValidity()), action_col, "CheckValidity");
    new KAction(i18n("Is &Solvable?"), acc->currentKey("IsSolvable"),
                this, SLOT(isSolvable()), action_col, "IsSolvable");

    updateUndoRedoState();
}


void LevelEditor::showValidityMessage(Map::MapValidity validity)
{
    if (validity == Map::IS_VALID)
    {
        KMessageBox::information(this, i18n("The level seems to be valid."));
    }
    else if (validity == Map::NO_KEEPER)
    {
        KMessageBox::error(this, i18n("The level is invalid.\nThere is no keeper!"));
    }
    else if (validity == Map::TOO_MANY_KEEPERS)
    {
        KMessageBox::error(this, i18n("The level is invalid.\nThere is more than one keeper!"));
    }
    else if (validity == Map::NO_GEMS)
    {
        KMessageBox::error(this, i18n("The level is invalid.\nThere are no gems!"));
    }
    else if (validity == Map::MORE_GEMS_THAN_GOALS)
    {
        KMessageBox::error(this, i18n("The level is invalid.\nThere are more gems than goals!"));
    }
    else if (validity == Map::MORE_GOALS_THAN_GEMS)
    {
        KMessageBox::error(this, i18n("The level is invalid.\nThere are more goals than gems!"));
    }
    else if (validity == Map::MAP_LEAKS)
    {
        KMessageBox::error(this, i18n("The level is invalid.\nThe map leaks (this means "
                                      "non wall fields have contact to outside fields)!"));
    }
    else if (validity == Map::MAP_SOLVED)
    {
        KMessageBox::error(this, i18n("The level is invalid.\nThe map is already solved!"));
    }
    else
    {
        KMessageBox::error(this, i18n("The level is invalid!"));
    }
}


void LevelEditor::updateUndoRedoState()
{
    m_undo_action->setEnabled(m_undo_pointer > 0);
    m_redo_action->setEnabled(m_undo_pointer + 1 < static_cast<int> (m_maps.size()));
}


void LevelEditor::insertMap()
{
    m_map = Map(m_map.width(), m_map.height(), m_map.pieces());

    ++m_undo_pointer;

    if (m_undo_pointer == static_cast<int> (m_maps.size()))
    {
        m_maps.push_back(m_map);
    }
    else if (CompressedMap(m_map) != m_maps[m_undo_pointer])
    {
        m_maps.erase(m_maps.begin() + m_undo_pointer, m_maps.end());
        m_maps.push_back(m_map);
    }

    updateUndoRedoState();
    m_was_saved = false;
}


void LevelEditor::updateOutside()
{
    int const width = m_map.width();
    int const height = m_map.height();
    int const size = width * height;

    for (int i = 0; i < size; ++i)
    {
        if (m_map.getPiece(i) == Map::OUTSIDE)
        {
            m_map.setPiece(i, Map::EMPTY);
        }
    }

    m_map = Map(m_map.width(), m_map.height(), m_map.pieces());
}
