/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_LEVEL_EDITOR_INC_GUARD_H
#define EASYSOK_LEVEL_EDITOR_INC_GUARD_H


#include <kmainwindow.h>

#include "compressed_map.h"
#include "level.h"
#include "theme.h"


class KAction;
class KRadioAction;
class MapWidget;




/**
 * Lets the user edit a level.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class LevelEditor : public KMainWindow
{
    Q_OBJECT

public:

    /**
     * Creates the window.
     *
     * @param level The level to edit.
     * @param theme The theme to use.
     * @param collection_nr The original collection nr.
     * @param level_nr The original level nr.
     * @param parent The parent widget.
     * @param name The internal name.
     */

    LevelEditor(Level const & level, int collection_nr, int level_nr, Theme * theme,
                QWidget * parent = 0, const char * name = 0);


    /**
     * Deletes the window an gives an opportunity to save a level if not already done.
     */

    virtual ~LevelEditor();


    /**
     * Sets a new original level and the collection and level number.
     *
     * @param level The new original level to edit.
     * @param collection_nr The original collection nr.
     * @param level_nr The original level nr.
     */

    void setOriginalLevel(Level const & level, int collection_nr, int level_nr);


    /**
     * Returns the edited level.
     */

    Level const & level() const;


    /**
     * Returns the original level.
     */

    Level const & originalLevel() const;


    /**
     * Returns the collection nr.
     */

    int collectionNr() const;


    /**
     * Returns the level nr.
     */

    int levelNr() const;


public slots:

    /**
     * Shows the help page of the editor.
     */

    void showEditorHelp();


    /**
     * Saves the level and exits.
     */

    void saveAndExit();


    /**
     * Saves the level.
     *
     * Returns true, if the level was saved.
     */

    bool save();


    /**
     * Exits without save.
     */

    void exitWithoutSave();


    /**
     * Undoes the last change.
     */

    void undo();


    /**
     * Redoes the last change.
     */

    void redo();


    /**
     * Called when the user has clicked a field.
     *
     * @param position The position the user has clicked.
     */

    void fieldClicked(QPoint position);


    /**
     * Called when the user has move a gem.
     *
     * @param from The start position of the gem.
     * @param end The end position of the gem.
     */

    void gemMoved(QPoint from, QPoint to);


    /**
     * Called when the user has move a keeper.
     *
     * @param from The start position of the keeper.
     * @param end The end position of the keeper.
     */

    void keeperMoved(QPoint from, QPoint to);


    /**
     * Called when the user rotates the map left.
     */

    void rotateLeft();


    /**
     * Called when the user rotateLeft the map right.
     */

    void rotateRight();


    /**
     * Called when the user mirrors the map horizontally.
     */

    void mirrorHorizontally();


    /**
     * Called when the user mirrors the map vertically.
     */

    void mirrorVertically();


    /**
     * Simpplifies the level.
     */

    void simplify();


    /**
     * Rounds the edges.
     */

    void fillEdges();


    /**
     * Fills all outside fields with walls.
     */

    void fillOutside();


    /**
     * Called, when the user wants to change the level name, author, etc.
     */

    void changeProperties();


    /**
     * Called when the user wants to change the size of the level.
     */

    void changeSize();


    /**
     * Adjusts the size of the map.
     */

    void adjustSize();


    /**
     * Creates a random empty map.
     */

    void createEmptyMap();


    /**
     * Creates random goals.
     */

    void createGoals();


    /**
     * Creates random gems.
     */

    void createGems();


    /**
     * Creates a solvable map (or at least tries so).
     */

    void createSolvableMap();


    /**
     * Called when the user checks the validity of the map.
     */

    void checkValidity();


    /**
     * Tries to solve the map.
     */

    void isSolvable();


    /**
     * Called, whne the active element changed.
     */

    void activeElementChanged();


    /**
     * Called when the keys should be configured.
     */

    void showKeyDialog();


    /**
     * Called when the toolbar should be configured.
     */

    void showToolbarDialog();


    /**
     * Called when the toolbar should be updated.
     */

    void updateToolbar();


    /**
     * Called when a mouse drag started.
     */

    void mouseDragStarted();


    /**
     * Called when a mouse drag is in action.
     *
     * @param from The last field.
     * @param to The new field.
     */

    void mouseDragged(QPoint from, QPoint to);


    /**
     * Called when a mouse drag ends.
     */

    void mouseDragEnded();


signals:

    /**
     * Called when the level of the editor should be saved.
     *
     * @param level_editor The level editor of the level which should be saved.
     */

    void saveLevel(LevelEditor * level_editor);


    /**
     * Called when the editor is exited.
     *
     * @param level_editor The level editor which has exited.
     */

    void exited(LevelEditor * level_editor);


private:

    /**
     * Ask the user to save a not saved level when exiting.
     *
     * @param cancel_possible If true, cancellation is possible.
     */

    void saveUnsavedChanges(bool cancel_possible);


    /**
     * Creates the accelerators.
     */

    void createAccels();


    /**
     * Creates the actions.
     */

    void createActions();


    /**
     * Displays a message depending on the validity.
     *
     * @param validity The validity of the map.
     */

    void showValidityMessage(Map::MapValidity validity);


    /**
     * Updates the state of the undo and redo actions.
     */

    void updateUndoRedoState();


    /**
     * Inserts the new map in the map vector and updates the undo/redo stat.
     */

    void insertMap();


    /**
     * Updates the outside fields.
     */

    void updateOutside();


    /**
     * The level.
     */

    mutable Level m_level;


    /**
     * The original level.
     */

    Level m_original_level;


    /**
     * The collection number.
     */

    int m_collection_nr;


    /**
     * The level number.
     */

    int m_level_nr;


    /**
     * The map.
     */

    Map m_map;


    /**
     * The map widget.
     */

    MapWidget * m_map_widget;


    /**
     * The theme to use.
     */

    Theme m_theme;


    /**
     * Defines the active elements.
     */

    enum ActiveElement
    {
        /**
         * An empty field.
         */

        EMPTY,


        /**
         * A wall field.
         */

        WALL,


        /**
         * A goal field.
         */

        GOAL,


        /**
         * A keeper field.
         */

        KEEPER,


        /**
         * A gem field.
         */

        GEM
    };


    /**
     * The active element.
     */

    ActiveElement m_active_element;


    /**
     * The undo/redo pointer.
     */

    int m_undo_pointer;


    /**
     * The maps vector.
     */

    std::vector<CompressedMap> m_maps;


    /**
     * The undo action.
     */

    KAction * m_undo_action;


    /**
     * The redo action.
     */

    KAction * m_redo_action;


    /**
     * The empty field element.
     */

    KRadioAction * m_empty_field_action;


    /**
     * The wall element.
     */

    KRadioAction * m_wall_field_action;


    /**
     * The keeper element.
     */

    KRadioAction * m_keeper_field_action;


    /**
     * The goal element.
     */

    KRadioAction * m_goal_field_action;


    /**
     * The gem fields element.
     */

    KRadioAction * m_gem_field_action;


    /**
     * If true, the actual map was saved.
     */

    bool m_was_saved;


    /**
     * If true, the exited signal was send.
     */

    bool m_exited_signaled;


    /**
     * If true, then we are in a mouse drag operation.
     */

    bool m_in_drag;


    /**
     * The pieces of the map from which to create the images.
     */

    static int s_pieces[65];
};


#endif
