/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "level_generator.h"

#include <algorithm>
#include <cassert>
#include <list>

#include <krandomsequence.h>

#include "hash.h"
#include "movements.h"




bool LevelGenerator::s_was_initialized = false;
std::vector<int> LevelGenerator::s_empty_pattern_lengths;
std::vector<int> LevelGenerator::s_empty_pattern_offsets;
std::vector<int> LevelGenerator::s_x_empty_patterns;
std::vector<int> LevelGenerator::s_y_empty_patterns;


Map LevelGenerator::createEmptyMap(int width, int height, double fill)
{
    init();

    assert(width >= 3);
    assert(height >= 3);
    assert(width < 128);
    assert(height < 128);

    std::vector<int> raw_pieces(width * height, Map::WALL);
    Map map(width, height, raw_pieces);

    int const inner_width = width - 2;
    int const inner_height = height - 2;
    int walls = inner_width * inner_height;
    int needed_walls = static_cast<int> (walls * fill);

    if (walls - needed_walls == 0)
    {
        --needed_walls;
    }

    KRandomSequence rand;
    int const number_of_patterns = static_cast<int> (s_empty_pattern_lengths.size());
    bool is_first_try = true;

    while (walls > needed_walls)
    {
        int const pattern = rand.getLong(number_of_patterns);
        int const length = s_empty_pattern_lengths[pattern];
        int const offset = s_empty_pattern_offsets[pattern];
        int const x = rand.getLong(inner_width) + 1;
        int const y = rand.getLong(inner_height) + 1;

        bool has_contact = false;
        bool is_valid = true;

        for (int i = 0; i < length; ++i)
        {
            int const index = offset + i;
            int const real_x = x + s_x_empty_patterns[index];
            int const real_y = y + s_y_empty_patterns[index];

            if ( (real_x > inner_width) || (real_y > inner_height) )
            {
                is_valid = false;

                break;
            }

            if (!has_contact)
            {
                if (map.getPiece(QPoint(real_x - 1, real_y)) == Map::EMPTY)
                {
                    has_contact = true;
                }
                else if (map.getPiece(QPoint(real_x + 1, real_y)) == Map::EMPTY)
                {
                    has_contact = true;
                }
                else if (map.getPiece(QPoint(real_x, real_y - 1)) == Map::EMPTY)
                {
                    has_contact = true;
                }
                else if (map.getPiece(QPoint(real_x, real_y + 1)) == Map::EMPTY)
                {
                    has_contact = true;
                }
            }

            if (map.getPiece(QPoint(real_x, real_y)) == Map::EMPTY)
            {
                is_valid = false;

                break;
            }
        }

        if (is_valid && (has_contact || is_first_try))
        {
            is_first_try = false;

            for (int i = 0; i < length; ++i)
            {
                int const index = offset + i;
                int const real_x = x + s_x_empty_patterns[index];
                int const real_y = y + s_y_empty_patterns[index];

                if (map.getPiece(QPoint(real_x, real_y)) == Map::WALL)
                {
                    --walls;
                    map.setPiece(QPoint(real_x, real_y), Map::EMPTY);
                }
            }
        }
    }

    for (int i = width * height - 1; i >= 0; --i)
    {
        if (map.getPiece(i) == Map::EMPTY)
        {
            map.setPiece(i, Map::KEEPER);

            break;
        }
    }

    return map;
}


Map LevelGenerator::createGoals(Map const & map, int goals)
{
    int const width = map.width();
    int const height = map.height();
    int const size = width * height;

    Map result(width, height, map.pieces());

    int empty = 0;

    for (int i = 0; i < size; ++i)
    {
        int piece = result.getPiece(i);

        if (Map::pieceContainsGoal(piece))
        {
            result.setPiece(i, piece - 1);
        }

        if (piece < Map::WALL)
        {
            ++empty;
        }
    }

    int const needed_goals = std::min(empty - 1, goals);
    int act_goals = 0;

    KRandomSequence rand;

    while (act_goals < needed_goals)
    {
        int const index= rand.getLong(size);
        int const piece = result.getPiece(index);

        if (piece >= Map::WALL)
        {
            continue;
        }

        if (!Map::pieceContainsGoal(piece))
        {
            result.setPiece(index, piece + 1);
            ++act_goals;
        }
    }

    return result;
}


Map LevelGenerator::createGems(Map const & map)
{
    int const width = map.width();
    int const height = map.height();
    int const size = width * height;

    Map result(width, height, map.pieces());
    result.calcDeadlocks();
    removeGems(result);

    int empty = 0;
    int goals = 0;

    for (int i = 0; i < size; ++i)
    {
        int const piece = result.getPiece(i);

        if ((piece < Map::WALL) && !result.isDeadlock(i) && !Map::pieceContainsKeeper(piece))
        {
            ++empty;
        }

        if (Map::pieceContainsGoal(piece))
        {
            ++goals;
        }
    }

    goals = std::min(empty, goals);

    KRandomSequence rand;

    int act_gems = 0;

    while (act_gems < goals)
    {
        int const index= rand.getLong(size);
        int const piece = result.getPiece(index);

        if ((piece >= Map::WALL) || result.isDeadlock(index) || Map::pieceContainsKeeper(piece))
        {
            continue;
        }

        if (!Map::pieceContainsGem(piece))
        {
            result.setPiece(index, piece - 2);
            ++act_gems;
        }
    }

    result.clearDeadlocks();

    return result;
}


void LevelGenerator::init()
{
    if (!s_was_initialized)
    {
        s_was_initialized = true;

        int index = 0;

        // #
        int length = 1;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        index += length;

        // ##
        length = 2;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        index += length;

        // #
        // #
        length = 2;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        index += length;

        // ##
        //  #
        length = 3;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        index += length;

        //  #
        // ##
        length = 3;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        index += length;

        // #
        // ##
        length = 3;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        index += length;

        // ##
        // #
        length = 3;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        index += length;

        // ###
        //  #
        length = 4;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(2);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        index += length;

        //  #
        // ###
        length = 4;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(2);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        index += length;

        // #
        // ##
        // #
        length = 4;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(2);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        index += length;

        //  #
        // ##
        //  #
        length = 4;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(2);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        index += length;

        // ###
        // ###
        length = 6;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(2);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(2);
        s_y_empty_patterns.push_back(1);
        index += length;

        // ##
        // ##
        // ##
        length = 6;
        s_empty_pattern_lengths.push_back(length);
        s_empty_pattern_offsets.push_back(index);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(0);
        s_y_empty_patterns.push_back(2);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(0);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(1);
        s_x_empty_patterns.push_back(1);
        s_y_empty_patterns.push_back(2);
        index += length;
    }
}


Movements LevelGenerator::createSolution(Map const & init_map, int pushes, int max_effort, Map & result_map)
{
    int const width = init_map.width();
    int const height = init_map.height();
    int const size = width * height;

    result_map = Map(width, height, init_map.pieces());
    removeGems(result_map);
    Movements moves;

    std::vector<int> non_goal_positions;

    for (int i = 0; i < size; ++i)
    {
        int const piece = result_map.getPiece(i);

        if ((piece < Map::WALL) && !Map::pieceContainsGoal(piece))
        {
            non_goal_positions.push_back(i);
        }
    }

    if (non_goal_positions.empty())
    {
        return moves;
    }

    KRandomSequence rand;
    int const keeper_pos = rand.getLong(static_cast<long> (non_goal_positions.size()));

    if (!result_map.containsKeeper(non_goal_positions[keeper_pos]))
    {
        result_map.setKeeper(non_goal_positions[keeper_pos]);
    }

    std::vector<int> gem_positions;

    for (int i = 0; i < size; ++i)
    {
        int const piece = result_map.getPiece(i);

        if (Map::pieceContainsGoal(piece))
        {
            if (Map::pieceContainsKeeper(piece))
            {
                return moves;
            }

            result_map.setPiece(i, Map::GEM_ON_GOAL);
            gem_positions.push_back(i);
        }
    }

    int const nr_of_gems = static_cast<int> (gem_positions.size());

    result_map.calcDeadlocks();
    std::vector<int> hash_positions;

    for (int i = 0; i < size; ++i)
    {
        int const piece = result_map.getPiece(i);

        if ((piece < Map::WALL) && !result_map.isDeadlock(i))
        {
            hash_positions.push_back(i);
        }
    }

    std::vector<Hash> visited;
    visited.push_back(Hash(result_map, hash_positions));

    int pushes_made = 0;
    int act_effort = 0;
    int const xy_offsets[4] = {-1, 1, -width, width};
    std::vector<std::list<int> > candidate_moves;
    std::vector<int> old_gem_positions;
    std::vector<int> old_gem_indices;

    while (pushes_made < pushes)
    {
        result_map.calcReachable();
        candidate_moves.push_back(std::list<int> ());

        for (int i = 0; i < nr_of_gems; ++i)
        {
            int const pos = gem_positions[i];

            for (int j = 0; j < 4; ++j)
            {
                int const nb_index = pos + xy_offsets[j];

                if (result_map.isReachable(nb_index) && !result_map.isDeadlock(nb_index) &&
                    result_map.canDropGem(nb_index))
                {
                    int keeper_index = nb_index + xy_offsets[j];

                    if (result_map.canDropKeeper(keeper_index))
                    {
                        if (rand.getLong(2) == 0)
                        {
                            candidate_moves.back().push_back(i * 4 + j);
                        }
                        else
                        {
                            candidate_moves.back().push_front(i * 4 + j);
                        }
                    }
                }
            }
        }

        if (candidate_moves.back().empty())
        {
            while (candidate_moves.back().empty())
            {
                candidate_moves.pop_back();

                if (candidate_moves.empty())
                {
                    return Movements();
                }

                visited.pop_back();
                --pushes_made;

                Move const backup_move = moves.prevMove();
                result_map.doUndoMove(backup_move, true);

                int const backup_moved_gem = old_gem_indices.back();
                old_gem_indices.pop_back();
                int const backup_gem_position = old_gem_positions.back();
                old_gem_positions.pop_back();
                gem_positions[backup_moved_gem] = backup_gem_position;
            }
        }

        while (true)
        {
            ++act_effort;

            if (act_effort >= max_effort)
            {
                return Movements();
            }

            int const raw_move = candidate_moves.back().front();
            int const moved_gem = raw_move >> 2;
            int const offset = xy_offsets[raw_move & 3];
            int const from_index = gem_positions[moved_gem] + offset;
            QPoint const from = result_map.getPoint(from_index);
            QPoint const to = result_map.getPoint(from_index + offset);
            Move const move(from, to, true);

            result_map.doMove(move, true);
            Hash const hash(result_map, hash_positions);

            candidate_moves.back().pop_front();

            if (std::find(visited.begin(), visited.end(), hash) != visited.end())
            {
                result_map.doUndoMove(move, true);

                while (candidate_moves.back().empty())
                {
                    if (candidate_moves.empty())
                    {
                        return Movements();
                    }
                    else
                    {
                        candidate_moves.pop_back();

                        if (candidate_moves.empty())
                        {
                            return Movements();
                        }

                        visited.pop_back();
                        --pushes_made;

                        Move const backup_move = moves.prevMove();
                        result_map.doUndoMove(backup_move, true);

                        int const backup_moved_gem = old_gem_indices.back();
                        old_gem_indices.pop_back();
                        int const backup_gem_position = old_gem_positions.back();
                        old_gem_positions.pop_back();
                        gem_positions[backup_moved_gem] = backup_gem_position;
                    }
                }
            }
            else
            {
                moves.addMove(move);
                old_gem_positions.push_back(gem_positions[moved_gem]);
                old_gem_indices.push_back(moved_gem);
                visited.push_back(hash);
                gem_positions[moved_gem] = from_index;
                ++pushes_made;

                break;
            }
        }
    }

    Movements real_moves;

    while (moves.hasPrevMove())
    {
        Move const move = moves.prevMove();

        real_moves.addMove(Move(move.to(), move.from(), true));
    }

    return real_moves;
}


void LevelGenerator::removeGems(Map & map)
{
    int const width = map.width();
    int const height = map.height();
    int const size = width * height;

    for (int i = 0; i < size; ++i)
    {
        int const piece = map.getPiece(i);

        if (Map::pieceContainsGem(piece))
        {
            map.setPiece(i, piece + 2);
        }
    }
}
