/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_LEVEL_GENERATOR_INC_GUARD_H
#define EASYSOK_LEVEL_GENERATOR_INC_GUARD_H


#include <vector>

#include "map.h"




/**
 * Here we have functions to generate levels.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class LevelGenerator
{
public:

    /**
     * Creates a map without gems or goals.
     *
     * @param width The width of the map (must be 3 or greater and smaller than 128).
     * @param height The height of the map (must be 3 or greater and smaller than 128).
     * @param fill A parameter (in the range [0:1] which determines how much walls
     *             will be in the map. The higher the value, the more walls.
     */

    static Map createEmptyMap(int width, int height, double fill);


    /**
     * Fills the map with random goals.
     *
     * All original goals are deleted.
     *
     * @param map The map to use.
     * @param goals The maximum number of goals.
     */

    static Map createGoals(Map const & map, int goals);


    /**
     * Fills the map with gems.
     *
     * All original gems are deleted.
     *
     * @param map The map to use.
     */

    static Map createGems(Map const & map);


    /**
     * Tries to create a map with the given number of pushes to find a solution.
     *
     * @param init_map The initial map.
     * @param pushes The desired number of pushes.
     * @param max_effort The maximum number of pushes to try until we consider to have failed.
     * @param result_map The resulting start map will be stored here.
     * @return The moves to solve the map or no moves at all if none could be made.
     */

    static Movements createSolution(Map const & init_map, int pushes, int max_effort, Map & result_map);


private:

    /**
     * Initializes the class.
     */

    static void init();


    /**
     * Removes all gems from the given map.
     *
     * @param map The map to use.
     */

    static void removeGems(Map & map);


    /**
     * If true, init has ben called.
     */

    static bool s_was_initialized;


    /**
     * The lengths of the empty patterns.
     */

    static std::vector<int> s_empty_pattern_lengths;


    /**
     * The offsets of the empty patterns.
     */

    static std::vector<int> s_empty_pattern_offsets;


    /**
     * The x-coordinates of the empty patterns.
     */

    static std::vector<int> s_x_empty_patterns;


    /**
     * The y-coordinates of the empty patterns.
     */

    static std::vector<int> s_y_empty_patterns;
};

#endif
