/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "main_window.h"

#include <algorithm>
#include <cassert>
#include <map>

#include <kaccel.h>
#include <kaction.h>
#include <kapp.h>
#include <kconfig.h>
#include <kdialogbase.h>
#include <kkeydialog.h>
#include <klineeditdlg.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kprinter.h>
#include <kstatusbar.h>
#include <kstddirs.h>
#include <ktip.h>

#include <qbuffer.h>
#include <qclipboard.h>
#include <qdatetime.h>
#include <qmessagebox.h>
#include <qpaintdevicemetrics.h>
#include <qpainter.h>
#include <qstringlist.h>
#include <qtextview.h>
#include <qtimer.h>
#include <qvbox.h>

#include "atomic_move.h"
#include "bookmarks.h"
#include "change_user_dialog.h"
#include "collection.h"
#include "collection_holder.h"
#include "collection_properties_dialog.h"
#include "compressed_map.h"
#include "create_user_dialog.h"
#include "delete_by_name_dialog.h"
#include "difficulty.h"
#include "duplicate_level_finder.h"
#include "game.h"
#include "hash.h"
#include "highscore_dialog.h"
#include "image_storer.h"
#include "import_user_dialog.h"
#include "info_dialog.h"
#include "ksokoban_bookmark_dialog.h"
#include "level_editor.h"
#include "level_properties_dialog.h"
#include "level_selection_dialog.h"
#include "map_widget.h"
#include "move_optimizer.h"
#include "pixmap_provider.h"
#include "progress_dialog.h"
#include "push_optimizer.h"
#include "reorder_dialog.h"
#include "set_user_dialog.h"
#include "solution_annotate_dialog.h"
#include "solution_delete_dialog.h"
#include "solution_holder.h"
#include "solution_optimize_dialog.h"
#include "solution_select_dialog.h"
#include "solver.h"
#include "solver_dialog.h"
#include "theme_holder.h"
#include "xsb_mime_source.h"




MainWindow::MainWindow(QWidget * parent, char const * name) :
    KMainWindow(parent, name),
    m_game(0),
    m_map(0),
    m_collection_menu(0),
    m_theme_menu(0),
    m_bookmark_set_menu(0),
    m_bookmark_goto_menu(0),
    m_state_timer(new QTimer(this)),
    m_recent_collections_menu(0)
{
    connect(m_state_timer, SIGNAL(timeout()), this, SLOT(saveState()));

    readConfig();
    setupActions();

    setupCollectionMenu();
    setupRecentCollectionsMenu();
    setupThemeMenu();
    setupBookmarkMenu();
    setStandardToolBarMenuEnabled(true);

    m_map_widget = new MapWidget(0, getPixmapProvider(), ThemeHolder::theme(m_theme), false, this);
    connect(m_map_widget, SIGNAL(fieldClicked(QPoint)), this, SLOT(moveKeeper(QPoint)));
    connect(m_map_widget, SIGNAL(keeperMoved(QPoint, QPoint)), this, SLOT(moveKeeper(QPoint, QPoint)));
    connect(m_map_widget, SIGNAL(gemMoved(QPoint, QPoint)), this, SLOT(moveGem(QPoint, QPoint)));
    connect(m_map_widget, SIGNAL(undo()), this, SLOT(undo()));
    connect(m_map_widget, SIGNAL(redo()), this, SLOT(redo()));
    setCentralWidget(m_map_widget);

    KStatusBar * status_bar = statusBar();
    status_bar->insertItem("", 0, 1);
    status_bar->insertItem("", 1, 0);
    status_bar->insertItem("", 2, 0, true);
    status_bar->setItemAlignment(0, AlignVCenter | AlignLeft);
    status_bar->setItemAlignment(1, AlignVCenter | AlignLeft);
    status_bar->setItemAlignment(2, AlignVCenter | AlignRight);

    createGUI();

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    if (!config->readBoolEntry("Toolbar visible")) {
	toolBar()->hide();
    }

    setLevel(m_collection_nr, m_level_nr, true);
    updateUserStatusBar();

    KTipDialog::showTip(this, KGlobal::dirs()->findResource("data", "easysok/tips"), false);

    QTimer::singleShot(0, this, SLOT(statusBarUpdateHack()));
}


void MainWindow::statusBarUpdateHack()
{
    // Make the status bar appear correctly.
    resize(width() - 1, height());
    resize(width() + 1, height());
}


MainWindow::~MainWindow()
{
    delete m_game;
    delete m_map;
}


bool MainWindow::queryExit()
{
    saveCurrentLevelState();
    writeConfig();

    return true;
}


void MainWindow::setLevel(int collection_nr, int level_nr, bool force_update, bool dont_save_act_level)
{
    assert(collection_nr >= 0);
    assert(collection_nr < CollectionHolder::numberOfCollections());
    assert(level_nr >= -1);
    assert(level_nr < CollectionHolder::collection(collection_nr)->numberOfLevels());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    if (level_nr == -1)
    {
        QString name = CollectionHolder::collection(collection_nr)->name();
        level_nr = config->readNumEntry("Last level in collection " + name, 0);
        int const max_nr = CollectionHolder::collection(collection_nr)->numberOfLevels() - 1;
        level_nr = std::min(std::max(0, level_nr), max_nr);
    }

    bool const is_new_collection = (m_collection_nr != collection_nr);

    if (is_new_collection)
    {
        m_recent_collections.insert(m_recent_collections.begin(), collection_nr);
        removeDoubleRecentCollections();
        setupRecentCollectionsMenu();
    }

    bool const is_new_level = (is_new_collection || (m_level_nr != level_nr)) || force_update;

    if (dont_save_act_level == false)
    {
        if (is_new_level)
        {
            if (m_game != 0)
            {
                m_game->emptyMoveQueue();
            }
        }

        if (is_new_collection)
        {
            config->writeEntry("Last level in collection " + actCollection()->name(), m_level_nr);
        }

        if (m_game != 0)
        {
            saveCurrentLevelState();
        }
    }

    m_collection_nr = collection_nr;
    m_level_nr = level_nr;

    m_collection_actions[m_collection_nr]->setChecked(true);
    m_collection_by_author_actions[m_collection_nr]->setChecked(true);
    m_collection_by_difficulty_actions[m_collection_nr]->setChecked(true);

    CompressedMap const & new_compressed_map = actLevel().compressedMap();
    Movements moves;

    if (SolutionHolder::hasLastAttempt(new_compressed_map))
    {
        moves = SolutionHolder::lastAttempt(new_compressed_map);
    }

    m_is_solved = false;
    m_last_was_virtual = false;
    m_was_solved = SolutionHolder::hasSolution(new_compressed_map);
    updateWasSolvedActions();

    if (m_was_solved)
    {
        int const index = SolutionHolder::getIndexForMap(new_compressed_map);
        assert(index != -1);

        int const move_solution = SolutionHolder::moveBestSolution(index);
        m_best_number_of_moves = SolutionHolder::movesInSolution(index, move_solution);

        int const push_solution = SolutionHolder::pushBestSolution(index);
        m_best_number_of_pushes = SolutionHolder::pushesInSolution(index, push_solution);

        int const linear_push_solution = SolutionHolder::linearPushBestSolution(index);
        m_best_number_of_linear_pushes = SolutionHolder::linearPushesInSolution(index, linear_push_solution);

        int const gem_change_solution = SolutionHolder::gemChangeBestSolution(index);
        m_best_number_of_gem_changes = SolutionHolder::gemChangesInSolution(index, gem_change_solution);
    }
    else
    {
        m_best_number_of_moves = 0;
        m_best_number_of_pushes = 0;
        m_best_number_of_linear_pushes = 0;
        m_best_number_of_gem_changes = 0;
    }

    delete m_map;
    m_map = new Map(actLevel().map());

    if (m_game == 0)
    {
        m_game = new Game(m_map, moves);
        m_map_widget->setKeeperDirection(m_game->keeperDirection());
        m_map_widget->setMap(m_map);

        connect(m_game, SIGNAL(mapChanged()), m_map_widget, SLOT(updateDisplay()));
        connect(m_game, SIGNAL(arrowsChanged(std::vector<Move> const &)),
                m_map_widget, SLOT(setArrows(std::vector<Move> const &)));
        connect(m_game, SIGNAL(keeperDirectionChanged(AtomicMove::AtomicMoveType)),
                m_map_widget, SLOT(setKeeperDirection(AtomicMove::AtomicMoveType)));
        connect(m_game, SIGNAL(virtualKeeperChanged(const QPoint &)),
                m_map_widget, SLOT(setVirtualKeeper(const QPoint &)));
        connect(m_game, SIGNAL(movesChanged(bool)), this, SLOT(updateStatusBar(bool)));
        connect(m_game, SIGNAL(wasSolved()), this, SLOT(solved()));
        connect(m_game, SIGNAL(wasUnsolved()), this, SLOT(unsolved()));
        connect(m_game, SIGNAL(wasSolvedByHand()), this, SLOT(solvedByHand()));

        m_game->setHonorDeadlock(m_honor_deadlocks);
        setShowArrows();
        m_game->setAnimation(m_animation_speed);
    }
    else
    {
        m_game->preventUpdate(true);
        m_game->setMapAndMoves(m_map, moves);
        m_game->preventUpdate(false);
        m_map_widget->setKeeperDirection(m_game->keeperDirection());
        m_map_widget->setMap(m_map);
        setShowArrows();
    }

    int const nr_of_collections = CollectionHolder::numberOfCollections();

    for (int i = 0; i < nr_of_collections; ++i)
    {
        const bool is_active = (i == m_collection_nr);

        m_collection_actions[i]->setChecked(is_active);
        m_collection_by_author_actions[i]->setChecked(is_active);
        m_collection_by_difficulty_actions[i]->setChecked(is_active);
    }

    KStatusBar * status_bar = statusBar();

    QString text = " ";
    QString const level_author = actLevel().authorLine();
    QString const collection_name = actCollection()->name();

    int const levels_in_collection = actCollection()->numberOfLevels();
    QString const level_name = actLevel().name();

    if (actCollection()->numberOfLevels() > 1)
    {
        if (level_name.isEmpty() || (level_name == collection_name))
        {
            text += i18n("Level %1 of %2").arg(m_level_nr + 1).arg(levels_in_collection) + " --- ";
        }
        else
        {
            QString add_text = i18n("%1 (%2 of %3)");
            add_text = add_text.arg(level_name).arg(m_level_nr + 1).arg(levels_in_collection);
            text += add_text + " --- ";
        }

        text += " ";
    }

    if (level_author.isEmpty() || !m_show_author)
    {
        text += collection_name;
    }
    else
    {
        text += i18n("%1 (%2)").arg(collection_name).arg(level_author);
    }

    if (m_show_difficulty)
    {
        text += ", ";
        text += i18n("Difficulty:") + " ";

        text += Difficulty::text(actLevel().difficulty());
    }

    text += " ";

    status_bar->changeItem(text, 0);

    updateStatusBar(true);
    updateUndoRedoActions();
    updateLevelActions();

    m_map_widget->updateDisplay();
}


void MainWindow::prevCollection()
{
    if (m_collection_nr > 0)
    {
        setLevel(m_collection_nr - 1, -1);
        m_collection_actions[m_collection_nr]->setChecked(true);
    }
}


void MainWindow::nextCollection()
{
    if (m_collection_nr < CollectionHolder::numberOfCollections() - 1)
    {
        setLevel(m_collection_nr + 1, -1);
        m_collection_actions[m_collection_nr]->setChecked(true);
    }
}


void MainWindow::prevLevel()
{
    if (m_level_nr > 0)
    {
        setLevel(m_collection_nr, m_level_nr - 1);
    }
}


void MainWindow::nextLevel()
{
    if (m_level_nr < actCollection()->numberOfLevels() - 1)
    {
        if (m_goto_any_level || m_was_solved)
        {
            setLevel(m_collection_nr, m_level_nr + 1);
        }
    }
}


void MainWindow::selectLevel()
{
    LevelSelectionDialog dialog(m_level_nr, lastLegalLevel());

    if (dialog.exec())
    {
        setLevel(m_collection_nr, dialog.levelSelected());
    }
}


void MainWindow::firstLevel()
{
    setLevel(m_collection_nr, 0);
}


void MainWindow::lastLevel()
{
    if (m_goto_any_level)
    {
        setLevel(m_collection_nr, actCollection()->numberOfLevels() - 1);
    }
    else
    {
        int const number_of_levels = actCollection()->numberOfLevels();

        for (int i = m_level_nr + 1; i < number_of_levels; ++i)
        {
            if (!SolutionHolder::hasSolution(actCollection()->level(i).compressedMap()))
            {
                setLevel(m_collection_nr, i);

                break;
            }
        }
    }
}


void MainWindow::moveLeft()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryMove(AtomicMove(AtomicMove::LEFT)))
        {
            moveVirtualLeft();
        }
    }
    else
    {
        m_game->tryMove(AtomicMove(AtomicMove::LEFT));
        updateUndoRedoActions();
    }
}


void MainWindow::moveRight()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryMove(AtomicMove(AtomicMove::RIGHT)))
        {
            moveVirtualRight();
        }
    }
    else
    {
        m_game->tryMove(AtomicMove(AtomicMove::RIGHT));
        updateUndoRedoActions();
    }
}


void MainWindow::moveUp()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryMove(AtomicMove(AtomicMove::UP)))
        {
            moveVirtualUp();
        }
    }
    else
    {
        m_game->tryMove(AtomicMove(AtomicMove::UP));
        updateUndoRedoActions();
    }
}


void MainWindow::moveDown()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryMove(AtomicMove(AtomicMove::DOWN)))
        {
            moveVirtualDown();
        }
    }
    else
    {
        m_game->tryMove(AtomicMove(AtomicMove::DOWN));
        updateUndoRedoActions();
    }
}


void MainWindow::moveFarLeft()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryFarMove(AtomicMove(AtomicMove::LEFT)))
        {
            moveVirtualLeft();
        }
    }
    else
    {
        m_game->tryFarMove(AtomicMove(AtomicMove::LEFT));
        updateUndoRedoActions();
    }
}


void MainWindow::moveFarRight()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryFarMove(AtomicMove(AtomicMove::RIGHT)))
        {
            moveVirtualRight();
        }
    }
    else
    {
        m_game->tryFarMove(AtomicMove(AtomicMove::RIGHT));
        updateUndoRedoActions();
    }
}


void MainWindow::moveFarUp()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryFarMove(AtomicMove(AtomicMove::UP)))
        {
            moveVirtualUp();
        }
    }
    else
    {
        m_game->tryFarMove(AtomicMove(AtomicMove::UP));
        updateUndoRedoActions();
    }
}


void MainWindow::moveFarDown()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryFarMove(AtomicMove(AtomicMove::DOWN)))
        {
            moveVirtualDown();
        }
    }
    else
    {
        m_game->tryFarMove(AtomicMove(AtomicMove::DOWN));
        updateUndoRedoActions();
    }
}


void MainWindow::moveVirtualLeft()
{
    m_last_was_virtual = !m_game->moveVirtualKeeper(AtomicMove(AtomicMove::LEFT));
    updateUndoRedoActions();
}


void MainWindow::moveVirtualRight()
{
    m_last_was_virtual = !m_game->moveVirtualKeeper(AtomicMove(AtomicMove::RIGHT));
    updateUndoRedoActions();
}


void MainWindow::moveVirtualUp()
{
    m_last_was_virtual = !m_game->moveVirtualKeeper(AtomicMove(AtomicMove::UP));
    updateUndoRedoActions();
}


void MainWindow::moveVirtualDown()
{
    m_last_was_virtual = !m_game->moveVirtualKeeper(AtomicMove(AtomicMove::DOWN));
    updateUndoRedoActions();
}


void MainWindow::pullMoveLeft()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryPullMove(AtomicMove(AtomicMove::LEFT)))
        {
            moveVirtualLeft();
        }
    }
    else
    {
        m_game->tryPullMove(AtomicMove(AtomicMove::LEFT));
        updateUndoRedoActions();
    }
}


void MainWindow::pullMoveRight()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryPullMove(AtomicMove(AtomicMove::RIGHT)))
        {
            moveVirtualRight();
        }
    }
    else
    {
        m_game->tryPullMove(AtomicMove(AtomicMove::RIGHT));
        updateUndoRedoActions();
    }
}


void MainWindow::pullMoveUp()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryPullMove(AtomicMove(AtomicMove::UP)))
        {
            moveVirtualUp();
        }
    }
    else
    {
        m_game->tryPullMove(AtomicMove(AtomicMove::UP));
        updateUndoRedoActions();
    }
}


void MainWindow::pullMoveDown()
{
    if (m_illegal_as_virtual)
    {
        if (m_last_was_virtual || !m_game->tryPullMove(AtomicMove(AtomicMove::DOWN)))
        {
            moveVirtualDown();
        }
    }
    else
    {
        m_game->tryPullMove(AtomicMove(AtomicMove::DOWN));
        updateUndoRedoActions();
    }
}


void MainWindow::moveKeeper(QPoint point)
{
    m_game->tryMoveKeeper(m_map->keeper(), point);
    updateUndoRedoActions();
}


void MainWindow::moveKeeper(QPoint from, QPoint to)
{
    m_game->tryMoveKeeper(from, to);
    updateUndoRedoActions();
}


void MainWindow::moveGem(QPoint from, QPoint to)
{
    m_game->tryMoveGem(from, to);
    updateUndoRedoActions();
}


void MainWindow::undo()
{
    m_game->tryUndo();
    updateUndoRedoActions();
}


void MainWindow::redo()
{
    m_game->tryRedo();
    updateUndoRedoActions();
}


void MainWindow::jumpToStart()
{
    m_game->jumpToStart();
    updateUndoRedoActions();
}


void MainWindow::jumpToEnd()
{
    m_game->jumpToEnd();
    updateUndoRedoActions();
}


void MainWindow::loadBestSolution()
{
    if (m_was_solved)
    {
        m_game->setMoves(SolutionHolder::movements(actLevel().compressedMap(), 0));
    }

    updateUndoRedoActions();
}


void MainWindow::loadASolution()
{
    if (!m_was_solved)
    {
        KMessageBox::error(this, i18n("This level was not solved yet!"));

        return;
    }

    CompressedMap const compressed_map = actLevel().compressedMap();
    int index = SolutionHolder::getIndexForMap(compressed_map);

    SolutionSelectDialog dialog(index, true, this);

    if (dialog.exec())
    {
        m_game->setMoves(SolutionHolder::movements(index, dialog.selectedSolution()));
        updateUndoRedoActions();
    }
}


void MainWindow::annotateSolution()
{
    if (m_was_solved)
    {
        SolutionAnnotateDialog dialog(actLevel().compressedMap());
        dialog.exec();
    }
    else
    {
        KMessageBox::error(this, i18n("This level was not solved yet!"));
    }
}


void MainWindow::setCollection(int index)
{
    assert(index >= 0);
    assert(index < CollectionHolder::numberOfCollections());

    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QString const collection_name = CollectionHolder::collection(index)->name();
    int const last_level = config->readNumEntry("Last level in collection " + collection_name, 0);

    setLevel(index, last_level);
}


void MainWindow::changeAnimations(int speed)
{
    m_animation_speed = speed;
    m_game->setAnimation(m_animation_speed);
}


void MainWindow::hideGemsChanged()
{
    setShowArrows();

    setTheme(m_theme);
}


void MainWindow::hideGoalsChanged()
{
    m_hide_goals = m_hide_goals_action->isChecked();

    setTheme(m_theme);
}


void MainWindow::outsideAsWallChanged()
{
    m_outside_as_wall = m_outside_as_wall_action->isChecked();

    setTheme(m_theme);
}


void MainWindow::honorDeadlockChanged()
{
    m_honor_deadlocks = m_honor_deadlock_action->isChecked();
    m_game->setHonorDeadlock(m_honor_deadlocks);
    m_map_widget->updateDisplay();
}


void MainWindow::setTheme(int theme)
{
    assert(theme >= 0);
    assert(theme < ThemeHolder::numberOfThemes());

    m_theme = theme;

    Theme * act_theme = ThemeHolder::theme(m_theme);
    act_theme->setHideGems(m_hide_gems);
    act_theme->setHideGoals(m_hide_goals);
    act_theme->setOutsideAsWall(m_outside_as_wall);

    m_map_widget->setPixmapProvider(getPixmapProvider(), act_theme);
}


void MainWindow::nextTheme()
{
    int new_theme = m_theme + 1;

    if (new_theme >= ThemeHolder::numberOfThemes())
    {
        new_theme = 0;
    }

    setTheme(new_theme);
    m_theme_actions[new_theme]->setChecked(true);
}


void MainWindow::previousTheme()
{
    int new_theme = m_theme - 1;

    if (new_theme < 0)
    {
        new_theme = ThemeHolder::numberOfThemes() - 1;
    }

    setTheme(new_theme);
    m_theme_actions[new_theme]->setChecked(true);
}


void MainWindow::playMoves()
{
    m_game->playMoves(true);
    updateUndoRedoActions();
}


void MainWindow::showKeyDialog()
{
    KKeyDialog::configureKeys(actionCollection(), xmlFile());
}


void MainWindow::showSolutions()
{
    if (!m_was_solved)
    {
        KMessageBox::error(this, i18n("This level was not solved yet!"));

        return;
    }

    CompressedMap const compressed_map = actLevel().compressedMap();
    int index = SolutionHolder::getIndexForMap(compressed_map);

    SolutionSelectDialog dialog(index, false, this);

    dialog.exec();
}


void MainWindow::showHighscores()
{
    HighscoreDialog dialog(m_collection_nr);

    if (dialog.exec())
    {
        int const selected_level = dialog.selectedLevel();

        if (selected_level <= lastLegalLevel())
        {
            setLevel(m_collection_nr, selected_level);
        }
    }
}


void MainWindow::showPushesChanged()
{
    m_show_arrows = m_show_arrow_action->isChecked();

    setShowArrows();

    m_map_widget->updateDisplay();
}


void MainWindow::setBookmark(int index)
{
    if (m_game->retroMode())
    {
        KMessageBox::error(this, i18n("Bookmarks in reto mode are not supported!"));

        return;
    }

    QString annotation;

    if (Bookmarks::hasBookmark(index))
    {
        KConfig * config = KApplication::kApplication()->config();
        KConfigGroupSaver group_saver(config, "Notification Messages");

        if (config->readBoolEntry("Bookmark overwrite", false))
        {
            QString const warning = i18n("Bookmark exists! Do you want to overwrite?");

            int const result = KMessageBox::warningContinueCancel(0, warning, i18n("Overwrite Bookmark"),
                                                                  i18n("OK"), "Bookmark overwrite", false);

            if (result == KMessageBox::Cancel)
            {
                return;
            }
        }

	annotation = Bookmarks::annotation(index);
    }

    KLineEditDlg dialog(i18n("Enter bookmark annotation"), annotation, this);

    if (dialog.exec())
    {
        Bookmarks::replaceBookmark(index, dialog.text(), actCollection()->name(), m_level_nr,
                                   actLevel().compressedMap(), m_game->moves());

        setupBookmarkMenuEntry(index);
    }
}


void MainWindow::gotoBookmark(int index)
{
    if (!Bookmarks::hasBookmark(index))
    {
        KMessageBox::error(0, i18n("Bookmark was not set!"));

        return;
    }

    QString collection_name = Bookmarks::collectionName(index);

    int const number_of_collections = CollectionHolder::numberOfCollections();
    int collection_nr = -1;

    for (int i = 0; i < number_of_collections; ++i)
    {
        if (CollectionHolder::collection(i)->name() == collection_name)
        {
            collection_nr = i;

            break;
        }
    }

    CompressedMap compressed_map = Bookmarks::map(index);

    if (collection_nr != -1)
    {
        Collection * collection = CollectionHolder::collection(collection_nr);
        int level_nr = Bookmarks::level(index);

        if ( (level_nr >= 0) && (level_nr < collection->numberOfLevels()) &&
             (collection->level(level_nr).compressedMap() == compressed_map) )
        {
            setLevel(collection_nr, level_nr);
            m_game->setMoves(Bookmarks::moves(index));
            jumpToEnd();

            return;
        }
    }

    for (int i = 0; i <  number_of_collections; ++i)
    {
        Collection * collection = CollectionHolder::collection(i);
        int const number_of_levels = collection->numberOfLevels();

        for (int j = 0; j < number_of_levels; ++j)
        {
            if (collection->level(j).compressedMap() == compressed_map)
            {
                setLevel(i, j);
                m_game->setMoves(Bookmarks::moves(index));
                jumpToEnd();

                return;
            }
        }
    }

    KMessageBox::error(0, i18n("Bookmark is corrupt!"));
}


void MainWindow::annotateBookmark(int index)
{
    if (!Bookmarks::hasBookmark(index))
    {
        KMessageBox::error(0, i18n("Bookmark was not set!"));

        return;
    }

    KLineEditDlg dialog(i18n("Enter new annotation"), Bookmarks::annotation(index), this);

    if (dialog.exec())
    {
        Bookmarks::replaceBookmark(index, dialog.text(), Bookmarks::collectionName(index),
                                   Bookmarks::level(index), Bookmarks::map(index), Bookmarks::moves(index));

        setupBookmarkMenuEntry(index);
    }
}


void MainWindow::importKSokobanBookmark(int index)
{
    bool found_ksokoban_bookmark = false;

    for (int i = 1; i <= 10; ++i)
    {
        if (Bookmarks::hasKSokobanBookmark(i))
        {
            found_ksokoban_bookmark = true;

            break;
        }
    }

    if (!found_ksokoban_bookmark)
    {
        KMessageBox::information(this, i18n("There are no KSokoban bookmarks to import!"));

        return;
    }

    KSokobanBookmarkDialog dialog(this);

    if (!dialog.exec())
    {
        return;
    }


    int bookmark = dialog.bookmark();

    if (!Bookmarks::hasKSokobanBookmark(bookmark))
    {
        KMessageBox::error(this, i18n("The selected KSokoban bookmark does not exists!"));

        return;
    }

    if (Bookmarks::hasBookmark(index))
    {
        KConfig * config = KApplication::kApplication()->config();
        KConfigGroupSaver group_saver(config, "Notification Messages");

        if (config->readBoolEntry("Bookmark overwrite", false))
        {
            QString const warning = i18n("Bookmark exists! Do you want to overwrite?");

            if (KMessageBox::warningContinueCancel(0, warning, i18n("Overwrite Bookmark"),
                                                   i18n("OK"), "Bookmark overwrite", false)
                == KMessageBox::Cancel)
            {
                return;
            }
        }
    }

    if (!Bookmarks::importKSokobanBookmark(bookmark, index))
    {
        KMessageBox::error(this, i18n("The KSokoban bookmark could not be imported!"));

        return;
    }

    setupBookmarkMenuEntry(index);
}


void MainWindow::minimizeMoves()
{
    MoveOptimizer optimizer;

    optimizeSolutions(0, CollectionHolder::numberOfCollections(), &optimizer);
}


void MainWindow::minimizeCollectionMoves()
{
    MoveOptimizer optimizer;

    optimizeSolutions(m_collection_nr, m_collection_nr + 1, &optimizer);
}


void MainWindow::minimizeLevelMoves()
{
    if (!m_was_solved)
    {
        return;
    }

    SolutionOptimizeDialog dialog(actLevel().compressedMap(), false, this);

    dialog.exec();
}


void MainWindow::minimizePushes()
{
    PushOptimizer optimizer;

    optimizeSolutions(0, CollectionHolder::numberOfCollections(), &optimizer);
}


void MainWindow::minimizeCollectionPushes()
{
    PushOptimizer optimizer;

    optimizeSolutions(m_collection_nr, m_collection_nr + 1, &optimizer);
}


void MainWindow::minimizeLevelPushes()
{
    if (!m_was_solved)
    {
        return;
    }

    SolutionOptimizeDialog dialog(actLevel().compressedMap(), true, this);

    dialog.exec();
}


void MainWindow::deleteSolutions()
{
    if (m_was_solved)
    {
        SolutionDeleteDialog dialog(actLevel().compressedMap());

        if (dialog.exec())
        {
            m_was_solved = SolutionHolder::hasSolution(actLevel().compressedMap());
            updateLevelActions();
        }
    }
}


void MainWindow::deleteSolutionsByName()
{
    DeleteByNameDialog dialog(this);

    if (!dialog.exec())
    {
        return;
    }

    QRegExp regexp(dialog.regExp());

    int deleted = 0;
    int found = 0;

    int const number_of_collections = CollectionHolder::numberOfCollections();

    for (int i = 0; i < number_of_collections; ++i)
    {
        Collection * collection = CollectionHolder::collection(i);
        int const number_of_levels = collection->numberOfLevels();

        for (int j = 0; j < number_of_levels; ++j)
        {
            CompressedMap const & compressed_map = collection->level(j).compressedMap();
            int const map_index = SolutionHolder::getIndexForMap(compressed_map);

            if ( (map_index != -1) && SolutionHolder::hasSolution(map_index) )
            {
                int const number_of_solutions = SolutionHolder::numberOfSolutions(map_index);

                for (int k = number_of_solutions - 1; k >= 0; --k)
                {
                    QString const annotation = SolutionHolder::infoOfSolution(map_index, k);

                    if (regexp.search(annotation, 0) != -1)
                    {
                        SolutionHolder::deleteSolution(map_index, k);
                        ++deleted;
                    }

                    ++found;
                }
            }
        }
    }

    KMessageBox::information(this, i18n("Deleted %1 of %2 solutions").arg(deleted).arg(found));
}


void MainWindow::stopAnimation()
{
    m_game->emptyMoveQueue();
    m_map_widget->updateDisplay();
}


void MainWindow::solved()
{
    m_is_solved = true;
}


void MainWindow::unsolved()
{
    m_is_solved = false;
}


void MainWindow::solvedByHand()
{
    Movements moves = m_game->moves();

    int number_of_pushes = m_game->numberOfPushes();
    int number_of_moves = m_game->numberOfMoves();
    int number_of_linear_pushes = moves.linearPushes();
    int number_of_gem_changes = moves.gemChanges();

    if (m_auto_optimize_pushes)
    {
        PushOptimizer optimizer(actLevel().map(), moves);
        moves = optimizer.moves();
        number_of_moves = optimizer.numberOfMoves();
        number_of_pushes = optimizer.numberOfPushes();
        number_of_linear_pushes = moves.linearPushes();
        number_of_gem_changes = moves.gemChanges();
    }
    else if (m_auto_optimize_moves)
    {
        MoveOptimizer optimizer(actLevel().map(), moves);
        moves = optimizer.moves();
        number_of_moves = optimizer.numberOfMoves();
        number_of_pushes = optimizer.numberOfPushes();
        number_of_linear_pushes = moves.linearPushes();
        number_of_gem_changes = moves.gemChanges();
    }

    if (!m_auto_send_to_server)
    {
        if (!m_was_solved)
        {
            QString msg = i18n("Congratulations! You solved the level for the first time!");
            KMessageBox::information(0, msg);
        }
        else if (number_of_pushes < m_best_number_of_pushes)
        {
            QString msg = i18n("Congratulations! You solved the level with fewer pushes than ever before!");
            KMessageBox::information(0, msg);
        }
        else if (number_of_moves < m_best_number_of_moves)
        {
            QString msg = i18n("Congratulations! You solved the level with fewer moves than ever before!");
            KMessageBox::information(0, msg);
        }
        else if (number_of_linear_pushes < m_best_number_of_linear_pushes)
        {
            QString msg = i18n("Congratulations! You solved the level with fewer linear "
                               "pushes than ever before!");
            KMessageBox::information(0, msg);
        }
        else if (number_of_gem_changes < m_best_number_of_gem_changes)
        {
            QString msg = i18n("Congratulations! You solved the level with fewer gem "
                               "changes than ever before!");
            KMessageBox::information(0, msg);
        }
    }
    else
    {
        sendSolutionToServer(actLevel().map(), moves, number_of_pushes, number_of_linear_pushes,
                             number_of_gem_changes, number_of_moves);
    }

    SolutionHolder::addSolution(actLevel().compressedMap(), moves, number_of_pushes,
                                number_of_linear_pushes, number_of_gem_changes, number_of_moves, "");

    if (number_of_pushes < m_best_number_of_pushes)
    {
        m_best_number_of_pushes = number_of_pushes;
    }

    if (number_of_moves < m_best_number_of_moves)
    {
        m_best_number_of_moves = number_of_moves;
    }

    if (number_of_linear_pushes < m_best_number_of_linear_pushes)
    {
        m_best_number_of_linear_pushes = number_of_linear_pushes;
    }

    if (number_of_gem_changes < m_best_number_of_gem_changes)
    {
        m_best_number_of_gem_changes = number_of_gem_changes;
    }

    m_is_solved = true;
    m_was_solved = true;

    updateWasSolvedActions();
    updateLevelActions();
    updateStatusBar(true);

    if (m_next_level_when_solved)
    {
        nextLevel();
    }
}


void MainWindow::autoOptChanged()
{
    m_auto_optimize_moves = m_auto_optimize_moves_action->isChecked();
    m_auto_optimize_pushes = m_auto_optimize_pushes_action->isChecked();
}


void MainWindow::retroModeChanged()
{
    if (m_retro_mode_action->isChecked() && m_map->containsGoal(m_map->keeper()))
    {
        KMessageBox::error(this, i18n("If you want to switch to the retro mode, make sure the "
                                      "keeper does not sits on a goal field!"));
        m_retro_mode_action->setChecked(false);
    }
    else
    {
        m_game->setRetroMode(m_retro_mode_action->isChecked());
    }
}


void MainWindow::editCurrentLevel()
{
    createLevelEditor(actLevel(), m_collection_nr, m_level_nr);
}


void MainWindow::insertNewLevel()
{
    std::vector<int> pieces(3 * 5, Map::WALL);
    pieces[6] = Map::KEEPER;
    pieces[7] = Map::GEM;
    pieces[8] = Map::GOAL;

    Map map(5, 3, pieces);
    Level level(map, QStringList(), QStringList(), "", "", "",  "", -1);

    createLevelEditor(level, m_collection_nr, -1);
}


void MainWindow::insertNewCollection()
{
    std::vector<int> pieces(3 * 5, Map::WALL);
    pieces[6] = Map::KEEPER;
    pieces[7] = Map::GEM;
    pieces[8] = Map::GOAL;

    Map map(5, 3, pieces);
    Level level(map, QStringList(), QStringList(), "", "", "",  "", -1);

    insertLevelIntoNewCollection(level);
}


void MainWindow::deleteCurrentLevel()
{
    if (actCollection()->numberOfLevels() == 1)
    {
        if ((CollectionHolder::numberOfPermanentCollections() == 1) &&
            !CollectionHolder::isTemporary(m_collection_nr))
        {
            KMessageBox::error(this, i18n("This is the last level in the collection.\n"
                                          "Deleting this level will also delete the collection.\n"
                                          "And this is the last permanent collection, so "
                                          "deleting it is not allowed."));

            return;
        }

        int result = KMessageBox::questionYesNo(this, i18n("This is the last level in the collection.\n"
                                                           "Deleting this level will also delete the "
                                                           "collection.\n"
                                                           "Do you really want to delete the level?"));
        if (result == KMessageBox::No)
        {
            return;
        }

        saveCurrentLevelState();
        XsbMimeSource * mime_source = new XsbMimeSource(actLevel());
        KApplication::clipboard()->setData(mime_source);

        CollectionHolder::removeCollection(m_collection_nr);
        m_collection_nr = std::min(CollectionHolder::numberOfCollections() - 1, m_collection_nr);
        setLevel(m_collection_nr, 0, true, true);
        setupCollectionMenu();
    }
    else
    {
        saveCurrentLevelState();
        XsbMimeSource * mime_source = new XsbMimeSource(actLevel());
        KApplication::clipboard()->setData(mime_source);

        actCollection()->removeLevel(m_level_nr);
        m_level_nr = std::min(actCollection()->numberOfLevels() - 1, m_level_nr);
        setLevel(m_collection_nr, m_level_nr, true, true);
    }
}


void MainWindow::deleteCurrentCollection()
{
    if ((CollectionHolder::numberOfPermanentCollections() == 1) &&
        !CollectionHolder::isTemporary(m_collection_nr))
    {
        KMessageBox::error(this, i18n("This is the last permanent collection, so "
                                      "deleting it is not allowed.\n"));

        return;
    }

    int result = KMessageBox::questionYesNo(this, i18n("Do you really want to delete the collection?"));

    if (result == KMessageBox::No)
    {
        return;
    }

    saveCurrentLevelState();
    CollectionHolder::removeCollection(m_collection_nr);
    m_collection_nr = std::max(0, m_collection_nr - 1);
    setLevel(m_collection_nr, 0, true, true);
    setupCollectionMenu();
}


void MainWindow::duplicateCurrentLevel()
{
    saveCurrentLevelState();
    actCollection()->insertLevel(actLevel(), m_level_nr);
    CollectionHolder::setModified();
    setLevel(m_collection_nr, 0, true, true);
    setupCollectionMenu();
}


void MainWindow::duplicateCurrentCollection()
{
    addCollection(*actCollection());
}


void MainWindow::copyCurrentLevel()
{
    XsbMimeSource * mime_source = new XsbMimeSource(actLevel());
    KApplication::clipboard()->setData(mime_source);
}


void MainWindow::pasteLevel()
{
    Level * level = levelFromClipboard();

    if (level == 0)
    {
        return;
    }

    saveCurrentLevelState();

    actCollection()->insertLevel(*level, m_level_nr);
    CollectionHolder::setModified();
    setLevel(m_collection_nr, m_level_nr, true, true);

    delete level;
}


void MainWindow::pasteLevelInNewCollection()
{
    Level * level = levelFromClipboard();

    if (level == 0)
    {
        return;
    }

    saveCurrentLevelState();

    int name_nr = 0;
    QString collection_name;

    do
    {
        ++name_nr;
        collection_name = i18n("Pasted %1").arg(name_nr);
    }
    while (CollectionHolder::indexFromName(collection_name) != -1);

    Collection * collection = new Collection(collection_name, level->authors(), level->emails(),
                                             level->homepage(), level->copyright(), level->info(),
                                             level->difficulty());

    collection->addLevel(*level);

    int collection_nr = CollectionHolder::addCollection(collection, true);


    setupCollectionMenu();
    setLevel(collection_nr, 0, true, true);

    delete level;
}


void MainWindow::makeCurrentCollectionNonTemporary()
{
    if (CollectionHolder::isTemporary(m_collection_nr))
    {
        CollectionHolder::setTemporary(m_collection_nr, false);
        setupCollectionMenu();

        KMessageBox::information(this, i18n("This collection is now non temporary."));
    }
    else
    {
        KMessageBox::error(this, i18n("This collection is already non temporary."));
    }

}


void MainWindow::solve()
{
    m_game->emptyMoveQueue();

    if (m_map->isSolved())
    {
        KMessageBox::information(this, i18n("The level is already solved!"));

        return;
    }

    if (m_map->numberOfGems() > 10)
    {
        if (KMessageBox::warningContinueCancel(this, i18n("The current map contains more than 10 gems.\n"
                                                          "Experience shows, that the solver is\n"
                                                          "almost always too dumb to solve maps of that\n"
                                                          "size. Do you really want to continue?"),
                                               QString::null, i18n("Continue"),
                                               "Warn when solving map with many gems",
                                               false) == KMessageBox::Cancel)
        {
            return;
        }
    }

    m_map_widget->tempDisableAutoHide(true);
    SolverDialog dialog(*m_map, this);

    if (dialog.exec() != QMessageBox::Cancel)
    {
        Solver const * solver = dialog.solver();
        Movements moves;

        QString data_text = i18n("Maximum search depth was: %1.").arg(solver->maxDepth()) + "\n";
        data_text += i18n("Number of examined positions was:") + " ";
        data_text += dialog.positionsExaminedAsText() + ".\n";


        if (solver->moves().isEmpty())
        {
            KMessageBox::information(this, i18n("No solution could be found!") + "\n\n" +
                                     i18n("You can view the best effort by using redo.") + data_text);
            moves = solver->bestEffort();
        }
        else
        {
            KMessageBox::information(this, i18n("Found a solution!") + "\n\n" +
                                     i18n("You can view the solution by using redo.") + "\n\n" + data_text);
            moves = solver->moves();
        }

        if (!moves.isEmpty())
        {
            Movements game_moves = m_game->moves();
            int const move_position = game_moves.movePointer();

            while (moves.hasNextMove())
            {
                game_moves.addMove(moves.nextMove());
            }

            game_moves.setMovePointer(move_position);
            m_game->setMoves(game_moves);
            updateUndoRedoActions();
        }
    }

    m_map_widget->tempDisableAutoHide(false);
}


void MainWindow::levelInfo()
{
    InfoDialog dialog(actLevel(), CollectionHolder::isTemporary(m_collection_nr), this);

    dialog.exec();
}


void MainWindow::collectionInfo()
{
    InfoDialog dialog(*actCollection(), CollectionHolder::isTemporary(m_collection_nr), this);

    dialog.exec();
}


void MainWindow::findDups()
{
    m_find_dups_timer = new QTimer;
    connect(m_find_dups_timer, SIGNAL(timeout()), this, SLOT(performFindDups()));

    m_duplicate_level_finder = new DuplicateLevelFinder();
    m_progress_dialog = new ProgressDialog(i18n("Find duplicate levels"), i18n("Percentage of levels checked"),
                                           m_duplicate_level_finder->steps(), this);

    m_find_dups_timer->start(0, true);

    m_progress_dialog->exec();

    if (m_duplicate_level_finder->finished())
    {
        if (m_duplicate_level_finder->foundDuplicates())
        {
            KMessageBox::information(this, i18n("No duplicate levels found!\n"));
        }
        else
        {
            KDialogBase dialog(this, 0, true, QString::null, KDialogBase::Close, KDialogBase::Close, true);
            new QTextView(m_duplicate_level_finder->text(), QString::null, dialog.makeVBoxMainWidget());

            dialog.resize(std::max(640, dialog.size().width()), dialog.size().height());
            dialog.exec();
        }
    }

    delete m_progress_dialog;
    delete m_duplicate_level_finder;
    delete m_find_dups_timer;
}


void MainWindow::performFindDups()
{
    if (m_duplicate_level_finder->process())
    {
        m_progress_dialog->close();
    }
    else
    {
        m_progress_dialog->advance();
        m_find_dups_timer->start(0, true);
    }
}


void MainWindow::isOldPosition()
{
    Movements moves = m_game->moves();
    moves.truncateToCurrent();
    moves.setToFirstPosition();

    Hash act_hash(*m_map);

    Map map(actLevel().map());

    bool found = false;

    while (moves.hasNextMove())
    {
        Move const move = moves.nextMove();

        if (move.stonePushed())
        {
            Hash hash(map);

            if (hash == act_hash)
            {
                found = true;
                break;
            }
        }

        map.doMove(move, false);
    }

    if (found)
    {
        KMessageBox::information(this, i18n("You have reached this position before."));
    }
    else
    {
        KMessageBox::information(this, i18n("<b></b>This position has <b>not</b> been reached before!"));
    }
}


void MainWindow::forceUpdateCollections()
{
    QDateTime min_date;
    min_date.setTime_t(0);

    if (!CollectionHolder::updateCollections(min_date, true))
    {
        KMessageBox::information(this, i18n("No new or changed collections are found."));
    }
}


void MainWindow::tipOfTheDay()
{
    KTipDialog::showTip(this, KGlobal::dirs()->findResource("data", "easysok/tips"), true);
}


void MainWindow::reorderCollectionsAndLevels()
{
    ReorderDialog dialog(this);

    if (dialog.exec())
    {
        setupCollectionMenu();
    }
}


void MainWindow::setEditedLevel(LevelEditor * level_editor)
{
    assert(level_editor != 0);
    assert(level_editor->level().map().isValid());

    int const nr_of_collections = CollectionHolder::numberOfCollections();
    int collection_nr = level_editor->collectionNr();
    int level_nr = level_editor->levelNr();
    Level original_level = level_editor->originalLevel();


    if (collection_nr < 0)
    {
        insertLevelIntoNewCollection(level_editor->level());
        level_editor->setOriginalLevel(level_editor->level(), nr_of_collections - 1, 0);

        return;
    }

    bool invalid_collection = (collection_nr >= nr_of_collections);

    if (level_nr < 0)
    {
        if (invalid_collection)
        {
            QString msg = i18n("The collection of the level has been deleted!\nDo you want to create "
                               "a new collection or discard the level?");
            int result = KMessageBox::warningYesNo(level_editor, msg, QString::null, i18n("Create New"),
                                                   i18n("Discard"));

            if (result == KMessageBox::Yes)
            {
                insertLevelIntoNewCollection(level_editor->level());
                level_editor->setOriginalLevel(level_editor->level(), nr_of_collections - 1, 0);
            }

            return;
        }
        else
        {
            Collection * collection = CollectionHolder::collection(collection_nr);
            level_nr = collection->numberOfLevels();

            collection->addLevel(level_editor->level());
            setLevel(collection_nr, level_nr);
            CollectionHolder::setModified();
            level_editor->setOriginalLevel(level_editor->level(), collection_nr, level_nr);

            return;
        }
    }

    bool invalid_level = false;
    bool level_changed = false;
    Collection * collection = 0;

    if (!invalid_collection)
    {
        collection = CollectionHolder::collection(collection_nr);
        invalid_level = (level_nr >= collection->numberOfLevels());

        if (!invalid_level)
        {
            level_changed = (original_level.compressedMap() != collection->level(level_nr).compressedMap());
        }
    }

    if (invalid_collection || invalid_level || level_changed)
    {
        std::vector<int> equal_collections;
        std::vector<int> equal_levels;

        for (int i = 0; i < nr_of_collections; ++i)
        {
            Collection * collection = CollectionHolder::collection(i);
            int const nr_of_levels = collection->numberOfLevels();

            for (int j = 0; j < nr_of_levels; ++j)
            {
                if (original_level.compressedMap() == collection->level(j).compressedMap())
                {
                    equal_collections.push_back(i);
                    equal_levels.push_back(j);
                }
            }
        }

        if (equal_collections.empty())
        {
            QString msg = i18n("The level you edited seems to have been deleted!\nDo you want to insert "
                               "it into a new collection or discard it?");
            int result = KMessageBox::warningYesNo(level_editor, msg, QString::null, i18n("Insert"),
                                                   i18n("Discard"));

            if (result == KMessageBox::Yes)
            {
                insertLevelIntoNewCollection(level_editor->level());
                level_editor->setOriginalLevel(level_editor->level(), nr_of_collections - 1, 0);
            }

            return;
        }
    }
    else
    {
        saveCurrentLevelState();
        collection->replaceLevel(level_editor->level(), level_nr);
        setLevel(m_collection_nr, m_level_nr, true, true);
        CollectionHolder::setModified();
        level_editor->setOriginalLevel(level_editor->level(), collection_nr, level_nr);
    }
}


void MainWindow::levelEditorExited(LevelEditor * level_editor)
{
    assert(std::find(m_editors.begin(), m_editors.end(), level_editor) != m_editors.end());

    m_editors.erase(std::find(m_editors.begin(), m_editors.end(), level_editor));
}


void MainWindow::createUser()
{
    CreateUserDialog dialog(this);

    dialog.exec();

    updateUserStatusBar();
}


void MainWindow::importUser()
{
    ImportUserDialog dialog(this);

    dialog.exec();

    updateUserStatusBar();
}


void MainWindow::setUser()
{
    SetUserDialog dialog(this);

    dialog.exec();

    updateUserStatusBar();
}


void MainWindow::changeUser()
{
    ChangeUserDialog dialog(this);

    dialog.exec();
}


void MainWindow::autoSendSolutions()
{
    m_auto_send_to_server = m_auto_send_to_server_action->isChecked();
}


void MainWindow::saveState()
{
    CollectionHolder::save();
    SolutionHolder::save();
    Bookmarks::save();
    KApplication::kApplication()->config()->sync();
}


void MainWindow::print()
{
    KPrinter printer;

    printer.setFullPage(true);

    if (!printer.setup(this))
    {
        return;
    }

    QPainter painter;

    painter.begin(&printer);

    QPaintDeviceMetrics metrics(&printer);

    int const width = metrics.width();
    int const height = metrics.height();

    Theme const * theme = ThemeHolder::theme(m_theme);

    QImage image = ImageStorer::createImage(*m_map, theme->suggestFieldSize(), *theme,
                                            m_game->keeperDirection(), true, false);

    double const x_scale = static_cast<double> (width) / image.width();
    double const y_scale = static_cast<double> (height) / image.height();

    if (x_scale > y_scale)
    {
        painter.scale(y_scale, y_scale);
        painter.drawImage(static_cast<int> ((width / y_scale - image.width()) / 2), 0, image);
    }
    else
    {
        painter.scale(x_scale, x_scale);
        painter.drawImage(0, static_cast<int> ((height / x_scale - image.height()) / 2), image);
    }

    painter.end();
}


void MainWindow::prevUnsolvedLevel()
{
    for (int i = m_level_nr - 1; i >= 0; --i)
    {
        if (!SolutionHolder::hasSolution(actCollection()->level(i).compressedMap()))
        {
            setLevel(m_collection_nr, i);

            return;
        }
    }

    KMessageBox::error(this, i18n("There is no previous unsolved level"));
}


void MainWindow::nextUnsolvedLevel()
{
    int const nr_of_levels = actCollection()->numberOfLevels();

    for (int i = m_level_nr + 1; i < nr_of_levels; ++i)
    {
        if (!SolutionHolder::hasSolution(actCollection()->level(i).compressedMap()))
        {
            setLevel(m_collection_nr, i);

            return;
        }
    }

    KMessageBox::error(this, i18n("There is no next unsolved level"));
}


void MainWindow::quit()
{
    setLevel(m_collection_nr, m_level_nr);
    close();
}


void MainWindow::focusOutEvent(QFocusEvent * event)
{
    m_map_widget->cursorVisible(true);

    KMainWindow::focusOutEvent(event);
}


void MainWindow::focusInEvent(QFocusEvent * event)
{
    m_map_widget->cursorVisible(false);

    KMainWindow::focusOutEvent(event);
}


void MainWindow::keyPressEvent(QKeyEvent * event)
{
    int const state = event->state();
    int const modifier = ( (state & ShiftButton ? SHIFT : 0) + (state & ControlButton ? CTRL : 0) +
                           (state & AltButton ? ALT : 0) );
    int const key = event->key() + modifier;

    KAccel const * acc = accel();

    
    /* The following code should be not needed, cause it's handled by the actions already. */
#if 0
    if (key == acc->currentKey("MoveLeft"))
    {
        moveLeft();
    }
    else if (key == acc->currentKey("MoveRight"))
    {
        moveRight();
    }
    else if (key == acc->currentKey("MoveUp"))
    {
        moveUp();
    }
    else if (key == acc->currentKey("MoveDown"))
    {
        moveDown();
    }
    else if (key == acc->currentKey("MoveFarLeft"))
    {
        moveFarLeft();
    }
    else if (key == acc->currentKey("MoveFarRight"))
    {
        moveFarRight();
    }
    else if (key == acc->currentKey("MoveFarUp"))
    {
        moveFarUp();
    }
    else if (key == acc->currentKey("MoveFarDown"))
    {
        moveFarDown();
    }
    else if (key == acc->currentKey("MoveVirtualLeft"))
    {
        moveVirtualLeft();
    }
    else if (key == acc->currentKey("MoveVirtualRight"))
    {
        moveVirtualRight();
    }
    else if (key == acc->currentKey("MoveVirtualUp"))
    {
        moveVirtualUp();
    }
    else if (key == acc->currentKey("MoveVirtualDown"))
    {
        moveVirtualDown();
    }
    else if (key == acc->currentKey("PullMoveLeft"))
    {
        pullMoveLeft();
    }
    else if (key == acc->currentKey("PullMoveRight"))
    {
        pullMoveRight();
    }
    else if (key == acc->currentKey("PullMoveUp"))
    {
        pullMoveUp();
    }
    else if (key == acc->currentKey("PullMoveDown"))
    {
        pullMoveDown();
    }
    else
#endif
    {
        event->ignore();
    }

    m_map_widget->cursorHidden();
}


void MainWindow::setShowArrows()
{
    m_hide_gems = m_hide_gems_action->isChecked();

    m_game->setShowArrows(!m_hide_gems && m_show_arrows);
}


void MainWindow::updateUserStatusBar()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QString user = config->readEntry("Current user", "");

    KStatusBar * status_bar = statusBar();

    if (m_show_user)
    {
        if (user.isEmpty())
        {
            status_bar->changeItem(i18n("User: <none>"), 1);
        }
        else
        {
            status_bar->changeItem(i18n("User: %1").arg(user), 1);
        }
    }
    else
    {
        status_bar->changeItem("", 1);
    }
}


void MainWindow::createLevelEditor(Level const & level, int collection_nr, int level_nr)
{
    LevelEditor * level_editor = new LevelEditor(level, collection_nr, level_nr,
                                                 ThemeHolder::theme(m_theme), 0);
    level_editor->show();
    connect(level_editor, SIGNAL(saveLevel(LevelEditor *)), this, SLOT(setEditedLevel(LevelEditor *)));
    connect(level_editor, SIGNAL(exited(LevelEditor *)), this, SLOT(levelEditorExited(LevelEditor *)));

    m_editors.push_back(level_editor);
}


void MainWindow::addCollection(Collection const & collection)
{
    Collection new_collection = collection;

    CollectionPropertiesDialog dialog(new_collection, this);

    if (!dialog.exec())
    {
        return;
    }

    new_collection.setAuthorEmailLine(dialog.author());
    new_collection.setHomepage(dialog.homepage());
    new_collection.setCopyright(dialog.copyright());
    new_collection.setName(dialog.name());
    new_collection.setInfo(dialog.info());
    saveCurrentLevelState();

    while ((CollectionHolder::indexFromName(new_collection.name()) != -1) ||
           new_collection.name().isEmpty())
    {
        if (!new_collection.name().isEmpty())
        {
            QString const message = i18n("A collection with the same name already exists. "
                                         "Do you want to overwrite it?");
            int result = KMessageBox::warningYesNoCancel(this, message, i18n("Overwrite collection"),
                                                         i18n("Overwrite"), i18n("Change Name"));

            if (result == KMessageBox::Cancel)
            {
                return;
            }
            else if (result == KMessageBox::Yes)
            {
                break;
            }
        }

        QString name_dialog_text = i18n("Change the collection name");

        if (new_collection.name().isEmpty())
        {
            name_dialog_text = i18n("Please give the collection a name.");
        }

        KLineEditDlg name_dialog(name_dialog_text, new_collection.name(), this);

        if (!name_dialog.exec())
        {
            return;
        }

        new_collection.setName(name_dialog.text());
    }

    m_collection_nr = CollectionHolder::addCollection(new  Collection(new_collection), false);
    setupCollectionMenu();
    setLevel(m_collection_nr, 0, true, true);
}


void MainWindow::insertLevelIntoNewCollection(Level const & level)
{
    Collection collection("", QStringList(), QStringList(), "", "", "", -1);
    collection.addLevel(level);

    addCollection(collection);
}


Level * MainWindow::levelFromClipboard()
{
    QMimeSource * mime_source = KApplication::clipboard()->data();

    if (mime_source == 0)
    {
        return 0;
    }

    Level * level = 0;

    if (mime_source->provides("application/easysok"))
    {
        QBuffer buffer(mime_source->encodedData("application/easysok"));
        buffer.open(IO_ReadOnly);
        QDataStream stream(&buffer);

        level = new Level(stream, 65536); // FIXME
    }
    else if (mime_source->provides("text/plain"))
    {
        QStringList lines = QStringList::split("\n", KApplication::clipboard()->text());

        level = new Level(lines, QStringList(), QStringList(), "", "", "", -1);
    }
    else
    {
        KMessageBox::error(this, i18n("The format is not supported."));

        return 0;
    }

    if (level->map().isValid())
    {
        return level;
    }
    else
    {
        KMessageBox::error(this, i18n("The inserted level is invalid."));

        return 0;
    }
}


void MainWindow::optimizeSolutions(int from, int to, Optimizer * optimizer)
{
    assert(from >= 0);
    assert(to >= from);
    assert(to <= CollectionHolder::numberOfCollections());

    int solutions = 0;
    int optimized = 0;

    for (int i = from; i < to; ++i)
    {
        Collection * collection = CollectionHolder::collection(i);
        int const number_of_levels = collection->numberOfLevels();

        for (int j = 0; j < number_of_levels; ++j)
        {
            Level const & level = collection->level(j);
            CompressedMap const & compressed_map = level.compressedMap();
            int const index = SolutionHolder::getIndexForMap(compressed_map);

            if (index == -1)
            {
                continue;
            }

            Map map(compressed_map);
            int const number_of_solutions = SolutionHolder::numberOfSolutions(index);

            std::vector<Movements> moves;
            std::vector<int> number_of_pushes;
            std::vector<int> number_of_linear_pushes;
            std::vector<int> number_of_gem_changes;
            std::vector<int> number_of_moves;
            QStringList infos;
            std::vector<QDateTime> date_times;

            for (int k = 0; k < number_of_solutions; ++k)
            {
                Movements const & old_moves = SolutionHolder::movements(index, k);
                int const old_number_of_pushes = SolutionHolder::pushesInSolution(index, k);
                int const old_number_of_moves = SolutionHolder::movesInSolution(index, k);

                optimizer->optimize(map, old_moves);
                moves.push_back(optimizer->moves());

                int const new_number_of_pushes = optimizer->numberOfPushes();
                int const new_number_of_moves = optimizer->numberOfMoves();
                int const new_number_of_linear_pushes = moves.back().linearPushes();
                int const new_number_of_gem_changes = moves.back().gemChanges();

                ++solutions;

                if (new_number_of_pushes < old_number_of_pushes)
                {
                    ++optimized;
                }
                else if (new_number_of_pushes == old_number_of_pushes)
                {
                    if (new_number_of_moves < old_number_of_moves)
                    {
                        ++optimized;
                    }
                }

                number_of_pushes.push_back(new_number_of_pushes);
                number_of_linear_pushes.push_back(new_number_of_linear_pushes);
                number_of_gem_changes.push_back(new_number_of_gem_changes);
                number_of_moves.push_back(new_number_of_pushes);
                infos.append(SolutionHolder::infoOfSolution(index, k));
                date_times.push_back(SolutionHolder::dateOfSolution(index, k));
            }

            for (int k = number_of_solutions - 1; k >= 0; --k)
            {
                SolutionHolder::deleteSolution(index, k);
            }

            for (int k = 0; k < number_of_solutions; ++k)
            {
                SolutionHolder::addSolution(index, moves[k], number_of_pushes[k],
                                            number_of_linear_pushes[k], number_of_gem_changes[k],
                                            number_of_moves[k], infos[k], date_times[k]);
            }
        }
    }

    KMessageBox::information(this, i18n("%1 of %2 solutions were optimized.").
                             arg(optimized).arg(solutions));
}


void MainWindow::updateWasSolvedActions()
{
    m_load_best_sol_action->setEnabled(m_was_solved);
    m_load_sol_action->setEnabled(m_was_solved);
    m_annotate_sol_action->setEnabled(m_was_solved);
    m_delete_sol_action->setEnabled(m_was_solved);
    m_min_moves_action->setEnabled(m_was_solved);
    m_min_pushes_action->setEnabled(m_was_solved);
}


void MainWindow::updateUndoRedoActions()
{
    bool const can_undo = m_game->moves().hasPrevMove();
    bool const can_redo = m_game->moves().hasNextMove();
    bool const can_play = can_undo || can_redo;

    if (can_undo != m_undo_action->isEnabled())
    {
        m_undo_action->setEnabled(can_undo);
    }

    if (can_redo != m_redo_action->isEnabled())
    {
        m_redo_action->setEnabled(can_redo);
    }

    if (can_undo != m_jump_start_action->isEnabled())
    {
        m_jump_start_action->setEnabled(can_undo);
    }

    if (can_redo != m_jump_end_action->isEnabled())
    {
        m_jump_end_action->setEnabled(can_redo);
    }

    if (can_play != m_play_action->isEnabled())
    {
        m_play_action->setEnabled(can_play);
    }
}


void MainWindow::updateLevelActions()
{
    bool const can_prev_level = (m_level_nr > 0);
    bool const can_next_level = ( (m_level_nr < actCollection()->numberOfLevels() - 1) &&
                                  (m_goto_any_level || m_was_solved) );

    m_prev_level_action->setEnabled(can_prev_level);
    m_next_level_action->setEnabled(can_next_level);
    m_first_level_action->setEnabled(can_prev_level);
    m_last_level_action->setEnabled(can_next_level);
    m_retro_mode_action->setChecked(m_game->retroMode());
}


PixmapProvider * MainWindow::getPixmapProvider()
{
    return new PixmapProvider(*ThemeHolder::theme(m_theme));
}


Collection * MainWindow::actCollection()
{
    return CollectionHolder::collection(m_collection_nr);
}


Level const & MainWindow::actLevel()
{
    return actCollection()->level(m_level_nr);
}


int MainWindow::lastLegalLevel() const
{
    Collection * collection = CollectionHolder::collection(m_collection_nr);
    int last_level = collection->numberOfLevels() - 1;

    if (!m_goto_any_level)
    {
        for (int i = m_level_nr; i <= last_level; ++i)
        {
            if (!SolutionHolder::hasSolution(collection->level(i).compressedMap()))
            {
                last_level = i;
            }
        }
    }

    return last_level;
}


void MainWindow::removeDoubleRecentCollections()
{
    int const nr_of_recent_collections = static_cast<int> (m_recent_collections.size());

    for (int i = nr_of_recent_collections - 1; i > 0; --i)
    {
        if (std::find(m_recent_collections.begin(), m_recent_collections.end(), m_recent_collections[i]) !=
            m_recent_collections.begin() + i)
        {
            m_recent_collections.erase(m_recent_collections.begin() + i);
        }
    }
}


void MainWindow::updateStatusBar(bool force)
{
    if (!force)
    {
        static QTime startTime = QTime::currentTime();
        static int last_offset = 0;
        int const current_offset = startTime.msecsTo(QTime::currentTime());

        if (current_offset < 0)
        {
            // We crossed a day
            startTime = QTime::currentTime();
            last_offset = 0;
        }
        else if ((current_offset - last_offset) < m_min_status_update_time)
        {
            return;
        }

        last_offset = current_offset;
    }

    KStatusBar * status_bar = statusBar();
    bool const retro_mode = m_retro_mode_action->isChecked();

    QString text =  " ";

    if (retro_mode)
    {
        text += i18n("Retro Mode");
    }
    else
    {
        text += i18n("Pushes:") + " " + QString::number(m_game->numberOfPushes());

        if (m_was_solved)
        {
            text += "(" + QString::number(m_best_number_of_pushes) + ")";
        }

        if (m_show_linear_pushes)
        {
            text += "  ";
            text += i18n("Linear Pushes:") + " " + QString::number(m_game->numberOfLinearPushes());

            if (m_was_solved)
            {
                text += "(" + QString::number(m_best_number_of_linear_pushes) + ")";
            }
        }

        if (m_show_gem_changes)
        {
            text += "  ";
            text += i18n("Gem Changes:") + " " + QString::number(m_game->numberOfGemChanges());

            if (m_was_solved)
            {
                text += "(" + QString::number(m_best_number_of_gem_changes) + ")";
            }
        }

        text += "  ";
        text += i18n("Moves:") + " " + QString::number(m_game->numberOfMoves());

        if (m_was_solved)
        {
            text += "(" + QString::number(m_best_number_of_moves) + ")";
        }

        if (m_is_solved)
        {
            text += "  " + i18n("SOLVED!");
        }
        else if (m_show_gems_left)
        {
            int const empty_goals = m_game->numberOfEmptyGoals();

            if (empty_goals == 1)
            {
                text += "  " + i18n("1 gem left");
            }
            else
            {
                text += "  " + i18n("%1 gems left").arg(empty_goals);
            }
        }
    }

    text += " ";
    status_bar->changeItem(text, 2);
}


void MainWindow::saveCurrentLevelState()
{
    if (m_game->retroMode())
    {
        m_game->setRetroMode(false);
    }

    SolutionHolder::setLastAttempt(actLevel().compressedMap(), m_game->moves());
}
