/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_MAIN_WINDOW_INC_GUARD_H
#define EASYSOK_MAIN_WINDOW_INC_GUARD_H


#include <vector>

#include <qstring.h>
#include <qstringlist.h>

#include <kmainwindow.h>
#include <kurl.h>

#include "level.h"


class Collection;
class DuplicateLevelFinder;
class Game;
class KAction;
class KActionMenu;
class KRadioAction;
class KToggleAction;
class KURL;
class Level;
class LevelEditor;
class Map;
class MapWidget;
class MngAnimationStorer;
class Movements;
class Optimizer;
class PixmapProvider;
class ProgressDialog;
class QMessageBox;
class QSignalMapper;
class QTimer;
class Solver;




/**
 * The main window.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class MainWindow : public KMainWindow
{
    Q_OBJECT

public:

    /**
     * Creates the window.
     *
     * @param parent The parent widget.
     * @param name The internal name of the widget.
     */

    MainWindow(QWidget * parent = 0, char const * name = 0);


    /**
     * Release all resources and writes back the current settings.
     */

    virtual ~MainWindow();


    /**
     * Changes the current level.
     *
     * @param collection_nr The number of the level collection.
     * @param level_nr The number of the level in the collection. If the number is -1,
     *                 we try to use the last visited level in the collection.
     * @param force_update If true we update even if the level has not changed.
     * @param dont_save_act_level If true, we don't save the current levels state.
     */

    void setLevel(int collection_nr, int level_nr, bool force_update = false,
                  bool dont_save_act_level = false);


    /**
     * The type of info to display with bookmarks.
     */

    enum BookmarkFormat
    {
        /**
         * Display date.
         */

        DATE = 1,


        /**
         * Display date and time.
         */

        DATE_AND_TIME = 2,


        /**
         * Display collection.
         */

        COLLECTION = 4,


        /**
         * Display collection and level.
         */

        COLLECTION_AND_LEVEL = 8,


        /**
         * Display annotation.
         */

        ANNOTATION = 16
    };


public slots:

    /**
     * Hack for a bug in KDE or Qt.
     */
    void statusBarUpdateHack();


    /**
     * Jumps to the previous collection.
     */

    void prevCollection();


    /**
     * Jumps to the next collection.
     */

    void nextCollection();


    /**
     * Jumps to the previous level in the collection.
     */

    void prevLevel();


    /**
     * Jumps to the next level in the collection.
     */

    void nextLevel();


    /**
     * Lets the user select a level.
     */

    void selectLevel();


    /**
     * Jumps to the first level in the collection.
     */

    void firstLevel();


    /**
     * Jumps to the last (allowed) level in the collection.
     */

    void lastLevel();


    /**
     * Moves the keep left.
     */

    void moveLeft();


    /**
     * Moves the keeper right.
     */

    void moveRight();


    /**
     * Moves the keeper up.
     */

    void moveUp();


    /**
     * Moves the keeper down.
     */

    void moveDown();


    /**
     * Moves a far left as possible.
     */

    void moveFarLeft();


    /**
     * Moves a far right as possible.
     */

    void moveFarRight();


    /**
     * Moves a far up as possible.
     */

    void moveFarUp();


    /**
     * Moves a far down as possible.
     */

    void moveFarDown();


    /**
     * Moves the virtual keeper left.
     */

    void moveVirtualLeft();


    /**
     * Moves the virtual keeper right.
     */

    void moveVirtualRight();


    /**
     * Moves the virtual keeper up.
     */

    void moveVirtualUp();


    /**
     * Moves the virtual keeper down.
     */

    void moveVirtualDown();


    /**
     * Moves the keeper left and tries to pull a gem.
     */

    void pullMoveLeft();


    /**
     * Moves the keeper right and tries to pull a gem.
     */

    void pullMoveRight();


    /**
     * Moves the keeper up and tries to pull a gem.
     */

    void pullMoveUp();


    /**
     * Moves the keeper down and tries to pull a gem.
     */

    void pullMoveDown();


    /**
     * Moves the keeper to the specified point if possible.
     *
     * @param point The point where the keeper should be moved.
     */

    void moveKeeper(QPoint point);


    /**
     * @overload
     */

    void moveKeeper(QPoint from, QPoint to);


    /**
     * Tries to move a gem from one position to another.
     *
     * @param from The start position of the gem.
     * @param to The end position of the gem.
     */

    void moveGem(QPoint from, QPoint to);


    /**
     * Undo.
     */

    void undo();


    /**
     * Redo.
     */

    void redo();


    /**
     * Jumps to the start position.
     */

    void jumpToStart();


    /**
     * Jumps to the end position.
     */

    void jumpToEnd();


    /**
     * Loads the best previous solution of this level.
     */

    void loadBestSolution();


    /**
     * Loads a previous solution.
     */

    void loadASolution();


    /**
     * Lets the user annotate a solution.
     */

    void annotateSolution();


    /**
     * Jumps to the level collection with the given index.
     *
     * @param index The index of the level collection.
     */

    void setCollection(int index);


    /**
     * Changes the animation speed.
     *
     * @param speed The new speed.
     */

    void changeAnimations(int speed);


    /**
     * Called the the state of the gem visibility is changed.
     */

    void hideGemsChanged();


    /**
     * Called the the state of the goal visibility is changed.
     */

    void hideGoalsChanged();


    /**
     * Called when the state of treating outside fields as walls is changed.
     */

    void outsideAsWallChanged();


    /**
     * Called when the state of the deadlock marking is changed.
     */

    void honorDeadlockChanged();


    /**
     * Sets a new theme.
     *
     * @param theme The index of the theme.
     */

    void setTheme(int theme);


    /**
     * Use the next theme.
     */

    void nextTheme();


    /**
     * Use the previous theme.
     */

    void previousTheme();


    /**
     * Imports a collection.
     */

    void importCollection();


    /**
     * Exports the current game.
     */

    void exportGame();


    /**
     * Send the current game.
     */

    void sendGame();


    /**
     * Imports a game.
     */

    void importGame();


    /**
     * Exports the current collection.
     */

    void exportCollection();


    /**
     * Sends the current collection.
     */

    void sendCollection();


    /**
     * Exports solutions.
     */

    void exportSolutions();


    /**
     * Sends solutions.
     */

    void sendSolutions();


    /**
     * Imports solutions.
     */

    void importSolutions();


    /**
     * Plays the moves till the end.
     */

    void playMoves();


    /**
     * Pops up the key configuration dialog.
     */

    void showKeyDialog();


    /**
     * Pops up the toolbar configuration dialog.
     */

    void showToolbarDialog();


    /**
     * Updates the toolbar.
     */

    void updateToolbar();


    /**
     * Shows the local solutions of the current level.
     */

    void showSolutions();


    /**
     * Shows the 'highscore' widget.
     */

    void showHighscores();


    /**
     * Called when show pushes changed.
     */

    void showPushesChanged();


    /**
     * Sets a bookmark.
     *
     * @param index The index of the bookmark.
     */

    void setBookmark(int index);


    /**
     * Goes to a specified bookmark.
     *
     * @param index The index of the bookmark.
     */

    void gotoBookmark(int index);


    /**
     * Changes the annotation fo a bookmark.
     *
     * @param index The index of the bookmark.
     */

    void annotateBookmark(int index);


    /**
     * Imports a KSokoban bookmark.
     *
     * @param index The index of the bookmark where the KSokoban bookmark will be stored.
     */

    void importKSokobanBookmark(int index);


    /**
     * Shows the options dialog.
     */

    void showOptionsDialog();


    /**
     * Minimize the moves in all solutions.
     */

    void minimizeMoves();


    /**
     * Minimize the moves in all solutions of the current collection.
     */

    void minimizeCollectionMoves();


    /**
     * Minimizes the moves in solutions of the current level.
     */

    void minimizeLevelMoves();


    /**
     * Minimize the moves in all solutions of the current collection.
     */

    void minimizePushes();


    /**
     * Minimize the pushes in all solutions.
     */

    void minimizeCollectionPushes();


    /**
     * Minimizes the pushes in solutions ofthe current level..
     */

    void minimizeLevelPushes();


    /**
     * Lets the user delete solutions of the current level.
     */

    void deleteSolutions();


    /**
     * Lets the user delete solutions by name.
     */

    void deleteSolutionsByName();


    /**
     * Called, when the current animation should be stopped.
     */

    void stopAnimation();


    /**
     * Called when the level was solved (either by hand or by redo/play etc.).
     */

    void solved();


    /**
     * Called when the level was unsolved (either by hand or by redo/play etc.).
     */

    void unsolved();


    /**
     * Called when the level was solved by hand.
     */

    void solvedByHand();


    /**
     * Updates the status bar.
     *
     * @param force If true, the update is forced. Otherwise it is only updated,
     *              when it wasn't updated a short time ago.
     */

    void updateStatusBar(bool force);


    /**
     * Called, when the config file has changed.
     */

    void configChanged();


    /**
     * Called, when the auto optimize solutions options have changed.
     */

    void autoOptChanged();


    /**
     * Called, when the retro mode was changed.
     */

    void retroModeChanged();


    /**
     * Called, when we want to edit the current level.
     */

    void editCurrentLevel();


    /**
     * Called when we want to insert a new level.
     */

    void insertNewLevel();


    /**
     * Called when we want to insert a new collection.
     */

    void insertNewCollection();


    /**
     * Called when we want to delete the current level.
     */

    void deleteCurrentLevel();


    /**
     * Called when we want to delete the current collection.
     */

    void deleteCurrentCollection();


    /**
     * Called when we duplicate the current level.
     */

    void duplicateCurrentLevel();


    /**
     * Called when we duplicate the current collection.
     */

    void duplicateCurrentCollection();


    /**
     * Called, when we copy a level.
     */

    void copyCurrentLevel();


    /**
     * Called when we paste a level.
     */

    void pasteLevel();


    /**
     * Called when we paste a level into a new collection.
     */

    void pasteLevelInNewCollection();


    /**
     * Makes the current collection non temporary.
     */

    void makeCurrentCollectionNonTemporary();


    /**
     * Solves a level from the current position.
     */

    void solve();


    /**
     * Exports the game as an image.
     */

    void exportGameImage();


    /**
     * Exports the game as an animation.
     */

    void exportGameAnimation();


    /**
     * Shows information about the current level.
     */

    void levelInfo();


    /**
     * Shows information about the current collection.
     */

    void collectionInfo();


    /**
     * Finds duplicated levels.
     */

    void findDups();


    /**
     * Does the real work of findDups();
     */

    void performFindDups();


    /**
     * An animation write step.
     */

    void animationWriteStep();


    /**
     * Determines, if the current position has been reached before.
     */

    void isOldPosition();


    /**
     * Forces the update of the collections.
     */

    void forceUpdateCollections();


    /**
     * Shows the tip of the day.
     */

    void tipOfTheDay();


    /**
     * Lets the user reorder collections and levels.
     */

    void reorderCollectionsAndLevels();


    /**
     * Sets a new level.
     *
     * @param level_editor The level editor to get the level from.
     */

    void setEditedLevel(LevelEditor * level_editor);


    /**
     * Called when a level editor is closed.
     *
     * @param level_editor The level editor which exits.
     */

    void levelEditorExited(LevelEditor * level_editor);


    /**
     * Creates a new user for global highscore files.
     */

    void createUser();


    /**
     * Imports a user.
     */

    void importUser();


    /**
     * Sets the current user.
     */

    void setUser();


    /**
     * Changes the current user.
     */

    void changeUser();


    /**
     * Called when the auto send to server action is changed.
     */

    void autoSendSolutions();


    /**
     * Sends the best solutions of the current level to the server.
     */

    void sendBestSolutionToServer();


    /**
     * Sends the best solutions of the collection to the server.
     */

    void sendBestCollectionSolutionToServer();


    /**
     * Sends the best solutions of all collections to the server.
     */

    void sendBestAllSolutionToServer();


    /**
     * Sends best solutions of all levels with solutions new than a given date.
     */

    void sendByDateToServer();


    /**
     * Compares your best solution with the others on the server.
     */

    void compareSolution();


    /**
     * Shows the best global solution.
     */

    void getBestSolutionFromServer();


    /**
     * Saves the state (bookmarks, levels, solutions).
     */

    void saveState();


    /**
     * Prints the current map.
     */

    void print();


    /**
     * Jumps to a previous unsolved level.
     */

    void prevUnsolvedLevel();


    /**
     * Jumps to the next unsolved level.
     */

    void nextUnsolvedLevel();


    /**
     * Quits the application.
     */

    void quit();


protected:

    /**
     * Writes the configuration before exiting.
     */
    virtual bool queryExit();


    /**
     * Used to make mouse pointer visible.
     */

    virtual void focusOutEvent(QFocusEvent * event);


    /**
     * Used to make mouse pointer visible.
     */

    virtual void focusInEvent(QFocusEvent * event);


    /**
     * Used to handle keeper movements.
     */

    virtual void keyPressEvent(QKeyEvent * event);


private:

    /**
     * Sets the showing of the arrows.
     *
     * Takes into account if arrows should be shown and if gems are visible.
     */

    void setShowArrows();


    /**
     * Sends the best solutions of the given levels.
     *
     * @param levels The levels to use.
     * @param only_one If true, only one level is used.
     */

    void sendSolutionsOfLevels(std::vector<Level const *> const & levels, bool only_one);


    /**
     * Removes double entries in the recent collections.
     */

    void removeDoubleRecentCollections();


    /**
     * Updates the recent collections menu.
     */

    void setupRecentCollectionsMenu();


    /**
     * Updates the current user in the status bar.
     */

    void updateUserStatusBar();


    /**
     * Tries to place the solution of the given level to the highscore server.
     *
     * @param map The map of the solution.
     * @param moves The moves of the solution.
     * @param nr_of_pushes The number of pushes in the solution.
     * @param nr_of_linear_pushes The number of linear pushes in the solution.
     * @param nr_of_gem_changes The number of gem changes in the solution.
     * @param nr_of_moves The number of moves in the solution.
     */

    void sendSolutionToServer(Map const & map, Movements const & moves, int nr_of_pushes,
                              int nr_of_linear_pushes, int nr_of_gem_changes, int nr_of_moves);


    /**
     * Creates a level editor.
     *
     * @param level The level to edit.
     * @param collection_nr The collection number of the level.
     * @param level_nr The level number of the level.
     */

    void createLevelEditor(Level const & level, int collection_nr, int level_nr);


    /**
     * Inserts the given level into a new collection.
     *
     * @param level The level to insert.
     */

    void insertLevelIntoNewCollection(Level const & level);


    /**
     * Returns a pasted level from the clipboard.
     *
     * Returns 0 if the clipboard does not contain a valid level.
     */

    Level * levelFromClipboard();


    /**
     * Returns the current game in text format.
     */

    QString currentGameToText();


    /**
     * Returns the text of solutions the user has selected or an empty string if the user canceled
     * the operation.
     */

    QString solutionsText();


    /**
     * Returns a url for saving.
     *
     * @param dialog_tag  A tag to remember the last dir for different save dialogs.
     * @param filter The filter(s) to use.
     */

    KURL getSaveUrl(QString const & dialog_tag, QString const & filter) const;


    /**
     * Lets the user export the given data as an xsb file.
     *
     * @param data The data to save.
     * @param dialog_tag A tag to remember the last dir for different save dialogs.
     */

    void saveDataAsXsbFile(QString const & data, QString const & dialog_tag) const;


    /**
     * Lets the user send data via mail.
     *
     * @param data The data to mail.
     */

    void mailData(QString const & data) const;


    /**
     * Loads an xsb file and returns the data as a list of strings.
     *
     * @param dialog_tag A tag to remember the last dir for different load dialogs.
     */

    QStringList loadDataFromXsbFile(QString const & dialog_tag) const;


    /**
     * Adds a new collection and ensures the name is valid and unique.
     *
     * @param collection The collection to add.
     */

    void addCollection(Collection const & collection);


    /**
     * Optimizes the solutions.
     *
     * @param from The index of the first collection to optimize.
     * @param to The indexx of one past the last collection to optimize.
     * @param optimizer The optimizer to use.
     */

    void optimizeSolutions(int from, int to, Optimizer * optimizer);


    /**
     * Updates the state of the action which depend on wether the level was solved before.
     */

    void updateWasSolvedActions();


    /**
     * Updates the state of the actions which depend of the undo/redo state.
     */

    void updateUndoRedoActions();


    /**
     * Updates the state of the action which depend on the current level.
     */

    void updateLevelActions();


    /**
     * Returns a pixmap provider for the actual theme.
     */

    PixmapProvider * getPixmapProvider();


    /**
     * Returns the current collection.
     */

    Collection * actCollection();


    /**
     * Returns the current level.
     */

    Level const & actLevel();


    /**
     * Returns the number of the last level, the player can jump to.
     */

    int lastLegalLevel() const;


    /**
     * Stores actual moves and solution of current level.
     */

    void saveCurrentLevelState();


    /**
     * Reads the configurations.
     */

    void readConfig();


    /**
     * Reads in configurations, which may have been changed by the options dialog.
     */

    void readChangedOptions();


    /**
     * Writes the configurtions.
     */

    void writeConfig();


    /**
     * Sets up the actions.
     */

    void setupActions();


    /**
     * Sets up the collection menu.
     */

    void setupCollectionMenu();


    /**
     * Sets up the theme menu.
     */

    void setupThemeMenu();



    /**
     * Sets up the bookmark menu.
     */

    void setupBookmarkMenu();


    /**
     * Sets up an entry in the bookmark menu.
     *
     * Note that it must be created before.
     *
     * @param index The index of the bookmark.
     */

    void setupBookmarkMenuEntry(int index);


    /**
     * Returns the image type of a url.
     *
     * @param url The url.
     */

    char const * imageType(KURL const & url) const;


    /**
     * The current game.
     */

    Game * m_game;


    /**
     * The current map.
     */

    Map * m_map;


    /**
     * The current collection number.
     */

    int m_collection_nr;


    /**
     * The current level number.
     */

    int m_level_nr;


    /**
     * The best number of pushes for the level so far.
     */

    int m_best_number_of_pushes;


    /**
     * The best number of linear pushes for the level so far.
     */

    int m_best_number_of_linear_pushes;


    /**
     * The best number of pushes for the level so far.
     */

    int m_best_number_of_gem_changes;


    /**
     * The best number of moves for the level so far.
     */

    int m_best_number_of_moves;


    /**
     * If true, the current level is solved.
     */

    int m_is_solved;


    /**
     * If true, the current level was solved or is solved.
     */

    bool m_was_solved;


    /**
     * If true, we move to the next level if we have solved one by hand.
     */

    bool m_next_level_when_solved;


    /**
     * If true, we automatically optimize the number of moves.
     */

    bool m_auto_optimize_moves;


    /**
     * If true, we automatically optimize the number of pushes.
     */

    bool m_auto_optimize_pushes;


    /**
     * The index of the actual theme.
     */

    int m_theme;


    /**
     * The map widget.
     */

    MapWidget * m_map_widget;


    /**
     * If true, we hide the gems.
     */

    bool m_hide_gems;


    /**
     * If true, we hide the goals.
     */

    bool m_hide_goals;


    /**
     * If true, we treat outside fields as walls.
     */

    bool m_outside_as_wall;


    /**
     * If true, we honor deadlocks.
     */

    bool m_honor_deadlocks;


    /**
     * If true, we show possible pushes.
     */

    bool m_show_arrows;


    /**
     * If true, we can also jump over not solved levels.
     */

    bool m_goto_any_level;


    /**
     * The animation speed.
     */

    int m_animation_speed;


    /**
     * The info display with the bookmarks.
     */

    BookmarkFormat m_bookmark_format;


    /**
     * The animation timer.
     */

    QTimer * m_animation_timer;


    /**
     * The animation storer.
     */

    MngAnimationStorer * m_animation_storer;


    /**
     * The progress dialog.
     */

    ProgressDialog * m_progress_dialog;


    /**
     * The find dups timer.
     */

    QTimer * m_find_dups_timer;


    /**
     * The duplicate level finder.
     */

    DuplicateLevelFinder * m_duplicate_level_finder;


    /**
     * The retro mode action.
     */

    KToggleAction * m_retro_mode_action;


    /**
     * The number of bookmarks.
     */

    int m_number_of_bookmarks;


    /**
     * The hide gems action.
     */

    KToggleAction * m_hide_gems_action;


    /**
     * The hide goals action.
     */

    KToggleAction * m_hide_goals_action;


    /**
     * The show outside fields as wall action.
     */

    KToggleAction * m_outside_as_wall_action;


    /**
     * The honor deadlock action.
     */

    KToggleAction * m_honor_deadlock_action;


    /**
     * The show arrows action.
     */

    KToggleAction * m_show_arrow_action;


    /**
     * The action menu of the collections.
     */

    KActionMenu * m_collection_menu;


    /**
     * The signal mapper for the collections.
     */

    QSignalMapper * m_collection_mapper;


    /**
     * The actions of the collections.
     */

    std::vector<KRadioAction *> m_collection_actions;


    /**
     * The collection by author menu
     */

    KActionMenu * m_collection_by_author_menu;


    /**
     * The actions of the collections by author.
     */

    std::vector<KRadioAction *> m_collection_by_author_actions;


    /**
     * The actions of the collections by author submenus.
     */

    std::vector<KActionMenu *> m_collection_by_author_submenus;


    /**
     * The collection by difficulty menu
     */

    KActionMenu * m_collection_by_difficulty_menu;


    /**
     * The actions of the collections by difficulty.
     */

    std::vector<KRadioAction *> m_collection_by_difficulty_actions;


    /**
     * The actions of the collections by difficulty submenus.
     */

    std::vector<KActionMenu *> m_collection_by_difficulty_submenus;


    /**
     * The action menu of the themes.
     */

    KActionMenu * m_theme_menu;


    /**
     * The signal mapper of the theme menu.
     */

    QSignalMapper * m_theme_mapper;


    /**
     * The actions of the themes.
     */

    std::vector<KRadioAction *> m_theme_actions;


    /**
     * The don't optimize action.
     */

    KToggleAction * m_dont_optimize_solutions_action;


    /**
     * The optimize moves action.
     */

    KToggleAction * m_auto_optimize_moves_action;


    /**
     * The optimize pushes action.
     */

    KToggleAction * m_auto_optimize_pushes_action;


    /**
     * The undo action.
     */

    KAction * m_undo_action;


    /**
     * The redo action.
     */

    KAction * m_redo_action;


    /**
     * The play action.
     */

    KAction * m_play_action;


    /**
     * The jump to start action.
     */

    KAction * m_jump_start_action;


    /**
     * The jump to end action.
     */

    KAction * m_jump_end_action;


    /**
     * The load best solution action.
     */

    KAction * m_load_best_sol_action;


    /**
     * The load solution action.
     */

    KAction * m_load_sol_action;


    /**
     * The annotate solution action.
     */

    KAction * m_annotate_sol_action;


    /**
     * The delete solution action.
     */

    KAction * m_delete_sol_action;


    /**
     * The minimize moves in solutions action.
     */

    KAction * m_min_moves_action;


    /**
     * The minimize pushes in solutions action.
     */

    KAction * m_min_pushes_action;


    /**
     * The previous level action.
     */

    KAction * m_prev_level_action;


    /**
     * The next level action.
     */

    KAction * m_next_level_action;


    /**
     * The first level action.
     */

    KAction * m_first_level_action;


    /**
     * The last level action.
     */

    KAction * m_last_level_action;


    /**
     * The bookmark set menu.
     */

    KActionMenu * m_bookmark_set_menu;


    /**
     * The bookmark goto menu.
     */

    KActionMenu * m_bookmark_goto_menu;


    /**
     * The bookmark annotate menu.
     */

    KActionMenu * m_bookmark_annotate_menu;


    /**
     * The bookmark import menu.
     */

    KActionMenu * m_bookmark_import_menu;


    /**
     * The bookmark set menu signal mapper.
     */

    QSignalMapper * m_bookmark_set_mapper;


    /**
     * The bookmark goto menu signal mapper.
     */

    QSignalMapper * m_bookmark_goto_mapper;


    /**
     * The bookmark annotate menu signal mapper.
     */

    QSignalMapper * m_bookmark_annotate_mapper;


    /**
     * The bookmark import menu signal mapper.
     */

    QSignalMapper * m_bookmark_import_mapper;


    /**
     * The bookmark set actions.
     */

    std::vector<KAction *> m_bookmark_set_actions;


    /**
     * The bookmark goto actions.
     */

    std::vector<KAction *> m_bookmark_goto_actions;


    /**
     * The bookmark annotate actions.
     */

    std::vector<KAction *> m_bookmark_annotate_actions;


    /**
     * The bookmark import actions.
     */

    std::vector<KAction *> m_bookmark_import_actions;


    /**
     * The bookmark set submenus.
     */

    std::vector<KActionMenu *> m_bookmark_set_submenus;


    /**
     * The bookmark goto submenus.
     */

    std::vector<KActionMenu *> m_bookmark_goto_submenus;


    /**
     * The bookmark annotate submenus.
     */

    std::vector<KActionMenu *> m_bookmark_annotate_submenus;


    /**
     * The bookmark import submenus.
     */

    std::vector<KActionMenu *> m_bookmark_import_submenus;


    /**
     * If true, we show the difficulty of the level in the status bar.
     */

    bool m_show_difficulty;


    /**
     * If true, we show the number of linear pushes in the status bar.
     */

    bool m_show_linear_pushes;


    /**
     * If true, we show the number of gem changes in the status bar.
     */

    bool m_show_gem_changes;


    /**
     * If true, we show the gems left in the status bar.
     */

    bool m_show_gems_left;


    /**
     * If true, we show the author of the level in the status bar.
     */

    bool m_show_author;


    /**
     * If true, we show the current user in the status bar.
     */

    bool m_show_user;


    /**
     * If true, we send solution automatically to the server.
     */

    bool m_auto_send_to_server;


    /**
     * The auto send to server action.
     */

    KToggleAction * m_auto_send_to_server_action;


    /**
     * Holds the current active level editors.
     */

    std::vector<LevelEditor *> m_editors;


    /**
     * The timer used for automatically storing of state.
     */

    QTimer * m_state_timer;


    /**
     * The auto save time.
     */

    int m_auto_save_time;


    /**
     * The number of recent collections.
     */

    int m_nr_of_recent_collections;


    /**
     * The recent collections.
     */

    std::vector<int> m_recent_collections;


    /**
     * The action menu of the recent collections.
     */

    KActionMenu * m_recent_collections_menu;


    /**
     * The signal mapper for the recent collections.
     */

    QSignalMapper * m_recent_collections_mapper;


    /**
     * The actions of the recent collections.
     */

    std::vector<KAction *> m_recent_collections_actions;


    /**
     * The old actions of the recent collections.
     */

    std::vector<KAction *> m_recent_collections_old_actions;


    /**
     * If true, illegal moves are treated as virtual moves.
     */

    bool m_illegal_as_virtual;


    /**
     * If true, the last move was a virtual move.
     */

    bool m_last_was_virtual;


    /**
     * The time difference between updating the status bar.
     */

    int m_min_status_update_time;
};


#endif
