/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "main_window.h"

#include "../config.h"

#include <cassert>
#include <map>
#include <utility>
#include <vector>

#include <qdatastream.h>
#include <qdatetime.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qregexp.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qtextstream.h>
#include <qtimer.h>

#include <kapp.h>
#include <kconfig.h>
#include <kfiledialog.h>
#include <klineeditdlg.h>
#include <klocale.h>
#include <kio/netaccess.h>
#include <kmessagebox.h>
#include <ktempfile.h>
#include <kurl.h>

#include "animation_storer_dialog.h"
#include "collection.h"
#include "collection_holder.h"
#include "compare_solution_dialog.h"
#include "create_solutions_dialog.h"
#include "export_solutions_dialog.h"
#include "game.h"
#include "image_storer.h"
#include "image_storer_dialog.h"
#include "import_solutions_dialog.h"
#include "level.h"
#include "map.h"
#include "map_widget.h"
#include "mng_animation_storer.h"
#include "move.h"
#include "movements.h"
#include "move_optimizer.h"
#include "progress_dialog.h"
#include "push_optimizer.h"
#include "query_dialog.h"
#include "select_date_dialog.h"
#include "send_solutions_dialog.h"
#include "server_connector.h"
#include "solution_holder.h"
#include "theme_holder.h"




void MainWindow::importGame()
{
    stopAnimation();

    QStringList lines = loadDataFromXsbFile("LoadGame");

    if (lines.isEmpty())
    {
        return;
    }

    Level level(lines, QStringList(), QStringList(),  "",  "", i18n("Loaded Game"), -1);

    if (!level.map().isValid())
    {
        KMessageBox::error(0, i18n("The map of the game is invalid!"));

        return;
    }

    Movements moves(level.map().keeper(), lines);

    if (!level.map().areValidMoves(moves))
    {
        KMessageBox::error(0, i18n("The moves of the game are invalid!"));

        return;
    }

    moves.setToFirstPosition();

    int collection_nr;
    int level_nr;
    CollectionHolder::findCompressedMap(level.compressedMap(), collection_nr, level_nr);

    if (collection_nr != -1)
    {
        setLevel(collection_nr, level_nr);
        m_game->setMoves(moves);
    }
    else
    {
        Collection * new_collection = new Collection(level.name(), level.authors(), level.emails(),
                                                     level.homepage(), level.copyright(), level.info(),
                                                     level.difficulty());

        new_collection->addLevel(level);

        if (new_collection->name().isEmpty())
        {
            new_collection->setName(i18n("Loaded Game"));
        }
        else
        {
            new_collection->setName(new_collection->name() + " (" + i18n("Loaded Game"));
        }

        if (CollectionHolder::indexFromName(new_collection->name()) != -1)
        {
            int suffix = 2;
            QString name;

            do
            {
                name = new_collection->name() + " "  + QString::number(suffix);
                ++suffix;
            }
            while (CollectionHolder::indexFromName(name) != -1);

            new_collection->setName(name);
        }

        int const collection_nr = CollectionHolder::addCollection(new_collection, true);
        setLevel(collection_nr, 0);
        m_game->setMoves(moves);
    }

    setupCollectionMenu();
}


void MainWindow::importCollection()
{
    stopAnimation();

    KURL url = KFileDialog::getOpenURL(":ImportCollection", i18n("*.xsb|Sokoban files\n*|All files"));

    if (!url.isValid() && url.isEmpty())
    {
        return;
    }

    QString filename;

    if (!KIO::NetAccess::download(url, filename))
    {
        KMessageBox::error(0, i18n("Could not open location!"));

        return;
    }

    QFileInfo file_info(filename);

    int const megabytes = static_cast<int> (file_info.size() / (1024 * 1024));

    if (megabytes >= 1)
    {
        if (KMessageBox::warningContinueCancel(this,
                                               i18n("The file '%1'\nhas a size of %2 megabyte.\n"
                                                    "This is very unusual. Do you want to "
                                                    "continue the loading?").arg(filename).arg(megabytes),
                                               QString::null, i18n("Continue"),
                                               "Warn because of big files", false) == KMessageBox::Cancel)
        {
            return;
        }
    }

    Collection collection(filename);

    if (collection.numberOfLevels() < 1)
    {
        KMessageBox::error(this, i18n("The collection contains no valid levels!"));

        return;
    }

    if (collection.name().isEmpty())
    {
        QFileInfo file_info(url.fileName());
        QString suggested_name = file_info.baseName();
        collection.setName(suggested_name);
    }

    while (CollectionHolder::indexFromName(collection.name()) != -1)
    {
        QString const msg = i18n("A collection with the same name already exists.\n"
                                 "Please enter a new name.");
        KLineEditDlg dialog(msg, collection.name(), this);

        if (!dialog.exec())
        {
            return;
        }

        collection.setName(dialog.text());
    }

    addCollection(collection);
}


void MainWindow::importSolutions()
{
    stopAnimation();

    ImportSolutionsDialog dialog;

    if (!dialog.exec())
    {
        return;
    }

    QStringList lines = loadDataFromXsbFile("ImportSolutions");

    if (lines.isEmpty())
    {
        return;
    }

    bool const import_current_level = dialog.importSolutionForCurrentLevel();
    bool const import_current_collection = dialog.importSolutionsForCurrentCollection();
    bool const optimize_pushes = dialog.optimizePushes();
    bool const optimize_moves = dialog.optimizeMoves();

    int const number_of_collections = CollectionHolder::numberOfCollections();
    int const start_collection = import_current_collection ? m_collection_nr : 0;
    int const end_collection = import_current_collection ? m_collection_nr + 1 : number_of_collections;

    QString const add_annotation = dialog.appendAnnotation();

    int invalid_maps = 0;
    int invalid_moves = 0;
    int ignored_maps = 0;
    int optimized_solutions = 0;
    int loaded_solutions = 0;

    bool const import_better = dialog.importOnlyBetterSolutions();
    bool const import_best = dialog.importOnlyBestSolution();
    bool const import_best_better = dialog.importOnlyBestBetterSolution();
    bool const import_all = (!import_better && !import_best && !import_best_better);

    std::map<CompressedMap, int> maps;

    while (!lines.isEmpty())
    {
        Level level(lines, QStringList(), QStringList(), "", "", "", -1);

        if (!level.map().isValid())
        {
            ++invalid_maps;

            continue;
        }

        Movements moves(level.map().keeper(), lines);

        int number_of_moves = 0;
        int number_of_pushes = 0;

        if (!level.map().areValidSolutionMoves(moves, number_of_pushes, number_of_moves))
        {
            ++invalid_moves;

            continue;
        }

        QString annotation = level.info();

        if (!add_annotation.isEmpty())
        {
            if (!annotation.isEmpty())
            {
                annotation += ' ';
            }

            annotation += add_annotation;
        }

        if (!import_all)
        {
            annotation = QChar(0) + annotation;
        }

        CompressedMap compressed_map(level.map());

        if (import_current_level && (compressed_map != actLevel().compressedMap()))
        {
            ++ignored_maps;

            continue;
        }
        else
        {
            bool map_found = false;

            for (int i = start_collection; i < end_collection; ++i)
            {
                Collection * collection = CollectionHolder::collection(i);
                int const number_of_levels = collection->numberOfLevels();

                for (int j = 0; j < number_of_levels; ++j)
                {
                    if (compressed_map == collection->level(j).compressedMap())
                    {
                        map_found = true;
                    }
                }
            }

            if (!map_found)
            {
                ++ignored_maps;

                continue;
            }
        }

        int new_number_of_pushes = number_of_pushes;
        int new_number_of_moves = number_of_moves;

        if (optimize_pushes)
        {
            PushOptimizer optimizer(level.map(), moves);

            moves = optimizer.moves();
            new_number_of_pushes = optimizer.numberOfPushes();
            new_number_of_moves = optimizer.numberOfMoves();
        }
        else if (optimize_moves)
        {
            MoveOptimizer optimizer(level.map(), moves);

            moves = optimizer.moves();
            new_number_of_pushes = optimizer.numberOfPushes();
            new_number_of_moves = optimizer.numberOfMoves();
        }

        if (new_number_of_pushes < number_of_pushes)
        {
            ++optimized_solutions;
        }
        else if ( (new_number_of_pushes == number_of_pushes) && (new_number_of_moves < number_of_moves) )
        {
            ++optimized_solutions;
        }

        number_of_pushes = new_number_of_pushes;
        number_of_moves = new_number_of_moves;
        int number_of_linear_pushes = moves.linearPushes();
        int number_of_gem_changes = moves.gemChanges();

        SolutionHolder::addSolution(compressed_map, moves, number_of_pushes, number_of_linear_pushes,
                                    number_of_gem_changes, number_of_moves, annotation);

        if (!import_all)
        {
            std::map<CompressedMap, int>::iterator it = maps.find(compressed_map);

            if (it == maps.end())
            {
                maps.insert(std::make_pair(compressed_map, 1));
            }
            else
            {
                ++it->second;
            }
        }
        else
        {
            ++loaded_solutions;
        }
    }

    if (!import_all)
    {
        std::map<CompressedMap, int>::iterator it = maps.begin();
        std::map<CompressedMap, int>::const_iterator end_it = maps.end();

        for (;it != end_it; ++it)
        {
            int const index = SolutionHolder::getIndexForMap(it->first);
            int const number_of_solutions = SolutionHolder::numberOfSolutions(index);
            int const imported = it->second;

            if (import_better)
            {
                int act_solutions = number_of_solutions;
                bool delete_solution = false;

                for (int i = 0; i < act_solutions; ++i)
                {
                    QString const info = SolutionHolder::infoOfSolution(index, 0);

                    if (delete_solution)
                    {
                        if (!info.isEmpty() && (info[0] == QChar(0)))
                        {
                            SolutionHolder::deleteSolution(index, i);
                            --i;
                            --act_solutions;
                        }
                    }
                    else
                    {
                        if (!info.isEmpty() && (info[0] == QChar(0)))
                        {
                            QString const new_info = info.right(info.length() - 1);
                            SolutionHolder::changeSolutionInfo(index, 0, new_info);
                            ++loaded_solutions;
                        }
                        else
                        {
                            delete_solution = true;
                        }
                    }
                }
            }
            else if (import_best)
            {
                int solutions_to_delete = imported - 1;

                for (int i = number_of_solutions - 1; i >= 0; --i)
                {
                    QString const info = SolutionHolder::infoOfSolution(index, 0);

                    if (!info.isEmpty() && (info[0] == QChar(0)))
                    {
                        if (solutions_to_delete > 1)
                        {
                            SolutionHolder::deleteSolution(index, i);
                            --solutions_to_delete;
                        }
                        else
                        {
                            QString const new_info = info.right(info.length() - 1);
                            SolutionHolder::changeSolutionInfo(index, 0, new_info);
                            ++loaded_solutions;
                        }
                    }
                }
            }
            else
            {
                QString const info = SolutionHolder::infoOfSolution(index, 0);

                if (!info.isEmpty() && (info[0] == QChar(0)))
                {
                    QString const new_info = info.right(info.length() - 1);
                    SolutionHolder::changeSolutionInfo(index, 0, new_info);
                    ++loaded_solutions;
                }

                for (int i = number_of_solutions - 1; i >= 1; --i)
                {
                    QString const info = SolutionHolder::infoOfSolution(index, 0);

                    if (!info.isEmpty() && (info[0] == QChar(0)))
                    {
                        SolutionHolder::deleteSolution(index, i);
                    }
                }
            }
        }
    }

    QString info;

    info += i18n("%1 of the solutions had an invalid map.").arg(invalid_maps) + '\n';
    info += i18n("%1 of the solutions were not really solutions.").arg(invalid_moves) + '\n';
    info += i18n("%1 of the solutions were for other levels.").arg(ignored_maps) + '\n';
    info += i18n("%1 of the solutions were further optimized.").arg(optimized_solutions) + '\n';
    info += i18n("%1 of the solutions were actually imported.").arg(loaded_solutions) + '\n';

    KMessageBox::information(this, info);
}


void MainWindow::exportGame()
{
    stopAnimation();

    if (m_game->retroMode())
    {
        KMessageBox::error(this, i18n("This is not supported in retro mode!"));

        return;
    }

    saveDataAsXsbFile(currentGameToText(), "game");
}


void MainWindow::sendGame()
{
    stopAnimation();

    if (m_game->retroMode())
    {
        KMessageBox::error(this, i18n("This is not supported in retro mode!"));

        return;
    }

    mailData(currentGameToText());
}


void MainWindow::exportCollection()
{
    stopAnimation();

    saveDataAsXsbFile(actCollection()->toText(), "collection");
}



void MainWindow::sendCollection()
{
    stopAnimation();

    mailData(actCollection()->toText());
}


void MainWindow::exportSolutions()
{
    stopAnimation();

    QString solutions = solutionsText();

    if (!solutions.isEmpty())
    {
        saveDataAsXsbFile(solutions, "solution");
    }
}


void MainWindow::sendSolutions()
{
    stopAnimation();

    QString solutions = solutionsText();

    if (!solutions.isEmpty())
    {
        mailData(solutions);
    }
}


QString MainWindow::currentGameToText()
{
    QString result = actLevel().toText(QStringList(), QStringList(), "", "", "", -1);

    result += Collection::levelNameAndCollection(actLevel(), m_level_nr, *actCollection());

    Movements short_moves = actLevel().map().expandMoves(m_game->moves(), false);
    result += short_moves.toText();

    return result;
}


QString MainWindow::solutionsText()
{
    ExportSolutionsDialog dialog(m_was_solved, this);

    if (!dialog.exec())
    {
        return "";
    }

    std::vector<int> collection_nrs;
    std::vector<int> level_nrs;

    if (dialog.exportCurrentLevel())
    {
        collection_nrs.push_back(m_collection_nr);
        level_nrs.push_back(m_level_nr);
    }
    else
    {
        int from_collection = 0;
        int to_collection = CollectionHolder::numberOfCollections();

        if (dialog.exportCollection())
        {
            from_collection = m_collection_nr;
            to_collection = m_collection_nr + 1;
        }

        for (int i = from_collection; i < to_collection; ++i)
        {
            Collection * collection = CollectionHolder::collection(i);
            int const number_of_levels = collection->numberOfLevels();

            for (int j = 0; j < number_of_levels; ++j)
            {
                CompressedMap const & compressed_map = collection->level(j).compressedMap();

                if (SolutionHolder::hasSolution(compressed_map))
                {
                    collection_nrs.push_back(i);
                    level_nrs.push_back(j);
                }
            }
        }
    }

    int const solution_nrs = static_cast<int> (collection_nrs.size());

    if (solution_nrs == 0)
    {
        KMessageBox::information(0, i18n("No solutions to save!"));

        return "";
    }

    CreateSolutionsDialog create_dialog(collection_nrs, level_nrs, dialog.solutionRegexp(),
                                        dialog.exportOnlyBest(), this);

    create_dialog.exec();

    if (create_dialog.finished())
    {
        return create_dialog.solutions();
    }
    else
    {
        return "";
    }

}


KURL MainWindow::getSaveUrl(QString const & dialog_tag, QString const & filter) const
{
    KURL url = KFileDialog::getSaveURL(":" + dialog_tag, filter);

    if (!url.isValid() || url.isEmpty())
    {
        return KURL();
    }

    if (KIO::NetAccess::exists(url))
    {
        KConfig * config = KApplication::kApplication()->config();
        KConfigGroupSaver group_saver(config, "Notification Messages");

        if (config->readBoolEntry("Overwrite " + dialog_tag, true))
        {
            int result = KMessageBox::warningContinueCancel(0, i18n("Do you want to overwrite "
                                                                    "the existing file?"),
                                                            i18n("Overwrite File"), i18n("OK"),
                                                            "Overwrite " + dialog_tag, false);

            if (result == KMessageBox::Cancel)
            {
                return KURL();
            }
        }
    }

    return url;
}


void MainWindow::saveDataAsXsbFile(QString const & data, QString const & dialog_tag) const
{
    KURL url = getSaveUrl(dialog_tag, i18n("*.xsb|Sokoban files\n*|All files"));

    if (url.isEmpty())
    {
        return;
    }

    KTempFile file;
    file.setAutoDelete(true);

    QTextStream & stream = *(file.textStream());

    stream << data;

    file.close();

    if (!KIO::NetAccess::upload(file.name(), url))
    {
        KMessageBox::error(0, i18n("Saving failed"));
    }
}


void MainWindow::mailData(QString const & data) const
{
    KApplication::kApplication()->invokeMailer("mailto:?subject=&body=" + data);
}


QStringList MainWindow::loadDataFromXsbFile(QString const & dialog_tag) const
{
    QStringList result;

    KURL url = KFileDialog::getOpenURL(":" + dialog_tag, i18n("*.xsb|Sokoban files\n*|All files"));

    if (!url.isValid() && url.isEmpty())
    {
        return result;
    }

    QString filename;

    if (!KIO::NetAccess::download(url, filename))
    {
        KMessageBox::error(0, i18n("Could not open location!"));

        return result;
    }

    QFileInfo file_info(filename);

    int const megabytes = static_cast<int> (file_info.size() / (1024 * 1024));

    if (megabytes >= 1)
    {
        if (KMessageBox::warningContinueCancel(const_cast<MainWindow *> (this),
                                               i18n("The file '%1'\nhas a size of %2 megabyte.\n"
                                                    "This is very unusual. Do you want to "
                                                    "continue the loading?").arg(filename).arg(megabytes),
                                               QString::null, i18n("Continue"),
                                               "Warn because of big files", false) == KMessageBox::Cancel)
        {
            return result;
        }
    }

    QFile file(filename);

    if (!file.open(IO_ReadOnly))
    {
        KMessageBox::error(0, i18n("Could not open copied file!"));

        return result;
    }

    QTextStream stream(&file);

    while (!stream.atEnd())
    {
        result.append(stream.readLine());
    }

    return result;
}


void MainWindow::exportGameImage()
{
    stopAnimation();

    ImageStorerDialog dialog(actLevel().map(), ThemeHolder::theme(m_theme), this);

    if (!dialog.exec())
    {
        return;
    }

    KURL url = getSaveUrl("image", i18n("*.png|PNG images") + "\n" + i18n("*.bmp|BMP images") + "\n" +
                                   i18n("*.xpm|XPM images") + "\n" + i18n("*|All files"));

    if (url.isEmpty())
    {
        return;
    }

    KTempFile file;
    file.setAutoDelete(true);

    char const * image_type = imageType(url);

    if (!image_type || (image_type == QString("MNG")))
    {
        KMessageBox::information(this, i18n("Unsupported format"));

        return;
    }

    if (!ImageStorer::store(file.file(), image_type, *m_map, dialog.pieceSize(), *ThemeHolder::theme(m_theme),
                            m_game->keeperDirection(), !dialog.transparentBackground(), dialog.lowQuality()))
    {
        KMessageBox::error(this, i18n("Saving failed"));

        return;
    }

    file.close();

    if (!KIO::NetAccess::upload(file.name(), url))
    {
        KMessageBox::error(0, i18n("Saving failed"));
    }
}


void MainWindow::exportGameAnimation()
{
    stopAnimation();

#ifndef EASYSOK_HAVE_MNG
    KMessageBox::error(this, i18n("Easysok was not compiled with mng-support!"));

    return;
#endif

    if (m_game->retroMode())
    {
        KMessageBox::error(this, i18n("This is not supported in retro mode!"));

        return;
    }

    AnimationStorerDialog dialog(actLevel().map(), ThemeHolder::theme(m_theme), this);

    if (!dialog.exec())
    {
        return;
    }

    KURL url = getSaveUrl("animation", i18n("*.mng|MNG animations\n*|All files"));

    if (url.isEmpty())
    {
        return;
    }

    Movements short_moves = actLevel().map().expandMoves(m_game->moves(), false);

#ifdef EASYSOK_HAVE_MNG

    m_animation_storer = new MngAnimationStorer(actLevel().map(), short_moves, *ThemeHolder::theme(m_theme),
                                                dialog.pieceSize(), !dialog.transparentBackground(),
                                                dialog.startDelay(), dialog.delay(), dialog.cylcle(),
                                                dialog.endDelay(), dialog.lowQuality(), url);
#endif

    m_animation_timer = new QTimer(this);
    connect(m_animation_timer, SIGNAL(timeout()), this, SLOT(animationWriteStep()));

    m_progress_dialog = new ProgressDialog(i18n("Save animation"), i18n("Saving game as animation"),
                                           short_moves.numberOfMoves(), this);

    m_animation_timer->start(0, true);

    m_progress_dialog->exec();

    delete m_animation_storer;
    delete m_animation_timer;
    delete m_progress_dialog;
}


void MainWindow::animationWriteStep()
{
#ifdef EASYSOK_HAVE_MNG
    MngAnimationStorer::Result result = m_animation_storer->proccess();

    m_progress_dialog->advance();

    if (result == MngAnimationStorer::FINISHED)
    {
        m_progress_dialog->close();
    }
    else if (result == MngAnimationStorer::ERROR)
    {
        KMessageBox::error(this, i18n("An error occured!"));
        m_progress_dialog->close();
    }
    else
    {
        m_animation_timer->start(0, true);
    }
#endif
}


char const * MainWindow::imageType(KURL const & url) const
{
    QString const filename = url.fileName();

    QRegExp png_regexp("[.]png$", false);
    QRegExp bmp_regexp("[.]bmp$", false);
    QRegExp xpm_regexp("[.]xpm$", false);

    if (png_regexp.search(filename, 0) != -1)
    {
        return "PNG";
    }
    else if (bmp_regexp.search(filename, 0) != -1)
    {
        return "BMP";
    }
    else if (xpm_regexp.search(filename, 0) != -1)
    {
        return "XPM";
    }
    else
    {
        return 0;
    }
}


void MainWindow::sendSolutionToServer(Map const & map, Movements const & moves, int nr_of_pushes,
                                      int nr_of_linear_pushes, int nr_of_gem_changes, int nr_of_moves)
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QString user = config->readEntry("Current user", "");

    if (user.isEmpty())
    {
        return;
    }

    config->setGroup("UserGroup-" + user);
    QString passwd = config->readEntry("Password", "");
    QString server = config->readEntry("Server", "");
    QString proxy = config->readEntry("Proxy", "");
    int proxy_port = config->readNumEntry("Proxy port", 8080);

    config->setGroup("");

    QString moves_text = map.expandMoves(moves, false).toText();
    moves_text = moves_text.right(moves_text.length() - 4);
    moves_text = moves_text.replace(QRegExp("\n"), "");

    QString query = "nickname=" + KURL::encode_string(user);
    query += "&passwd=" + KURL::encode_string(passwd);
    query += "&nrpushes=" + KURL::encode_string(QString::number(nr_of_pushes));
    query += "&nrlinearpushes=" + KURL::encode_string(QString::number(nr_of_linear_pushes));
    query += "&nrgemchanges=" + KURL::encode_string(QString::number(nr_of_gem_changes));
    query += "&nrmoves=" + KURL::encode_string(QString::number(nr_of_moves));
    query += "&map=" + KURL::encode_string(map.toServerFormat());
    query += "&moves=" + KURL::encode_string(moves_text);

    QueryDialog dialog(i18n("Sending Solution"), i18n("Trying send the solution to the server ..."),
                       KURL(server), proxy, proxy_port, "add_score.php", query, this);

    if (!dialog.exec() || (dialog.result() != ServerConnector::OK))
    {
        KMessageBox::error(this, dialog.resultText());
    }
    else
    {
        QStringList data = QStringList::split('\n', dialog.data(), true);

        if (data.count() < 4)
        {
            KMessageBox::error(this, i18n("Could not get result from server!"));
        }
        else
        {
            bool const best_pushes = (data[0] == "yes");
            bool const best_linear_pushes = (data[1] == "yes");
            bool const best_gem_changes = (data[2] == "yes");
            bool const best_moves = (data[3] == "yes");

            QString msg;

            if (best_pushes)
            {
                msg = i18n("<b></b>The solution is your <b>best</b> in terms of the\n"
                           "<b>number of pushes</b>.<br>");
            }
            else
            {
                msg = i18n("<b></b>The solution is <b>not your best</b> in terms\n"
                           "of the <b>number of pushes</b>.<br>");
            }

            if (best_linear_pushes)
            {
                msg += i18n("The solution is your <b>best</b> in terms of the\n"
                           "<b>number of linear pushes</b>.<br>");
            }
            else
            {
                msg += i18n("The solution is <b>not your best</b> in terms\n"
                           "of the <b>number of linear pushes</b>.<br>");
            }

            if (best_gem_changes)
            {
                msg += i18n("The solution is your <b>best</b> in terms of the\n"
                           "<b>number of gem changes</b>.<br>");
            }
            else
            {
                msg += i18n("The solution is <b>not your best</b> in terms\n"
                           "of the <b>number of gem changes</b>.<br>");
            }

            if (best_moves)
            {
                msg += i18n("The solution is your <b>best</b> in terms of the\n"
                           "<b>number of moves</b>.<br>");
            }
            else
            {
                msg += i18n("The solution is <b>not your best</b> in terms\n"
                           "of the <b>number of moves</b>.<br>");
            }

            KMessageBox::information(this, msg);
        }
    }
}


void MainWindow::sendSolutionsOfLevels(std::vector<Level const *> const & levels, bool only_one)
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QString user = config->readEntry("Current user", "");

    if (user.isEmpty())
    {
        KMessageBox::error(this, i18n("You must first set a user!"));

        return;
    }

    if (levels.empty())
    {
        KMessageBox::information(this, i18n("No levels with solutions found"));

        return;
    }

    config->setGroup("UserGroup-" + user);
    QString passwd = config->readEntry("Password", "");
    QString server = config->readEntry("Server", "");
    QString proxy = config->readEntry("Proxy", "");
    int proxy_port = config->readNumEntry("Proxy port", 8080);

    config->setGroup("");

    SendSolutionsDialog dialog(server, proxy, proxy_port, user, passwd, levels, this);

    dialog.exec();

    if (dialog.result() != ServerConnector::OK)
    {
        KMessageBox::error(this, ServerConnector::resultText(dialog.result()));
    }
    else
    {
        QString msg;

        if (only_one)
        {
            bool const best_pushes = dialog.bestPushes();
            bool const best_linear_pushes = dialog.bestLinearPushes();
            bool const best_gem_changes = dialog.bestGemChanges();
            bool const best_moves = dialog.bestMoves();

            if (best_pushes)
            {
                msg += i18n("<b></b>One solution was your <b>best</b> in terms of the\n"
                            "<b>number of pushes</b>.<br>");
            }

            if (best_linear_pushes)
            {
                msg += i18n("<b></b>One solution was your <b>best</b> in terms of the\n"
                            "<b>number of linear pushes</b>.<br>");
            }

            if (best_gem_changes)
            {
                msg += i18n("<b></b>One solution was your <b>best</b> in terms of the\n"
                            "<b>number of gem changes</b>.<br>");
            }

            if (best_moves)
            {
                msg += i18n("<b></b>One solution was your <b>best</b> in terms of the\n"
                            "<b>number of moves</b>.<br>");
            }

            if (msg.isEmpty())
            {
                msg = i18n("No improvements were made.");
            }
        }
        else
        {
            msg = i18n("<b>%1</b> solutions were best in term of the\n"
                       "number of <b>pushes</b>.<br>").arg(dialog.bestPushes());
            msg += i18n("<b>%1</b> solutions were best in term of the\n"
                        "number of <b>linear pushes</b>.<br>").arg(dialog.bestLinearPushes());
            msg += i18n("<b>%1</b> solutions were best in term of the\n"
                        "number of <b>gem changes</b>.<br>").arg(dialog.bestGemChanges());
            msg += i18n("<b>%1</b> solutions were best in term of the\n"
                        "number of <b>moves</b>.<br>").arg(dialog.bestMoves());
        }

        KMessageBox::information(this, msg);
    }
}


void MainWindow::sendBestSolutionToServer()
{
    if (!m_was_solved)
    {
        KMessageBox::error(this, i18n("You have not solved this level yet!"));

        return;
    }

    std::vector<Level const *> levels;
    levels.push_back(&(actLevel()));

    sendSolutionsOfLevels(levels, true);
}


void MainWindow::sendBestCollectionSolutionToServer()
{
    std::vector<Level const *> levels;

    Collection * collection = actCollection();
    int const nr_of_levels = collection->numberOfLevels();

    for (int i = 0; i < nr_of_levels; ++i)
    {
        levels.push_back(&(collection->level(i)));
    }

    sendSolutionsOfLevels(levels, false);
}


void MainWindow::sendBestAllSolutionToServer()
{
    std::vector<Level const *> levels;

    int const nr_of_collections = CollectionHolder::numberOfCollections();

    for (int i = 0; i < nr_of_collections; ++i)
    {
        Collection * collection = CollectionHolder::collection(i);
        int const nr_of_levels = collection->numberOfLevels();

        for (int j = 0; j < nr_of_levels; ++j)
        {
            levels.push_back(&(collection->level(j)));
        }
    }

    sendSolutionsOfLevels(levels, false);
}


void MainWindow::sendByDateToServer()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QString user = config->readEntry("Current user", "");

    if (user.isEmpty())
    {
        KMessageBox::error(this, i18n("You must first set a user!"));

        return;
    }

    SelectDateDialog dialog(this);

    if (!dialog.exec())
    {
        return;
    }

    QDateTime from_date(dialog.date());

    std::vector<Level const *> levels;

    int const nr_of_collections = CollectionHolder::numberOfCollections();

    for (int i = 0; i < nr_of_collections; ++i)
    {
        Collection * collection = CollectionHolder::collection(i);
        int const nr_of_levels = collection->numberOfLevels();

        for (int j = 0; j < nr_of_levels; ++j)
        {
            int const level_index = SolutionHolder::getIndexForMap(collection->level(j).compressedMap());

            if (level_index != -1)
            {
                int const nr_of_solutions = SolutionHolder::numberOfSolutions(level_index);

                for (int k = 0; k < nr_of_solutions; ++k)
                {
                    if (SolutionHolder::dateOfSolution(level_index, k) > from_date)
                    {
                        levels.push_back(&(collection->level(j)));

                        break;
                    }
                }
            }
        }
    }

    sendSolutionsOfLevels(levels, false);
}


void MainWindow::compareSolution()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QString user = config->readEntry("Current user", "");

    if (user.isEmpty())
    {
        KMessageBox::error(this, i18n("You must first set a user!"));

        return;
    }

    config->setGroup("UserGroup-" + user);
    QString passwd = config->readEntry("Password", "");
    QString server = config->readEntry("Server", "");
    QString proxy = config->readEntry("Proxy", "");
    int proxy_port = config->readNumEntry("Proxy port", 8080);
    config->setGroup("");

    std::vector<Level const *> levels;
    levels.push_back(&(actLevel()));

    SendSolutionsDialog send_dialog(server, proxy, proxy_port, user, passwd, levels, this);

    send_dialog.exec();

    if (send_dialog.result() != ServerConnector::OK)
    {
        KMessageBox::error(this, ServerConnector::resultText(send_dialog.result()));
    }

    QString query = "nickname=" + KURL::encode_string(user);
    query += "&map=" + KURL::encode_string(actLevel().map().toServerFormat());

    QueryDialog dialog(i18n("Comparing Solution"),
                       i18n("Trying to compare your solution with that on the server ..."),
                       KURL(server), proxy, proxy_port, "get_score.php", query, this);

    if (!dialog.exec() || (dialog.result() != ServerConnector::OK))
    {
        if (dialog.result() == ServerConnector::LEVEL_NOT_SOLVED)
        {
            KMessageBox::error(this, i18n("You must first send your solution to the server!"));
        }
        else
        {
            KMessageBox::error(this, dialog.resultText());
        }
    }
    else
    {
        QStringList data = QStringList::split('\n', dialog.data(), true);

        if (data.count() < 25)
        {
            KMessageBox::error(this, i18n("Could not get result from server!"));
        }
        else
        {
            int const my_pushmin_pushes = data[0].toInt();
            int const my_pushmin_moves = data[1].toInt();
            int const my_linearmin_pushes = data[2].toInt();
            int const my_linearmin_moves = data[3].toInt();
            int const my_gemchanges_pushes = data[4].toInt();
            int const my_gemchanges_moves = data[5].toInt();
            int const my_movemin_pushes = data[6].toInt();
            int const my_movemin_moves = data[7].toInt();
            int const nr_of_solutions = data[8].toInt();
            int const better_pushes = data[9].toInt();
            int const equal_pushes = data[10].toInt();
            int const better_linear_pushes = data[11].toInt();
            int const equal_linear_pushes = data[12].toInt();
            int const better_gem_changes = data[13].toInt();
            int const equal_gem_changes = data[14].toInt();
            int const better_moves = data[15].toInt();
            int const equal_moves = data[16].toInt();
            int const best_pushmin_pushes = data[17].toInt();
            int const best_pushmin_moves = data[18].toInt();
            int const best_linearmin_pushes = data[19].toInt();
            int const best_linearmin_moves = data[20].toInt();
            int const best_gemchanges_pushes = data[21].toInt();
            int const best_gemchanges_moves = data[22].toInt();
            int const best_movemin_pushes = data[23].toInt();
            int const best_movemin_moves = data[24].toInt();

            CompareSolutionDialog dialog(my_pushmin_pushes, my_pushmin_moves, my_linearmin_pushes,
                                         my_linearmin_moves, my_gemchanges_pushes, my_gemchanges_moves,
                                         my_movemin_pushes, my_movemin_moves, best_pushmin_pushes,
                                         best_pushmin_moves, best_linearmin_pushes, best_linearmin_moves,
                                         best_gemchanges_pushes, best_gemchanges_moves, best_movemin_pushes,
                                         best_movemin_moves, nr_of_solutions, better_pushes, equal_pushes,
                                         better_linear_pushes, equal_linear_pushes, better_gem_changes,
                                         equal_gem_changes, better_moves, equal_moves, this);

            dialog.exec();
        }
    }
}


void MainWindow::getBestSolutionFromServer()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    QString user = config->readEntry("Current user", "");

    if (user.isEmpty())
    {
        KMessageBox::error(this, i18n("You must first set a user!"));

        return;
    }

    config->setGroup("UserGroup-" + user);
    QString passwd = config->readEntry("Password", "");
    QString server = config->readEntry("Server", "");
    QString proxy = config->readEntry("Proxy", "");
    int proxy_port = config->readNumEntry("Proxy port", 8080);
    config->setGroup("");

    std::vector<Level const *> levels;
    levels.push_back(&(actLevel()));

    SendSolutionsDialog send_dialog(server, proxy, proxy_port, user, passwd, levels, this);

    send_dialog.exec();

    if (send_dialog.result() != ServerConnector::OK)
    {
        KMessageBox::error(this, ServerConnector::resultText(send_dialog.result()));
    }

    QString query = "map=" + KURL::encode_string(actLevel().map().toServerFormat());

    QueryDialog dialog(i18n("Get Best Solution"),
                       i18n("Trying to get the best solution from the server ..."),
                       KURL(server), proxy, proxy_port, "get_highscore.php", query, this);

    if (!dialog.exec() || (dialog.result() != ServerConnector::OK))
    {
        KMessageBox::error(this, dialog.resultText());
    }
    else
    {
        QStringList data = QStringList::split('\n', dialog.data(), true);

        if (data.count() < 8)
        {
            KMessageBox::error(this, i18n("Could not get result from server!"));
        }
        else
        {
            int const pushmin_pushes = data[0].toInt();
            int const pushmin_moves = data[1].toInt();
            int const linear_pushmin_pushes = data[2].toInt();
            int const linear_pushmin_moves = data[3].toInt();
            int const gemchanges_pushes = data[4].toInt();
            int const gemchanges_moves = data[5].toInt();
            int const movemin_pushes = data[6].toInt();
            int const movemin_moves = data[7].toInt();

            if (pushmin_pushes >= 1000000)
            {
                KMessageBox::information(this, i18n("This level was not solved!"));
            }
            else
            {
                QString msg = i18n("<p>The best <b>push optimal</b> solution of this level consists of "
                                   "%1 pushes and %2 moves.</p>");
                msg = msg.arg(pushmin_pushes).arg(pushmin_moves);
                msg += i18n("<p>The best <b>linear push optimal</b> solution of this level consists of "
                            "%1 pushes and %2 moves.</p>");
                msg = msg.arg(linear_pushmin_pushes).arg(linear_pushmin_moves);
                msg += i18n("<p>The best <b>gem change optimal</b> solution of this level consists of "
                            "%1 gem changes and %2 moves.</p>");
                msg = msg.arg(gemchanges_pushes).arg(gemchanges_moves);
                msg += i18n("<p>The best <b>move optimal</b> solution of this level consists of "
                            "%1 pushes and %2 moves.</p>");
                msg = msg.arg(movemin_pushes).arg(movemin_moves);

                KMessageBox::information(this, msg);
            }
        }
    }
}
