/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


// Here we have parked the whole setup stuff (creating actions, accels, menus etc).


#include "main_window.h"

#include <config.h>

#include <cassert>

#include <qsignalmapper.h>
#include <qtimer.h>
#include <qtooltip.h>
#include <qwhatsthis.h>

#include <kaction.h>
#include <kapp.h>
#include <kconfig.h>
#include <kedittoolbar.h>
#include <klocale.h>
#include <kpopupmenu.h>
#include <kstdaccel.h>
#include <kstdaction.h>

#include "bookmarks.h"
#include "collection.h"
#include "collection_holder.h"
#include "configuration_dialog.h"
#include "difficulty.h"
#include "game.h"
#include "map_widget.h"
#include "theme.h"
#include "theme_holder.h"




void MainWindow::showOptionsDialog()
{
    ConfigurationDialog dialog;

    connect(&dialog, SIGNAL(configurationChanged()), this, SLOT(configChanged()));

    dialog.exec();
}


void MainWindow::showToolbarDialog()
{
    KEditToolbar dialog(actionCollection(), "easysokui.rc");
    connect(&dialog, SIGNAL(newToolbarConfig()), this, SLOT(updateToolbar()));

    if (dialog.exec())
    {
        createGUI();
    }
}


void MainWindow::updateToolbar()
{
    createGUI();
}


void MainWindow::configChanged()
{
    setupBookmarkMenu();
    m_game->configChanged();
    m_map_widget->configChanged();
    updateLevelActions();
    readChangedOptions();
    setupRecentCollectionsMenu();
}


void MainWindow::readConfig()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_collection_nr = config->readNumEntry("Collection", 0);
    int const max_collections = CollectionHolder::numberOfCollections();
    m_collection_nr = std::min(std::max(0, m_collection_nr), max_collections - 1);

    m_level_nr = config->readNumEntry("Level", 0);
    m_level_nr = std::min(std::max(0, m_level_nr), actCollection()->numberOfLevels() - 1);

    m_theme = config->readNumEntry("Theme", 0);
    m_theme = std::max(std::min(m_theme, ThemeHolder::numberOfThemes() - 1), 0);

    m_hide_gems = config->readBoolEntry("Hide gems", false);
    ThemeHolder::theme(m_theme)->setHideGems(m_hide_gems);

    m_hide_goals = config->readBoolEntry("Hide goals", false);
    ThemeHolder::theme(m_theme)->setHideGoals(m_hide_goals);

    m_outside_as_wall = config->readBoolEntry("Outside as wall", false);
    ThemeHolder::theme(m_theme)->setOutsideAsWall(m_outside_as_wall);

    m_honor_deadlocks = config->readBoolEntry("Honor deadlocks", false);

    m_show_arrows = config->readBoolEntry("Show possible pushes", false);

    m_animation_speed = config->readNumEntry("Animation speed", 1);
    m_animation_speed = std::min(std::max(m_animation_speed, 0), 3);

    m_auto_optimize_moves = config->readBoolEntry("Automatically Optimize Moves", false);

    m_auto_optimize_pushes = config->readBoolEntry("Automatically Optimize Pushes", true);

    m_auto_send_to_server = config->readBoolEntry("Automatically send solutions to server", false);

    QRect default_geometry(0, 0, 640, 480);
    setGeometry(config->readRectEntry("Main window geometry", &default_geometry));

    m_recent_collections.resize(0);

    for (int i = 1; i <= 11; ++i)
    {
        int collection = config->readNumEntry("Recent collection " + QString::number(i), -1);

        if ((collection >= 0) && (collection < max_collections))
        {
            m_recent_collections.push_back(collection);
        }
    }

    removeDoubleRecentCollections();

    readChangedOptions();
}


void MainWindow::readChangedOptions()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_goto_any_level = config->readBoolEntry("Goto any level", true);

    m_next_level_when_solved = config->readBoolEntry("Jump to next level when solved", true);
    m_illegal_as_virtual = config->readBoolEntry("Illegal moves as virtual", false);
    m_show_difficulty = config->readBoolEntry("Show difficulty", true);
    m_show_linear_pushes = config->readBoolEntry("Show linear pushes", false);
    m_show_gem_changes = config->readBoolEntry("Show gem changes", false);
    m_show_gems_left = config->readBoolEntry("Show gems left", true);
    m_show_author = config->readBoolEntry("Show author", true);
    m_show_user = config->readBoolEntry("Show user", true);
    m_min_status_update_time = config->readNumEntry("Min status update time", 100);

    m_nr_of_recent_collections = config->readNumEntry("Number of recent collections", 5);
    m_nr_of_recent_collections = std::min(std::max(1, m_nr_of_recent_collections), 10);

    m_auto_save_time = config->readNumEntry("Auto save time", 15);

    if (m_auto_save_time > 0)
    {
        m_state_timer->start(m_auto_save_time * 60000);
    }
    else
    {
        m_state_timer->stop();
    }

    if (!m_collection_actions.empty())
    {
        updateUserStatusBar();
        setLevel(m_collection_nr, m_level_nr);
    }
}


void MainWindow::writeConfig()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    config->writeEntry("Collection", m_collection_nr);
    config->writeEntry("Level", m_level_nr);
    config->writeEntry("Theme", m_theme);
    config->writeEntry("Hide gems", m_hide_gems);
    config->writeEntry("Hide goals", m_hide_goals);
    config->writeEntry("Outside as wall", m_outside_as_wall);
    config->writeEntry("Honor deadlocks", m_honor_deadlocks);
    config->writeEntry("Show possible pushes", m_show_arrows);
    config->writeEntry("Goto any level", m_goto_any_level);
    config->writeEntry("Bookmark format", m_bookmark_format);
    config->writeEntry("Number of bookmarks", m_number_of_bookmarks);
    config->writeEntry("Jump to next level when solved", m_next_level_when_solved);
    config->writeEntry("Animation speed", m_animation_speed);
    config->writeEntry("Automatically Optimize Moves", m_auto_optimize_moves);
    config->writeEntry("Automatically Optimize Pushes", m_auto_optimize_pushes);
    config->writeEntry("Main window geometry", geometry());
    config->writeEntry("Automatically send solutions to server", m_auto_send_to_server);
    config->writeEntry("Number of recent collections", m_nr_of_recent_collections);

    int const nr_of_recent_collections = static_cast<int> (m_recent_collections.size());

    for (int i = 0; i < nr_of_recent_collections; ++i)
    {
        config->writeEntry("Recent collection " + QString::number(i + 1), m_recent_collections[i]);
    }

    config->writeEntry("Toolbar visible", toolBar()->isVisible());
}


void MainWindow::setupActions()
{
    KActionCollection * action_col = actionCollection();

    KStdAction::keyBindings(this, SLOT(showKeyDialog()), action_col);
    KStdAction::configureToolbars(this, SLOT(showToolbarDialog()), action_col);
    KStdAction::preferences(this, SLOT(showOptionsDialog()), action_col);


    // The file menu

    new KAction(i18n("&Open Game ..."), "fileopen", KStdAccel::key(KStdAccel::Open),
                this, SLOT(importGame()), action_col, "ImportGame");
    new KAction(i18n("&Import Collection ..."), 0,
                this, SLOT(importCollection()), action_col, "ImportCollection");
    new KAction(i18n("I&mport Solutions ..."), CTRL | Key_I,
                this, SLOT(importSolutions()), action_col, "ImportSolutions");
    new KAction(i18n("&Save Game ..."), "filesave", KStdAccel::key(KStdAccel::Save),
                this, SLOT(exportGame()), action_col, "ExportGame");
    new KAction(i18n("S&end Game ..."), "mail_generic", 0,
                this, SLOT(sendGame()), action_col, "SendGame");
    new KAction(i18n("&Export Current Collection ..."), 0,
                this, SLOT(exportCollection()), action_col, "ExportCollection");
    new KAction(i18n("Sen&d Current Collection ..."), "mail_generic", 0,
                this, SLOT(sendCollection()), action_col, "SendCollection");
    new KAction(i18n("E&xport Solutions ..."), 0,
                this, SLOT(exportSolutions()), action_col, "ExportSolutions");
    new KAction(i18n("Send So&lutions ..."), "mail_generic", 0,
                this, SLOT(sendSolutions()), action_col, "SendSolutions");
    new KAction(i18n("Export &Current Map As Image ..."), 0,
                this, SLOT(exportGameImage()), action_col, "ExportGameImage");
    new KAction(i18n("Export Game As &Animation ..."), 0,
                this, SLOT(exportGameAnimation()), action_col, "ExportGameAnimation");
    new KAction(i18n("&Print ..."), "fileprint", KStdAccel::key(KStdAccel::Print),
                this, SLOT(print()), action_col, "Print");
    new KAction(i18n("&Quit"), "exit", KStdAccel::key(KStdAccel::Quit),
                this, SLOT(quit()), action_col, "Quit");


    // The edit menu

    new KAction(i18n("&Edit Current Level ..."), "edit", CTRL | Key_E,
                this, SLOT(editCurrentLevel()), action_col, "EditCurrentLevel");
    new KAction(i18n("&Insert New Level ..."), CTRL | Key_L,
                this, SLOT(insertNewLevel()), action_col, "InsertNewLevel");
    new KAction(i18n("I&nsert New Collection ..."), CTRL | Key_N,
                this, SLOT(insertNewCollection()), action_col, "InsertNewCollection");
    new KAction(i18n("&Delete Current Level"), "editcut", KStdAccel::key(KStdAccel::Cut),
                this, SLOT(deleteCurrentLevel()), action_col, "DeleteCurrentLevel");
    new KAction(i18n("De&lete Current Collection ..."), "editdelete", 0,
                this, SLOT(deleteCurrentCollection()), action_col, "DeleteCollection");
    new KAction(i18n("&Copy Current Level"), "editcopy", KStdAccel::key(KStdAccel::Copy),
                this, SLOT(copyCurrentLevel()), action_col, "CopyCurrentLevel");
    new KAction(i18n("&Paste Level"), "editpaste", KStdAccel::key(KStdAccel::Paste),
                this, SLOT(pasteLevel()), action_col, "PasteLevel");
    new KAction(i18n("Paste Level in &New Collection"), 0,
                this, SLOT(pasteLevelInNewCollection()), action_col, "PasteLevelNew");
    new KAction(i18n("D&uplicate Current Level"), CTRL | Key_D,
                this, SLOT(duplicateCurrentLevel()), action_col, "DuplicateLevel");
    new KAction(i18n("Duplica&te Current Collection ..."), SHIFT | Key_D,
                this, SLOT(duplicateCurrentCollection()), action_col, "DuplicateCollection");
    new KAction(i18n("Make Current Collection Non &Temporary"), 0,
                this, SLOT(makeCurrentCollectionNonTemporary()), action_col, "MakeNonTemporary");
    new KAction(i18n("&Reorder Collections and Levels ..."), 0,
                this, SLOT(reorderCollectionsAndLevels()), action_col, "ReorderCollectionLevels");


    // The level menu

    new KAction(i18n("&Previous Collection"), SHIFT | Key_P,
                this, SLOT(prevCollection()), action_col, "PrevCollection");
    new KAction(i18n("&Next Collection"), SHIFT | Key_N,
                this, SLOT(nextCollection()), action_col, "NextCollection");
    m_prev_level_action = new KAction(i18n("P&revious Level"), "back", Key_P,
                                      this, SLOT(prevLevel()), action_col, "PrevLevel");
    m_next_level_action= new KAction(i18n("N&ext Level"), "forward", Key_N,
                                     this, SLOT(nextLevel()), action_col, "NextLevel");
    new KAction(i18n("Pre&vious Unsolved Level"), "forward", SHIFT | CTRL | Key_P,
                this, SLOT(prevUnsolvedLevel()), action_col, "PrevUnsolvedLevel");
    new KAction(i18n("Ne&xt Unsolved Level"), "forward", SHIFT | CTRL | Key_N,
                this, SLOT(nextUnsolvedLevel()), action_col, "NextUnsolvedLevel");
    m_first_level_action = new KAction(i18n("&First Level"), "start", Key_F,
                                       this, SLOT(firstLevel()), action_col, "FirstLevel");
    m_last_level_action = new KAction(i18n("&Last Level"), "finish", Key_B,
                                      this, SLOT(lastLevel()), action_col, "LastLevel");
    new KAction(i18n("&Go To A Level ..."), "goto", Key_G,
                this, SLOT(selectLevel()), action_col, "SelectLevel");
    new KAction(i18n("&Show Level Information ..."), Key_I,
                this, SLOT(levelInfo()), action_col, "LevelInfo");
    new KAction(i18n("Show Collection &Information ..."), SHIFT | Key_I,
                this, SLOT(collectionInfo()), action_col, "CollectionInfo");
    new KAction(i18n("&Find Duplicate Levels ..."), 0,
                this, SLOT(findDups()), action_col, "FindDups");
    new KAction(i18n("Force &Update of Collections ..."), 0,
                this, SLOT(forceUpdateCollections()), action_col, "ForceUpdateCollections");


    // Solutions menu

    new KAction(i18n("&Solutions Of Current Level"), Key_S,
                this, SLOT(showSolutions()), action_col, "Solutions");
    new KAction(i18n("Best Solutions Of &Collection"), Key_H,
                this, SLOT(showHighscores()), action_col, "Highscores");
    m_load_best_sol_action = new KAction(i18n("Load &Best Solution"), Key_L,
                                         this, SLOT(loadBestSolution()), action_col, "LoadBestSolution");
    m_load_sol_action = new KAction(i18n("&Load A Solution ..."), SHIFT | Key_L,
                                    this, SLOT(loadASolution()), action_col, "LoadASolution");
    m_annotate_sol_action = new KAction(i18n("&Annotate Solution ..."), SHIFT | Key_A,
                                        this, SLOT(annotateSolution()), action_col, "AnnotateSolution");
    m_delete_sol_action = new KAction(i18n("&Delete Solutions Of Current Level ..."), "editdelete", 0,
                                      this, SLOT(deleteSolutions()), action_col, "DeleteSolutions");
    new KAction(i18n("Delete Solutions By A&nnotation ..."), "editdelete", 0,
                this, SLOT(deleteSolutionsByName()), action_col, "DeleteSolutionsByName");
    new KAction(i18n("S&end Best Solutions Of Level ..."), SHIFT | Key_E,
                this, SLOT(sendBestSolutionToServer()), action_col, "SendBest");
    new KAction(i18n("Sen&d Best Solutions Of Collection ..."), 0,
                this, SLOT(sendBestCollectionSolutionToServer()), action_col, "SendBestCollection");
    new KAction(i18n("Send Best Solutions Of &All Collections ..."), 0,
                this, SLOT(sendBestAllSolutionToServer()), action_col, "SendBestAll");
    new KAction(i18n("Send Best Solutions By Da&te ..."), 0,
                this, SLOT(sendByDateToServer()), action_col, "SendByDate");
    new KAction(i18n("S&how Best Solution"), SHIFT | Key_B,
                this, SLOT(getBestSolutionFromServer()), action_col, "GetBest");
    new KAction(i18n("C&ompare Best Solution"), Key_C,
                this, SLOT(compareSolution()), action_col, "CompareSolution");
    m_auto_send_to_server_action = new KToggleAction(i18n("A&utomatically Send Solutions"), 0, this,
                                                     SLOT(autoSendSolutions()), action_col,
                                                     "AutoSendSolutions");
    m_auto_send_to_server_action->setChecked(m_auto_send_to_server);

    new KAction(i18n("Se&t Current User ..."), 0,
                this, SLOT(setUser()), action_col, "SetUser");
    new KAction(i18n("Create Ne&w User ..."), 0,
                this, SLOT(createUser()), action_col, "CreateUser");
    new KAction(i18n("&Import User ..."), 0,
                this, SLOT(importUser()), action_col, "ImportUser");
    new KAction(i18n("Chan&ge User Data ..."), 0,
                this, SLOT(changeUser()), action_col, "ChangeUser");


    // The movements menu

    m_retro_mode_action = new KToggleAction(i18n("&Retro Mode"), "retro", 0,
                                            this, SLOT(retroModeChanged()), action_col, "RetroMode");
    m_retro_mode_action->setChecked(false);

    new KAction(i18n("Move left"), Key_Left,
                this, SLOT(moveLeft()), action_col, "MoveLeft");
    new KAction(i18n("Move right"), Key_Right,
                this, SLOT(moveRight()), action_col, "MoveRight");
    new KAction(i18n("Move up"), Key_Up,
                this, SLOT(moveUp()), action_col, "MoveUp");
    new KAction(i18n("Move down"), Key_Down,
                this, SLOT(moveDown()), action_col, "MoveDown");
    new KAction(i18n("Move maximum left"), SHIFT | Key_Left,
                this, SLOT(moveFarLeft()), action_col, "MoveFarLeft");
    new KAction(i18n("Move maximum right"), SHIFT | Key_Right,
                this, SLOT(moveFarRight()), action_col, "MoveFarRight");
    new KAction(i18n("Move maximum up"), SHIFT | Key_Up,
                this, SLOT(moveFarUp()), action_col, "MoveFarUp");
    new KAction(i18n("Move maximum down"), SHIFT | Key_Down,
                this, SLOT(moveFarDown()), action_col, "MoveFarDown");
    new KAction(i18n("Move virtually left"), CTRL | Key_Left,
                this, SLOT(moveVirtualLeft()), action_col, "MoveVirtualLeft");
    new KAction(i18n("Move virtually right"), CTRL | Key_Right,
                this, SLOT(moveVirtualRight()), action_col, "MoveVirtualRight");
    new KAction(i18n("Move virtually up"), CTRL | Key_Up,
                this, SLOT(moveVirtualUp()), action_col, "MoveVirtualUp");
    new KAction(i18n("Move virtually down"), CTRL | Key_Down,
                this, SLOT(moveVirtualDown()), action_col, "MoveVirtualDown");
    new KAction(i18n("Move left and pull gem"), SHIFT | ALT | Key_Left,
                this, SLOT(pullMoveLeft()), action_col, "PullMoveLeft");
    new KAction(i18n("Move right and pull gem"), SHIFT | ALT | Key_Right,
                this, SLOT(pullMoveRight()), action_col, "PullMoveRight");
    new KAction(i18n("Move up and pull gem"), SHIFT | ALT | Key_Up,
                this, SLOT(pullMoveUp()), action_col, "PullMoveUp");
    new KAction(i18n("Move down and pull gem"), SHIFT | ALT | Key_Down,
                this, SLOT(pullMoveDown()), action_col, "PullMoveDown");
    m_undo_action = new KAction(i18n("&Undo Move"), "undo", Key_U,
                                this, SLOT(undo()), action_col, "UndoMove");
    m_redo_action = new KAction(i18n("&Redo Move"), "redo", Key_R,
                                this, SLOT(redo()), action_col, "RedoMove");
    m_jump_start_action = new KAction(i18n("Jump To &Start"), "top", Key_Escape,
                                      this, SLOT(jumpToStart()), action_col, "JumpToStart");
    m_jump_end_action = new KAction(i18n("Jump To &End"), "bottom", Key_E,
                                    this, SLOT(jumpToEnd()), action_col, "JumpToEnd");
    m_play_action = new KAction(i18n("&Play Moves"), "run", SHIFT | Key_Space,
                                this, SLOT(playMoves()), action_col, "PlayMoves");
    new KAction(i18n("Sto&p Animation"), "stop", Key_Space,
                this, SLOT(stopAnimation()), action_col, "StopAnimation");

    QSignalMapper * animation_mapper = new QSignalMapper(this);
    connect(animation_mapper, SIGNAL(mapped(int)), this, SLOT(changeAnimations(int)));

    KRadioAction * no_animation_action = new KRadioAction(i18n("&No Animations"), Key_1, animation_mapper,
                                                          SLOT(map()), action_col, "NoAnimations");
    no_animation_action->setExclusiveGroup("animation");
    no_animation_action->setChecked(m_animation_speed == 0);
    animation_mapper->setMapping(no_animation_action, 0);

    KRadioAction * fast_animation_action = new KRadioAction(i18n("&Fastest Animations"), Key_2,
                                                            animation_mapper, SLOT(map()), action_col,
                                                            "FastAnimations");
    fast_animation_action->setExclusiveGroup("animation");
    fast_animation_action->setChecked(m_animation_speed == 1);
    animation_mapper->setMapping(fast_animation_action, 1);

    KRadioAction * normal_animation_action = new KRadioAction(i18n("No&rmal Animations"), Key_3,
                                                              animation_mapper, SLOT(map()), action_col,
                                                              "NormalAnimations");
    normal_animation_action->setExclusiveGroup("animation");
    normal_animation_action->setChecked(m_animation_speed == 2);
    animation_mapper->setMapping(normal_animation_action, 2);


    KRadioAction * slow_animation_action = new KRadioAction(i18n("&Slow Animations"), Key_4,
                                                            animation_mapper, SLOT(map()), action_col,
                                                            "SlowAnimations");
    slow_animation_action->setExclusiveGroup("animation");
    slow_animation_action->setChecked(m_animation_speed == 3);
    animation_mapper->setMapping(slow_animation_action, 3);


    // The optimization menu

    new KAction(i18n("&Solve ..."), "idea", SHIFT | Key_S,
                this, SLOT(solve()), action_col, "Solve");
    new KAction(i18n("&Reached Position Before?"), CTRL | Key_B,
                this, SLOT(isOldPosition()), action_col, "ReachedBefore");

    m_dont_optimize_solutions_action = new KToggleAction(i18n("&Don't Automatically Optimize Solutions"), 0,
                                                         this, SLOT(autoOptChanged()), action_col,
                                                         "DontOptimizeSolutions");
    m_dont_optimize_solutions_action->setExclusiveGroup("optimize");
    m_dont_optimize_solutions_action->setChecked(!m_auto_optimize_moves && !m_auto_optimize_pushes);

    m_auto_optimize_moves_action = new KToggleAction(i18n("&Automatically Minimze Moves In Solutions"), 0,
                                                     this, SLOT(autoOptChanged()), action_col,
                                                     "AutoMinimizeMoves");
    m_auto_optimize_moves_action->setExclusiveGroup("optimize");
    m_auto_optimize_moves_action->setChecked(m_auto_optimize_moves);

    m_auto_optimize_pushes_action = new KToggleAction(i18n("Automatically Minimze &Pushes In Solutions"), 0,
                                                      this, SLOT(autoOptChanged()), action_col,
                                                      "AutoMinimizePushes");
    m_auto_optimize_pushes_action->setExclusiveGroup("optimize");
    m_auto_optimize_pushes_action->setChecked(m_auto_optimize_pushes);

    new KAction(i18n("Minimize &Moves In All Solutions"), 0,
                this, SLOT(minimizeMoves()), action_col, "MinimizeMoves");
    new KAction(i18n("M&inimize Moves In Solutions of Current Collection"), 0,
                this, SLOT(minimizeCollectionMoves()), action_col, "MinimizeCollectionMoves");
    m_min_moves_action = new KAction(i18n("Mi&nimize Moves In Solutions of Current Level ..."), 0,
                                     this, SLOT(minimizeLevelMoves()), action_col, "MinimizeLevelMoves");
    new KAction(i18n("Minimize P&ushes In All Solutions"), 0,
                this, SLOT(minimizePushes()), action_col, "MinimizePushes");
    new KAction(i18n("Minimize Pu&shes In Solutions of Current Collection"), 0,
                this, SLOT(minimizeCollectionPushes()), action_col, "MinimizeCollectionPushes");
    m_min_pushes_action = new KAction(i18n("Minimize Pushes In &Solutions of Current Level ..."), 0,
                                      this, SLOT(minimizeLevelPushes()), action_col, "MinimizeLevelPushes");


    // The view menu

    new KAction(i18n("&Previous Theme"), SHIFT | Key_T,
                this, SLOT(previousTheme()), action_col, "PrevTheme");
    new KAction(i18n("&Next Theme"), Key_T,
                this, SLOT(nextTheme()), action_col, "NextTheme");
    m_hide_gems_action = new KToggleAction(i18n("Hide All G&ems"), 0,
                                           this, SLOT(hideGemsChanged()), action_col, "HideGems");
    m_hide_gems_action->setChecked(m_hide_gems);

    m_hide_goals_action = new KToggleAction(i18n("Hide All G&oals"), 0,
                                           this, SLOT(hideGoalsChanged()), action_col, "HideGoals");
    m_hide_goals_action->setChecked(m_hide_goals);

    m_outside_as_wall_action = new KToggleAction(i18n("&Treat Outside Fields As Walls"), 0, this,
                                                 SLOT(outsideAsWallChanged()), action_col, "OutsideAsWall");
    m_outside_as_wall_action->setChecked(m_outside_as_wall);

    m_honor_deadlock_action = new KToggleAction(i18n("&Honor Dead Fields"), "deadlock", Key_D, this,
                                                SLOT(honorDeadlockChanged()), action_col, "HonorDeadlock");
    m_honor_deadlock_action->setChecked(m_honor_deadlocks);

    m_show_arrow_action = new KToggleAction(i18n("&Show Possible Pushes"), "arrows", Key_A, this,
                                            SLOT(showPushesChanged()), action_col, "ShowPushes");
    m_show_arrow_action->setChecked(m_show_arrows);



    // Help menu

    new KAction(i18n("Ti&p Of The Day ..."), 0,
                this, SLOT(tipOfTheDay()), action_col, "TipOfTheDay");
}


void MainWindow::setupCollectionMenu()
{
    int const entries = static_cast<int> (m_collection_actions.size());

    for (int i = 0; i < entries; ++i)
    {
        m_collection_menu->remove(m_collection_actions[i]);
        delete m_collection_actions[i];
    }

    m_collection_actions.clear();

    int const author_entries = static_cast<int> (m_collection_by_author_submenus.size());

    for (int i = 0; i < author_entries; ++i)
    {
	delete m_collection_by_author_submenus[i];
    }

    m_collection_by_author_submenus.clear();
    m_collection_by_author_actions.clear();

    int const difficulty_entries = static_cast<int> (m_collection_by_difficulty_submenus.size());

    for (int i = 0; i < difficulty_entries; ++i)
    {
	delete m_collection_by_difficulty_submenus[i];
    }

    m_collection_by_difficulty_submenus.clear();
    m_collection_by_difficulty_actions.clear();

    if (m_collection_menu == 0)
    {
        m_collection_menu = new  KActionMenu(i18n("&Collections"), actionCollection(), "Collections");
        m_collection_menu->popupMenu()->insertTearOffHandle();

        m_collection_by_author_menu = new KActionMenu(i18n("Collections By &Author"), actionCollection(),
                                                           "CollectionsByAuthor");
        m_collection_by_author_menu->popupMenu()->insertTearOffHandle();

        m_collection_by_difficulty_menu = new KActionMenu(i18n("Collections By &Difficulty"),
                                                          actionCollection(), "CollectionsByDifficulty");
        m_collection_by_difficulty_menu->popupMenu()->insertTearOffHandle();

        m_collection_mapper = new QSignalMapper(m_collection_menu);
        connect(m_collection_mapper, SIGNAL(mapped(int)), this, SLOT(setCollection(int)));
    }

    int const number_of_collections = CollectionHolder::numberOfCollections();

    QStringList authors;
    std::vector<int> has_difficulty(12, 0);

    for (int i = 0; i < number_of_collections; ++i)
    {
	if (!CollectionHolder::collection(i)->authors().isEmpty())
	{
	    QString const first_author = CollectionHolder::collection(i)->authors()[0];
	    QString const first_surename = QStringList::split(" ", first_author).last();
	    QString author = CollectionHolder::collection(i)->authors().join(" " + i18n("and") + " ");
	    author = first_surename + " " + author;

	    if (authors.contains(author) == 0)
	    {
		authors.append(author);
	    }
	}
	else
	{
	    if (authors.contains(i18n("Unknown author")) == 0)
	    {
		authors.append(i18n("*** Unknown author"));
	    }
	}

        has_difficulty[CollectionHolder::collection(i)->difficulty() + 1] = 1;
    }

    authors.sort();
    int const nr_of_authors = authors.count();

    for (int i = 0; i < nr_of_authors; ++i)
    {
        QStringList author_strings = QStringList::split(" ", authors[i]);
        author_strings.remove(author_strings.begin());
        authors[i] = author_strings.join(" ");
	KActionMenu * menu = new KActionMenu(authors[i], m_collection_by_author_menu);
	m_collection_by_author_submenus.push_back(menu);
        m_collection_by_author_menu->insert(menu);
        menu->popupMenu()->insertTearOffHandle();
    }

    for (int i = 0; i <= 11; ++i)
    {
        if (has_difficulty[i])
        {
	    KActionMenu * menu = new KActionMenu(Difficulty::text(i - 1), m_collection_by_difficulty_menu);
            m_collection_by_difficulty_submenus.push_back(menu);
            m_collection_by_difficulty_menu->insert(menu);
            menu->popupMenu()->insertTearOffHandle();
        }
        else
        {
            m_collection_by_difficulty_submenus.push_back(0);
        }
    }

    for (int i = 0; i < number_of_collections; ++i)
    {
        QString level_name = CollectionHolder::collection(i)->name();
        QString author;

	if (CollectionHolder::collection(i)->authors().isEmpty())
	{
	    author = i18n("Unknown author");
	}
	else
	{
	    author = CollectionHolder::collection(i)->authors().join(" " + i18n("and") + " ");
	}

        int const number_of_levels = CollectionHolder::collection(i)->numberOfLevels();

        if (number_of_levels > 1)
        {
            if (CollectionHolder::isTemporary(i))
            {
                level_name += " (" + i18n("%1 levels, temporary").arg(number_of_levels) + ")";
            }
            else
            {
                level_name += " (" + i18n("%1 levels").arg(number_of_levels) + ")";
            }
        }
        else
        {
            if (CollectionHolder::isTemporary(i))
            {
                level_name += " (" + i18n("1 level, temporary") + ")";
            }
            else
            {
                level_name += " (" + i18n("1 level") + ")";
            }
        }

        KRadioAction * action = new KRadioAction(level_name, 0, m_collection_mapper, SLOT(map()), this);
        m_collection_mapper->setMapping(action, i);
        m_collection_menu->insert(action);
        m_collection_actions.push_back(action);

        int const author_index = authors.findIndex(author);
        assert(author_index != -1);

        action = new KRadioAction(level_name, 0, m_collection_mapper, SLOT(map()),
				  m_collection_by_author_submenus[author_index]);
        m_collection_mapper->setMapping(action, i);
        m_collection_by_author_submenus[author_index]->insert(action);
        m_collection_by_author_actions.push_back(action);

	int const diff_index = CollectionHolder::collection(i)->difficulty() + 1;
        action = new KRadioAction(level_name, 0, m_collection_mapper, SLOT(map()),
				  m_collection_by_difficulty_submenus[diff_index]);
        m_collection_mapper->setMapping(action, i);
        m_collection_by_difficulty_submenus[diff_index]->insert(action);
        m_collection_by_difficulty_actions.push_back(action);
    }
}


void MainWindow::setupRecentCollectionsMenu()
{
    int const old_recent_entries_actions = static_cast<int> (m_recent_collections_old_actions.size());

    for (int i = 0; i < old_recent_entries_actions; ++i)
    {
        delete m_recent_collections_old_actions[i];
    }

    m_recent_collections_old_actions.resize(0);

    int const recent_entries_actions = static_cast<int> (m_recent_collections_actions.size());

    for (int i = 0; i < recent_entries_actions; ++i)
    {
        m_recent_collections_menu->remove(m_recent_collections_actions[i]);
        m_recent_collections_old_actions.push_back(m_recent_collections_actions[i]);
    }

    m_recent_collections_actions.resize(0);

    if (m_recent_collections_menu == 0)
    {
        m_recent_collections_menu = new KActionMenu(i18n("&Recent Collections"), actionCollection(),
                                                    "RecentCollections");
        m_recent_collections_menu->popupMenu()->insertTearOffHandle();

        m_recent_collections_mapper = new QSignalMapper(m_recent_collections_menu);
        connect(m_recent_collections_mapper, SIGNAL(mapped(int)), this, SLOT(setCollection(int)));
    }

    int const recent_entries = std::min(m_nr_of_recent_collections,
                                        static_cast<int> (m_recent_collections.size()));

    int const number_of_collections = CollectionHolder::numberOfCollections();

    for (int i = 0; i < recent_entries; ++i)
    {
        int const collection_nr = m_recent_collections[i];

        if ((collection_nr >= 0) && (collection_nr < number_of_collections))
        {
            KAction * action = new KAction(CollectionHolder::collection(collection_nr)->name(), 0,
                                           m_recent_collections_mapper, SLOT(map()), this);
            m_recent_collections_mapper->setMapping(action, collection_nr);
            m_recent_collections_menu->insert(action);
            m_recent_collections_actions.push_back(action);
        }
    }
}


void MainWindow::setupThemeMenu()
{
    int const entries = static_cast<int> (m_theme_actions.size());

    for (int i = 0; i < entries; ++i)
    {
        m_theme_menu->remove(m_theme_actions[i]);
        delete m_theme_actions[i];
    }

    m_theme_actions.clear();

    if (m_theme_menu == 0)
    {
        m_theme_menu = new KActionMenu(i18n("&Theme"), actionCollection(), "Theme");
        m_theme_menu->popupMenu()->insertTearOffHandle();
        m_theme_mapper = new QSignalMapper(m_theme_menu);
        connect(m_theme_mapper, SIGNAL(mapped(int)), this, SLOT(setTheme(int)));
    }

    int const number_of_themes = ThemeHolder::numberOfThemes();

    for (int i = 0; i < number_of_themes; ++i)
    {
        QString theme_name = ThemeHolder::theme(i)->name();
        KRadioAction * action = new KRadioAction(theme_name, 0, m_theme_mapper, SLOT(map()), this);
        action->setExclusiveGroup("themes");
        action->setChecked(i == m_theme);
        m_theme_mapper->setMapping(action, i);
        m_theme_menu->insert(action);
        m_theme_actions.push_back(action);
    }
}


void MainWindow::setupBookmarkMenu()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_bookmark_format = static_cast<BookmarkFormat> (config->readNumEntry("Bookmark format", 1));
    m_number_of_bookmarks = config->readNumEntry("Number of bookmarks", 10);

    int const submenus = static_cast<int> (m_bookmark_set_submenus.size());
    bool const clear_entries = m_bookmark_set_submenus.empty();

    for (int i = 0; i < submenus; ++i)
    {
        m_bookmark_set_menu->remove(m_bookmark_set_submenus[i]);
        delete m_bookmark_set_submenus[i];

        m_bookmark_goto_menu->remove(m_bookmark_goto_submenus[i]);
        delete m_bookmark_goto_submenus[i];

        m_bookmark_annotate_menu->remove(m_bookmark_annotate_submenus[i]);
        delete m_bookmark_annotate_submenus[i];

        m_bookmark_import_menu->remove(m_bookmark_import_submenus[i]);
        delete m_bookmark_import_submenus[i];
    }

    m_bookmark_set_submenus.clear();
    m_bookmark_goto_submenus.clear();
    m_bookmark_annotate_submenus.clear();
    m_bookmark_import_submenus.clear();

    if (clear_entries)
    {
        int const entries = static_cast<int> (m_bookmark_set_actions.size());

        for (int i = 0; i < entries; ++i)
        {
            m_bookmark_set_menu->remove(m_bookmark_set_actions[i]);
            delete m_bookmark_set_actions[i];

            m_bookmark_goto_menu->remove(m_bookmark_goto_actions[i]);
            delete m_bookmark_goto_actions[i];

            m_bookmark_annotate_menu->remove(m_bookmark_annotate_actions[i]);
            delete m_bookmark_annotate_actions[i];

            m_bookmark_import_menu->remove(m_bookmark_import_actions[i]);
            delete m_bookmark_import_actions[i];
        }
    }

    m_bookmark_set_actions.clear();
    m_bookmark_goto_actions.clear();
    m_bookmark_annotate_actions.clear();
    m_bookmark_import_actions.clear();

    if (m_bookmark_set_menu == 0)
    {
        m_bookmark_set_menu = new KActionMenu(i18n("&Set Bookmark"), "bookmark_add",
                                              actionCollection(), "SetBookmark");
        m_bookmark_set_menu->popupMenu()->insertTearOffHandle();

        m_bookmark_goto_menu = new KActionMenu(i18n("&Goto Boomark"), "bookmark",
                                               actionCollection(), "GotoBookmark");
        m_bookmark_goto_menu->popupMenu()->insertTearOffHandle();

        m_bookmark_annotate_menu = new KActionMenu(i18n("&Annotate Bookmark"), "bookmark",
                                                   actionCollection(), "AnnotateBookmark");
        m_bookmark_annotate_menu->popupMenu()->insertTearOffHandle();

        m_bookmark_import_menu = new KActionMenu(i18n("&Import KSokoban Bookmark ..."), "bookmark_add",
                                                 actionCollection(), "ImportKSokobanBookmark");
        m_bookmark_import_menu->popupMenu()->insertTearOffHandle();

        m_bookmark_set_mapper = new QSignalMapper(m_bookmark_set_menu);
        connect(m_bookmark_set_mapper, SIGNAL(mapped(int)), this, SLOT(setBookmark(int)));

        m_bookmark_goto_mapper = new QSignalMapper(m_bookmark_goto_menu);
        connect(m_bookmark_goto_mapper, SIGNAL(mapped(int)), this, SLOT(gotoBookmark(int)));

        m_bookmark_annotate_mapper = new QSignalMapper(m_bookmark_annotate_menu);
        connect(m_bookmark_annotate_mapper, SIGNAL(mapped(int)), this, SLOT(annotateBookmark(int)));

        m_bookmark_import_mapper = new QSignalMapper(m_bookmark_import_menu);
        connect(m_bookmark_import_mapper, SIGNAL(mapped(int)), this, SLOT(importKSokobanBookmark(int)));
    }

    int const submenus_border = 20;
    bool use_submenus = (m_number_of_bookmarks > submenus_border);

    if (use_submenus)
    {
        int const submenus = (m_number_of_bookmarks - 1) / 10 + 1;

        for (int i = 0; i < submenus; ++i)
        {
            int const from = i * 10 + 1;
            int const to = std::min((i * 10 + 10), m_number_of_bookmarks);

            QString name = QString("%1 .. %2").arg(from).arg(to);

            m_bookmark_set_submenus.push_back(new KActionMenu(name, m_bookmark_set_menu));
            m_bookmark_set_menu->insert(m_bookmark_set_submenus.back());

            m_bookmark_goto_submenus.push_back(new KActionMenu(name, m_bookmark_goto_menu));
            m_bookmark_goto_menu->insert(m_bookmark_goto_submenus.back());

            m_bookmark_annotate_submenus.push_back(new KActionMenu(name, m_bookmark_annotate_menu));
            m_bookmark_annotate_menu->insert(m_bookmark_annotate_submenus.back());

            m_bookmark_import_submenus.push_back(new KActionMenu(name, m_bookmark_import_menu));
            m_bookmark_import_menu->insert(m_bookmark_import_submenus.back());
        }
    }

    for (int i = 0; i < m_number_of_bookmarks; ++i)
    {
        if (use_submenus)
        {
            int const submenu = i / 10;
            KAction * set_action = new KAction("", 0, m_bookmark_set_mapper, SLOT(map()),
                                               m_bookmark_set_submenus[submenu]);
            m_bookmark_set_actions.push_back(set_action);
            m_bookmark_set_submenus[submenu]->insert(set_action);
            m_bookmark_set_mapper->setMapping(set_action, i);

            KAction * goto_action = new KAction("", 0, m_bookmark_goto_mapper, SLOT(map()),
                                                 m_bookmark_goto_submenus[submenu]);
            m_bookmark_goto_actions.push_back(goto_action);
            m_bookmark_goto_submenus[submenu]->insert(goto_action);
            m_bookmark_goto_mapper->setMapping(goto_action, i);

            KAction * annotate_action = new KAction("", 0, m_bookmark_annotate_mapper, SLOT(map()),
                                                    m_bookmark_annotate_submenus[submenu]);
            m_bookmark_annotate_actions.push_back(annotate_action);
            m_bookmark_annotate_submenus[submenu]->insert(annotate_action);
            m_bookmark_annotate_mapper->setMapping(annotate_action, i);

            KAction * import_action = new KAction("", 0, m_bookmark_import_mapper, SLOT(map()),
                                                  m_bookmark_import_submenus[submenu]);
            m_bookmark_import_actions.push_back(import_action);
            m_bookmark_import_submenus[submenu]->insert(import_action);
            m_bookmark_import_mapper->setMapping(import_action, i);
        }
        else
        {
            KAction * set_action = new KAction("","bookmark_add", 0, m_bookmark_set_mapper,
                                               SLOT(map()), m_bookmark_set_menu);
            m_bookmark_set_actions.push_back(set_action);
            m_bookmark_set_menu->insert(set_action);
            m_bookmark_set_mapper->setMapping(set_action, i);

            KAction * goto_action = new KAction("", "bookmark", 0, m_bookmark_goto_mapper,
                                                SLOT(map()), m_bookmark_goto_menu);
            m_bookmark_goto_actions.push_back(goto_action);
            m_bookmark_goto_menu->insert(goto_action);
            m_bookmark_goto_mapper->setMapping(goto_action, i);

            KAction * annotate_action = new KAction("", "bookmark", 0, m_bookmark_annotate_mapper,
                                                    SLOT(map()), m_bookmark_annotate_menu);
            m_bookmark_annotate_actions.push_back(annotate_action);
            m_bookmark_annotate_menu->insert(annotate_action);
            m_bookmark_annotate_mapper->setMapping(annotate_action, i);

            KAction * import_action = new KAction("", "bookmark_add", 0, m_bookmark_import_mapper,
                                                  SLOT(map()), m_bookmark_import_menu);
            m_bookmark_import_actions.push_back(import_action);
            m_bookmark_import_menu->insert(import_action);
            m_bookmark_import_mapper->setMapping(import_action, i);
        }

        setupBookmarkMenuEntry(i);
    }
}


void MainWindow::setupBookmarkMenuEntry(int index)
{
    QString name = QString::number(index + 1) + ' ';

    bool has_date = m_bookmark_format & DATE;
    bool has_time = m_bookmark_format & DATE_AND_TIME;
    bool has_collection = m_bookmark_format & COLLECTION;
    bool has_level = m_bookmark_format & COLLECTION_AND_LEVEL;
    bool has_annotation = m_bookmark_format & ANNOTATION;
    bool has_date_or_time = has_date || has_time;
    bool has_collection_or_level = has_collection || has_level;

    if (Bookmarks::hasBookmark(index))
    {
        if (has_annotation)
        {
            name += Bookmarks::annotation(index);
        }

        if (has_annotation && (has_collection_or_level || has_date_or_time))
        {
            name += " (";
        }

        if (has_collection)
        {
            name += Bookmarks::collectionName(index);
        }
        else if (has_level)
        {
            QString const collection_name =Bookmarks::collectionName(index);
            name += i18n("Level %1 of %2").arg(Bookmarks::level(index) + 1).arg(collection_name);
        }

        if (has_date_or_time && has_collection_or_level && !has_annotation)
        {
            name += " (";
        }

        if (has_date_or_time && has_collection_or_level && has_annotation)
        {
            name += ", ";
        }

        if (has_date)
        {
            name += KGlobal::locale()->formatDate(Bookmarks::date(index).date(), true);
        }
        else if (has_time)
        {
            name += KGlobal::locale()->formatDateTime(Bookmarks::date(index).date(), true);
        }

        if ( (has_annotation && (has_collection_or_level || has_date_or_time)) ||
             (has_collection_or_level && has_date_or_time) )
        {
            name += ")";
        }
    }
    else
    {
        name = "(" + i18n("unused") + ")";
    }

    m_bookmark_set_actions[index]->setText(name);
    m_bookmark_goto_actions[index]->setText(name);
    m_bookmark_annotate_actions[index]->setText(name);
    m_bookmark_import_actions[index]->setText(name);
}
