/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "map.h"

#include <algorithm>
#include <cassert>
#include <set>

#include <qstringlist.h>

#include "compressed_map.h"
#include "hash.h"




char const Map::s_piece_to_text[8] = {'@', '+', '$', '*', ' ', '.', '#', ' '};
bool const Map::s_piece_contains_gem[8] = {false, false, true, true, false, false, false, false};
bool const Map::s_piece_contains_goal[8] = {false, true, false, true, false, true, false, false};
bool const Map::s_can_drop_keeper[24] = {true, true, false, false, true, true, false, false,
                                         false, false, false, false, false, false, false, false,
                                         false, false, false, false, false, false, false, false};
bool const Map::s_can_drop_gem[8] = {true, true, false, false, true, true, false, false};
QRegExp * Map::s_map_regexp = 0;


Map::Map(Map const & map)
{
    init(map);
}


Map::Map(int width, int height, std::vector<int> const & pieces) :
    m_width(width),
    m_height(height),
    m_size(width * height),
    m_keeper(0, 0),
    m_deadlocks_valid(false),
    m_reachable_valid(false),
    m_empty_goals_valid(true),
    m_validity_valid(false)
{
    assert(m_width > 0);
    assert(m_height > 0);
    assert(m_width < 128);
    assert(m_height < 128);

    m_pieces = new int[m_size];

    for (int i = 0; i < m_size; ++i)
    {
        m_pieces[i] = pieces[i];
    }

    createOutsidePieces();
    setupOffsets();
    setupKeeperAndEmptyGoals();
}


Map::Map(QDataStream & stream) :
    m_deadlocks_valid(false),
    m_reachable_valid(false),
    m_empty_goals_valid(true),
    m_validity_valid(false),
    m_pieces(0)
{
    CompressedMap compressed_map(stream);

    m_width = compressed_map.width();
    m_height = compressed_map.height();
    m_size = m_width * m_height;
    m_keeper = getPoint(compressed_map.keeperIndex());
    m_empty_goals = compressed_map.numberOfEmptyGoals();

    setPieces(compressed_map);
    setupOffsets();
}


Map::Map(QStringList & lines) :
    m_width(0),
    m_height(0),
    m_size(0),
    m_deadlocks_valid(false),
    m_reachable_valid(false),
    m_empty_goals_valid(true),
    m_validity_valid(false),
    m_pieces(0)
{
    while (!lines.isEmpty() && !isMapLine(lines.first()))
    {
        lines.remove(lines.begin());
    }

    if (!lines.isEmpty())
    {
        int max_width = 0;
        QStringList map_lines;

        while (!lines.isEmpty() && isMapLine(lines.first()))
        {
            QString line = lines.first();
            lines.remove(lines.begin());

            while (line[line.length() - 1] == ' ')
            {
                line = line.left(line.length() - 1);
            }

            max_width = std::max(max_width, static_cast<int> (line.length()));
            map_lines.append(line);
        }

        m_width = max_width;
        m_height = map_lines.count();
        m_size = m_width * m_height;
        m_pieces = new int[m_size];

        for (int i = 0; i < m_size; ++i)
        {
            m_pieces[i] = EMPTY;
        }

        for (int y = 0; y < m_height; ++y)
        {
            char const * act_line = map_lines[y].latin1();
            int const act_width = map_lines[y].length();

            for (int x = 0; x < act_width; ++x)
            {
                char const act_char = act_line[x];

                for (int i = 0; i < 8; ++i)
                {
                    if (s_piece_to_text[i] == act_char)
                    {
                        m_pieces[x + y * m_width] = i;

                        break;
                    }
                }
            }
        }
    }

    createOutsidePieces();
    setupOffsets();
    setupKeeperAndEmptyGoals();
}


Map::Map(CompressedMap const & compressed_map) :
    m_deadlocks_valid(false),
    m_reachable_valid(false),
    m_empty_goals_valid(true),
    m_validity_valid(false),
    m_pieces(0)
{
    m_width = compressed_map.width();
    m_height = compressed_map.height();
    m_size = m_width * m_height;
    m_keeper = getPoint(compressed_map.keeperIndex());
    m_empty_goals = compressed_map.numberOfEmptyGoals();

    setPieces(compressed_map);
    setupOffsets();
}


Map::~Map()
{
    delete[] m_pieces;
}


Map::Map & Map::operator = (Map const & map)
{
    delete[] m_pieces;
    init(map);

    return *this;
}


void Map::writeToStream(QDataStream & stream)
{
    CompressedMap compressed_map(*this);
    compressed_map.writeToStream(stream);
}


int Map::width() const
{
    return m_width;
}


int Map::height() const
{
    return m_height;
}


int Map::numberOfGems() const
{
    int nr_of_gems = 0;

    for (int i = 0; i < m_size; ++i)
    {
        if (containsGem(i))
        {
            ++nr_of_gems;
        }
    }

    return nr_of_gems;
}


int Map::numberOfEmptyGoals() const
{
    if (!m_empty_goals_valid)
    {
        setupNumberOfEmptyGoals();
    }

    return m_empty_goals;
}


std::vector<int> Map::pieces() const
{
    return std::vector<int> (m_pieces, m_pieces + m_size);
}


bool Map::isValid() const
{
    return validity() == IS_VALID;
}


Map::MapValidity Map::validity() const
{
    if (m_validity_valid)
    {
        return m_validity;
    }

    m_validity = IS_VALID;
    int keepers = 0;
    int goals = 0;
    int gems = 0;

    for (int i = 0; i < m_size; ++i)
    {
        int const piece = getPiece(i);

        if (pieceContainsKeeper(piece))
        {
            ++keepers;
        }

        if (pieceContainsGem(piece))
        {
            ++gems;
        }

        if (pieceContainsGoal(piece))
        {
            ++goals;
        }
    }

    if (keepers < 1)
    {
        m_validity =  NO_KEEPER;
    }
    else if (keepers > 1)
    {
        m_validity = TOO_MANY_KEEPERS;
    }

    if (gems < 1)
    {
        m_validity = NO_GEMS;
    }

    if (goals < gems)
    {
        m_validity = MORE_GEMS_THAN_GOALS;
    }
    else if (goals > gems)
    {
        m_validity = MORE_GOALS_THAN_GEMS;
    }

    if (m_validity != IS_VALID)
    {
        return m_validity;
    }

    for (int i = 0; i < m_size; ++i)
    {
        if (getPiece(i) == OUTSIDE)
        {
            for (int j = 0; j < 4; ++j)
            {
                int const nb_index = i + m_xy_offsets[j];

                if (isValidIndex(nb_index)) // It doesn't matter if we cross the sides of the map!
                {
                    Piece const piece = getPiece(nb_index);

                    if ( (piece != OUTSIDE) && (piece != WALL) )
                    {
                        m_validity = MAP_LEAKS;

                        return m_validity;
                    }
                }
            }
        }
    }

    for (int x = 0; x < m_width; ++x)
    {
        int const piece1 = getPiece(QPoint(x, 0));
        int const piece2 = getPiece(QPoint(x, m_height - 1));

        if ( (piece1 != Map::OUTSIDE) && (piece1 != Map::WALL) )
        {
            m_validity = MAP_LEAKS;

            return m_validity;
        }

        if ( (piece2 != Map::OUTSIDE) && (piece2 != Map::WALL) )
        {
            m_validity = MAP_LEAKS;

            return m_validity;
        }
    }

    for (int y = 0; y < m_height; ++y)
    {
        int const piece1 = getPiece(QPoint(0, y));
        int const piece2 = getPiece(QPoint(m_width - 1, y));

        if ( (piece1 != Map::OUTSIDE) && (piece1 != Map::WALL) )
        {
            m_validity = MAP_LEAKS;

            return m_validity;
        }

        if ( (piece2 != Map::OUTSIDE) && (piece2 != Map::WALL) )
        {
            m_validity = MAP_LEAKS;

            return m_validity;
        }
    }

    if (isSolved())
    {
        m_validity = MAP_SOLVED;

        return m_validity;
    }

    m_validity = IS_VALID;

    return m_validity;
}


bool Map::isSolved() const
{
    return numberOfEmptyGoals() == 0;
}


int Map::getIndex(QPoint position) const
{
    assert(isValidPosition(position));

    return position.x() + m_width * position.y();
}


QPoint Map::getPoint(int index) const
{
    assert(isValidIndex(index));

    int const y = index / m_width;
    int const x = index - y * m_width;

    return QPoint(x, y);
}


bool Map::isValidPosition(QPoint position) const
{
    bool const x_valid = ((position.x() >= 0) && (position.x() < m_width));
    bool const y_valid = ((position.y() >= 0) && (position.y() < m_height));

    return x_valid && y_valid;
}


bool Map::isValidIndex(int index) const
{
    return (index >= 0) && (index < m_size);
}


Map::Piece Map::getPiece(QPoint position) const
{
    assert(isValidPosition(position));

    return getPiece(getIndex(position));
}


Map::Piece Map::getPiece(int index) const
{
    assert(isValidIndex(index));

    return static_cast<Piece> (m_pieces[index] & PIECE);
}


void Map::setPiece(QPoint position, int piece)
{
    assert(isValidPosition(position));

    setPiece(getIndex(position), piece);
}


void Map::setPiece(int index, int piece)
{
    assert(isValidIndex(index));
    assert(piece >= KEEPER);
    assert(piece <= OUTSIDE);

    m_pieces[index] = (m_pieces[index] & CLEAR_PIECE) + piece;
    m_reachable_valid = false;
    m_deadlocks_valid = false;
    m_empty_goals_valid = false;
    m_validity_valid = false;
}


void Map::mirrorVertically()
{
    int const width_half = (m_width + 1) / 2;

    for (int x = 0; x < width_half; ++x)
    {
        for (int y = 0; y < m_height; ++y)
        {
            std::swap(m_pieces[x + m_width * y], m_pieces[m_width - x - 1 + m_width * y]);
        }
    }

    m_keeper = QPoint(m_width - m_keeper.x() - 1, m_keeper.y());
}


void Map::mirrorHorizontally()
{
    int const height_half = (m_height + 1) / 2;

    for (int y = 0; y < height_half; ++y)
    {
        for (int x = 0; x < m_width; ++x)
        {
            std::swap(m_pieces[x + m_width * y], m_pieces[x + m_width * (m_height - y - 1)]);
        }
    }

    m_keeper = QPoint(m_keeper.x(), m_height - m_keeper.y() - 1);
}


void Map::rotateLeft()
{
    rotateRight();
    rotateRight();
    rotateRight();
}


void Map::rotateRight()
{
    int * new_pieces = new int[m_size];

    int const width_half = (m_width + 1) / 2;
    int const height_half = (m_height + 1) / 2;

    for (int x = 0; x < width_half; ++x)
    {
        int const x2 = m_width - x - 1;

        for (int y = 0; y < height_half; ++y)
        {
            int const y2 = m_height - y - 1;

            new_pieces[y2 + x * m_height] = m_pieces[x + m_width * y];
            new_pieces[y + x * m_height] = m_pieces[x + m_width * y2];
            new_pieces[y + x2 * m_height] = m_pieces[x2 + m_width * y2];
            new_pieces[y2 + x2 * m_height] = m_pieces[x2 + m_width * y];
        }
    }

    m_keeper = QPoint(m_height - m_keeper.y() - 1, m_keeper.x());
    std::swap(new_pieces, m_pieces);
    delete[] new_pieces;
    std::swap(m_width, m_height);
    setupOffsets();
}


bool Map::isCrossed(QPoint position) const
{
    assert(isValidPosition(position));

    return isCrossed(getIndex(position));
}


bool Map::isCrossed(int index) const
{
    assert(isValidIndex(index));

    return m_pieces[index] & CROSSED;
}


void Map::uncrossAll()
{
    for (int i = 0; i < m_size; ++i)
    {
        m_pieces[i] &= CLEAR_CROSSED;
    }
}


bool Map::isReachable(QPoint position) const
{
    assert(isValidPosition(position));

    return isReachable(getIndex(position));
}


bool Map::isReachable(int index) const
{
    assert(isValidIndex(index));

    return m_pieces[index] & REACHABLE;
}


void Map::calcReachable() const
{
    if (!m_reachable_valid)
    {
        calcReachable(m_keeper);
        m_reachable_valid = true;
    }
}


void Map::calcReachable(QPoint position) const
{
    assert(isValidPosition(position));

    calcReachable(getIndex(position));
}


void Map::calcReachable(int index) const
{
    assert(isValidIndex(index));

    clearReachable();

    // This is a DFS

    m_pieces[index] |= REACHABLE;

    static int * indices = 0;
    static int indices_size;

    if ((indices == 0) || (indices_size < m_size * 4))
    {
        delete[] indices;
        indices_size = m_size * 4;
        indices = new int[indices_size];
    }

    indices[0] = index << 3;

    int pos = 0;

    while (pos >= 0)
    {
    l1:
        int const index = indices[pos];
        int const act_direction = index & 7;

        if (act_direction < 4)
        {
//            int const act_index = index >> 3;
            int const new_index = (index >> 3) + m_xy_offsets[act_direction];
            ++indices[pos];

            if (s_can_drop_keeper[m_pieces[new_index] & 23])
            {
                ++pos;
                indices[pos] = new_index << 3;
                m_pieces[new_index] |= REACHABLE;
            }

            goto l1; // This gives about 10 percent faster code on my machine. And because
                     // this routine will be called quite heavily by the solver, I
                     // will go with goto.
        }
        else
        {
            --pos;
        }
    }

    m_reachable_valid = false;
}


void Map::clearReachable() const
{
    int const size = m_size;

    for (int i = 0; i < size; ++i)
    {
        m_pieces[i] &= CLEAR_REACHABLE;
    }

    m_reachable_valid = false;
}


bool Map::isDeadlock(QPoint position) const
{
    assert(isValidPosition(position));

    return isDeadlock(getIndex(position));
}


bool Map::isDeadlock(int index) const
{
    assert(isValidIndex(index));

    return m_pieces[index] & DEADLOCK;
}


void Map::clearDeadlocks() const
{
    for (int i = 0; i < m_size; ++i)
    {
        m_pieces[i] &= CLEAR_DEADLOCK;
    }
}


void Map::calcDeadlocks() const
{
    if (m_deadlocks_valid)
    {
        return;
    }

    calcTrivialDeadlocks();

    // We first move right and then down.
    for (int direction = 1; direction < 4; direction += 2)
    {
        int const offset = m_xy_offsets[direction];
        int const other_offset1 = m_xy_offsets[2 - (direction & 2)];
        int const other_offset2 = m_xy_offsets[3 - (direction & 2)];

        for (int i = 0; i < m_size; ++i)
        {
            int const next = i + offset;

            if (isDeadlock(i) && !isDeadlock(next) && isPossibleDeadlock(next))
            {
                int act = next;

                do
                {
                    if ( (getPiece(act + other_offset1) != WALL) && (getPiece(act + other_offset2) != WALL) )
                    {
                        break;
                    }

                    if (isDeadlock(act))
                    {
                        for (int j = i + offset; j < act; j += offset)
                        {
                            m_pieces[j] |= DEADLOCK;
                        }

                        break;
                    }

                    act += offset;
                }
                while (isPossibleDeadlock(act));
            }
        }
    }

    m_deadlocks_valid = true;
}


void Map::crossDeadlocks() const
{
    for (int i = 0; i < m_size; ++i)
    {
        if (isDeadlock(i))
        {
            m_pieces[i] |= CROSSED;
        }
    }
}


QPoint Map::keeper() const
{
    return m_keeper;
}


void Map::setKeeper(QPoint position)
{
    assert(isValidPosition(position));
    assert(canDropKeeper(position));

    int const old_index = getIndex(m_keeper);

    if (containsKeeper(old_index))
    {
        m_pieces[old_index] += 4;
    }

    int const new_index = getIndex(position);
    m_keeper = position;

    if (!containsKeeper(new_index))
    {
        assert(canDropKeeper(new_index));

        m_pieces[new_index] -= 4;
    }

    m_reachable_valid = false;
}


void Map::setKeeper(int index)
{
    assert(isValidIndex(index));
    assert(canDropKeeper(index));

    int const old_index = getIndex(m_keeper);

    if (containsKeeper(old_index))
    {
        m_pieces[old_index] += 4;
    }

    m_keeper = getPoint(index);

    if (!containsKeeper(index))
    {
        assert(canDropKeeper(index));

        m_pieces[index] -= 4;
    }

    m_reachable_valid = false;
}


void Map::setKeeperToFirstReachable()
{
    calcReachable();

    for (int i = m_width + 1; i < m_size; ++i)
    {
        if (isReachable(i))
        {
            setKeeper(i);

            return;
        }
    }
}


void Map::moveGem(QPoint from, QPoint to)
{
    assert(isValidPosition(from));
    assert(isValidPosition(to));

    moveGem(getIndex(from), getIndex(to));
}


void Map::moveGem(int from, int to)
{
    assert(isValidIndex(from));
    assert(isValidIndex(to));
    assert(containsGem(from));

    if (from == to)
    {
        return;
    }

    m_pieces[from] += 2;

    assert(canDropGem(to));
    assert(!containsKeeper(to));

    m_pieces[to] -= 2;
    m_reachable_valid = false;

    if (containsGoal(from))
    {
        ++m_empty_goals;
    }

    if (containsGoal(to))
    {
        --m_empty_goals;
    }
}


bool Map::containsKeeper(QPoint position) const
{
    assert(isValidPosition(position));

    return containsKeeper(getIndex(position));
}


bool Map::containsKeeper(int index) const
{
    assert(isValidIndex(index));

    return pieceContainsKeeper(getPiece(index));
}


bool Map::containsGem(QPoint position) const
{
    assert(isValidPosition(position));

    return containsGem(getIndex(position));
}


bool Map::containsGem(int index) const
{
    assert(isValidIndex(index));

    return pieceContainsGem(getPiece(index));
}


bool Map::containsGoal(QPoint position) const
{
    assert(isValidPosition(position));

    return containsGoal(getIndex(position));
}


bool Map::containsGoal(int index) const
{
    assert(isValidIndex(index));

    return pieceContainsGoal(getPiece(index));
}


bool Map::canDropKeeper(QPoint position) const
{
    assert(isValidPosition(position));

    return canDropKeeper(getIndex(position));
}


bool Map::canDropKeeper(int index) const
{
    assert(isValidIndex(index));

    return canDropKeeperOnPiece(getPiece(index));
}


bool Map::canDropGem(QPoint position) const
{
    assert(isValidPosition(position));

    return canDropGem(getIndex(position));
}


bool Map::canDropGem(int index) const
{
    assert(isValidIndex(index));

    return canDropGemOnPiece(getPiece(index));
}


void Map::doMove(Move const & move, bool retro_mode)
{
    assert(!move.stonePushed() || isValidPushMove(move, retro_mode));
    assert(move.stonePushed() || isValidNonPushMove(move));

    if (move.stonePushed())
    {
        QPoint const diff_sign = move.diffSign();

        if (retro_mode)
        {
            setKeeper(move.to());
            moveGem(move.from() - diff_sign, move.to() - diff_sign);
        }
        else
        {
            setKeeper(move.from());
            moveGem(move.from() + diff_sign, move.to() + diff_sign);
            setKeeper(move.to());
        }
    }
    else
    {
        setKeeper(move.to());
    }
}


void Map::doUndoMove(Move const & move, bool retro_mode)
{
    if (move.stonePushed())
    {
        QPoint const diff_sign = move.diffSign();

        if (retro_mode)
        {
            setKeeper(move.to());
            moveGem(move.to() - diff_sign, move.from() - diff_sign);
            setKeeper(move.from());
        }
        else
        {
            setKeeper(move.from());
            moveGem(move.to() + diff_sign, move.from() + diff_sign);
        }
    }
    else
    {
        setKeeper(move.from());
    }
}


bool Map::isValidMove(Move const & move, bool retro_mode) const
{
    if (move.stonePushed())
    {
        return isValidPushMove(move, retro_mode);
    }
    else
    {
        return isValidNonPushMove(move);
    }
}


bool Map::isValidNonPushMove(Move const & move) const
{
    assert(!move.stonePushed());

    if (!canDropKeeper(move.to()))
    {
        return false;
    }

    calcReachable();

    return isReachable(move.to());
}


bool Map::isValidPushMove(Move const & move, bool retro_mode) const
{
    assert(move.stonePushed());

    if (move.isAtomicMove())
    {
        return isValidAtomicPushMove(move, retro_mode);
    }
    else
    {
        QPoint const diff_sign = move.diffSign();
        QPoint from = move.from();

        if (!isValidAtomicPushMove(Move(from, from + diff_sign, true), retro_mode))
        {
            return false;
        }

        int const moves = (from - move.to()).manhattanLength();
        from += diff_sign;

        for (int i = 0; i < moves; ++i)
        {
            from += diff_sign;

            if (retro_mode)
            {
                if (!canDropGem(from - 2 * diff_sign))
                {
                    return false;
                }
            }
            else
            {
                if (!canDropGem(from))
                {
                    return false;
                }
            }
        }
    }

    return true;
}


bool Map::isValidAtomicPushMove(Move const & move, bool retro_mode) const
{
    assert(move.stonePushed());
    assert(move.isAtomicMove());

    QPoint const diff = move.diff();

    if (retro_mode)
    {
        if (!containsGem(move.from() - diff))
        {
            return false;
        }

        if (!canDropGem(move.to() - diff))
        {
            return false;
        }
    }
    else
    {
        if (!containsGem(move.from() + diff))
        {
            return false;
        }

        if (!canDropGem(move.to() + diff))
        {
            return false;
        }
    }

    return true;
}


Movements Map::getShortestPath(QPoint from, QPoint to) const
{
    assert(canDropKeeper(from));
    assert(canDropKeeper(to));

    if (from == to)
    {
        return Movements();
    }

    int const unknown = -2;
    int const forbidden = -1;
    int const unprocessed = m_size;

    // This is a BFS

    std::vector<int> path_map(m_size, unknown);

    static std::vector<int> last_tests;
    last_tests.clear();

    static std::vector<int> new_tests;
    new_tests.clear();

    int const to_index = getIndex(to);
    int index = getIndex(from);

    path_map[index] = 0;
    last_tests.push_back(index);

    int act_dist = 1;

    while (true)
    {
        int const last_tests_size = last_tests.size();

        for (int i = 0; i < last_tests_size; ++i)
        {
            index = last_tests[i];

            for (int j = 0; j < 4; ++j)
            {
                int const nb_index = index + m_xy_offsets[j];

                if (path_map[nb_index] == unknown)
                {
                    if (canDropKeeper(nb_index))
                    {
                        new_tests.push_back(nb_index);
                        path_map[nb_index] = act_dist;
                    }
                    else
                    {
                        path_map[nb_index] = forbidden;
                    }
                }
                else if (path_map[nb_index] == unprocessed)
                {
                    new_tests.push_back(nb_index);
                    path_map[nb_index] = act_dist;
                }
            }
        }

        // Test if there is no path
        if (last_tests.empty())
        {
            break;
        }

        // Test if we have already found the path
        if (std::find(last_tests.begin(), last_tests.end(), to_index) != last_tests.end())
        {
            break;
        }

        new_tests.swap(last_tests);
        new_tests.clear();

        ++act_dist;
    }

    if (path_map[getIndex(to)] == unprocessed)
    {
        return Movements();
    }
    else
    {
        std::vector<Move> path;

        QPoint act_point = to;

        while (act_point != from)
        {
            int const index = getIndex(act_point);
            act_dist = path_map[index] - 1;

            QPoint next_point(m_size, m_size);

            for (int j = 0; j < 4; ++j)
            {
                int nb_index = index + m_xy_offsets[j];

                if (path_map[nb_index] == act_dist)
                {
                    next_point =getPoint(nb_index);
                    break;
                }
            }

            assert(next_point != QPoint(m_size, m_size));

            path.push_back(Move(next_point, act_point, false));

            act_point = next_point;
        }

        std::reverse(path.begin(), path.end());

        return path;
    }
}


Movements Map::getShortestPathForGem(QPoint from, QPoint to, bool retro_mode) const
{
    assert(containsGem(from));
    assert(canDropGem(to));

    Movements result;

    if (isDeadlock(from) || isDeadlock(to))
    {
        return result;
    }

    int from_index = getIndex(from);

    Map test_map = *this;
    test_map.m_pieces[from_index] +=2; // Remove gem
    std::vector<int> distance_map = test_map.getDistanceMap(getIndex(to), m_size * 4, retro_mode);

    int act_dist = m_size * 4;

    calcReachable();

    for (int i = 0; i < 4; ++i)
    {
        if (isReachable(from_index + m_xy_offsets[i]))
        {
            int const map_index = i + 4 * from_index;

            if (distance_map[map_index] < act_dist)
            {
                act_dist = distance_map[map_index];
            }
        }
    }

    if (act_dist == m_size * 4)
    {
        return result;
    }

    int const retro_fac = retro_mode ? -1 : 1;

    test_map = *this;

    while (act_dist > 0)
    {
        for (int i = 0; i < 4; ++i)
        {
            int const offset = m_xy_offsets[i];
            int const next_from_index = from_index - retro_fac * offset;
            int const keeper_pos = from_index + offset;
            int const map_index = i + 4 * next_from_index;

            if ( (test_map.isReachable(keeper_pos)) &&
                 (distance_map[map_index] == act_dist - 1) &&
                 (distance_map[i + 4 * from_index] == act_dist) )
            {
                --act_dist;

                QPoint const act_from = getPoint(from_index);
                QPoint const act_to = getPoint(next_from_index);

                if (retro_mode)
                {
                    test_map.setKeeper(next_from_index + offset);
                    test_map.moveGem(act_from, act_to);
                }
                else
                {
                    test_map.setKeeper(keeper_pos);
                    test_map.moveGem(act_from, act_to);
                    test_map.setKeeper(from_index);
                }

                test_map.calcReachable();
                result.addMove(Move(act_from, act_to, true));
                from_index = next_from_index;

                break;
            }

            assert(i < 3);
        }
    }

    return result.gemMovesToKeeperMoves(keeper(), retro_mode);
}


std::vector<int> Map::getDistanceMap(QPoint const & position, int unsolvable,
                                     bool retro_mode) const
{
    assert(isValidPosition(position));

    return getDistanceMap(getIndex(position), unsolvable, retro_mode);
}


std::vector<int> Map::getDistanceMap(int index, int unsolvable, bool retro_mode) const
{
    // This is a BFS
    std::vector<int> result(4 * m_size, unsolvable);

    std::vector<int> last_positions;
    std::vector<int> last_directions;

    int act_dist = 0;

    calcReachable();

    Map test_map = *this;

    int const retro_fac1 = retro_mode ? -1 : 1;
    int const retro_fac2 = retro_mode ? -2 : 1;
    int const retro_fac3 = retro_mode ? 1 : 0;

    for (int i = 0; i < 4; ++i)
    {
        int const new_pos = index + retro_fac1 * m_xy_offsets[i];

        if (canDropKeeper(new_pos))
        {
            result[i + 4 * index] = act_dist;
            last_positions.push_back(index);
            last_directions.push_back(i);
        }
    }

    while (!last_positions.empty())
    {
        assert(last_positions.size() == last_directions.size());

        ++act_dist;
        std::vector<int> new_positions;
        std::vector<int> new_directions;
        int const size = static_cast<int> (last_positions.size());

        for (int i = 0; i < size; ++i)
        {
            int const old_pos = last_positions[i];
            int const old_direction = last_directions[i];
            int const new_pos = old_pos + m_xy_offsets[old_direction];
            int const new_keeper_pos = new_pos + retro_fac2 * m_xy_offsets[old_direction];

            if (canDropGem(new_pos) && canDropKeeper(new_keeper_pos))
            {
                test_map.setPiece(new_pos, Map::GEM);

                for (int j = 0; j < 4; ++j)
                {
                    if (test_map.canDropKeeper(new_pos + m_xy_offsets[j]))
                    {
                        int new_index = new_pos * 4 + j;

                        if (result[new_index] == unsolvable)
                        {
                            test_map.calcReachable(new_pos + m_xy_offsets[j]);

                            if (test_map.isReachable(new_keeper_pos))
                            {
                                result[new_index] = act_dist;
                                new_positions.push_back(new_pos);
                                new_directions.push_back(j ^ retro_fac3);
                            }
                        }
                    }
                }

                test_map.setPiece(new_pos, getPiece(new_pos));
            }
        }

        last_positions.swap(new_positions);
        last_directions.swap(new_directions);
    }

    return result;
}


Movements Map::expandMove(Move const & move, bool retro_mode) const
{
    assert(isValidMove(move, retro_mode));

    if (move.isAtomicMove())
    {
        Movements result;

        result.addMove(move);

        return result;
    }
    else
    {
        if (move.stonePushed())
        {
            Movements result;
            QPoint from = move.from();

            QPoint diff_sign = move.diffSign();
            int const number_of_atomic_moves = (move.to() - from).manhattanLength();

            for (int i = 0; i < number_of_atomic_moves; ++i)
            {
                result.addMove(Move(from, from + diff_sign, true));
                from += diff_sign;
            }

            return result;

        }
        else
        {
            return getShortestPath(move.from(), move.to());
        }
    }
}


Movements Map::expandUndoMove(Move const & move) const
{
    if (move.isAtomicMove())
    {
        Movements result;

        result.addMove(move);

        return result;
    }
    else
    {
        if (move.stonePushed())
        {
            Movements result;
            QPoint to = move.to();

            QPoint diff_sign = move.diffSign();
            int const number_of_atomic_moves = (to - move.from()).manhattanLength();

            for (int i = 0; i < number_of_atomic_moves; ++i)
            {
                result.addMove(Move(to - diff_sign, to, true));
                to -= diff_sign;
            }

            return result;
        }
        else
        {
            Movements moves = getShortestPath(move.from(), move.to());
            moves.reverse();

            return moves;
        }
    }
}


Movements Map::expandMoves(Movements moves, bool retro_mode) const
{
    moves.setToFirstPosition();
    Movements result;

    Map map = *this;
    map.calcReachable();

    while (moves.hasNextMove())
    {
        Move move = moves.nextMove();

        if (move.from() != map.keeper())
        {
            Movements keeper_moves = map.getShortestPath(map.keeper(), move.from());

            if (keeper_moves.isEmpty())
            {
                return Movements();
            }

            while (keeper_moves.hasNextMove())
            {
                result.addMove(keeper_moves.nextMove());
            }

            map.setKeeper(move.from());
        }

        if (move.stonePushed())
        {
            if (!map.isValidPushMove(move, retro_mode))
            {
                return Movements();
            }

            map.calcReachable();
            QPoint const diff = move.diffSign();
            QPoint from = move.from();

            while (from != move.to())
            {
                result.addMove(Move(from, from + diff, true));
                from += diff;
            }

            map.doMove(move, retro_mode);
        }
        else
        {
            Movements keeper_moves = map.getShortestPath(move.from(), move.to());

            if (keeper_moves.isEmpty())
            {
                return Movements();
            }

            while (keeper_moves.hasNextMove())
            {
                result.addMove(keeper_moves.nextMove());
            }

            map.setKeeper(move.to());
        }
    }

    return result;
}


Movements Map::collapseMoves(Movements moves) const
{
    moves.setToFirstPosition();

    Movements result;
    bool last_was_push = moves.peekNextMove().stonePushed();
    QPoint last_diff = moves.peekNextMove().diff();
    QPoint last_from = moves.peekNextMove().from();
    QPoint last_to;

    while (moves.hasNextMove())
    {
        Move move = moves.nextMove();
        assert(move.isAtomicMove());

        if (move.stonePushed())
        {
            QPoint const diff = move.diff();

            if (last_was_push)
            {
                if (diff != last_diff)
                {
                    result.addMove(Move(last_from, move.from(), true));
                    last_from = move.from();
               }
            }
            else
            {
                if (last_from != move.from())
                {
                    result.addMove(Move(last_from, move.from(), false));
                    last_from = move.from();
                }
            }

            last_was_push = true;
            last_diff = diff;
         }
        else
        {
            if (last_was_push)
            {
                result.addMove(Move(last_from, move.from(), true));
                last_from = move.from();
            }

            last_was_push = false;
        }

        last_to = move.to();
    }

    result.addMove(Move(last_from, last_to, last_was_push));

    return result;
}


bool Map::areValidMoves(Movements const & moves) const
{
    bool is_solution;
    int number_of_pushes;
    int number_of_moves;

    return areValidSolutionMovesImpl(moves, is_solution, number_of_pushes, number_of_moves);
}


bool Map::areValidSolutionMoves(Movements const & moves, int & number_of_pushes, int & number_of_moves) const
{
    bool is_solution;

    if (areValidSolutionMovesImpl(moves, is_solution, number_of_pushes, number_of_moves))
    {
        return is_solution;
    }
    else
    {
        return false;
    }
}


QString Map::toText() const
{
    QString result;

    for (int y = 0; y < m_height; ++y)
    {
        QString line;

        for (int x = 0; x < m_width; ++x)
        {
            line += s_piece_to_text[getPiece(QPoint(x, y))];
        }

        line = ("#" + line).stripWhiteSpace();
        line = line.right(line.length() - 1);
        result += line + '\n';
    }

    return result;
}


QString Map::toServerFormat() const
{
    QString result;

    for (int y = 0; y < m_height; ++y)
    {
        QString line;

        for (int x = 0; x < m_width; ++x)
        {
            line += QString::number(getPiece(QPoint(x, y)));
        }

        if (y < m_height - 1)
        {
            line += '8';
        }

        result += line;
    }

    return result;
}


bool Map::isMapLine(QString const & line)
{
    if (s_map_regexp == 0)
    {
        s_map_regexp = new QRegExp("^ *#[# .$*@+]* *$");
    }

    return s_map_regexp->search(line, 0) == 0;
}


bool Map::pieceContainsKeeper(int piece)
{
    assert(piece >= KEEPER);
    assert(piece <= OUTSIDE);

    return piece <= KEEPER_ON_GOAL;
}


bool Map::pieceContainsGem(int piece)
{
    assert(piece >= KEEPER);
    assert(piece <= OUTSIDE);

    return s_piece_contains_gem[piece];
}


bool Map::pieceContainsGoal(int piece)
{
    assert(piece >= KEEPER);
    assert(piece <= OUTSIDE);

    return s_piece_contains_goal[piece];
}


bool Map::canDropKeeperOnPiece(int piece)
{
    assert(piece >= KEEPER);
    assert(piece <= OUTSIDE);

    return s_can_drop_keeper[piece];
}


bool Map::canDropGemOnPiece(int piece)
{
    assert(piece >= KEEPER);
    assert(piece <= OUTSIDE);

    return s_can_drop_gem[piece];
}


bool Map::areValidSolutionMovesImpl(Movements const & moves, bool & is_solution,
                                    int & number_of_pushes, int & number_of_moves) const
{
    Map test_map = *this;
    Movements test_moves = moves;
    test_moves.setToFirstPosition();

    is_solution = false;
    number_of_pushes = 0;
    number_of_moves = 0;

    while (test_moves.hasNextMove())
    {
        Move move = test_moves.nextMove();
        assert(move.isAtomicMove());

        if (test_map.isValidMove(move, false))
        {
            test_map.doMove(move, false);

            ++number_of_moves;

            if (move.stonePushed())
            {
                ++number_of_pushes;
            }
        }
        else
        {
            return false;
        }
    }

    is_solution = test_map.isSolved();

    return true;
}


Map Map::adjustSize() const
{
    int const old_width = width();
    int const old_height = height();

    int upper_border = 0;

    for (int y = 0; y < old_height; ++y)
    {
        bool is_empty = true;

        for (int x = 0; x < old_width; ++x)
        {
            if (getPiece(QPoint(x, y)) != Map::OUTSIDE)
            {
                is_empty = false;

                break;
            }
        }

        if (!is_empty)
        {
            break;
        }

        ++upper_border;
    }

    int lower_border = 0;

    for (int y = old_height - 1; y >= 0; --y)
    {
        bool is_empty = true;

        for (int x = 0; x < old_width; ++x)
        {
            if (getPiece(QPoint(x, y)) != Map::OUTSIDE)
            {
                is_empty = false;

                break;
            }
        }

        if (!is_empty)
        {
            break;
        }

        ++lower_border;
    }

    int left_border = 0;

    for (int x = 0; x < old_width; ++x)
    {
        bool is_empty = true;

        for (int y = 0; y < old_height; ++y)
        {
            if (getPiece(QPoint(x, y)) != Map::OUTSIDE)
            {
                is_empty = false;

                break;
            }
        }

        if (!is_empty)
        {
            break;
        }

        ++left_border;
    }

    int right_border = 0;

    for (int x = old_width - 1; x >= 0; --x)
    {
        bool is_empty = true;

        for (int y = 0; y < old_height; ++y)
        {
            if (getPiece(QPoint(x, y)) != Map::OUTSIDE)
            {
                is_empty = false;

                break;
            }
        }

        if (!is_empty)
        {
            break;
        }

        ++right_border;
    }

    int const new_width = old_width - left_border - right_border;
    int const new_height = old_height - upper_border - lower_border;
    int const new_size = new_width * new_height;
    int * new_pieces = new int[new_size];

    for (int y = 0; y < new_height; ++y)
    {
        for (int x = 0; x < new_width; ++x)
        {
            QPoint const position = QPoint(x + left_border, y + upper_border);
            new_pieces[x + new_width * y] = getPiece(position);
        }
    }

    return Map(new_width, new_height, std::vector<int>(new_pieces, new_pieces + new_size));
}


Map Map::fillEdges() const
{
    static int offset_indices[4] = {0, 2, 1, 3};

    Map map = *this;

    for (int i = 0; i < m_size; ++i)
    {
        if (map.getPiece(i) == OUTSIDE)
        {
            bool fill = false;

            for (int j = 0; j < 4; ++j)
            {
                int const offset1 = m_xy_offsets[offset_indices[j]];
                int const offset2 = m_xy_offsets[offset_indices[(j + 1) & 3]];
                int const index1 = i + offset1;
                int const index2 = i + offset2;

                if ( (index1 < 0) || (index1 >= m_size) || (index2 < 0) || (index2 >= m_size) )
                {
                    continue;
                }

                if ( (map.getPiece(index1) == WALL) && (map.getPiece(index2) == WALL) )
                {
                    switch(map.getPiece(i + offset1 + offset2))
                    {
                    case OUTSIDE:
                    case WALL:
                        break;

                    default:
                        fill = true;
                        break;
                    }
                }

                if (fill)
                {
                    map.setPiece(i,  WALL);
                }
            }
        }
    }

    return map;
}


Map Map::simplify() const
{
    int const new_width = width() + 2;
    int const new_height = height() + 2;
    int const new_size = new_width * new_height;
    int const new_max = new_size - new_width;
    std::vector<int> new_pieces(new_size, OUTSIDE);
    std::vector<int> unmoveable_gems(new_size, 0);

    for (int x = 0; x < m_width; ++x)
    {
        for (int y = 0; y < m_height; ++y)
        {
            new_pieces[x + 1 + (y + 1) * new_width] = m_pieces[x + y * m_width];
        }
    }

    int xy_offsets[4] = {-1, 1, -new_width, new_width};

    Map map(new_width, new_height, new_pieces);
    Map without_gems = map;

    map.calcDeadlocks();

    for (int i = new_width + 1; i < new_max; ++i)
    {
        if (map.getPiece(i) == GEM_ON_GOAL)
        {
            bool unmoveable = true;

            for (int j = 0; j < 4; ++j)
            {
                int const index = i + xy_offsets[j];
                int const other_index = i + xy_offsets[j ^ 1];

                if (map.canDropKeeper(other_index) && map.canDropGem(index) && !map.isDeadlock(index))
                {
                    unmoveable = false;
                    break;
                }
            }

            if (unmoveable)
            {
                unmoveable_gems[i] = 1;
            }
        }
    }

    bool was_moveable;

    do
    {
        was_moveable = false;

        for (int i = new_width + 1; i < new_max; ++i)
        {
            if (unmoveable_gems[i] )
            {
                bool is_moveable = false;

                for (int j = 0; j < 4; ++j)
                {
                    int const index = i + xy_offsets[j];
                    int const other_index = i + xy_offsets[j ^ 1];

                    if ( (map.canDropGem(other_index) || (unmoveable_gems[other_index] == 0)) &&
                         ((map.canDropGem(index) || unmoveable_gems[index] == 0) && !map.isDeadlock(index)) )
                    {
                        unmoveable_gems[i] = 0;
                        is_moveable = true;
                        break;
                    }
                }

                if (is_moveable)
                {
                    was_moveable = true;
                }
            }
        }
    }
    while (was_moveable);

    for (int i = new_width + 1; i < new_max; ++i)
    {
        if (unmoveable_gems[i])
        {
            map.setPiece(i, WALL);
        }
    }

    for (int i = new_width + 1; i < new_max; ++i)
    {
        if (without_gems.containsGem(i))
        {
            without_gems.setPiece(i, EMPTY);
        }
    }

    without_gems.calcReachable();

    for (int i = new_width + 1; i < new_max; ++i)
    {
        if (!without_gems.isReachable(i) && (map.getPiece(i) != WALL))
        {
            map.setPiece(i, OUTSIDE);
        }
    }

    bool simplified;

    do
    {
        simplified = false;

        for (int i = new_width + 1; i < new_size; ++i)
        {
            if (map.getPiece(i) == WALL)
            {
                bool can_be_deleted = true;

                for (int j = 0; j < 4; ++j)
                {
                    switch (map.getPiece(i + xy_offsets[j]))
                    {
                    case WALL:
                    case OUTSIDE:
                        break;

                    default:
                        can_be_deleted = false;
                        break;
                    }
                }

                if (can_be_deleted)
                {
                    map.setPiece(i, OUTSIDE);
                    simplified = true;
                }
            }
        }
    }
    while (simplified);

    for (int i = 0; i < new_size; ++i)
    {
        if (map.getPiece(i) == OUTSIDE)
        {
            map.setPiece(i, EMPTY);
        }
    }

    map.createOutsidePieces();

    return map.adjustSize();
}


void Map::calcTrivialDeadlocks() const
{
    for (int i = 0; i < m_size; ++i)
    {
        if (isPossibleDeadlock(i))
        {
            if ( (getPiece(i - 1) == WALL) || (getPiece(i + 1) == WALL) )
            {
                if ( (getPiece(i - m_width) == WALL) || (getPiece(i + m_width) == WALL) )
                {
                    m_pieces[i] |= DEADLOCK;
                }
            }
        }
    }
}


bool Map::isPossibleDeadlock(int index) const
{
    assert(isValidIndex(index));

    int const piece = getPiece(index);

    return (piece == EMPTY) || (piece == GEM) || (piece == KEEPER);
}


void Map::setupOffsets()
{
    m_xy_offsets[0] = -1;
    m_xy_offsets[1] = 1;
    m_xy_offsets[2] = -m_width;
    m_xy_offsets[3] = m_width;
}


void Map::setupNumberOfEmptyGoals() const
{
    m_empty_goals = 0;

    for (int i = 0; i < m_size; ++i)
    {
        int const piece = getPiece(i);

        if (pieceContainsGoal(piece) && !pieceContainsGem(piece))
        {
            ++m_empty_goals;
        }
    }

    m_empty_goals_valid = true;
}


void Map::setupKeeperAndEmptyGoals()
{
    m_empty_goals = 0;

    for (int i = 0; i < m_size; ++i)
    {
        int const piece = getPiece(i);

        if (pieceContainsGoal(piece) && !pieceContainsGem(piece))
        {
            ++m_empty_goals;
        }

        if (pieceContainsKeeper(piece))
        {
            m_keeper = getPoint(i);
        }
    }
}


void Map::createOutsidePieces()
{
    for (int x = 0; x < m_width; ++x)
    {
        createOutsidePiecesHelper(x, 0);
        createOutsidePiecesHelper(x, m_height - 1);
    }

    for (int y = 0; y < m_height; ++y)
    {
        createOutsidePiecesHelper(0, y);
        createOutsidePiecesHelper(m_width - 1, y);
    }
}


void Map::createOutsidePiecesHelper(int x, int y)
{
    int const index = x + m_width * y;

    if (getPiece(index) != EMPTY)
    {
        return;
    }

    m_pieces[index] = OUTSIDE;

    if (x > 0)
    {
        createOutsidePiecesHelper(x - 1, y);
    }

    if (y > 0)
    {
        createOutsidePiecesHelper(x, y - 1);
    }

    if (x + 1 < m_width)
    {
        createOutsidePiecesHelper(x + 1, y);
    }

    if (y + 1 < m_height)
    {
        createOutsidePiecesHelper(x, y + 1);
    }
}


void Map::setPieces(CompressedMap const & compressed_map)
{
    std::vector<int> tmp_pieces;
    compressed_map.setPieces(tmp_pieces);
    assert(static_cast<int> (tmp_pieces.size())== m_size);

    m_pieces = new int[m_size];

    for (int i = 0; i < m_size; ++i)
    {
        m_pieces[i] = tmp_pieces[i];
    }
}


void Map::init(Map const & map)
{
    m_width = map.m_width;
    m_height = map.m_height;
    m_size = map.m_size;
    m_keeper = map.m_keeper;
    m_validity = map.m_validity;
    m_empty_goals = map.m_empty_goals;
    m_deadlocks_valid = map.m_deadlocks_valid;
    m_reachable_valid = map.m_reachable_valid;
    m_empty_goals_valid = map.m_empty_goals_valid;
    m_validity_valid = map.m_validity_valid;

    m_pieces = new int[m_size];

    for (int i = 0; i < m_size; ++i)
    {
        m_pieces[i] = map.m_pieces[i];
    }

    for (int i = 0; i < 4; ++i)
    {
        m_xy_offsets[i] = map.m_xy_offsets[i];
    }
}
