/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_MAP_INC_GUARD_H
#define EASYSOK_MAP_INC_GUARD_H


#include <vector>

#include <qglobal.h>
#include <qpoint.h>
#include <qregexp.h>
#include <qstring.h>

#include "movements.h"


class CompressedMap;
class Move;
class QDataStream;
class QStringList;




/**
 * This class represents a sokoban map.
 *
 * Note that there exists two version of most functions.
 * One taking a point (x-, y-values) and one an index (x + y * width()).
 *
 * @warning The height and width of the map must be smaller than 128.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class Map
{

public:


    /**
     * Here we define the pieces of a field.
     */

    enum Piece
    {
        /**
         * A keeper on an empty field.
         */

        KEEPER = 0,


        /**
         * A keeper on a goal field.
         */

        KEEPER_ON_GOAL = 1,


        /**
         * A gem on an empty field.
         */

        GEM = 2,


        /**
         * A gem on an empty field.
         */

        GEM_ON_GOAL = 3,


        /**
         * An empty field.
         */

        EMPTY = 4,


        /**
         * A goal field.
         */

        GOAL = 5,


        /**
         * A wall field.
         */

        WALL = 6,


        /**
         * An outside field.
         */

        OUTSIDE = 7
   };


    /**
     * Defines some reasons, why a map is not valid.
     */

    enum MapValidity
    {
        /**
         * The map is valid.
         */

        IS_VALID,


        /**
         * The map contains no keeper.
         */

        NO_KEEPER,


        /**

         * The map contains more than one keeper.
         */

        TOO_MANY_KEEPERS,


        /**
         * The map contains no gems.
         */

        NO_GEMS,


        /**
         * The map contains more gems than goals.
         */

        MORE_GEMS_THAN_GOALS,


        /**
         * The map contains more goals than gems.
         */

        MORE_GOALS_THAN_GEMS,


        /**
         * The map is not closed (an outside field next to a non wall field).
         */

        MAP_LEAKS,


        /**
         * The map is already solved.
         */

        MAP_SOLVED,


        /**
         * Any other reason why the map is invalid.
         */

        MAP_INVALID
    };


    /**
     * The copy constructor.
     */

    Map(Map const & map);


    /**
     * Creates a new map.
     *
     * @param width The width of the map.
     * @param height The height of the map.
     * @param pieces A vector containing the pieces (a combination of FieldContent flags).
     */

    Map(int width, int height, std::vector<int> const & pieces);


    /**
     * Constructs a map from a data stream.
     *
     * @param stream The stream to use.
     */

    Map(QDataStream & stream);


    /**
     * Constructs the map from a list of lines in xsb format.
     *
     * The constructor removes all lines in the string list, which were before the
     * actual map and the lines of the actual map itself.
     * Be sure to call isValid() afterwards.
     *
     * @param lines The list with the lines.
     */

    Map(QStringList & lines);


    /**
     * Constructs a map from a compressed map.
     *
     * @param compressed_map The compressed map.
     */

    Map(CompressedMap const & compressed_map);


    /**
     * Releases the resources.
     */

    ~Map();


    /**
     * The copy assignment operator.
     */

    Map & operator = (Map const & map);


    /**
     * Writes the map to a data stream.
     *
     * @param stream The stream to use.
     */

    void writeToStream(QDataStream & stream);


    /**
     * Returns the width of the map.
     */

    int width() const;


    /**
     * Returns the height of the map.
     */

    int height() const;


    /**
     * Returns a vector with the raw piece codes.
     */

    std::vector<int> pieces() const;


    /**
     * Returns true, if the map is valid.
     */

    bool isValid() const;


    /**
     * Returns the validity of the map.
     *
     * You should call this after creation of the map, when you can not be sure if the map
     * is valid.
     */

    MapValidity validity() const;


    /**
     * Returns true, if the map is solved.
     */

    bool isSolved() const;


    /**
     * Returns the number of gems.
     */

    int numberOfGems() const;


    /**
     * Returns the number of goals without gems.
     */

    int numberOfEmptyGoals() const;


    /**
     * Returns the index for the given position.
     *
     * @param position The position for which to create the index.
     */

    int getIndex(QPoint position) const;


    /**
     * Returns the point for the given index.
     *
     * @param index.
     */

    QPoint getPoint(int index) const;


    /**
     * Returns true, if the given position is valid (corresponds to a field of the map).
     *
     * @param position The position to test.
     */

    bool isValidPosition(QPoint position) const;


    /**
     * Returns true, if the index is valid.
     *
     * @param index The index to test.
     */

    bool isValidIndex(int index) const;


    /**
     * Returns the piece at the given position.
     *
     * @param position The position of the piece.
     */

    Piece getPiece(QPoint position) const;


    /**
     * @overload
     */

    Piece getPiece(int index) const;


    /**
     * Sets the piece at the given position.
     *
     * If you want to set the keeper, you better use setKeeper().
     *
     * @param position The position.
     * @param piece The Piece to set.
     */

    void setPiece(QPoint position, int piece);


    /**
     * @overload
     */

    void setPiece(int index, int piece);


    /**
     * Mirrors the map vertically.
     */

    void mirrorVertically();


    /**
     * Mirrors the map horizontally.
     */

    void mirrorHorizontally();


    /**
     * Rotates the map left by 90 deg.
     *
     * This is implemented by 3 times rotateRight().
     */

    void rotateLeft();


    /**
     * Rotates the map right by 90 deg.
     */

    void rotateRight();


    /**
     * Returns true, if the piece at the given position is crossed.
     */

    bool isCrossed(QPoint position) const;


    /**
     * @overload
     */

    bool isCrossed(int index) const;


    /**
     * Clears all crossed fields.
     */

    void uncrossAll();


    /**
     * Returns true, if the field is reachable.
     *
     * Note that you should have called calcReachable() before.
     */

    bool isReachable(QPoint position) const;


    /**
     * @overload
     */

    bool isReachable(int index) const;


    /**
     * Calculates all reachable fields.
     *
     * Note that the position of the keeper is important.
     */

    void calcReachable() const;


    /**
     * Calculates all reachable fields if the keeper stands on the specified position.
     *
     * @param position The position of the keeper.
     */

    void calcReachable(QPoint position) const;


    /**
     * @overload
     */

    void calcReachable(int index) const;


    /**
     * Clears all reachable fields.
     */

    void clearReachable() const;


    /**
     * Returns true, if the given field is a deadlock field.
     *
     * @param position The position of the field to test.
     */

    bool isDeadlock(QPoint position) const;


    /**
     * @overload
     */

    bool isDeadlock(int index) const;


    /**
     * Clears all deadlocks.
     */

    void clearDeadlocks() const;


    /**
     * Calculates all deadlock fields.
     */

    void calcDeadlocks() const;


    /**
     * Crosses all deadlocks.
     */

    void crossDeadlocks() const;


    /**
     * Returns the position of the keeper.
     */

    QPoint keeper() const;


    /**
     * Sets the position of the keeper.
     *
     * Removes the keeper from the old position (if it still was at the old position).
     *
     * Note that the keeper is 'beamed'.
     *
     * @param position The new position of the keeper.
     */

    void setKeeper(QPoint position);


    /**
     * @overload
     */

    void setKeeper(int index);


    /**
     * Sets the keeper to the first reachable field.
     */

    void setKeeperToFirstReachable();


    /**
     * Moves a gem from one to another position.
     *
     * @param from The original position of the gem.
     * @param to The new position of the gem (may be equal to from).
     */

    void moveGem(QPoint from, QPoint to);


    /**
     * @overload
     */

    void moveGem(int from, int to);


    /**
     * Returns true, if the piece at the given position contains a keeper.
     *
     * @param position The position to test.
     */

    bool containsKeeper(QPoint position) const;


    /**
     * @overload
     */

    bool containsKeeper(int index) const;


    /**
     * Returns true, if the piece at the given position contains a gem.
     *
     * @param position The position to test.
     */

    bool containsGem(QPoint position) const;


    /**
     * @overload
     */

    bool containsGem(int index) const;


    /**
     * Returns true, if the piece at the given position contains a goal.
     *
     * @param position The position to test.
     */

    bool containsGoal(QPoint position) const;


    /**
     * @overload
     */

    bool containsGoal(int index) const;


    /**
     * Returns true, if a keeper can be dropped on the field.
     *
     * Note that a field with a keeper is also considered a valid keeper drop field.
     *
     * @param position The field to test.
     */

    bool canDropKeeper(QPoint position) const;


    /**
     * @overload
     */

    bool canDropKeeper(int index) const;


    /**
     * Returns true, if a gem can be dropped to the given position.
     *
     * @param position The position to test.
     */

    bool canDropGem(QPoint position) const;


    /**
     * @overload
     */

    bool canDropGem(int index) const;


    /**
     * Performs the given move.
     *
     * @param move The move to perform.
     * @param retro_mode If true, the move is in retro mode.
     */

    void doMove(Move const & move, bool retro_mode);


    /**
     * Performs the given undo move.
     *
     * An undo move is the reverse of a normal move.
     *
     * @param move The undo move to perform.
     * @param retro_mode If true, the move is in retro mode.
     */

    void doUndoMove(Move const & move, bool retro_mode);


    /**
     * Returns true, if the given move is valid.
     *
     * @param move The move to test.
     * @param retro_mode If true, the move is in retro mode.
     */

    bool isValidMove(Move const & move, bool retro_mode) const;


    /**
     * Returns true, if the given (not push move) is valid.
     *
     * @param move The move to test.
     */

    bool isValidNonPushMove(Move const & move) const;


    /**
     * Returns true, if the given stone push move is valid (meaning the stone can be pushed).
     *
     * @param move The move to test.
     * @param retro_mode If true, the move is in retro mode.
     */

    bool isValidPushMove(Move const & move, bool retro_mode) const;


    /**
     * Returns true, if the given stone push move is valid (meaning the stone can be pushed).
     *
     * @param move The move to test.
     * @param retro_mode If true, the move is in retro mode.
     */

    bool isValidAtomicPushMove(Move const & move, bool retro_mode) const;


    /**
     * Gets the shortes path for a keeper (if it exists) between two points.
     *
     * If no path exists, we return Movements with no moves.
     *
     * @param from The start point.
     * @param to The end point.
     */

    Movements getShortestPath(QPoint from, QPoint to) const;


    /**
     * Gets the shortes path (if it exists) for moving a gem from one position to another.
     *
     * If no path exists, we return Movements with no moves.
     *
     * Note that the position of the keeper matters.
     *
     * @param from The gem start point.
     * @param to The gem end point.
     * @param retro If true, we are in retro mode.
     */

    Movements getShortestPathForGem(QPoint from, QPoint to, bool retro) const;


    /**
     * Returns the distance map for a given field.
     *
     * The value of the returned map at [d + 4 * gem_pos] is the number of
     * moves required, to move a gem at position gem_pos to the given field, when the
     * keeper stands left (d = 0), right (d = 1), above (d = 2) or down (d = 3) the
     * gem.
     *
     * @param position The position of the field to reach.
     * @param max_distance The distance to indicate unreachable positions.
     * @param retro_mode If true, the move is in retro mode.
     */

    std::vector<int> getDistanceMap(QPoint const & position, int unsolvable, bool retro_mode = false) const;


    /**
     * @overload
     */

    std::vector<int> getDistanceMap(int index, int unsolvable, bool retro_mode = false) const;


    /**
     * Returns a sequence of atomic moves for a given move.
     *
     * The move must be legal.
     *
     * @param move The move to make atomic.
     * @param retro_mode If true, the move is in retro mode.
     */

    Movements expandMove(Move const & move, bool retro_mode) const;


    /**
     * Returns a sequence of atomic moves for a given undo move.
     *
     * The undo move must be legal.
     *
     * @param move The move undo to make atomic.
     */

    Movements expandUndoMove(Move const & move) const;


    /**
     * Returns the movements given in an explicit form (meaning all moves are simple moves).
     *
     * If the moves are illegal starting from the current map, an empty moves list ist returned.
     *
     * @param moves The moves to expand.
     * @param retro_mode If true, the move is in retro mode.
     */

    Movements expandMoves(Movements moves, bool retro_mode) const;


    /**
     * Returns the movements given in a collapsed form.
     *
     * This means simple keeper moves are combined und ston pushes in one direction also.
     *
     * Note the we don't check, if the moves are valid, so it may be the best to call
     * expandMoves first.
     *
     * @param moves The moves to collapse (must be all simple moves).
     */

    Movements collapseMoves(Movements moves) const;


    /**
     * Returns true, if the given moves are valid starting from this map.
     *
     * Note that the moves must be atomic!
     *
     * @param moves The moves to check.
     */

    bool areValidMoves(Movements const & moves) const;


    /**
     * Returns true, if the given moves are valid starting from this map and solve this map.
     *
     * Note that the moves must be atomic!
     *
     * @param moves The moves to check.
     * @param number_of_pushes Here we store the number of pushes.
     * @param number_of_moves Here we store the number of moves.
     */

    bool areValidSolutionMoves(Movements const & moves, int & number_of_pushes, int & number_of_moves) const;


    /**
     * Return a new map, which has the minimum possible size of the current map.
     */

    Map adjustSize() const;


    /**
     * Returns a version of the map with filled edges.
     */

    Map fillEdges() const;


    /**
     * Returns a simplified version of the map.
     *
     * In the simplified version we deleted as much as possible walls and gems, without
     * changing the solution of the map.
     */

    Map simplify() const;


    /**
     * Returns a human readable representation of the map in xsokoban format.
     */

    QString toText() const;


    /**
     * Returns a representation of the map which is understandable by the highscore server.
     */

    QString toServerFormat() const;


    /**
     * Returns true, if the given string is a map line.
     *
     * @param line The string of the line.
     */

    static bool isMapLine(QString const & line);


    /**
     * Returns true, if the piece contains a keeper.
     *
     * @param piece The piece to test.
     */

    static bool pieceContainsKeeper(int piece);


    /**
     * Returns true, if the piece contains a gem.
     *
     * @param piece The piece to test.
     */

    static bool pieceContainsGem(int piece);


    /**
     * Returns true, of the piece contains a goal.
     *
     * @param piece The piece to test.
     */

    static  bool pieceContainsGoal(int piece);


    /**
     * Returns true, if the keeper can be dropped in the piece.
     *
     * @param piece The piece to test.
     */

    static bool canDropKeeperOnPiece(int piece);


    /**
     * Returns true, if a gem can be dropped on the given piece.
     *
     * A field with the keeper is also considered a valid gem drop field.
     *
     * @param piece The piece to test.
     */

    static bool canDropGemOnPiece(int piece);


private:

    /**
     * Implementation of areValidSolutionMoves().
     *
     * Also used in areValidMoves().
     *
     * Note that the moves must be atomic!
     *
     * @param moves The moves to check.
     * @param is_solution Set to true, if the moves are a solution.
     * @param number_of_pushes Here we store the number of pushes.
     * @param number_of_moves Here we store the number of moves.
     */

    bool areValidSolutionMovesImpl(Movements const & moves, bool & is_solution,
                                   int & number_of_pushes, int & number_of_moves) const;


    /**
     * Defines some additional flags for a piece.
     */

    enum PieceFlags
    {
        /**
         * The field is crossed.
         */

        CROSSED = 8,


        /**
         * The field is reachable.
         */

        REACHABLE = 16,


        /**
         * The field is a deadlock field.
         */

        DEADLOCK = 32
    };


     /**
     * Used to map out the different informations on the fields.
     */

    enum Mask
    {
        /**
         * Mask out the piece.
         */

        PIECE = 7,


        /**
         * All attributes.
         */

        ALL = PIECE + CROSSED + REACHABLE + DEADLOCK,


        /**
         * Clear all pieces.
         */

        CLEAR_PIECE = ALL - PIECE,


        /**
         * Clear crossed.
         */

        CLEAR_CROSSED = ALL - CROSSED,


        /**
         * Clear reachable.
         */

        CLEAR_REACHABLE = ALL - REACHABLE,


        /**
         * Clear deadlocks.
         */

        CLEAR_DEADLOCK = ALL - DEADLOCK
    };


    /**
     * Calculates all trivial deadlocks (positions in which a stone can not be moved).
     */

    void calcTrivialDeadlocks() const;


    /**
     * Returns true, if the field is a possible deadlock field.
     *
     * This means it must be EMPTY, GEM or KEEPER.
     *
     * @param index The field to test.
     */

    bool isPossibleDeadlock(int index) const;


    /**
     * Sets up the offsets.
     */

    void setupOffsets();


    /**
     * Sets up the numer of empty goals.
     */

    void setupNumberOfEmptyGoals() const;


    /**
     * Finds and set the keeper and the number of empty goals.
     */

    void setupKeeperAndEmptyGoals();


    /**
     * Creates the outside pieces.
     */

    void createOutsidePieces();


    /**
     * Helper function needed by createOutsidePieces().
     */

    void createOutsidePiecesHelper(int x, int y);


    /**
     * Sets the pieces from the compressed map.
     * The pieces must *not* have been set before.
     *
     * @param compressed_map The compressed map to use.
     */

    void setPieces(CompressedMap const & compressed_map);


    /**
     * Initialized the map with the given map.
     *
     * @param map The map to use.
     */

    void init(Map const & map);


    /**
     * The width of the map.
     */

    int m_width;


    /**
     * The height of the map.
     */

    int m_height;


    /**
     * This is m_width * m_height.
     */

    int m_size;


    /**
     * Here we store the position of the keeper.
     */


    QPoint m_keeper;


    /**
     * The validity of the map.
     */

    mutable MapValidity m_validity;


    /**
     * The number of empty goals.
     */

    mutable int m_empty_goals;


    /**
     * If true, we have already calculated the deadlocks.
     */

    mutable bool m_deadlocks_valid;


    /**
     * If true, the reachable fields are up to date.
     */

    mutable bool m_reachable_valid;


    /**
     * If true, the numer of empty goals is valid.
     */

    mutable bool m_empty_goals_valid;


    /**
     * If true, the validity is valid :)
     */

    mutable bool m_validity_valid;


    /**
     * Here we store the pieces and additional information of the fields.
     * This is not a vector, because performance is paramount when accessing
     * the fields.
     */

    mutable int * m_pieces;


    /**
     * Here we store the offset to go one in x- or y-direction.
     */

    int m_xy_offsets[4];


    /**
     * Used to convert compressed piece values to characters used by xsokoban.
     */

    static char const s_piece_to_text[8];


    /**
     * Used to determine, if a piece contains a gem.
     */

    static bool const s_piece_contains_gem[8];


    /**
     * Used to determine, if a piece contains a goal.
     */

    static bool const s_piece_contains_goal[8];


    /**
     * Used to determine, if we can drop the keeper on the piece.
     */

    static bool const s_can_drop_keeper[24];


    /**
     * Used to determine, if we can drop a gem on the piece.
     */

    static bool const s_can_drop_gem[8];


    /**
     * The regexp used to identify map lines.
     */

    static QRegExp * s_map_regexp;
};


#endif
