/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "map_widget.h"

#include <algorithm>
#include <cassert>

#include <qtimer.h>

#include <kapp.h>
#include <kconfig.h>
#include <kcursor.h>
#include <kglobalsettings.h>

#include "map.h"
#include "piece_image.h"
#include "pixmap_provider.h"



MapWidget::MapWidget(Map const * map, PixmapProvider * pixmap_provider, Theme const * theme,
                     bool disable_auto_hide, QWidget * parent, char const * name) :
    QCanvasView(0, parent, name),
    m_canvas(640, 480),
    m_pixmap_provider(pixmap_provider),
    m_theme(theme),
    m_map(map),
    m_square_size(20),
    m_map_width(-1),
    m_map_height(-1),
    m_size(0),
    m_valid_click(false),
    m_in_drag(false),
    m_keeper_direction(AtomicMove::UP),
    m_virtual_keeper_pos(QPoint(-1, -1)),
    m_virtual_keeper_item(0),
    m_moving_item(0),
    m_timer(new QTimer(this)),
    m_repeat_timer(new QTimer(this)),
    m_cursor_hidden(false),
    m_disable_auto_hide(disable_auto_hide),
    m_cursor_timer(new QTimer(this)),
    m_temp_disable_auto_hide(false)
{
    assert(pixmap_provider != 0);
    assert(theme != 0);

    connect(m_timer, SIGNAL(timeout()), this, SLOT(switchToBetterScaling()));
    connect(m_repeat_timer, SIGNAL(timeout()), this, SLOT(mouseRepeat()));
    connect(m_cursor_timer, SIGNAL(timeout()), this, SLOT(cursorHidden()));

    configChanged();
    setMap(map);

    m_canvas.setDoubleBuffering(true);
    setCanvas(&m_canvas);
}

MapWidget::~MapWidget()
{
    deleteItems(m_moving_item);
    deleteItems();
    deletePixmaps();
    delete m_pixmap_provider;
}


void MapWidget::setMap(Map const * map)
{
    m_map = map;

    if (map == 0)
    {
        return;
    }

    m_map_width = map->width();
    m_map_height = map->height();
    m_size = m_map_height * m_map_width;
    m_pieces = map->pieces();

    int const old_square_size = m_square_size;

    setVirtualKeeper(QPoint(-1, -1));

    deleteItems();

    updateGeometry();

    if (old_square_size != m_square_size)
    {
        deletePixmaps();
    }

    ensureFieldVisible(map->keeper());

    newItems();

    m_force_update = true;
    m_canvas.retune(m_square_size, 1);

    if (m_scaling_mode == 2)
    {
        m_pixmap_provider->setSmoothScaling(false);
        m_timer->stop();
        m_timer->start(m_switch_time, true);
    }
}


void MapWidget::setPixmapProvider(PixmapProvider * pixmap_provider, Theme const * theme,
                                  bool update_background)
{
    assert(pixmap_provider != 0);
    assert(theme != 0);

    m_theme = theme;

    if (m_pixmap_provider != pixmap_provider)
    {
        delete m_pixmap_provider;
    }

    if ((m_scaling_mode == 0) && pixmap_provider->smoothScaling())
    {
        pixmap_provider->setSmoothScaling(false);
    }
    else if ((m_scaling_mode == 1) && !pixmap_provider->smoothScaling())
    {
        pixmap_provider->setSmoothScaling(true);
    }

    m_pixmap_provider = pixmap_provider;

    if (update_background)
    {
        if (pixmap_provider->hasBackgroundImage())
        {
            m_canvas.setBackgroundPixmap(pixmap_provider->backgroundImage());
        }
        else
        {
            QPixmap empty_pixmap = QPixmap(64, 64);
            empty_pixmap.fill(pixmap_provider->backgroundColor());
            m_canvas.setBackgroundPixmap(empty_pixmap);
        }
    }

    if (m_map == 0)
    {
        return;
    }

    deleteItems();
    deletePixmaps();

    updateGeometry();

    newItems();

    if (!m_arrows.empty())
    {
        setArrows(m_arrows);
    }

    updateDisplay();
}


AtomicMove::AtomicMoveType MapWidget::keeperDirection()
{
    return m_keeper_direction;
}


void MapWidget::setKeeperDirection(AtomicMove::AtomicMoveType direction)
{
    m_keeper_direction = direction;
    m_force_update = true;
}


void MapWidget::setArrows(std::vector<Move> const & arrows)
{
    deleteArrows();

    m_arrows = arrows;

    int const number_of_arrows = static_cast<int> (arrows.size());

    for (int i = 0; i < number_of_arrows; ++i)
    {
        Move const act_move = arrows[i];
        addArrow(act_move.from(), act_move.to());
    }

    m_force_update = true;
}


void MapWidget::updateDisplay()
{
    if (m_map == 0)
    {
        return;
    }

    int pieces_changed = 0;

    for (int i = 0; i < m_size; ++i)
    {
        bool const is_crossed = m_map->isCrossed(i);
        int const piece = m_map->getPiece(i);
        int const piece_and_crossed = piece + (is_crossed << 3);

        if (piece_and_crossed != m_pieces[i])
        {
            m_pieces[i] = piece_and_crossed;
            QPoint const point = m_map->getPoint(i);

            createItems(m_items[i], point, m_square_size * point.x() + m_x_offset,
                        m_square_size * point.y() + m_y_offset, 1);

            ++pieces_changed;

            if (piece < 2) // A field with a keeper
            {
                ensureFieldVisible(m_map->getPoint(i));
            }
        }
    }

    if ((pieces_changed > 0) || m_force_update)
    {
        m_force_update = false;
        m_canvas.update();
    }
}


QPoint const & MapWidget::virtualKeeper() const
{
    return m_virtual_keeper_pos;
}


void MapWidget::setVirtualKeeper(QPoint const & position)
{
    if (position != QPoint(-1, -1))
    {
        QPoint old_pos = m_virtual_keeper_pos;

        if (old_pos != position)
        {
            if (m_virtual_keeper_pos.x() == -1)
            {
                old_pos = m_map->keeper();
            }

            if (old_pos.x() - position.x() > 1)
            {
                old_pos.setX(position.x() - 1);
            }
            else if (old_pos.x() - position.x() < -1)
            {
                old_pos.setX(position.x() + 1);
            }

            if (old_pos.y() - position.y() > 1)
            {
                old_pos.setY(position.y() - 1);
            }
            else if (old_pos.y() - position.y() < -1)
            {
                old_pos.setY(position.y() + 1);
            }

            m_virtual_keeper_direction = Move(old_pos, position, false).atomicMove().type();
        }

        createItems(m_virtual_keeper_item, Theme::LEFT_VIRTUAL_KEEPER + m_virtual_keeper_direction, position,
                    m_square_size * position.x() + m_x_offset,
                    m_square_size * position.y() + m_y_offset, 500);

        ensureFieldVisible(position);
    }
    else
    {
        deleteItems(m_virtual_keeper_item);
    }

    m_virtual_keeper_pos = position;
    m_force_update = true;
}


void MapWidget::tempDisableAutoHide(bool disabled)
{
    m_temp_disable_auto_hide = disabled;

    if (disabled)
    {
        cursorVisible(false);
    }
}


void MapWidget::configChanged()
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_scaling_mode = config->readNumEntry("Scaling mode", 2);
    m_switch_time = config->readNumEntry("Scaling switch time", 150);
    m_min_square_size = config->readNumEntry("Minimum field size", 10);
    m_auto_hide_cursor = config->readNumEntry("Cursor hide", true);
    m_mouse_repeat_delay = config->readNumEntry("Mouse repeat delay", 1000);
    m_mouse_repeat_rate = std::max(config->readNumEntry("Mouse repeat rate", 4), 1);
    m_cursor_hide_delay = config->readNumEntry("Cursor hide delay", 5) * 1000;

    if (m_auto_hide_cursor && !m_disable_auto_hide)
    {
        m_cursor_hidden = false;
        m_cursor_timer->start(m_cursor_hide_delay, true);
    }

    viewport()->setMouseTracking(m_auto_hide_cursor && !m_disable_auto_hide);
    unsetCursor();

    setPixmapProvider(m_pixmap_provider, m_theme);
    setMap(m_map);

    if (m_map != 0)
    {
        updateDisplay();
    }
}


void MapWidget::switchToBetterScaling()
{
    if (!m_pixmap_provider->smoothScaling())
    {
        m_pixmap_provider->setSmoothScaling(true);
        setPixmapProvider(m_pixmap_provider, m_theme, false);
    }
}


void MapWidget::mouseRepeat()
{
    if (m_mouse_repeat_state == MidButton)
    {
        redo();
        m_repeat_timer->start(1000 / m_mouse_repeat_rate, true);
    }
    else if (m_mouse_repeat_state == RightButton)
    {
        undo();
        m_repeat_timer->start(1000 / m_mouse_repeat_rate, true);
    }
}


void MapWidget::cursorVisible(bool stop_autohide)
{
    unsetCursor();
    m_cursor_hidden = false;

    if (!stop_autohide && m_auto_hide_cursor && !m_disable_auto_hide)
    {
        m_cursor_timer->start(m_cursor_hide_delay, true);
    }
    else
    {
        m_cursor_timer->stop();
    }
}


void MapWidget::cursorHidden()
{
    if (m_auto_hide_cursor && !m_disable_auto_hide && !m_temp_disable_auto_hide)
    {
        setCursor(KCursor::blankCursor());
        m_cursor_hidden = true;
    }
}


void MapWidget::resizeEvent(QResizeEvent * event)
{
    QCanvasView::resizeEvent(event);

    if (m_map != 0)
    {
        int const old_square_size = m_square_size;

        deleteItems();

        updateGeometry();

        if (old_square_size != m_square_size)
        {
            deletePixmaps();
        }

        newItems();

        if (!m_arrows.empty())
        {
            setArrows(m_arrows);
        }

        m_canvas.retune(m_square_size, 1);
        m_canvas.update();
    }
}


void MapWidget::contentsMousePressEvent(QMouseEvent * event)
{
    if (m_valid_click)
    {
        return;
    }

    m_mouse_repeat_state = 0;
    m_in_drag = false;
    m_valid_click = false;

    QPoint const position(event->x(), event->y());

    if (isValidPosition(position) && (event->button() == LeftButton))
    {
        m_valid_click = true;
        m_click_position = position;
        m_last_mouse_position = position;
        m_click_field = getFieldFromPosition(position);
        m_click_offset = QPoint(position.x() - (m_click_field.x() * m_square_size + m_x_offset),
                                position.y() - (m_click_field.y() * m_square_size + m_y_offset));
        m_last_drag_position = m_click_field;

        if (!m_map->containsGem(m_click_field) && !m_map->containsKeeper(m_click_field))
        {
            emit mouseDragStarted();
        }
    }
    else if (event->button() == MidButton)
    {
        emit redo();
        m_mouse_repeat_state = MidButton;
        m_repeat_timer->start(m_mouse_repeat_delay, true);
    }
    else if (event->button() == RightButton)
    {
        emit undo();
        m_mouse_repeat_state = RightButton;
        m_repeat_timer->start(m_mouse_repeat_delay, true);
    }
}


void MapWidget::contentsMouseMoveEvent(QMouseEvent * event)
{
    if (m_auto_hide_cursor && !m_disable_auto_hide && m_cursor_hidden)
    {
        unsetCursor();
        m_cursor_hidden = false;
        m_cursor_timer->start(m_cursor_hide_delay, true);
    }

    QPoint const position(event->x(), event->y());

    if (!m_valid_click)
    {
        if (event->state() == LeftButton)
        {
            QPoint act_field = getFieldFromPosition(position);

            if (act_field != m_last_drag_position)
            {
                emit mouseDragged(m_last_drag_position, act_field);
                m_last_drag_position = act_field;
            }
        }

        return;
    }

    if (!m_in_drag)
    {
        int const threshold = KGlobalSettings::dndEventDelay();

        if ( (std::abs(m_click_position.x() - position.x()) > threshold) ||
             (std::abs(m_click_position.y() - position.y()) > threshold) )
        {
            bool const is_gem = m_map->containsGem(m_click_field) && !m_theme->hideGems();
            bool const is_keeper = m_map->containsKeeper(m_click_field);

            if (is_gem || is_keeper)
            {
                m_in_drag = true;
                m_is_keeper_drag = is_keeper;

                int type = is_gem ? Theme::MOVING_GEM : Theme::LEFT_MOVING_KEEPER + m_keeper_direction;

                createItems(m_moving_item, type, QPoint(0, 0), m_click_position.x() - m_click_offset.x(),
                            m_click_position.y() - m_click_offset.y(), 1000);
            }
            else
            {
                emit mouseDragged(m_last_drag_position, m_last_drag_position);

                QPoint const act_field = getFieldFromPosition(position);

                if (act_field != m_last_drag_position)
                {
                    emit mouseDragged(m_last_drag_position, act_field);
                    m_last_drag_position = act_field;
                }

                m_valid_click = false;
            }
        }
    }

    if (m_in_drag && !m_moving_item.empty())
    {
        moveMovingItemPosition(position - m_last_mouse_position);
        m_canvas.update();

        if (m_has_scrollbars)
        {
            ensureVisible(event->x(), event->y(), m_square_size * 2, m_square_size * 2);
        }
    }

    m_last_mouse_position = position;
}


void MapWidget::contentsMouseReleaseEvent(QMouseEvent * event)
{
    if (!m_moving_item.empty())
    {
        deleteItems(m_moving_item);
        m_canvas.update();
    }

    if (m_valid_click)
    {
        if (m_in_drag)
        {
            QPoint const position(event->x() - m_click_offset.x() + m_square_size / 2,
                                  event->y() - m_click_offset.y() + m_square_size / 2);
            QPoint const field = getFieldFromPosition(position);

            if (field != m_click_field)
            {
                if ( (field.x() >= 0) && (field.y() >= 0) && (field.x() < m_map_width) &&
                     (field.y() < m_map_height) )
                {
                    if (m_is_keeper_drag)
                    {
                        emit keeperMoved(m_click_field, field);
                    }
                    else
                    {
                        emit gemMoved(m_click_field, field);
                    }
                }
            }
        }
        else
        {
            emit fieldClicked(m_click_field);
        }
    }
    else if (event->button() == LeftButton)
    {
        QPoint const act_field = getFieldFromPosition(QPoint(event->x(), event->y()));

        if (act_field != m_last_drag_position)
        {
            emit mouseDragged(m_last_drag_position, act_field);
        }

        emit mouseDragEnded();
    }

    m_valid_click = false;
    m_mouse_repeat_state = 0;
}


void MapWidget::deleteItems(std::vector<QCanvasSprite *> & items) const
{
    if (!items.empty())
    {
        int const nr_of_items = static_cast<int> (items.size());

        for (int i = 0; i < nr_of_items; ++i)
        {
            delete items[i];
        }

        items.resize(0);
    }
}


void MapWidget::createItems(std::vector<QCanvasSprite *> & items, int type,
                            QPoint const & position, int x, int y, int z) const
{

    std::vector<int> indices = m_theme->imageIndices(static_cast<Theme::ImageType> (type), position, *m_map);

    createItems(items, indices, x, y, z);
}


void MapWidget::createItems(std::vector<QCanvasSprite *> & items, QPoint const & position,
                            int x, int y, int z) const
{
    createItems(items, m_theme->imageIndicesFromPosition(m_keeper_direction, position, *m_map), x, y, z);
}


void MapWidget::createItems(std::vector<QCanvasSprite *> & items, std::vector<int> image_indices,
                            int x, int y, int z) const
{
    if (!items.empty())
    {
        deleteItems(items);
    }

    int const nr_of_images = static_cast<int> (image_indices.size());
    items.resize(nr_of_images);

    for (int i = 0; i < nr_of_images; ++i)
    {
        int const index = image_indices[i];

        if (m_pixmaps[index] == 0)
        {
            m_pixmaps[index] = createPixmapArray(m_pixmap_provider->createPixmap(index, m_square_size));
        }

         QCanvasSprite * sprite = new QCanvasSprite(m_pixmaps[index], &m_canvas);
         items[i] = sprite;

         QPoint const offset = m_pixmap_provider->offset(index, m_square_size);

         sprite->setX(x + offset.x());
         sprite->setY(y + offset.y());
         sprite->setZ(z + i);
         sprite->show();
    }
}


void MapWidget::updateGeometry()
{
    double const real_width = m_map_width + m_theme->leftBorder() + m_theme->rightBorder();
    double const real_height = m_map_height + m_theme->upperBorder() + m_theme->lowerBorder();
    double const size_for_width = std::max(static_cast<double> (m_min_square_size), width() / real_width);
    double const size_for_height = std::max(static_cast<double> (m_min_square_size), height() / real_height);

    m_square_size = static_cast<int> (std::min(size_for_width, size_for_height));

    int const left_border = static_cast<int> (m_theme->leftBorder() * m_square_size);
    int const right_border = static_cast<int> (m_theme->rightBorder() * m_square_size);
    int const upper_border = static_cast<int> (m_theme->upperBorder() * m_square_size);
    int const lower_border = static_cast<int> (m_theme->lowerBorder() * m_square_size);

    m_x_offset = (std::max(0, (width() - m_square_size * m_map_width - left_border - right_border) / 2) +
                  left_border);
    m_y_offset = (std::max(0, (height() - m_square_size * m_map_height - upper_border - lower_border) / 2) +
                  right_border);

    int const canvas_width = m_square_size * m_map_width + 2 * m_x_offset - left_border + right_border;
    int const canvas_height = m_square_size * m_map_height + 2 * m_y_offset - upper_border + lower_border;

    m_canvas.resize(canvas_width, canvas_height);

    if ( (canvas_width > width()) || (canvas_height > height()) )
    {
        setVScrollBarMode(AlwaysOn);
        setHScrollBarMode(AlwaysOn);
        m_has_scrollbars = true;
    }
    else
    {
        setHScrollBarMode(AlwaysOff);
        setVScrollBarMode(AlwaysOff);
        m_has_scrollbars = false;
    }
}


void MapWidget::deletePixmaps()
{
    int const nr_of_pixmaps = static_cast<int> (m_pixmaps.size());

    for (int i = 0; i < nr_of_pixmaps; ++i)
    {
        delete m_pixmaps[i];
    }

    m_pixmaps.clear();
}


void MapWidget::deleteArrows()
{
    int const number_of_arrows = static_cast<int> (m_arrow_items.size());

    for (int i = 0; i < number_of_arrows; ++i)
    {
        deleteItems(m_arrow_items[i]);
    }

    m_arrow_items.clear();
}


void MapWidget::addArrow(QPoint from, QPoint to)
{
    int const size_half = m_square_size / 2;
    int direction;
    int x_offset = 0;
    int y_offset = 0;

    if (from.x() < to.x())
    {
        direction = 1;
        x_offset = size_half;
    }
    else if (from.x() > to.x())
    {
        direction = 0;
        x_offset = -size_half;
    }
    else if (from.y() < to.y())
    {
        direction = 3;
        y_offset = size_half;
    }
    else
    {
        direction = 2;
        y_offset = -size_half;
    }

    std::vector<QCanvasSprite *> new_arrow;
    createItems(new_arrow, Theme::LEFT_ARROW + direction, from,
                from.x() * m_square_size + x_offset + m_x_offset,
                from.y() * m_square_size + y_offset + m_y_offset, 200);
    m_arrow_items.push_back(new_arrow);
}


void MapWidget::ensureFieldVisible(QPoint const & position)
{
    if (m_has_scrollbars)
    {
        ensureVisible(m_square_size * position.x() + m_x_offset, m_square_size * position.y() + m_y_offset,
                      m_square_size * 2, m_square_size * 2);
    }
}


void MapWidget::deleteItems()
{
    deleteItems(m_virtual_keeper_item);
    deleteItems(m_moving_item);
    deleteArrows();

    int const nr_of_items = static_cast<int> (m_items.size());

    for (int i = 0; i < nr_of_items; ++i)
    {
        deleteItems(m_items[i]);
    }

    m_items.clear();
}


void MapWidget::newItems()
{
    m_items.resize(m_size);
    m_pixmaps.resize(m_theme->nrOfPieceImages(), 0);

    for (int y = 0; y < m_map_height; ++y)
    {
        for (int x = 0; x < m_map_width; ++x)
        {
            int const index = x + y * m_map_width;
            int const piece = m_map->getPiece(index);
            bool const is_crossed = m_map->isCrossed(index);

            createItems(m_items[index], QPoint(x, y), x * m_square_size + m_x_offset,
                        y * m_square_size + m_y_offset, 1);

            m_pieces[index] = piece + (is_crossed << 3);
        }
    }

    setVirtualKeeper(m_virtual_keeper_pos);
}


bool MapWidget::isValidPosition(QPoint position) const
{
    bool const valid_x = ((position.x() >= m_x_offset) &&
                          (position.x() < m_x_offset + m_map_width * m_square_size));
    bool const valid_y = ((position.y() >= m_y_offset) &&
                          (position.y() < m_y_offset + m_map_height * m_square_size));

    return valid_x && valid_y;
}


QPoint MapWidget::getFieldFromPosition(QPoint position) const
{
    return QPoint((position.x() - m_x_offset) / m_square_size, (position.y() - m_y_offset) / m_square_size);
}


QCanvasPixmapArray * MapWidget::createPixmapArray(QPixmap const & pixmap) const
{
    QList<QPixmap> pixmaps;
    QList<QPoint> points;

    pixmaps.append(new QPixmap(pixmap));
    pixmaps.setAutoDelete(true);
    points.append(new QPoint(0, 0));
    points.setAutoDelete(true);

    QCanvasPixmapArray * result = new QCanvasPixmapArray(pixmaps, points);

    return result;
}


void MapWidget::moveMovingItemPosition(QPoint const & offset)
{
    int const nr_of_sprites = static_cast<int> (m_moving_item.size());

    for (int i = 0; i < nr_of_sprites; ++i)
    {
        m_moving_item[i]->moveBy(offset.x(), offset.y());
    }
}
