/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASY_SOK_MAP_WIDGET_INC_GUARD_H
#define EASY_SOK_MAP_WIDGET_INC_GUARD_H


#include <vector>

#include <qcanvas.h>
#include <qpixmap.h>
#include <qpoint.h>

#include "atomic_move.h"
#include "move.h"
#include "theme.h"


class Map;
class PixmapProvider;
class QTimer;
class Theme;




/**
 * A widget to show a map.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class MapWidget : public QCanvasView
{
    Q_OBJECT

public:

    /**
     * Creates the widget with the given map.
     *
     * The widget does not takes ownership of the map or the pixmap provider.
     *
     * @param map The map to display (or 0, if you don't want to display something).
     * @param pixmap_provider The pixmap provider to use.
     * @param theme The theme of the pixmap provider.
     * @param disable_auto_hide If true, the auto hide feature of the map is disabled.
     * @param parent The parent widget.
     * @param name The internal widget name.
     */

    MapWidget(Map const * map, PixmapProvider * pixmap_provider, Theme const * theme,
              bool disable_auto_hide, QWidget * parent = 0, char const * name = 0);


    /**
     * Releasy all resources.
     */

    virtual ~MapWidget();


    /**
     * Sets a map to display.
     *
     * Note that you have to call updateDisplay() to update the display.
     *
     * All arrows and the virtual keeper will be hidden.
     *
     * @param map The map to display.
     */

    void setMap(Map const * map);


    /**
     * Sets the pixmap provider to use.
     *
     * The old pixmap provider will be deleted!
     *
     * @param pixmap_provider The pixmap provider to use.
     * @param theme The theme of the pixmap provider.
     * @param update_background If true, the background is also changed.
     */

    void setPixmapProvider(PixmapProvider * pixmap_provider, Theme const * theme,
                           bool update_background = true);


    /**
     * Returns the current direction of the keeper.
     */

    AtomicMove::AtomicMoveType keeperDirection();


    /**
     * Returns the position of the virtual keeper (QPoint(-1, -1) means hidden).
     */

    QPoint const & virtualKeeper() const;


    /**
     * Controlls if automatic hiding of the cursor is disabled or not.
     *
     * @param disabled If true, auto hiding is disabled.
     */

    void tempDisableAutoHide(bool disabled);


public slots:

    /**
     * Sets the direction of the keeper.
     *
     * @param direction The direction of the keeper.
     */

    void setKeeperDirection(AtomicMove::AtomicMoveType direction);


    /**
     * Sets a vector with arrows for atomic moves.
     *
     * Call updateDisplay() afterwards.
     *
     * @param arrows A vector of moves for which arrorws are drawn.
     */

    void setArrows(std::vector<Move> const & arrows);


    /**
     * Call to update the display.
     */

    void updateDisplay();


    /**
     * Sets the virtual keeper.
     *
     * QPoint(-1,-1) means to hide the virtual keeper.
     *
     * @param position The position of the virtual keeper.
     */

    void setVirtualKeeper(const QPoint & position);


    /**
     * Call this, when the configuration file was changed.
     */

    void configChanged();


    /**
     * This is called, when we should switch to better scaling.
     */

    void switchToBetterScaling();


    /**
     * This is called for mouse repeat.
     */

    void mouseRepeat();


    /**
     * This is called, if the mouse cursor should be displayed.
     *
     * @param stop_autohide If true, the auto hide feature is stopped.
     */

    void cursorVisible(bool stop_autohide);


    /**
     * This is called, if the mouse cursor hide delay expired.
     */

    void cursorHidden();


signals:

    /**
     * This is emitted, whener the user clicks on an field in the map.
     *
     * @param point The point where the user clicked.
     */

    void fieldClicked(QPoint point);


    /**
     * This is emitted if a gem was dragged.
     *
     * @param from The original position of the gem.
     * @param The end position of the gem.
     */

    void gemMoved(QPoint from, QPoint to);


    /**
     * This is emitted if a keeper was dragged.
     *
     * @param from The original position of the keeper.
     * @param to The original position of the keeper.
     */

    void keeperMoved(QPoint from, QPoint to);


    /**
     * This is emitted if a mouse drag started.
     */

    void mouseDragStarted();


    /**
     * This is emitted if the mouse was moved with the left mouse button hold
     * and we are not dragging a keeper or a gem.
     *
     * @param from The last position (as an index of the field).
     * @param to The new position.
     */

    void mouseDragged(QPoint from, QPoint to);


    /**
     * This is emitted if a mouse drag ended.
     */

    void mouseDragEnded();


    /**
     * This is emitted when undo was requested.
     */

    void undo();


    /**
     * This is emitted when redo was requested.
     */

    void redo();


protected:

    /**
     * Adjustes the size of the pixmaps.
     */

    virtual void resizeEvent(QResizeEvent * event);


    /**
     * Used to let the used move keeper and gems with the mouse.
     */

    virtual void contentsMousePressEvent(QMouseEvent * event);


    /**
     * Used to let the used move keeper and gems with the mouse.
     */

    virtual void contentsMouseMoveEvent(QMouseEvent * event);


    /**
     * Used to let the used move keeper and gems with the mouse.
     */

    virtual void contentsMouseReleaseEvent(QMouseEvent * event);


private:


    /**
     * Ensures, that the field at the given position is visible.
     *
     * @param position The position of the field.
     */

    void ensureFieldVisible(QPoint const & position);


    /**
     * Deletes the given vector of items.
     *
     * @param items The vector with the items to delete;
     */

    void deleteItems(std::vector<QCanvasSprite *> & items) const;


    /**
     * Fills the given vector of items.
     *
     * If the vector is not empty, it will be cleared before.
     *
     * @param items The vector with the items to create.
     * @param type The type of the image
     * @param direction The direction of the keeper.
     * @param position The position of the field in the map.
     * @param x The x-coordinate.
     * @param x The y-coordinate.
     * @param x The z-coordinate.
     */

    void createItems(std::vector<QCanvasSprite *> & items, int type,
                     QPoint const & position, int x, int y, int z) const;


    /**
     * @overload
     */

    void createItems(std::vector<QCanvasSprite *> & items, QPoint const & position, int x, int y, int z) const;


    /**
     * @overload
     */

    void createItems(std::vector<QCanvasSprite *> & items, std::vector<int> image_indices,
                     int x, int y, int z) const;


    /**
     * Places the items at the given position taking the offset into account and
     * making them visible.
     *
     * @param items The vector with the items to create.
     * @param x The x-coordinate.
     * @param x The y-coordinate.
     * @param x The z-coordinate.
     */

    void placeItems(std::vector<QCanvasSprite *> & items, int x, int y, int z);


    /**
     * Updates the square size, offsets and so on.
     */

    void updateGeometry();


    /**
     * Deletes the currently used pixmaps.
     */

    void deletePixmaps();


    /**
     * Deletes the currently used arrows.
     */

    void deleteArrows();


    /**
     * Adds a new arrow.
     *
     * @param from The source of the arrow.
     * @param to The destination of the arrow.
     */

    void addArrow(QPoint from, QPoint to);


    /**
     * Deletes the items.
     */

    void deleteItems();


    /**
     * Make new items.
     */

    void newItems();


    /**
     * Returns true, if the given mouse position is valid.
     *
     * @param position The position.
     */

    bool isValidPosition(QPoint position) const;


    /**
     * Converts mouse coordinates to field coordinates.
     *
     * @param position The mouse coordinates.
     */

    QPoint getFieldFromPosition(QPoint position) const;


    /**
     * Moves the moving item.
     *
     * @param offsets The offset to move.
     */

    void moveMovingItemPosition(QPoint const & offset);


    /**
     * Creates a QCanvasPixmapArray from a QPixmap.
     *
     * @param pixmap The pixmap.
     */

    QCanvasPixmapArray * createPixmapArray(QPixmap const & pixmap) const;


    /**
     * The current canvas.
     */

    mutable QCanvas m_canvas;


    /**
     * The current pixmap provider.
     */

    PixmapProvider * m_pixmap_provider;


    /**
     * The current theme.
     */

    Theme const * m_theme;


    /**
     * The current map.
     */

    Map const * m_map;


    /**
     * The currently used piece type.
     */

    std::vector<int> m_pieces;


    /**
     * The x-offset.
     */

    int m_x_offset;


    /**
     * The y-offset.
     */

    int m_y_offset;


    /**
     * The size of the squares.
     */

    int m_square_size;


    /**
     * The width of the map.
     */

    int m_map_width;


    /**
     * The height of the map.
     */

    int m_map_height;


    /**
     * m_map_height * m_map_width
     */

    int m_size;


    /**
     * This is true, if we force an update.
     */

    bool m_force_update;


    /**
     * If true, the click position was valid.
     */

    bool m_valid_click;


    /**
     * If true, we are in a drag operation.
     */

    bool m_in_drag;


    /**
     * If true, we are in a keeper drag operation and else in a gem drag operation.
     */

    bool m_is_keeper_drag;


    /**
     * Here we store the keeper direction.
     */

    AtomicMove::AtomicMoveType m_keeper_direction;


    /**
     * The mouse position when we clicked.
     */

    QPoint m_click_position;


    /**
     * The field, where the mouse was clicked.
     */

    QPoint m_click_field;


    /**
     * The offset between the click point and the position of the clicked item.
     */

    QPoint m_click_offset;


    /**
     * The last mouse position.
     */

    QPoint m_last_mouse_position;


    /**
     * The position of the virtual keeper.
     */

    QPoint m_virtual_keeper_pos;


    /**
     * The item for the virtual keeper.
     */

    std::vector<QCanvasSprite *> m_virtual_keeper_item;


    /**
     * The item for the moving keeper or gem.
     */

    std::vector<QCanvasSprite *> m_moving_item;


    /**
     * Here we store the arrow items.
     */

    std::vector<std::vector<QCanvasSprite *> > m_arrow_items;


    /**
     * The currently used canvas items.
     */

    std::vector<std::vector<QCanvasSprite *> > m_items;


    /**
     * The pixmaps.
     */

    mutable std::vector<QCanvasPixmapArray *> m_pixmaps;

    /**
     * The arrows.
     */

    std::vector<Move> m_arrows;


    /**
     * The timer used.
     */

    QTimer * m_timer;


    /**
     * The timer used for mouse repeat.
     */

    QTimer * m_repeat_timer;


    /**
     * The button state for the mouse repeat.
     */

    int m_mouse_repeat_state;


    /**
     * The mouse repeat delay in ms.
     */

    int m_mouse_repeat_delay;


    /**
     * The mouse repeat rate.
     */

    int m_mouse_repeat_rate;


    /**
     * The scaling mode (0 = fast, 1 = good, 2 = timed good).
     */

    int m_scaling_mode;


    /**
     * The time to switch between fast and good scaling.
     */

    int m_switch_time;


    /**
     * The minimum size of a field.
     */

    int m_min_square_size;


    /**
     * If true, we have scrollbars.
     */

    bool m_has_scrollbars;


    /**
     * If true, we hide the mouse cursor.
     */

    bool m_cursor_hidden;


    /**
     * If true, we automartically hide the cursor.
     */

    bool m_auto_hide_cursor;


    /**
     * If true, we disable auto hide forever.
     */

    bool m_disable_auto_hide;


    /**
     * The cursor hide delay.
     */

    int m_cursor_hide_delay;


    /**
     * The timer used for mouse cursor hiding.
     */

    QTimer * m_cursor_timer;


    /**
     * The current direction of the virtual keeper.
     */

    int m_virtual_keeper_direction;


    /**
     * The last drag position.
     */

    QPoint m_last_drag_position;


    /**
     * If true, we temporarily disable auto hiding the cursor.
     */

    bool m_temp_disable_auto_hide;
};


#endif
