/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "mng_animation_storer.h"


#ifdef EASYSOK_HAVE_MNG


#include <algorithm>
#include <cassert>
#include <cstring>

#include <qbuffer.h>
#include <qcolor.h>
#include <qcstring.h>
#include <qfile.h>
#include <qimage.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qdatastream.h>

#include <kglobal.h>
#include <kio/netaccess.h>
#include <kstddirs.h>
#include <ktempfile.h>

#include "theme.h"




MngAnimationStorer::MngAnimationStorer(Map const & map, Movements const & moves, Theme const & theme,
                                       int piece_size, int use_background, int start_delay, int delay,
                                       bool cycle, int end_delay, bool low_quality, KURL const & url) :
    m_map(map),
    m_moves(moves),
    m_theme(theme),
    m_pixmap_provider(theme),
    m_piece_size(piece_size),
    m_keeper_direction(0),
    m_use_background(use_background),
    m_use_background_image(false),
    m_low_quality(low_quality),
    m_delay(delay),
    m_url(url)
{
    m_temp_file = new KTempFile;
    m_temp_file->setAutoDelete(true);

    m_offset = QPoint(static_cast<int> (theme.leftBorder() * m_piece_size),
                      static_cast<int> (theme.upperBorder() * m_piece_size));
    int const right_border = static_cast<int> (theme.rightBorder() * m_piece_size);
    int const lower_border = static_cast<int> (theme.lowerBorder() * m_piece_size);
    int const width = map.width() * m_piece_size + m_offset.x() + right_border;
    int const height = map.height() * m_piece_size + m_offset.y() + lower_border;

    m_moves.setToFirstPosition();

    m_handle = mng_initialize(reinterpret_cast<mng_ptr> (this), allocMNG, freeMNG, MNG_NULL);

    mng_setcb_openstream(m_handle, openMNG);
    mng_setcb_closestream(m_handle, closeMNG);
    mng_setcb_writedata(m_handle, writeMNG);

    mng_create(m_handle);
    mng_putchunk_mhdr(m_handle, width, height, 1000, 0, 30, 30000, 0);

    if (cycle)
    {
        mng_putchunk_term(m_handle, 3, 0, end_delay, 0x7fffffff);
    }

    if (!theme.backgroundImage().isEmpty() && m_use_background)
    {
        QString const filename = KGlobal::dirs()->findResource("wallpaper", theme.backgroundImage());
        QImage background_image;

        if (background_image.load(filename))
        {
            m_use_background_image = true;
            createPngObject(1, background_image);
            m_background_width = background_image.width();
            m_background_height = background_image.height();
        }
    }

    int const nr_of_images = theme.nrOfPieceImages();

    for (int i = 0; i < nr_of_images; ++i)
    {
        QPixmap pixmap = m_pixmap_provider.createPixmap(i, piece_size);
        QImage image = pixmap.convertToImage();
        createPngObject(i + 2, image);
    }

    if (!m_use_background)
    {
        mng_putchunk_back(m_handle, 65535, 65535, 65535, 0, 0, 0);
        mng_putchunk_fram(m_handle, false, 4, 0, 0, 2, 0, 2, 0, start_delay, 0, 0,
                          0, width, 0, height, 0, 0);
    }
    else
    {
        mng_putchunk_fram(m_handle, false, 2, 0, 0, 2, 0, 0, 0, start_delay, 0, 0, 0, 0, 0, 0, 0, 0);

        if (m_use_background_image)
        {
            // The BACK header with tiled images didn't worked for me, so we do it by hand.

            int const background_max_x = (width - 1) / m_background_width;
            int const background_max_y = (height - 1) / m_background_height;

            for (int y = 0; y <= background_max_y; ++y)
            {
                for (int x = 0; x <= background_max_x; ++x)
                {
                    paintObject(1, QPoint(x * m_background_width, y * m_background_height));
                }
            }
        }
        else
        {
            mng_putchunk_back(m_handle, qRed(theme.backgroundColor().rgb()) * 256,
                              qGreen(theme.backgroundColor().rgb()) * 256,
                              qBlue(theme.backgroundColor().rgb()) * 256, 1, 0, 0);
        }
    }

    for (int y = 0; y < map.height(); ++y)
    {
        for (int x = 0; x < map.width(); ++x)
        {
            paintPiece(QPoint(x, y));
        }
    }
}


MngAnimationStorer::~MngAnimationStorer()
{
    mng_putchunk_mend(m_handle);
    mng_write(m_handle);

    mng_cleanup(&m_handle);
    m_temp_file->close();

    KIO::NetAccess::upload(m_temp_file->name(), m_url);

    delete m_temp_file;
}


MngAnimationStorer::Result MngAnimationStorer::proccess()
{
    if (!m_moves.hasNextMove())
    {
        return FINISHED;
    }

    bool is_first_move = m_moves.movePointer() == 0;

    Move const move = m_moves.nextMove();
    QPoint const diff = move.diff();

    // We have to upate the display.

    QPoint p1 = move.from();
    QPoint p2 = move.to();

    QRect bounding_rect = getBoundingRect(p1).unite(getBoundingRect(p2));

    if (move.stonePushed())
    {
        bounding_rect = bounding_rect.unite(getBoundingRect(p2 + diff));
    }

    m_keeper_direction = move.atomicMove().type();

    m_map.doMove(move, false);

    bounding_rect = bounding_rect.unite(getBoundingRect(p1)).unite(getBoundingRect(p2));

    if (move.stonePushed())
    {
        bounding_rect = bounding_rect.unite(getBoundingRect(p2 + diff));
    }

    bounding_rect = bounding_rect.normalize();

    QPoint const real_p1 = QPoint(bounding_rect.left(), bounding_rect.top());
    QPoint const real_p2 = QPoint(bounding_rect.right() + 1, bounding_rect.bottom() + 1);

    if (m_use_background)
    {
        int const background_x1 = real_p1.x() / m_background_width;
        int const background_x2 = (real_p2.x() - 1) / m_background_width;
        int const background_y1 = real_p1.y() / m_background_height;
        int const background_y2 = (real_p2.y() - 1) / m_background_height;

        if (is_first_move)
        {
            if (mng_putchunk_fram(m_handle, false, 2, 0, 0, 2, 0, 2, 0, m_delay, 0, 0,
                                  real_p1.x(), real_p2.x(), real_p1.y(), real_p2.y(), 0, 0) != MNG_NOERROR)
            {
                return ERROR;
            }
        }
        else
        {
            if (mng_putchunk_fram(m_handle, false, 2, 0, 0, 0, 0, 2, 0, 0, 0, 0,
                                  real_p1.x(), real_p2.x(), real_p1.y(), real_p2.y(), 0, 0) != MNG_NOERROR)
            {
                return ERROR;
            }
        }

        if (m_use_background_image)
        {
            for (int y = background_y1; y <= background_y2; ++y)
            {
                for (int x = background_x1; x <= background_x2; ++x)
                {
                    paintObject(1, QPoint(x * m_background_width, y * m_background_height));
                }
            }
        }
    }
    else
    {
        if (is_first_move)
        {
            if (mng_putchunk_fram(m_handle, false, 4, 0, 0, 2, 0, 2, 0, m_delay, 0, 0,
                                  real_p1.x(), real_p2.x(), real_p1.y(), real_p2.y(), 0, 0) != MNG_NOERROR)
            {
                return ERROR;
            }
        }
        else
        {
            if (mng_putchunk_fram(m_handle, false, 4, 0, 0, 0, 0, 2, 0, 0, 0, 0,
                                  real_p1.x(), real_p2.x(), real_p1.y(), real_p2.y(), 0, 0) != MNG_NOERROR)
            {
                return ERROR;
            }
        }
    }

    paintPiece(move.from());
    paintPiece(move.to());

    if (move.stonePushed())
    {
        paintPiece(move.to() + diff);
    }

    return OK;
}


QRect MngAnimationStorer::getBoundingRect(QPoint const & position) const
{
    std::vector<int> image_indices = m_theme.imageIndicesFromPosition(m_keeper_direction, position, m_map);
    int const nr_of_indices = static_cast<int> (image_indices.size());

    if (nr_of_indices > 0)
    {
        QPoint offset = m_pixmap_provider.offset(image_indices[0], m_piece_size) + m_offset;
        QPoint screen_position = QPoint(position.x() * m_piece_size, position.y() * m_piece_size);
        QRect rect(screen_position + offset, QSize(m_piece_size, m_piece_size));

        for (int i = 1; i < nr_of_indices; ++i)
        {
            offset = m_pixmap_provider.offset(image_indices[i], m_piece_size);
            QRect add_rect(screen_position + offset, QSize(m_piece_size, m_piece_size));

            rect = rect.unite(add_rect);
        }

        return rect;
    }
    else
    {
        return QRect();
    }
}


void MngAnimationStorer::createPngObject(int index, QImage const & image)
{
    mng_putchunk_defi(m_handle, index, 1, 0, false, 0, 0, false, 0, 0, 0, 0);

    QBuffer buffer;
    buffer.open(IO_WriteOnly);

    QImageIO image_io(&buffer, "PNG");
    QImage new_image = image;

    if (m_low_quality)
    {
        new_image = new_image.convertDepth(8, Qt::DiffuseAlphaDither | Qt::DiffuseDither |
                                           Qt::AutoColor | Qt::AvoidDither);
    }

    image_io.setImage(new_image);
    image_io.write();

    buffer.close();
    buffer.open(IO_ReadOnly);
    buffer.at(8);

    QDataStream stream(&buffer);

    char raw_chunk_name[5] = {0, 0, 0, 0, 0};

    while (true)
    {
        Q_UINT32 chunk_length;
        stream >> chunk_length;

        stream.readRawBytes(raw_chunk_name, 4);
        QString chunk_name(raw_chunk_name);

        if (chunk_name == "IHDR")
        {
            Q_UINT32 width;
            stream >> width;

            Q_UINT32 height;
            stream >> height;

            Q_UINT8 bit_depth;
            stream >> bit_depth;

            Q_UINT8 color_type;
            stream >> color_type;

            Q_UINT8 compression_method;
            stream >> compression_method;

            Q_UINT8 filter_method;
            stream >> filter_method;

            Q_UINT8 interlace_method;
            stream >> interlace_method;

            mng_putchunk_ihdr(m_handle, width, height, bit_depth, color_type, compression_method,
                              filter_method, interlace_method);
        }
        else if (chunk_name == "IDAT")
        {
            QByteArray data(chunk_length);
            stream.readRawBytes(data.data(), chunk_length);

            mng_putchunk_idat(m_handle, chunk_length, reinterpret_cast<mng_ptr> (data.data()));
        }
        else if (chunk_name == "IEND")
        {
            mng_putchunk_iend(m_handle);

            break;
        }
        else if (chunk_name == "PLTE")
        {
            QByteArray data(chunk_length);
            stream.readRawBytes(data.data(), chunk_length);

            mng_putchunk_plte(m_handle, chunk_length / 3, reinterpret_cast<mng_palette8e *> (data.data()));
        }
        else if (chunk_name == "tRNS")
        {
            QByteArray data(256);
            stream.readRawBytes(data.data(), chunk_length);

            mng_putchunk_trns(m_handle, false, false, 3, chunk_length,
                              reinterpret_cast<mng_uint8 *> (data.data()), 0, 0, 0, 0,
                              0, reinterpret_cast<mng_uint8 *> (data.data()));
        }
        else
        {
            for (unsigned int i = 0; i < chunk_length; ++i)
            {
                stream.readRawBytes(raw_chunk_name, 1);
            }
        }

        stream.readRawBytes(raw_chunk_name, 4);
    }
}


void MngAnimationStorer::paintObject(int index, QPoint const & position)
{
    mng_putchunk_move(m_handle, index, index, 0, position.x(), position.y());
    mng_putchunk_show(m_handle, false, index, index, 0);
}


void MngAnimationStorer::paintPiece(QPoint const & position)
{
    std::vector<int> image_indices = m_theme.imageIndicesFromPosition(m_keeper_direction, position, m_map);
    int const nr_of_indices = static_cast<int> (image_indices.size());

    for (int i = 0; i < nr_of_indices; ++i)
    {
        QPoint const offset = m_pixmap_provider.offset(image_indices[i], m_piece_size);

        paintObject(image_indices[i] + 2, QPoint(position.x() * m_piece_size + offset.x() + m_offset.x(),
                                                 position.y() * m_piece_size + offset.y() + m_offset.y()));
    }
}


mng_ptr MngAnimationStorer::allocMNG(mng_size_t length)
{
    char * buffer = new char[length];

    std::memset(buffer, 0, length);

    return reinterpret_cast<mng_ptr> (buffer);
}


void MngAnimationStorer::freeMNG(mng_ptr pointer, mng_size_t)
{
    delete[] reinterpret_cast<char *> (pointer);
}


mng_bool MngAnimationStorer::openMNG(mng_handle)
{
    return MNG_TRUE;
}


mng_bool MngAnimationStorer::closeMNG(mng_handle)
{
    return MNG_TRUE;
}


mng_bool MngAnimationStorer::writeMNG(mng_handle handle, mng_ptr buffer, mng_uint32 size, mng_uint32 * written)
{
    MngAnimationStorer * storer;
    storer = reinterpret_cast<MngAnimationStorer *> (mng_get_userdata(handle));

    *written = storer->m_temp_file->file()->writeBlock(reinterpret_cast<char *> (buffer), size);

    return *written == size;
}


#endif
