/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_MNG_ANIMATION_STORER_INC_GUARD_H
#define EASYSOK_MNG_ANIMATION_STORER_INC_GUARD_H


#include "../config.h"


#ifdef EASYSOK_HAVE_MNG


#define MNG_BUILD_SO // Needed for writing capabilities
#include <libmng.h>

#include <qpoint.h>
#include <qrect.h>

#include <kurl.h>

#include "map.h"
#include "movements.h"
#include "pixmap_provider.h"


class KTempFile;
class QImage;
class QPoint;
class Theme;




/**
 * This class lets you store a game as MNG-animation.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class MngAnimationStorer
{

public:

    /**
     * Releases the resources.
     */

    virtual ~MngAnimationStorer();


    /**
     * Creates the object.
     *
     * @param map The map to use.
     * @param move The <e>atomic</e> moves to display.
     * @param theme The theme to use.
     * @param piece_size The size (width and height) of a single piece.
     * @param use_background If true, we use the background image of the theme if there is one.
     * @param start_delay The start delay (in miliseconds) for the first frame.
     * @param delay The delay (in miliseconds) between two frames.
     * @param cycle If true, we go to the first frame if at end.
     * @param end_delay The time (in miliseconds) between the last frame and the first frame
     *                  in cyling.
     * @param low_quality If true, we only use low quality images.
     * @param url The url to save the animation.
     */

    MngAnimationStorer(Map const & map, Movements const & moves, Theme const & theme, int piece_size,
                       int use_background, int start_delay, int delay, bool cycle, int end_delay,
                       bool low_quality, KURL const & url);



    /**
     * Defines the result of proccess().
     */

    enum Result
    {
        /**
         * Everythings fine, but you have to call proccess() at least one time more.
         */

        OK,


        /**
         * We are ready.
         */

        FINISHED,


        /**
         * An error occured.
         */

        ERROR
    };


    /**
     * Call this as long as OK is returned.
     */

    virtual Result proccess();


private:

    /**
     * Returns the bound rectangle of the piece at the given position.
     *
     * @param position The position of the piece.
     */

    QRect getBoundingRect(QPoint const & position) const;


    /**
     * Creates a object in png format for the given image.
     *
     * @param index The index of the object to create.
     * @param image The image to convert to png.
     */

    void createPngObject(int index, QImage const & image);


    /**
     * Puts an object at the given position.
     *
     * @param index The index of the object.
     * @param position The position of the object.
     */

    void paintObject(int index, QPoint const & position);


    /**
     * Paints an item of a map.
     *
     * @param direction The direction of the keeper.
     * @param position The position of the piece.
     */

    void paintPiece(QPoint const & position);


    /**
     * The callback function for memory allocation.
     *
     * @param length The lenght of the allocated buffer.
     * @return A pointer to the allocated buffer.
     */

    static mng_ptr allocMNG(mng_size_t length);


    /**
     * The callback function for freeing memory.
     *
     * @param pointer The pointer to the buffer to free.
     * @param length The length of the buffer.
     */

    static void freeMNG(mng_ptr pointer, mng_size_t length);


    /**
     * The callback functions for opening a stream (this is a dummy function).
     *
     * @param handle The handle to use.
     */

    static mng_bool openMNG(mng_handle handle);


    /**
     * The callback function for closing a stream (this is a dummy function).
     *
     * @param handle The handle to use.
     */

    static mng_bool closeMNG(mng_handle handle);


    /**
     * The callback function for writing to a stream.
     *
     * @param handle The handle to use.
     * @param buffer The buffer to write.
     * @param size The size of the buffer to write.
     * @param written Here we write the number of bytes written.
     */

    static mng_bool writeMNG(mng_handle handle, mng_ptr buffer, mng_uint32 size, mng_uint32 * written);


    /**
     * The current map.
     */

    Map m_map;


    /**
     * The moves.
     */

    Movements m_moves;


    /**
     * The theme used.
     */

    Theme const & m_theme;


    /**
     * The pixmap provider.
     */

    PixmapProvider m_pixmap_provider;


    /**
     * The piece size.
     */

    int m_piece_size;


    /**
     * The offset.
     */

    QPoint m_offset;


    /**
     * The direction of the keeper.
     */

    int m_keeper_direction;


    /**
     * The libmng handle.
     */

    mng_handle m_handle;


    /**
     * The width of the background image.
     */

    int m_background_width;


    /**
     * The height of the background image.
     */

    int m_background_height;


    /**
     * If true, we use the theme background.
     */

    bool m_use_background;


    /**
     * If true, we use a background image.
     */

    bool m_use_background_image;


    /**
     * If true, we use low quality images.
     */

    bool m_low_quality;


    /**
     * The delay to use.
     */

    int m_delay;


    /**
     * The url to save.
     */

    KURL m_url;


    /**
     * The temp file to use.
     */

    KTempFile * m_temp_file;
};


#else


struct MngAnimationStorer {
};


#endif


#endif
