/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "move.h"


#include <algorithm>
#include <cassert>

#include <qdatastream.h>




Move::Move() :
    m_from(QPoint(0, 0)),
    m_to(QPoint(0, 0))
{
}


Move::Move(QPoint from, AtomicMove atomic_move, bool stone_pushed) :
    m_from(from),
    m_to(from + atomic_move.diff()),
    m_stone_pushed(stone_pushed)
{
}


Move::Move(QPoint from, QPoint to, bool stone_pushed) :
    m_from(from),
    m_to(to),
    m_stone_pushed(stone_pushed)
{
    assert(!m_stone_pushed || (m_from.x() == m_to.x()) || (m_from.y() == m_to.y()));
}


Move::Move(QDataStream & stream)
{
    Q_UINT32 code;
    stream >> code;

    assert((code >> 29) == 0);

    int const from_x = (code & (127 << 1)) >> 1;
    int const from_y = (code & (127 << 8)) >> 8;
    int const to_x = (code & (127 << 15)) >> 15;
    int const to_y = (code & (127 << 22)) >> 22;

    assert(from_x < 128);
    assert(from_y < 128);
    assert(to_x < 128);
    assert(to_y < 128);

    m_from = QPoint(from_x, from_y);
    m_to = QPoint(to_x, to_y);
    m_stone_pushed = code & 1;
}


QPoint Move::from() const
{
    return m_from;
}


QPoint Move::to() const
{
    return m_to;
}


bool Move::stonePushed() const
{
    return m_stone_pushed;
}


bool Move::isAtomicMove() const
{
    return (m_from - m_to).manhattanLength() == 1;
}


AtomicMove Move::atomicMove() const
{
    assert(isAtomicMove());

    AtomicMove::AtomicMoveType type =  AtomicMove::LEFT;

    QPoint const act_diff = diff();

    if (act_diff.x() > 0)
    {
        type =  AtomicMove::RIGHT;
    }
    else if (act_diff.y() < 0)
    {
        type =  AtomicMove::UP;
    }
    else if (act_diff.y() > 0)
    {
        type =  AtomicMove::DOWN;
    }

    return AtomicMove(type);
}


bool Move::operator == (Move const & other_move) const
{
    return ( (m_from == other_move.m_from) &&
             (m_to == other_move.m_to) &&
             (m_stone_pushed == other_move.m_stone_pushed));
}


bool Move::operator != (Move const & other_move) const
{
    return !(*this == other_move);
}


QPoint Move::diff() const
{
    return m_to - m_from;
}


QPoint Move::diffSign() const
{
    return QPoint(sign(m_to.x() - m_from.x()), sign(m_to.y() - m_from.y()));

}

void Move::reverse()
{
    std::swap(m_from, m_to);
}


QString Move::toText() const
{
    QString result = '(' + QString::number(m_from.x()) + ", " + QString::number(m_from.y());

    if (m_stone_pushed)
    {
        result += ") ->* (";
    }
    else
    {
        result += ") -> (";
    }

    result += QString::number(m_to.x()) + ", " + QString::number(m_to.y()) + ')';

    return result;
}


void Move::writeToStream(QDataStream & stream) const
{
    assert(m_from.x() >= 0);
    assert(m_from.y() >= 0);
    assert(m_to.x() >= 0);
    assert(m_to.y() >= 0);
    assert(m_from.x() < 128);
    assert(m_from.y() < 128);
    assert(m_to.x() < 128);
    assert(m_to.y() < 128);

    Q_UINT32 code = m_stone_pushed + (m_from.x() << 1) + (m_from.y() << 8) +
                    (m_to.x() << 15) + (m_to.y() << 22);
    stream << code;
}


int Move::sign(int number) const
{
    if (number > 0)
    {
        return 1;
    }
    else if (number < 0)
    {
        return -1;
    }
    else
    {
        return 0;
    }
}
