/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_MOVE_INC_GUARD_H
#define EASYSOK_MOVE_INC_GUARD_H


#include <qpoint.h>
#include <qstring.h>

#include "atomic_move.h"


class QDataStream;




/**
 * This class represents a single move of the keeper.
 *
 * A single move may consist of more than one atomic move (an
 * atomic move is a move one field left, up, right or down).
 * If you want to get the atomic moves for a move, look at the documentation
 * of Map (because they may depend on the layout of the map).
 *
 * For a non pushing move (a move were the keeper does not pushes a stone)
 * there are no restrictions, while for a pushing move
 * we are only allowed to move in one direction.
 * For example:
 * @code
 * Move move1(QPoint(1, 0), QPoint(7, 12), false); // This is OK.
 * Move move2(QPoint(1, 0), QPoint(3, 0), true);   // This is also OK.
 * Move move3(QPoint(1, 0), QPoint(1, 7), true);   // And so is this.
 * Move move4(QPoint(1, 0), QPoint(7, 12), true);  // But this is an error!
 * @endcode
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */


class Move
{

public:

    /**
     * Constructs an empty move (a move from (0, 0) to (0, 0)).
     */

    Move();


    /**
     * Constructs the move.
     *
     * @param from The start field.
     * @param atomic_move The AtomicMove made from the start field.
     * @param stone_pushed If true, we pushed a stone during the move.
     */

    Move(QPoint from, AtomicMove atomic_move, bool stone_pushed);


    /**
     * Constructs the move.
     *
     * @param from The start field of the move.
     * @param to The end field of the move.
     * @param stone_pushed If true, we pushed a stone during the move.
     */

    Move(QPoint from, QPoint to, bool stone_pushed);


    /**
     * Constructs the move by reading it from a data stream.
     *
     * @param stream The stream to use.
     */

    Move(QDataStream & stream);


    /**
     * Returns true, if the move is empty (meaning no movement of the keeper is involved).
     */

    bool isEmpty() const;


    /**
     * Returns the from point of the move.
     */

    QPoint from() const;


    /**
     * Returns the to point of the move.
     */

    QPoint to() const;


    /**
     * Returns true, if the move was a stone move.
     */

    bool stonePushed() const;


    /**
     * Returns true, if the move is an atomic move (the keeper moves only one field).
     */

    bool isAtomicMove() const;


    /**
     * Returns a code for the move.
     *
     * Note that the move <e>must</e> be an atomic move!
     */

    AtomicMove atomicMove() const;


    /**
     * Returns true, if two moves are equal.
     */

    bool operator == (Move const & other_move) const;


    /**
     * Returns true, if two moves are unequal.
     */

    bool operator != (Move const & other_move) const;


    /**
     * Returns to() - from().
     */

    QPoint diff() const;


    /**
     * Returns the sign (-1, 0 or 1) of every element in diff().
     */

    QPoint diffSign() const;


    /**
     * Swaps from and to.
     */

    void reverse();


    /**
     * Returns a human readable representation of the move,
     */

    QString toText() const;


    /**
     * Writes the move to a QDataStream.
     *
     * Note that the x and y componets of the from and to position of the move must be in the
     * range [0:128[. This should be no problem since the map is restricted to this size also.
     *
     * @param stream The stream to use.
     */

    void writeToStream(QDataStream & stream) const;


private:

    /**
     * Is there really no sign function? Hard to belive.
     */

    int sign(int number) const;


    /**
     * The from point.
     */

    QPoint m_from;


    /**
     * The to point.
     */

    QPoint m_to;


    /**
     * Are we pushing a stone?
     */

    bool m_stone_pushed;
};


#endif
