/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "movements.h"

#include <algorithm>
#include <cassert>

#include <qdatastream.h>
#include <qstringlist.h>
#include <qregexp.h>

#include "compressed_movements.h"




QRegExp * Movements::s_moves_regexp = 0;



Movements::Movements() :
    m_pos(0)
{
}


Movements::Movements(CompressedMovements const & compressed_moves) :
    m_pos(0)
{
    compressed_moves.setMoves(m_moves);
}


Movements::Movements(std::vector<Move> const & moves) :
    m_moves(moves),
    m_pos(0)
{
}


Movements::Movements(QDataStream & stream) :
    m_pos(0)
{
    Q_UINT32 number_of_moves;
    stream >> number_of_moves;

    m_moves.reserve(number_of_moves);

    for (Q_UINT32 i = 0; i < number_of_moves; ++i)
    {
        m_moves.push_back(Move(stream));
    }
}


Movements::Movements(QPoint keeper_start, QStringList & lines) :
    m_pos(0)
{
    QPoint from = keeper_start;
    QPoint diff;

    while (!lines.isEmpty() && (lines.first() != "+-+-"))
    {
        lines.remove(lines.begin());
    }

    if (!lines.isEmpty())
    {
        lines.remove(lines.begin());
    }

    if (s_moves_regexp == 0)
    {
        s_moves_regexp = new QRegExp("^[uUdDrRlL]+$");
    }

    while (!lines.isEmpty() && (s_moves_regexp->search(lines.first(), 0) == 0))
    {
        QString const act_line = lines.first();
        lines.remove(lines.begin());

        int const moves = act_line.length();

        for (int i = 0; i < moves; ++i)
        {
            bool is_push = false;

            switch (act_line[i].latin1())
            {
            case 'U':
                is_push = true;
            case 'u':
                diff = QPoint(0, -1);
                break;

            case 'D':
                is_push = true;
            case 'd':
                diff = QPoint(0, 1);
                break;

            case 'L':
                is_push = true;
            case 'l':
                diff = QPoint(-1, 0);
                break;

            case 'R':
                is_push = true;
            case 'r':
                diff = QPoint(1, 0);
                break;

            default:
                m_moves = std::vector<Move>();
                return;
            }

            m_moves.push_back(Move(from, from + diff, is_push));
            from += diff;
        }
    }
}


bool Movements::isEmpty() const
{
    return m_moves.empty();
}


int Movements::numberOfMoves() const
{
    return static_cast<int> (m_moves.size());
}


void Movements::setToFirstPosition()
{
    m_pos = 0;
}


void Movements::setToLastPosition()
{
    m_pos = m_moves.size();
}


bool Movements::hasNextMove() const
{
    return m_pos < m_moves.size();
}


Move Movements::peekNextMove() const
{
    assert(hasNextMove());

    return m_moves[m_pos];
}


Move Movements::nextMove()
{
    assert(hasNextMove());

    return m_moves[m_pos++];
}


bool Movements::hasPrevMove() const
{
    return m_pos > 0;
}


Move Movements::peekPrevMove() const
{
    assert(hasPrevMove());

    return m_moves[m_pos - 1];
}


Move Movements::prevMove()
{
    assert(hasPrevMove());

    return m_moves[--m_pos];
}


Move const & Movements::move(int index) const
{
    assert(index >= 0);
    assert(index < numberOfMoves());

    return m_moves[index];
}


void Movements::addMove(Move const & move)
{
    if (!hasNextMove() || (m_moves[m_pos] != move))
    {
        truncateToCurrent();
        m_moves.push_back(move);
    }

    ++m_pos;
}


void Movements::addMovements(Movements const & moves)
{
    truncateToCurrent();

    int const nr_to_add = moves.numberOfMoves();
    m_moves.reserve(m_pos + nr_to_add);

    for (int i = 0; i < nr_to_add; ++i)
    {
        m_moves.push_back(moves.m_moves[i]);
    }
}


void Movements::truncateToCurrent()
{
    m_moves.resize(m_pos);
}


int Movements::movePointer() const
{
    return static_cast<int> (m_pos);
}


void Movements::setMovePointer(int move_pointer)
{
    assert(move_pointer >= 0);
    assert(move_pointer <= static_cast<int> (m_moves.size()));

    m_pos = move_pointer;
}


void Movements::reverse()
{
    std::reverse(m_moves.begin(), m_moves.end());
}


Movements Movements::gemMovesToKeeperMoves(QPoint const & keeper, bool retro_mode) const
{
    QPoint act_keeper = keeper;
    Movements result;

    int const moves = numberOfMoves();

    if (retro_mode)
    {
        for (int i = 0; i < moves; ++i)
        {
            Move const move = m_moves[i];
            QPoint const diff = move.diff();
            QPoint const from_keeper = move.to();

            if (act_keeper != from_keeper)
            {
                result.addMove(Move(act_keeper, from_keeper, false));
            }

            act_keeper = move.to() + diff;
            result.addMove(Move(from_keeper, act_keeper, true));
        }
    }
    else
    {
        for (int i = 0; i < moves; ++i)
        {
            Move const move = m_moves[i];
            QPoint const diff = move.diff();
            QPoint const from_keeper = move.from() - diff;

            if (act_keeper != from_keeper)
            {
                result.addMove(Move(act_keeper, from_keeper, false));
            }

            result.addMove(Move(from_keeper, move.from(), true));
            act_keeper = move.from();
        }
    }

    return result;
}


QString Movements::toText() const
{
    static int const max_chars_per_line = 80;
    static char buffer[max_chars_per_line + 2];
    static char const right_chars[] = {'r', 'R'};
    static char const left_chars[] = {'l', 'L'};
    static char const down_chars[] = {'d', 'D'};
    static char const up_chars[] = {'u', 'U'};

    Movements moves = *this;
    moves.setToFirstPosition();

    QString result = "+-+-\n";
    int pos = 0;

    while (moves.hasNextMove())
    {
        Move const move = moves.nextMove();
        assert(move.isAtomicMove());

        int const stone_pushes = move.stonePushed();

        if (move.from().x() < move.to().x())
        {
            buffer[pos] = right_chars[stone_pushes];
        }
        else if (move.from().x() > move.to().x())
        {
            buffer[pos] = left_chars[stone_pushes];
        }
        else if (move.from().y() < move.to().y())
        {
            buffer[pos] = down_chars[stone_pushes];
        }
        else
        {
            buffer[pos] = up_chars[stone_pushes];
        }

        ++pos;

        if ((pos == max_chars_per_line) || !moves.hasNextMove())
        {
            buffer[pos] = '\n';
            buffer[pos + 1] = 0;
            result += buffer;
            pos = 0;
        }
    }

    return result;
}


void Movements::writeToStream(QDataStream & stream) const
{
    Q_UINT32 number_of_moves = static_cast<Q_UINT32> (m_moves.size());
    stream << number_of_moves;

    for (Q_UINT32 i = 0; i < number_of_moves; ++i)
    {
        m_moves[i].writeToStream(stream);
    }
}


int Movements::moves() const
{
    return static_cast<int> (m_moves.size());
}


int Movements::pushes() const
{
    int const nr_of_moves = moves();
    int pushes = 0;

    for (int i = 0; i < nr_of_moves; ++i)
    {
        Move const & move = m_moves[i];

        if (move.stonePushed())
        {
            pushes += move.diff().manhattanLength();
        }
    }

    return pushes;
}


int Movements::linearPushes() const
{
    int const nr_of_moves = moves();
    int pushes = 0;
    bool last_was_push = false;
    QPoint last_diff(0, 0);

    for (int i = 0; i < nr_of_moves; ++i)
    {
        Move const & move = m_moves[i];

        if (move.stonePushed())
        {
            QPoint diff_sign = move.diffSign();

            if (!last_was_push || (diff_sign != last_diff))
            {
                ++pushes;
                last_diff = diff_sign;
            }

            last_was_push = true;
        }
        else
        {
            last_was_push = false;
        }
    }

    return pushes;
}


int Movements::gemChanges() const
{
    int const nr_of_moves = moves();
    int gem_changes = 0;
    QPoint last_gem = QPoint(0, 0);

    for (int i = 0; i < nr_of_moves; ++i)
    {
        Move const & move = m_moves[i];

        if (move.stonePushed())
        {
            QPoint const gem_from = move.from() + move.diffSign();
            QPoint const gem_to = gem_from + move.diff();

            if (gem_from != last_gem)
            {
                ++gem_changes;
            }

            last_gem = gem_to;
        }
    }

    return gem_changes;
}


bool Movements::operator == (Movements const & other_moves) const
{
    return m_moves == other_moves.m_moves;
}


bool Movements::operator != (Movements const & other_moves) const
{
    return m_moves != other_moves.m_moves;
}
