/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_MOVEMENTS_INC_GUARD_H
#define EASYSOK_MOVEMENTS_INC_GUARD_H


#include <vector>

#include <qpoint.h>
#include <qregexp.h>

#include "move.h"


class CompressedMovements;
class QDataStream;
class QStringList;




/**
 * This class represents a sequence of moves.
 *
 * You can use this class in the spirit of a bidirectional iterator or a vector.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 * @see move
 */

class Movements
{

public:

    /**
     * Creates an empty Movements object.
     */

    Movements();


    /**
     * Constructs the objects from compressed movements.
     *
     * @param compressed_moves The compressed moves.
     */

    Movements(CompressedMovements const & compressed_moves);


    /**
     * Constructs the object with the given moves.
     *
     * The move pointer points to the first move.
     *
     * @param moves A vector containing the moves.
     */

    Movements(std::vector<Move> const & moves);


    /**
     * Constructs the object from a data stream.
     *
     * The move pointer points to the first move.
     *
     * @param stream The stream to use.
     */

    Movements(QDataStream & stream);


    /**
     * Constructs the movements from lines in sokogen format.
     *
     * Removes all lines from the QStringList, which were taken for building up the moves.
     *
     * @param keeper_start The start position of the keeper.
     * @param lines The lines holding the moves.
     */

    Movements(QPoint keeper_start, QStringList & lines);


    /**
     * Returns true, if there are no moves.
     */

    bool isEmpty() const;


    /**
     * The number of moves.
     */

    int numberOfMoves() const;


    /**
     * Sets the move pointer to the first move.
     *
     * This means, that hasPrevMove() will return false after this call.
     */

    void setToFirstPosition();


    /**
     * Sets the move pointer to the last move.
     *
     * This means, that hasNextMove() will return false after this call.
     */

    void setToLastPosition();


    /**
     * Returns true, if there is a next move.
     */

    bool hasNextMove() const;


    /**
     * Returns the next move, but unlike nextMove() does not changes the
     * move pointer.
     */

    Move peekNextMove() const;


    /**
     * Returns the next move and increments the move pointer.
     *
     * This allows you traverse the movements forwards.
     */

    Move nextMove();


    /**
     * Returns true, if there is a previous move.
     */

    bool hasPrevMove() const;


    /**
     * Returns the previous move, but unlike prevMove() does not changes the
     * move pointer.
     */

    Move peekPrevMove() const;


    /**
     * Returns the previous move and decrements the move pointer.
     *
     * This allows you traverse the movements backwards.
     */

    Move prevMove();


    /**
     * Returns the move with the given index.
     *
     * @param index The index of the move.
     */

    Move const & move(int index) const;


    /**
     * Adds a move.
     *
     * If the next move and the move to add are identical, nothing happens, apart from
     * the fact, the we jump to the next move.
     * If not, the added move is also the last move.
     *
     * @param move The move to add.
     */

    void addMove(Move const & move);


    /**
     * Adds moves.
     *
     * The last move of the added move is also the last move.
     *
     * @param move The moves to add.
     */

    void addMovements(Movements const & moves);


    /**
     * Truncates the movements up to the current move.
     *
     * This means, hasNextMove() will return false.
     */

    void truncateToCurrent();


    /**
     * Returns the move pointer as an index.
     */

    int movePointer() const;


    /**
     * Sets the mover pointer.
     *
     * The pointer must be in the range [0:numberOfMoves()[.
     *
     * @param move_pointer The new move pointer.
     */

    void setMovePointer(int move_pointer);


    /**
     * Reverses the moves.
     *
     * The last move is the first move and so on.
     *
     * Note that the moves itself are the same (so from() of the first move will be from() of
     * the last move).
     */

    void reverse();


    /**
     * Converts a sequence of gems moves to a sequence of keeper moves.
     *
     * This assumes, the current moves are gem moves (holding the positions of the
     * moves gem instead of the keeper).
     *
     * Note that the moves must be atomic!
     *
     * @param keeper The start position of the keeper.
     * @param retro_mode If true, we are in retro mode.
     */

    Movements gemMovesToKeeperMoves(QPoint const & keeper, bool retro_mode = false) const;


    /**
     * Returns a sokogen compatible form of the moves.
     *
     * Note that all moves must be atomic.
     */

    QString toText() const;


    /**
     * Writes the movements to a stream.
     *
     * @param stream The stream to use.
     */

    void writeToStream(QDataStream & stream) const;


    /**
     * Returns the number of moves.
     */

    int moves() const;


    /**
     * Returns the number of pushes.
     */

    int pushes() const;


    /**
     * Returns the number of linear pushes.
     */

    int linearPushes() const;


    /**
     * Returns the number of gem changes (the number of times a different gem has been pushed).
     */

    int gemChanges() const;


    /**
     * Return false, if two movements are different
     */

    bool operator == (Movements const & other_moves) const;


    /**
     * Return true, if two movements are different
     */

    bool operator != (Movements const & other_moves) const;


private:

    /**
     * The movements.
     */

    std::vector<Move> m_moves;


    /**
     * The pointer to the next move.
     */

    size_t m_pos;


    /**
     * The regexp for a valid sokogen moves lines.
     */

    static QRegExp * s_moves_regexp;
};


#endif
