/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "piece_image.h"

#include <cassert>

#include <qdom.h>

#include "dom_helper.h"
#include "image_effect.h"
#include "piece_image_layer.h"




PieceImage::PieceImage(QDomElement const & dom_element)
{
    assert(dom_element.tagName() == "PieceImage");

    m_size = DomHelper::getInteger(dom_element, "size", 1);
    m_x_offset = DomHelper::getDouble(dom_element, "xoffset", 0);
    m_y_offset = DomHelper::getDouble(dom_element, "yoffset", 0);

    QDomNodeList childs = dom_element.childNodes();
    int const nr_of_childs = static_cast<int> (childs.count());

    if (nr_of_childs >= 1)
    {
        int offset = 0;
        QDomNode node = childs.item(0);

        if (node.isElement())
        {
            QDomElement element = node.toElement();

            if (element.tagName() == "Effect")
            {
                m_effect = PieceImageEffect(element);
                ++offset;
            }

            for (int i = offset; i < nr_of_childs; ++i)
            {
                node = childs.item(i);

                if (node.isElement())
                {
                    element = node.toElement();

                    if (element.tagName() == "Layer")
                    {
                        m_layers.push_back(PieceImageLayer(element));
                    }
                    else
                    {
                        assert(false);
                    }
                }
                else
                {
                    assert(false);
                }
            }
        }
    }
}


QImage PieceImage::createImage() const
{
    QImage image;
    int const nr_of_layers = static_cast<int> (m_layers.size());

    image = QImage(m_size, m_size, 32);
    image.fill(qRgba(0, 0, 0, 0));
    image.setAlphaBuffer(true);

    for (int i = 0; i < nr_of_layers; ++i)
    {
        QImage add_image = m_layers[i].createImage(m_size);
        ImageEffect::blendOnLower(0, 0, add_image, image);
    }

    return m_effect.apply(image);
}


double PieceImage::xOffset() const
{
    return m_x_offset;
}


double PieceImage::yOffset() const
{
    return m_y_offset;
}


int PieceImage::size() const
{
    return m_size;
}


bool PieceImage::operator == (PieceImage const & other_piece_image) const
{
    return ( (m_effect == other_piece_image.m_effect) && (m_layers == other_piece_image.m_layers) &&
             (m_size == other_piece_image.m_size) && (m_x_offset == other_piece_image.m_x_offset) &&
             (m_y_offset == other_piece_image.m_y_offset) );
}
