/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "piece_image_effect.h"

#include <cassert>

#include <qdom.h>

#include <kimageeffect.h>

#include "dom_helper.h"
#include "image_effect.h"




PieceImageEffect::PieceImageEffect()
{
}


PieceImageEffect::PieceImageEffect(QDomElement const & dom_element)
{
    assert(dom_element.tagName() == "Effect");

    QDomNodeList childs = dom_element.childNodes();

    int const nr_of_effects = childs.count();

    for (int i = 0; i < nr_of_effects; ++i)
    {
        QDomNode node = childs.item(i);

        if (node.isElement())
        {
            QDomElement element = node.toElement();
            QString tag_name = element.tagName();

            if (tag_name == "Move")
            {
                m_effects.push_back(MOVE);
                m_parameters.push_back(DomHelper::getInteger(element, "xoffset", 0));
                m_parameters.push_back(DomHelper::getInteger(element, "yoffset", 0));
            }
            else if (tag_name == "RotateLeft")
            {
                m_effects.push_back(ROTATE_LEFT);
            }
            else if (tag_name == "RotateRight")
            {
                m_effects.push_back(ROTATE_RIGHT);
            }
            else if (tag_name == "UTurn")
            {
                m_effects.push_back(UTURN);
            }
            else if (tag_name == "HMirror")
            {
                m_effects.push_back(HMIRROR);
            }
            else if (tag_name == "VMirror")
            {
                m_effects.push_back(VMIRROR);
            }
            else if (tag_name == "BlendRgb")
            {
                m_effects.push_back(BLEND_RGB);
                putFactor(element, "factor", 256);
                putColor(element);
            }
            else if (tag_name == "BlendRgba")
            {
                m_effects.push_back(BLEND_RGBA);
                putFactor(element, "factor", 256);
                putColor(element);
            }
            else if (tag_name == "BlendAlpha")
            {
                m_effects.push_back(BLEND_ALPHA);
                putFactor(element, "factor", 256);
                putColor(element);
            }
            else if (tag_name == "Colorize")
            {
                m_effects.push_back(COLORIZE);
                putFactor(element, "factor", 256);
                putColor(element);
            }
            else if (tag_name == "Scale")
            {
                m_effects.push_back(SCALE);
                m_parameters.push_back(DomHelper::getInteger(element, "width", 1));
                m_parameters.push_back(DomHelper::getInteger(element, "height", 1));
            }
            else if (tag_name == "Crop")
            {
                m_effects.push_back(CROP);
                m_parameters.push_back(DomHelper::getInteger(element, "xoffset", 0));
                m_parameters.push_back(DomHelper::getInteger(element, "yoffset", 0));
                m_parameters.push_back(DomHelper::getInteger(element, "width", 1));
                m_parameters.push_back(DomHelper::getInteger(element, "height", 1));
            }
            else
            {
                assert(false);
            }
        }
    }
}


QImage PieceImageEffect::apply(QImage const & image) const
{
    QImage result = image;

    int const nr_of_effects = static_cast<int> (m_effects.size());
    int param_index = 0;

    for (int i = 0; i < nr_of_effects; ++i)
    {
        switch(m_effects[i])
        {
        case MOVE:
            move(result, m_parameters[param_index], m_parameters[param_index + 1]);
            param_index += 2;
            break;

        case ROTATE_LEFT:
            rotate270(result);
            break;

        case ROTATE_RIGHT:
            rotate90(result);
            break;

        case UTURN:
            rotate180(result);
            break;

        case HMIRROR:
            horizontalMirror(result);
            break;

        case VMIRROR:
            verticalMirror(result);
            break;

        case BLEND_RGB:
            blendRgb(result, m_parameters[param_index], m_parameters[param_index + 1],
                     m_parameters[param_index + 2], m_parameters[param_index + 3]);
            param_index +=5;
            break;

        case BLEND_RGBA:
            blendRgba(result, m_parameters[param_index], m_parameters[param_index + 1],
                      m_parameters[param_index + 2], m_parameters[param_index + 3],
                      m_parameters[param_index + 4]);
            param_index +=5;
            break;

        case BLEND_ALPHA:
            blendAlpha(result, m_parameters[param_index], m_parameters[param_index + 4]);
            param_index +=5;
            break;

        case COLORIZE:
            colorize(result, m_parameters[param_index], m_parameters[param_index + 1],
                     m_parameters[param_index + 2], m_parameters[param_index + 3]);
            param_index +=5;
            break;

        case SCALE:
            scale(result, m_parameters[param_index], m_parameters[param_index + 1]);
            param_index += 2;
            break;

        case CROP:
            crop(result, m_parameters[param_index], m_parameters[param_index + 1],
                 m_parameters[param_index + 2], m_parameters[param_index + 3]);
            param_index +=4;
            break;

        default:
            assert(false);
            break;
        }
    }

    return result;
}


bool PieceImageEffect::operator == (PieceImageEffect const & other_effect) const
{
    return ( (m_effects == other_effect.m_effects) && (m_parameters == other_effect.m_parameters) );
}


void PieceImageEffect::putFactor(QDomElement const & element, QString const & name, int scale)
{
    m_parameters.push_back(static_cast<int> (DomHelper::getDouble(element, name, 0.0) * scale));
}


void PieceImageEffect::putColor(QDomElement const & element)
{
    m_parameters.push_back(DomHelper::getInteger(element, "red", 0));
    m_parameters.push_back(DomHelper::getInteger(element, "green", 0));
    m_parameters.push_back(DomHelper::getInteger(element, "blue", 0));
    m_parameters.push_back(DomHelper::getInteger(element, "alpha", 255));
}


void PieceImageEffect::rotate90(QImage & image) const
{
    assert(image.width() > 0);
    assert(image.height() > 0);

    int const width = image.width();
    int const width_half = (width + 1) / 2;
    int const height = image.height();
    int const height_half = (height + 1) / 2;

    QImage result(height, width, 32);
    result.setAlphaBuffer(true);

    for (int x = 0; x < width_half; ++x)
    {
        int const other_x = width - x - 1;

        for (int y = 0; y < height_half; ++y)
        {
            int const other_y = height - y - 1;

            QRgb const p1 = image.pixel(x, y);
            QRgb const p2 = image.pixel(other_x, y);
            QRgb const p3 = image.pixel(other_x, other_y);
            QRgb const p4 = image.pixel(x, other_y);

            result.setPixel(other_y, x, p1);
            result.setPixel(other_y, other_x, p2);
            result.setPixel(y, other_x, p3);
            result.setPixel(y, x, p4);
        }
    }

    image = result;
}


void PieceImageEffect::rotate180(QImage & image) const
{
    image = image.mirror(true,true);
}


void PieceImageEffect::rotate270(QImage & image) const
{
    assert(image.width() > 0);
    assert(image.height() > 0);

    int const width = image.width();
    int const width_half = (width + 1) / 2;
    int const height = image.height();
    int const height_half = (height + 1) / 2;

    QImage result(height, width, 32);
    result.setAlphaBuffer(true);

    for (int x = 0; x < width_half; ++x)
    {
        int const other_x = width - x - 1;

        for (int y = 0; y < height_half; ++y)
        {
            int const other_y = height - y - 1;

            QRgb const p1 = image.pixel(x, y);
            QRgb const p2 = image.pixel(other_x, y);
            QRgb const p3 = image.pixel(other_x, other_y);
            QRgb const p4 = image.pixel(x, other_y);

            result.setPixel(y, other_x, p1);
            result.setPixel(y, x, p2);
            result.setPixel(other_y, x, p3);
            result.setPixel(other_y, other_x, p4);
        }
    }

    image = result;
}


void PieceImageEffect::horizontalMirror(QImage & image) const
{
    image = image.mirror(true, false);
}


void PieceImageEffect::verticalMirror(QImage & image) const
{
    image = image.mirror(false, true);
}


void PieceImageEffect::blendRgb(QImage & image, int factor, int red, int green, int blue) const
{
    ImageEffect::blend(image, factor, qRgb(red, green, blue), true, false);
}


void PieceImageEffect::blendRgba(QImage & image, int factor, int red, int green, int blue, int alpha) const
{
    ImageEffect::blend(image, factor, qRgba(red, green, blue, alpha), true, true);
}


void PieceImageEffect::blendAlpha(QImage & image, int factor, int alpha) const
{
    ImageEffect::blend(image, factor, qRgba(0, 0, 0, alpha), false, true);
}


void PieceImageEffect::colorize(QImage & image, int factor, int red, int green, int blue) const
{
    ImageEffect::colorize(image, factor, qRgb(red, green, blue));
}


void PieceImageEffect::crop(QImage & image, int x_offset, int y_offset, int width, int height) const
{
    assert(image.depth() == 32);

    int const old_width = image.width();
    int const old_height = image.height();

    QImage new_image(old_width, old_height, 32);
    new_image.fill(qRgba(0, 0, 0, 0));
    new_image.setAlphaBuffer(true);

    image.setAlphaBuffer(false);
    image = image.copy(x_offset, y_offset, width, height);

    for (int y = 0; y < height; ++y)
    {
        QRgb * src_scanline = reinterpret_cast<QRgb *> (image.scanLine(y));
        QRgb * dst_scanline = reinterpret_cast<QRgb *> (new_image.scanLine(y));

        for (int x = 0; x < width; ++x)
        {
            dst_scanline[x] = src_scanline[x];
        }
    }

    image = new_image;
}


void PieceImageEffect::move(QImage & image, int x_offset, int y_offset) const
{
    crop(image, -x_offset, -y_offset, image.width(), image.height());
}


void PieceImageEffect::scale(QImage & image, int width, int height) const
{
    ImageEffect::scale(image, width, height);
}
