/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_PIECE_IMAGE_EFFECT_INC_GUARD_H
#define EASYSOK_PIECE_IMAGE_EFFECT_INC_GUARD_H


#include <vector>

#include <qimage.h>


class QDomElement;




/**
 * This class performs some effects on images used for PieceImage objects.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class PieceImageEffect
{

public:

    /**
     * Creates an empty object with no effects.
     */

    PieceImageEffect();


    /**
     * Creates the effect from an dom element.
     *
     * @param dom_element The dom element to use.
     */

    PieceImageEffect(QDomElement const & dom_element);


    /**
     * Returns a new version of a given image, with all effects specified applied.
     *
     * @param image The image to use.
     */

    QImage apply(QImage const & image) const;


    /**
     * Returns true if the two piece image effects are equal.
     *
     * @param other_effect The other effect.
     */

    bool operator == (PieceImageEffect const & other_effect) const;


private:

    /**
     * The type of effects.
     */

    enum EffectType
    {
        /**
         * Move effect.
         */

        MOVE,


        /**
         * Rotate left effect.
         */

        ROTATE_LEFT,


        /**
         * Rotate right effect.
         */

        ROTATE_RIGHT,


        /**
         * Uturn effect (rotate 180 degrees).
         */

        UTURN,


        /**
         * Horizontal mirror effect.
         */

        HMIRROR,


        /**
         * Vertical mirror effect.
         */

        VMIRROR,


        /**
         * Blend rgb effect.
         */

        BLEND_RGB,


        /**
         * Blend rgba effect.
         */

        BLEND_RGBA,


        /**
         * Blend alpha effect.
         */

        BLEND_ALPHA,


        /**
         * Colorize effect.
         */

        COLORIZE,


        /**
         * Scale effect.
         */

        SCALE,


        /**
         * Crop effect.
         */

        CROP
    };


    /**
     * Puts a to an integer converted double factor to the parameter list.
     *
     * @param dom_element The dom_element to use.
     * @param name The name of the attribute to use.
     * @param scale The scale factor multiplied with the value before converted to an integer.
     */

    void putFactor(QDomElement const & element, QString const & name, int scale);


    /**
     * Puts the color (rgba values) to the parameter list.
     *
     * @param dom_element The dom_element to use.
     */

    void putColor(QDomElement const & element);


    /**
     * Does rotate 90.
     */

    void rotate90(QImage & image) const;


    /**
     * Does rotate 180.
     */

    void rotate180(QImage & image) const;


    /**
     * Does rotate 270.
     */

    void rotate270(QImage & image) const;


    /**
     * Does horizontal mirroring.
     */

    void horizontalMirror(QImage & image) const;


    /**
     * Does vertical mirroring.
     */

    void verticalMirror(QImage & image) const;


    /**
     * Blends the image to a color.
     *
     * @param image The image to use.
     * @param factor The ammount of blending (in the range [0:256]).
     * @param red The red component of the blend color.
     * @param green The green component of the blend color.
     * @param blue The blue component of the blend color.
     */

    void blendRgb(QImage & image, int factor, int red, int green, int blue) const;


    /**
     * Blends the image to a color.
     *
     * @param image The image to use.
     * @param factor The ammount of blending (in the range [0:256]).
     * @param red The red component of the blend color.
     * @param green The green component of the blend color.
     * @param blue The blue component of the blend color.
     * @param alpha The alpha component of the blend color.
     */

    void blendRgba(QImage & image, int factor, int red, int green, int blue, int alpha) const;


    /**
     * Blends the alpha channel of the image to a specific value.
     *
     * @param image The image to use.
     * @param factor The ammount of blending (in the range [0:256]).
     * @param alpha The alpha component to blend to.
     */

    void blendAlpha(QImage & image, int factor, int alpha) const;


    /**
     * Colorizes the image to a color.
     *
     * @param image The image to use.
     * @param factor The ammount of colorizing (in the range [0:256]).
     * @param red The red component of the colorize color.
     * @param green The green component of the colorize color.
     * @param blue The blue component of the colorize color.
     */

    void colorize(QImage & image, int factor, int red, int green, int blue) const;


    /**
     * Crops the image.
     *
     * @param image The image to use.
     * @param x_offset The x-offset.
     * @param y_offset The y-offset.
     * @param width The new width.
     * @param height The new height.
     */

    void crop(QImage & image, int x_offset, int y_offset, int width, int height) const;


    /**
     * Moves the image.
     *
     * @param image The image to use.
     * @param x_offset The x-offset.
     * @param y_offset The y-offset.
     */

    void move(QImage & image, int x_offset, int y_offset) const;


    /**
     * Scales the image.
     *
     * @param image The image to use.
     * @param width The new width.
     * @param height The new height.
     */

    void scale(QImage & image, int width, int height) const;


    /**
     * The effects.
     */

    std::vector<int> m_effects;


    /**
     * The parameters of the effects.
     */

    std::vector<int> m_parameters;
};


#endif
