/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "pixmap_provider.h"

#include <cassert>

#include <qwmatrix.h>

#include <kglobal.h>
#include <kstddirs.h>

#include "image_effect.h"
#include "piece_image.h"
#include "theme.h"




PixmapProvider::PixmapProvider(Theme const & theme) :
    m_theme(theme),
    m_background_color(theme.backgroundColor()),
    m_smooth_scaling(true)
{
    if (!theme.backgroundImage().isEmpty())
    {
        QString const filename = KGlobal::dirs()->findResource("wallpaper", theme.backgroundImage());
        m_background_image.load(filename);
    }

    m_pixmaps.resize(theme.nrOfPieceImages(), 0);
}


PixmapProvider::~PixmapProvider()
{
    int const nr_of_pixmaps = static_cast<int> (m_pixmaps.size());

    for (int i = 0; i < nr_of_pixmaps; ++i)
    {
        delete m_pixmaps[i];
    }
}


bool PixmapProvider::hasBackgroundImage() const
{
    return !m_background_image.isNull();
}


QPixmap const & PixmapProvider::backgroundImage() const
{
    return m_background_image;
}


QColor PixmapProvider::backgroundColor() const
{
    return m_background_color;
}


QPixmap PixmapProvider::createPixmap(int index, int size) const
{
    assert(index >= 0);
    assert(index < static_cast<int> (m_pixmaps.size()));

    if (m_pixmaps[index] == 0)
    {
        m_pixmaps[index] = new QPixmap(convertToPixmap(m_theme.pieceImage(index)));
    }

    return scale(*m_pixmaps[index], size);
}


QPoint PixmapProvider::offset(int index, int size) const
{
    assert(index >= 0);
    assert(index < static_cast<int> (m_pixmaps.size()));

    PieceImage const & piece_image = m_theme.pieceImage(index);

    return QPoint(static_cast<int> (size * piece_image.xOffset()),
                  static_cast<int> (size * piece_image.yOffset()));
}


bool PixmapProvider::smoothScaling() const
{
    return m_smooth_scaling;
}


void PixmapProvider::setSmoothScaling(bool smooth_scaling)
{
    m_smooth_scaling = smooth_scaling;
}


QPixmap PixmapProvider::convertToPixmap(PieceImage const & image) const
{
    QPixmap pixmap;
    pixmap.convertFromImage(image.createImage());

    return pixmap;
}


QPixmap PixmapProvider::scale(QPixmap const & pixmap, int size) const
{
    if (pixmap.width() == size)
    {
        return pixmap;
    }

    QPixmap scaled_pixmap(size, size);

    if (m_smooth_scaling)
    {
        QImage image = pixmap.convertToImage();
        ImageEffect::scale(image, size, size);
        scaled_pixmap.convertFromImage(image);
    }
    else
    {
        double const sx = static_cast<double> (size) / pixmap.width();
        double const sy = static_cast<double> (size) / pixmap.height();

        QWMatrix scale_matrix;
        scale_matrix.scale(sx, sy);

        scaled_pixmap = pixmap.xForm(scale_matrix);
    }

    return scaled_pixmap;
}
