/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_PIXMAP_PROVIDER_INC_GUARD_H
#define EASYSOK_PIXMAP_PROVIDER_INC_GUARD_H


#include <vector>

#include <qcolor.h>
#include <qpixmap.h>
#include <qpoint.h>

#include "theme.h"


class PieceImage;




/**
 * This class can provide pixmaps from a theme.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class PixmapProvider
{

public:

    /**
     * Constructs the object.
     *
     * @param theme The theme to use. Note that this theme object has to live as long as the
     *              pixmap provider itself lives.
     */

    PixmapProvider(Theme const & theme);


    /**
     * Releases the created pixmaps.
     */

    ~PixmapProvider();


    /**
     * Returns true, if we have a background image.
     */

    bool hasBackgroundImage() const;


    /**
     * Returns the background image.
     */

    QPixmap const & backgroundImage() const;


    /**
     * Returns the background color.
     */

    QColor backgroundColor() const;


    /**
     * Returns the given piece image in the desired size.
     *
     * @param index The index of the piece image.
     * @param size The desired size of the image.
     */

    QPixmap createPixmap(int index, int size) const;


    /**
     * Returns the offset of the given piece image in the desired size.
     *
     * @param index The index of the piece image.
     * @param size The desired size of the image.
     */

    QPoint offset(int index, int size) const;


    /**
     * Returns true, if we use smooth scaling.
     */

    bool smoothScaling() const;


    /**
     * @param smooth_scaling If true, we use smooth scaling.
     */

    void setSmoothScaling(bool smooth_scaling);


private:

    /**
     * Converts an immage to a pixmap.
     *
     * @param image The image to convert.
     */

    QPixmap convertToPixmap(PieceImage const & image) const;


    /**
     * Scales the pixmap.
     *
     * @param pixmap The pixmap.
     * @param size The needed size.
     */

    QPixmap scale(QPixmap const & pixmap, int size) const;


    /**
     * The theme to use.
     */

    Theme const & m_theme;


    /**
     * The background image to use.
     */

    QPixmap m_background_image;


    /**
     * The background color to use.
     */

    QColor m_background_color;


    /**
     * The vector of the pixmaps;
     */

    mutable std::vector<QPixmap *> m_pixmaps;


    /**
     * If true, we use smooth scaling.
     */

    bool m_smooth_scaling;
};


#endif
