/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "push_optimizer.h"

#include <cassert>
#include <map>
#include <utility>
#include <vector>

#include "hash.h"
#include "map.h"
#include "move_optimizer.h"




PushOptimizer::PushOptimizer() :
    m_number_of_moves(0),
    m_number_of_pushes(0)
{
}


PushOptimizer::PushOptimizer(Map const & map, Movements const & moves)
{
    optimize(map, moves);
}


void PushOptimizer::optimize(Map const & map, Movements const & moves)
{
    m_number_of_moves = 0;
    m_number_of_pushes = 0;

    if (moves.isEmpty())
    {
        return;
    }

    Map act_map = map;

    // First optimize the moves.
    m_moves = act_map.expandMoves(moves, false);
    assert(!m_moves.isEmpty());
    m_moves = act_map.collapseMoves(m_moves);
    m_moves = act_map.expandMoves(m_moves, false);
    m_moves.setToFirstPosition();

    std::map<Hash, int> hash_mapping;

    std::vector<int> doubles;
    int number_of_doubles = 0;

    std::vector<Move> only_pushes;

    while (m_moves.hasNextMove())
    {
        Move move = m_moves.nextMove();

        if (move.stonePushed())
        {
            only_pushes.push_back(move);

            Hash from_hash(act_map);

            act_map.doMove(move, false);

            Hash to_hash(act_map);

            std::map<Hash, int>::const_iterator const it = hash_mapping.find(to_hash);

            if (it != hash_mapping.end())
            {
                doubles.push_back(it->second);
                ++number_of_doubles;
            }
            else
            {
                doubles.push_back(-1);
            }

            hash_mapping.insert(std::make_pair(from_hash, static_cast<int> (only_pushes.size() - 1)));
        }
        else
        {
            act_map.doMove(move, false);
        }
    }

    if (number_of_doubles > 0)
    {
        Movements new_moves = Movements();
        m_number_of_pushes = static_cast<int> (only_pushes.size());

        QPoint last_from = map.keeper();

        for (int i = 0; i < m_number_of_pushes; ++i)
        {
            Move push_move = only_pushes[i];
            int copy_index = -1;

            for (int j = m_number_of_pushes - 1; j > i; --j)
            {
                if (doubles[j] == i)
                {
                    copy_index = j;

                    break;
                }
            }

            if (copy_index != -1)
            {
                i = copy_index;
                push_move = only_pushes[i];

                if (push_move.to() != last_from)
                {
                    new_moves.addMove(Move(last_from, push_move.to(), false));
                    last_from = push_move.to();
                }
            }
            else
            {
                if (push_move.from() != last_from)
                {
                    new_moves.addMove(Move(last_from, push_move.from(), false));
                }

                new_moves.addMove(push_move);
                last_from = push_move.to();
            }
        }

        m_moves = new_moves;
    }

    MoveOptimizer move_optimizer(map, m_moves);
    m_moves = move_optimizer.moves();
    m_number_of_pushes = move_optimizer.numberOfPushes();
    m_number_of_moves = move_optimizer.numberOfMoves();
}


Movements const & PushOptimizer::moves() const
{
    return m_moves;
}


int PushOptimizer::numberOfMoves() const
{
    return m_number_of_moves;
}


int PushOptimizer::numberOfPushes() const
{
    return m_number_of_pushes;
}
