/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "reorder_list_view.h"

#include <cassert>

#include <qstring.h>

#include <klocale.h>
#include <kmessagebox.h>

#include "collection.h"
#include "collection_holder.h"
#include "level.h"




ReorderListView::ReorderListView(QWidget * parent, char const * name) :
    KListView(parent, name)
{
    addColumn(i18n("Collection"));
    setRootIsDecorated(true);
    setSorting(-1);
    setDragEnabled(true);
    setAcceptDrops(true);
    setMultiSelection(false);

    int const nr_of_collections = CollectionHolder::numberOfCollections();

    for (int i = nr_of_collections - 1; i >= 0; --i)
    {
        Collection * collection = CollectionHolder::collection(i);
        int const nr_of_levels = collection->numberOfLevels();
        QString name = collection->name();

        if (CollectionHolder::isTemporary(i))
        {
            name += " " + i18n("(temporary)");
        }

        QListViewItem * collection_item = new QListViewItem(this, name, QString::number(i));

        for (int j = nr_of_levels - 1; j >= 0; --j)
        {
            QString name = collection->level(j).name();

            if (name.isEmpty())
            {
                name = i18n("Level %1 of %2").arg(j + 1).arg(collection->name());
            }

            new QListViewItem(collection_item, name, QString::number(i), QString::number(j));
        }
    }

    connect(this, SIGNAL(moved()), this, SLOT(afterMove()));
}


bool ReorderListView::updateCollectionHolder()
{
    std::vector<Collection *> collections;
    std::vector<int> is_temporary;

    QListViewItem * collection_item = firstChild();

    bool moved_level_to_temporary = false;

    while (collection_item != 0)
    {
        if (!collection_item->text(2).isEmpty())
        {
            KMessageBox::error(this, i18n("You have moved a level out of a collection.\n"
                                          "This is not allowed!"));

            return false;
        }

        int const collection_index = collection_item->text(1).toInt();
        bool const temporary = CollectionHolder::isTemporary(collection_index);
        Collection * old_collection = CollectionHolder::collection(collection_index);
        Collection * collection = new Collection(old_collection->name(), old_collection->authors(),
                                                 old_collection->emails(), old_collection->homepage(),
                                                 old_collection->copyright(), old_collection->info(),
                                                 old_collection->difficulty());

        QListViewItem * level_item = collection_item->firstChild();

        while (level_item != 0)
        {
            if (level_item->text(2).isEmpty())
            {
                KMessageBox::error(this, i18n("You have moved a collection into another collection.\n"
                                              "This is not allowed!"));

                return false;
            }

            int const level_collection_index = level_item->text(1).toInt();
            Collection * level_collection = CollectionHolder::collection(level_collection_index);

            int const level_index = level_item->text(2).toInt();
            collection->addLevel(level_collection->level(level_index));

            if (temporary && !CollectionHolder::isTemporary(level_collection_index))
            {
                moved_level_to_temporary = true;
            }

            level_item = level_item->nextSibling();
        }

        collections.push_back(collection);
        is_temporary.push_back(temporary);

        collection_item = collection_item->nextSibling();
    }

    if (moved_level_to_temporary)
    {
        if (KMessageBox::warningContinueCancel(this,
                                               i18n("You have moved at least one level from a non\n"
                                                    "temporary collection to a temporary. This level(s)\n"
                                                    "will therefore be deleted if you exit EasySok.\n"
                                                    "Do you really want to continue?"),
                                               i18n("Warning"), i18n("Continue"),
                                               "Level moved to temp") == KMessageBox::Cancel)
        {
            return false;
        }
    }

    int const nr_of_collections = static_cast<int> (collections.size());
    assert(nr_of_collections == CollectionHolder::numberOfCollections());

    for (int i = 0; i < nr_of_collections; ++i)
    {
        CollectionHolder::removeCollection(0);
    }

    for (int i = 0; i < nr_of_collections; ++i)
    {
        CollectionHolder::addCollection(collections[i], is_temporary[i]);
    }

    return true;
}


bool ReorderListView::acceptDrag(QDropEvent * event) const
{
    if (event->source() != viewport())
    {
        return false;
    }

    QListViewItem * current_item = currentItem();

    if (current_item == 0)
    {
        return false;
    }

    bool const is_collection = (current_item->text(2).isEmpty());

    QListViewItem * parent;
    QListViewItem * after_me;

    const_cast<ReorderListView *> (this)->findDrop(event->pos(), parent, after_me);

    if (is_collection)
    {
        return parent == 0;
    }
    else
    {
        return parent != 0;
    }
}


void ReorderListView::afterMove()
{
    clearSelection();
}
