/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "send_solutions_dialog.h"

#include <algorithm>
#include <cassert>

#include <qstringlist.h>
#include <qtimer.h>

#include <klocale.h>

#include "level.h"
#include "movements.h"
#include "solution_holder.h"




SendSolutionsDialog::SendSolutionsDialog(KURL const & server, QString const & proxy, int proxy_port,
                                         QString const & user, QString const & passwd,
                                         std::vector<Level const *> const & levels, QWidget * parent) :
    QMessageBox(i18n("Send Solutions"), i18n("Getting timeout from the server"), Information,
                Cancel | Default, NoButton, NoButton, parent, 0, true),
    m_server(server),
    m_proxy(proxy),
    m_proxy_port(proxy_port),
    m_user(user),
    m_passwd(passwd),
    m_levels(levels),
    m_result(ServerConnector::OK),
    m_timeout(-1),
    m_server_connector(0),
    m_best_pushes(0),
    m_best_linear_pushes(0),
    m_best_gem_changes(0),
    m_best_moves(0),
    m_is_best_pushes(false),
    m_is_best_linear_pushes(false),
    m_is_best_gem_changes(false),
    m_is_best_moves(false),
    m_query_index(0)
{
    assert(!m_levels.empty());

    m_build_timer = new QTimer(this);
    connect(m_build_timer, SIGNAL(timeout()), this, SLOT(buildQuery()));

    m_build_timer->start(100, true);
}


bool SendSolutionsDialog::finished() const
{
    return (m_query_index == static_cast<int> (m_queries.count()));
}


ServerConnector::Result SendSolutionsDialog::result() const
{
    return m_result;
}


int SendSolutionsDialog::bestPushes() const
{
    return m_best_pushes;
}


int SendSolutionsDialog::bestLinearPushes() const
{
    return m_best_linear_pushes;
}


int SendSolutionsDialog::bestGemChanges() const
{
    return m_best_gem_changes;
}


int SendSolutionsDialog::bestMoves() const
{
    return m_best_moves;
}


void SendSolutionsDialog::buildQuery()
{
    if (m_build_timer == 0)
    {
        return;
    }

    if (m_query_index == static_cast<int> (m_levels.size()))
    {
        delete m_build_timer;
        m_build_timer = 0;

        m_query_index = 0;

        m_query_timer = new QTimer(this);
        connect(m_query_timer, SIGNAL(timeout()), this, SLOT(query()));

        if (m_new_level.empty())
        {
            accept();

            return;
        }

        m_query_timer->start(0, true);

        return;
    }

    std::vector<Movements> best_solutions;

    int const level_index = SolutionHolder::getIndexForMap(m_levels[m_query_index]->compressedMap());
    int nr_of_solutions;

    if ((level_index != -1) && ((nr_of_solutions = SolutionHolder::numberOfSolutions(level_index)) > 0))
    {
        int best_push_index = SolutionHolder::pushBestSolution(level_index);
        Movements best_push_moves = SolutionHolder::movements(level_index, best_push_index);
        best_solutions.push_back(best_push_moves);

        int best_linear_push_index = SolutionHolder::linearPushBestSolution(level_index);
        Movements best_linear_push_moves = SolutionHolder::movements(level_index, best_linear_push_index);

        if (std::find(best_solutions.begin(), best_solutions.end(), best_linear_push_moves) ==
            best_solutions.end())
        {
            best_solutions.push_back(best_linear_push_moves);
        }

        int best_gem_changes_index = SolutionHolder::gemChangeBestSolution(level_index);

        Movements best_gem_changes_moves = SolutionHolder::movements(level_index, best_gem_changes_index);

        if (std::find(best_solutions.begin(), best_solutions.end(), best_gem_changes_moves) ==
            best_solutions.end())
        {
            best_solutions.push_back(best_gem_changes_moves);
        }

        int best_move_index = SolutionHolder::moveBestSolution(level_index);

        Movements best_move_moves = SolutionHolder::movements(level_index, best_move_index);

        if (std::find(best_solutions.begin(), best_solutions.end(), best_move_moves) ==
            best_solutions.end())
        {
            best_solutions.push_back(best_move_moves);
        }

        int const nr_of_best_solutions = static_cast<int> (best_solutions.size());

        Map const & map = m_levels[m_query_index]->map();

        for (int i = 0; i < nr_of_best_solutions; ++i)
        {
            Movements const & moves = best_solutions[i];
            QString moves_text = map.expandMoves(moves, false).toText();
            moves_text = moves_text.right(moves_text.length() - 4);
            moves_text = moves_text.replace(QRegExp("\n"), "");

            QString query = "nickname=" + KURL::encode_string(m_user);
            query += "&passwd=" + KURL::encode_string(m_passwd);
            query += "&map=" + KURL::encode_string(map.toServerFormat());
            query += "&moves=" + KURL::encode_string(moves_text);
            query += "&nrpushes=" + KURL::encode_string(QString::number(moves.pushes()));
            query += "&nrlinearpushes=" + KURL::encode_string(QString::number(moves.linearPushes()));
            query += "&nrgemchanges=" + KURL::encode_string(QString::number(moves.gemChanges()));
            query += "&nrmoves=" + KURL::encode_string(QString::number(moves.moves()));

            m_queries.append(query);

            m_new_level.push_back(i == 0);
        }
    }

    ++m_query_index;

    m_build_timer->start(m_query_index % 10 ? 0 : 10, true);
}


void SendSolutionsDialog::query()
{
    if (m_server_connector != 0)
    {
        if (m_timeout == -1)
        {
            if (m_server_connector->result() != ServerConnector::OK)
            {
                m_result = m_server_connector->result();
                accept();

                return;
            }
            else
            {
                QStringList data = QStringList::split('\n', m_server_connector->data(), true);

                if (data.count() < 1)
                {
                    m_result = ServerConnector::ABORTED;
                    accept();

                    return;
                }
                else
                {
                    m_timeout = std::max(0, data[0].toInt());
                }
            }
        }
        else
        {
            if (m_new_level.front())
            {
                m_is_best_pushes = false;
                m_is_best_linear_pushes = false;
                m_is_best_gem_changes = false;
                m_is_best_moves = false;
            }

            ServerConnector::Result result = m_server_connector->result();

            if (result == ServerConnector::OK)
            {
                QStringList data = QStringList::split('\n', m_server_connector->data(), true);

                if (data.count() < 4)
                {
                    m_result = ServerConnector::ABORTED;
                    accept();

                    return;
                }

                m_is_best_pushes = m_is_best_pushes || (data[0] == "yes");
                m_is_best_linear_pushes = m_is_best_linear_pushes || (data[1] == "yes");
                m_is_best_gem_changes = m_is_best_gem_changes || (data[2] == "yes");
                m_is_best_moves = m_is_best_moves || (data[3] == "yes");
            }
            else if (result != ServerConnector::MAP_DOES_NOT_EXISTS)
            {
                m_result = result;
                accept();

                return;
            }

            ++m_query_index;

            if ((m_query_index == static_cast<int> (m_queries.count())) ||
                (m_new_level[m_query_index]))
            {
                if (m_is_best_pushes)
                {
                    ++m_best_pushes;
                }

                if (m_is_best_linear_pushes)
                {
                    ++m_best_linear_pushes;
                }

                if (m_is_best_gem_changes)
                {
                    ++m_best_gem_changes;
                }

                if (m_is_best_moves)
                {
                    ++m_best_moves;
                }
            }

            if (m_query_index == static_cast<int> (m_queries.count()))
            {
                accept();

                return;
            }
        }

        delete m_server_connector;
        m_server_connector = 0;
        m_query_timer->start(m_timeout * 1000, true);
    }
    else
    {
        if (m_timeout == -1)
        {
            m_server_connector = new ServerConnector(m_server, m_proxy, m_proxy_port,
                                                     "get_timeout.php", "", this);
        }
        else
        {
            m_server_connector = new ServerConnector(m_server, m_proxy, m_proxy_port,
                                                     "add_score.php", m_queries[m_query_index]);

            setText(i18n("Sending solutions to server (%1 percent finished).").
                    arg(static_cast<int> (m_query_index * 100.0 / (m_queries.count() + 1))));
            adjustSize();
        }

        connect(m_server_connector, SIGNAL(finished()), this, SLOT(query()));
    }
}
