/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef EASYSOK_SEND_SOLUTIONS_DIALOG_INC_GUARD_H
#define EASYSOK_SEND_SOLUTIONS_DIALOG_INC_GUARD_H


#include <vector>

#include <qmessagebox.h>
#include <qstring.h>
#include <qstringlist.h>

#include <kurl.h>

#include "server_connector.h"


class Level;
class QTimer;




/**
 * The dialog shown when we send the best solutions of the given levels to the server.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class SendSolutionsDialog : public QMessageBox
{
    Q_OBJECT

public:

    /**
    * Creates the dialog.
    *
    * @param server The server to use.
    * @param proxy The proxy to use.
    * @param proxy_port The proxy port.
    * @param user The user to use.
    * @param passwd The password to use.
    * @param level The levels for which to send the best solutions.
    * @param parent The parent of the dialog.
    */

    SendSolutionsDialog(KURL const & server, QString const & proxy, int proxy_port, QString const & user,
                        QString const & passwd, std::vector<Level const *> const & levels,
                        QWidget * parent = 0);


    /**
     * Returns true, if all levels have been send.
     */

    bool finished() const;


    /**
     * Returns the return code of the last query.
     */

    ServerConnector::Result result() const;


    /**
     * Returns the number of solutions send with best pushes.
     */

    int bestPushes() const;


    /**
     * Returns the number of solutions send with best linear pushes.
     */

    int bestLinearPushes() const;


    /**
     * Returns the number of solutions send with best gem changes.
     */

    int bestGemChanges() const;

    /**
     * Returns the number of solutions send with best moves.
     */

    int bestMoves() const;


public slots:

    /**
     * Builds the queries.
     */

    void buildQuery();


    /**
     * Performs the next query.
     */

    void query();


private:

    /**
     * The server.
     */

    KURL m_server;


    /**
     * The proxy.
     */

    QString m_proxy;


    /**
     * The proxy port.
     */

    int m_proxy_port;


    /**
     * The user.
     */

    QString m_user;


    /**
     * The password
     */

    QString m_passwd;


    /**
     * The levels.
     */

    std::vector<Level const *> m_levels;


    /**
     * The queries.
     */

    QStringList m_queries;


    /**
     * The last query return code.
     */

    ServerConnector::Result m_result;


    /**
     * Is that the query of a different level than that before?.
     */

    std::vector<int> m_new_level;


    /**
     * The timeout.
     */

    int m_timeout;


    /**
     * The build timer to use.
     */

    QTimer * m_build_timer;


    /**
     * The query timer to use.
     */

    QTimer * m_query_timer;


    /**
     * The server connector.
     */

    ServerConnector * m_server_connector;


    /**
     * The number of best push solutions.
     */

    int m_best_pushes;


    /**
     * The number of best linear pushes.
     */

    int m_best_linear_pushes;


    /**
     * The number of best gem changes.
     */

    int m_best_gem_changes;


    /**
     * The number of best moves;
     */

    int m_best_moves;


    /**
     * If true, a new best push solution for the current level has been found.
     */

    bool m_is_best_pushes;


    /**
     * If true, a new best linear push solution for the current level has been found.
     */

    bool m_is_best_linear_pushes;


    /**
     * If true, a new best gem change solution for the current level has been found.
     */

    bool m_is_best_gem_changes;


    /**
     * If true, a new best move solution for the current level has been found.
     */

    bool m_is_best_moves;


    /**
     * The query index.
     */

    int m_query_index;
};


#endif
