/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "server_connector.h"

#include <qsocket.h>
#include <qstringlist.h>
#include <qtextstream.h>
#include <qtimer.h>

#include <klocale.h>
#include <kurl.h>




ServerConnector::ServerConnector(KURL const & server, QString const & proxy, int proxy_port,
                                 QString const & script, QString const & query, QObject * parent) :
    QObject(parent),
    m_script(script),
    m_query(query),
    m_result(ABORTED),
    m_path(server.path(1)),
    m_host(server.host()),
    m_port(server.port() ? server.port() : 80)
{
    if (!proxy.isEmpty())
    {
        m_host = proxy;
        m_port = proxy_port;
        m_path = server.url(1);
    }

    m_socket = new QSocket(this);

    connect(m_socket, SIGNAL(readyRead()), this, SLOT(read()));
    connect(m_socket, SIGNAL(connectionClosed()), this, SLOT(closed()));
    connect(m_socket, SIGNAL(connected()), this, SLOT(connected()));


    QTimer * timer = new QTimer(this);
    connect(timer, SIGNAL(timeout()), this, SLOT(connectToHost()));

    timer->start(100, true);
}


ServerConnector::Result ServerConnector::result() const
{
    return m_result;

}


QString ServerConnector::resultText() const
{
    return resultText(m_result);
}


QString ServerConnector::resultText(ServerConnector::Result result)
{
    switch (result)
    {
    case OK:
        return i18n("Everything went fine!");

    case ABORTED:
        return i18n("The operation was aborted!");

    case COULD_NOT_CONNECT_SERVER:
        return i18n("Could not connect to server!");

    case COULD_NOT_CONNECT_DATABASE:
        return i18n("Could not connect to database!");

    case INVALID_NICKNAME:
        return i18n("The nickname is invalid!");

    case USER_NOT_FOUND:
        return i18n("The user with the given nickname was not found!");

    case INVALID_PASSWORD:
        return i18n("The password is wrong!");

    case COULD_NOT_UPDATE_DATABASE:
        return i18n("The database could not be updated!");

    case USER_ALREADY_EXISTS:
        return i18n("The user already exists!");

    case MAP_DOES_NOT_EXISTS:
        return i18n("The map of the current level is not stored on the server!");

    case SOLUTION_IS_INVALID:
        return i18n("The solution is invalid!");

    case LEVEL_NOT_SOLVED:
        return i18n("You have not solved the level!");

    case TOO_MANY_USERS:
        return i18n("The maximum number of allowed users has been reached on the server!");
        break;

    case ILLEGAL_QUERY:
        return i18n("The query was illegal!");

    default:
        return i18n("An unknown error occured!");
    }
}


QString ServerConnector::data() const
{
    return m_data;
}


int ServerConnector::version() const
{
    return m_version;
}


void ServerConnector::read()
{
    while (m_socket->canReadLine())
    {
        m_server_data += m_socket->readLine();
    }
}


void ServerConnector::closed()
{
    read();

    m_server_data += m_socket->readLine();
    QStringList parts = QStringList::split("SokobanHighscoreServer\n", m_server_data, true);

    if (parts.count() < 2)
    {
        m_result = COULD_NOT_CONNECT_SERVER;
    }
    else
    {
        parts.remove(parts.begin());
        parts = QStringList::split('\n', parts.join("SokobanHighscoreServer\n"), true);

        bool ok;
        m_version = parts.first().toInt(&ok);

        if (!ok)
        {
            m_result = USER_NOT_FOUND;
        }

        parts.remove(parts.begin());

        m_result = static_cast<Result> (parts.first().toInt(&ok));

        if (!ok)
        {
            m_result = COULD_NOT_CONNECT_SERVER;
        }

        parts.remove(parts.begin());

        m_data = parts.join("\n");
    }

    emit finished();
}


void ServerConnector::error()
{
    m_result = COULD_NOT_CONNECT_SERVER;

    finished();
}


void ServerConnector::connected()
{
    QTextStream stream(m_socket);

    QString header = "POST " + m_path + m_script + " HTTP/1.0\n";
    header += "Host: " + m_host + '\n';
    header += "Content-type: application/x-www-form-urlencoded\n";
    header += "Content-length: " + QString::number(m_query.length()) + "\n\n";

    stream << header << m_query << '\n';
}


void ServerConnector::connectToHost()
{
    m_socket->connectToHost(m_host, m_port);
}
