/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_SERVER_CONNECTOR_INC_GUARD_H
#define EASYSOK_SERVER_CONNECTOR_INC_GUARD_H


#include <qobject.h>
#include <qstring.h>


class KURL;
class QSocket;




/**
 * Sends queries to the highscore server and returns the result.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class ServerConnector : public QObject
{
    Q_OBJECT

public:

    /**
     * Creates the connector
     *
     * @param server The url of the server.
     * @param proxy The host of the proxy to use (empty if we don't use one).
     * @param proxy_port The port of the proxy.
     * @param script The script to use.
     * @param query The query to use.
     * @param parent The parent of the connector.
     */

    ServerConnector(KURL const & server, QString const & proxy, int proxy_port, QString const & script,
                    QString const & query, QObject * parent = 0);


    /**
     * Defines the result of the query operation.
     */

    enum Result
    {
        /**
         * Everything was ok.
         */

        OK = 0,


        /**
         * The query was aborted.
         */

        ABORTED = 1,


        /**
         * The server could not be connected.
         */

        COULD_NOT_CONNECT_SERVER = 2,


        /**
         * The databae could not be connected.
         */

        COULD_NOT_CONNECT_DATABASE = 3,


        /**
         * The nickname is invalid.
         */

        INVALID_NICKNAME = 4,


        /**
         * The user is not found.
         */

        USER_NOT_FOUND = 5,


        /**
         * The password is invalid.
         */

        INVALID_PASSWORD = 6,


        /**
         * The database could not be updated.
         */

        COULD_NOT_UPDATE_DATABASE = 7,


        /**
         *  The user already exists.
         */

        USER_ALREADY_EXISTS = 8,


        /**
         * The map does not exists.
         */

        MAP_DOES_NOT_EXISTS = 9,


        /**
         * The solution is invalid.
         */

        SOLUTION_IS_INVALID = 10,


        /**
         * You haven't solved the level.
         */

        LEVEL_NOT_SOLVED = 11,


        /**
         * To many users.
         */

        TOO_MANY_USERS = 12,


        /**
         * The query wass illegal.
         */

        ILLEGAL_QUERY = 13
    };


    /**
     * Returns the result of the query.
     */

    Result result() const;


    /**
     * Returns a human readable version of the result.
     */

    QString resultText() const;


    /**
     * Returns a human readable version of the result.
     *
     * @param result The result to use.
     */

    static QString resultText(Result result);


    /**
     * Returns the data from the query.
     */

    QString data() const;


    /**
     * Returns the version of the server.
     */

    int version() const;


signals:

    /**
     * Is emitted when the query is finished.
     */

    void finished();


protected slots:

    /**
     * Called to read from the socket.
     */

    void read();


    /**
     * Called when the socket was closed.
     */

    void closed();


    /**
     * Called when an error occured.
     */

    void error();


    /**
     * Called when the connection is established.
     */

    void connected();


    /**
     * Called when we should connect to the server.
     */

    void connectToHost();


private:

    /**
     * The data from the server.
     */

    QString m_server_data;


    /**
     * The socket to use.
     */

    QSocket * m_socket;


    /**
     * The script.
     */

    QString m_script;


    /**
     * The query.
     */

    QString m_query;


    /**
     * The result.
     */

    Result m_result;


    /**
     * The server path.
     */

    QString m_path;


    /**
     * The server host.
     */

    QString m_host;


    /**
     * The server port.
     */

    int m_port;


    /**
     * The data received.
     */

    QString m_data;


    /**
     * The server version.
     */

    int m_version;
};


#endif
