/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "solution_holder.h"

#include <cassert>
#include <utility>

#include <qfile.h>
#include <qdatastream.h>

#include <kglobal.h>
#include <kstddirs.h>




std::map<CompressedMap, int> SolutionHolder::s_map_indices;
std::vector<std::vector<CompressedMovements> > SolutionHolder::s_solutions;
std::vector<CompressedMovements> SolutionHolder::s_last_attempts;
std::vector<int> SolutionHolder::s_last_attempt_positions;
std::vector<std::vector<int> > SolutionHolder::s_pushes;
std::vector<std::vector<int> > SolutionHolder::s_linear_pushes;
std::vector<std::vector<int> > SolutionHolder::s_gem_changes;
std::vector<std::vector<int> > SolutionHolder::s_moves;
std::vector<std::vector<QDateTime> > SolutionHolder::s_dates;
std::vector<QStringList > SolutionHolder::s_infos;
bool SolutionHolder::s_was_modified = false;


void SolutionHolder::load()
{
    assert(s_solutions.empty());

    QString data_file = KGlobal::dirs()->findResource("data", "easysok/solutions.dat");

    if (!data_file.isEmpty())
    {
        QFile file(data_file);
        QStringList data;

        if (file.open(IO_ReadOnly))
        {
            QDataStream stream(&file);

            Q_INT32 version;
            stream >> version;

            Q_INT32 number_of_maps;
            stream >> number_of_maps;

            for (int i = 0; i < number_of_maps; ++i)
            {
                s_map_indices.insert(std::make_pair(CompressedMap(stream), i));
                s_solutions.push_back(std::vector<CompressedMovements>());
                s_pushes.push_back(std::vector<int>());
                s_linear_pushes.push_back(std::vector<int>());
                s_gem_changes.push_back(std::vector<int>());
                s_moves.push_back(std::vector<int>());
                s_dates.push_back(std::vector<QDateTime>());
                s_infos.push_back(QStringList());

                Q_INT32 number_of_solutions;
                stream >> number_of_solutions;

                for (Q_INT32 j = 0; j < number_of_solutions; ++j)
                {
                    if (version > 2)
                    {
                        s_solutions[i].push_back(CompressedMovements(stream));
                    }
                    else
                    {
                        s_solutions[i].push_back(CompressedMovements(Movements(stream)));
                    }

                    Q_INT32 pushes;
                    stream >> pushes;
                    s_pushes[i].push_back(pushes);

                    if (version > 0)
                    {
                        Q_INT32 linear_pushes;
                        stream >> linear_pushes;

                        // Needed,  because the number of linear pushes was calculated wrong
                        // in version < 4.
                        if (version > 3)
                        {
                            s_linear_pushes[i].push_back(linear_pushes);
                        }
                        else
                        {
                            Movements movements(s_solutions[i].back());
                            s_linear_pushes[i].push_back(movements.linearPushes());
                            s_was_modified = true;
                        }
                    }
                    else
                    {
                        Movements movements(s_solutions[i].back());
                        s_linear_pushes[i].push_back(movements.linearPushes());
                        s_was_modified = true;
                    }

                    if (version > 1)
                    {
                        Q_INT32 gem_changes;
                        stream >> gem_changes;
                        s_gem_changes[i].push_back(gem_changes);
                    }
                    else
                    {
                        Movements movements(s_solutions[i].back());
                        s_gem_changes[i].push_back(movements.gemChanges());
                        s_was_modified = true;
                    }

                    Q_INT32 moves;
                    stream >> moves;
                    s_moves[i].push_back(moves);

                    QDateTime date_time;
                    stream >> date_time;
                    s_dates[i].push_back(date_time);

                    QString info;
                    stream >> info;
                    s_infos[i].append(info);
                }

                if (version > 2)
                {
                    s_last_attempts.push_back(CompressedMovements(stream));
                }
                else
                {
                    s_last_attempts.push_back(CompressedMovements(Movements(stream)));
                }

                Q_INT32 move_pointer;
                stream >> move_pointer;

                s_last_attempt_positions.push_back(static_cast<int> (move_pointer));
            }
        }
    }
}


void SolutionHolder::save()
{
    if (!s_was_modified)
    {
        return;
    }

    QString data_file = KGlobal::dirs()->saveLocation("data", "easysok/", true) + "solutions.dat";

    if (!data_file.isEmpty())
    {
        QFile file(data_file);

        if (file.open(IO_WriteOnly | IO_Truncate))
        {
            QDataStream stream(&file);

            Q_INT32 const version = 4;
            stream << version;

            Q_INT32 const number_of_maps = static_cast<Q_INT32> (s_solutions.size());
            stream << number_of_maps;

            std::map<CompressedMap, int>::const_iterator it = s_map_indices.begin();
            std::map<CompressedMap, int>::const_iterator const end_it = s_map_indices.end();

            for (; it != end_it; ++it)
            {
                it->first.writeToStream(stream);

                int const index = it->second;

                Q_INT32 const number_of_solutions = static_cast<int> (s_solutions[index].size());
                stream << number_of_solutions;

                for (Q_INT32 i = 0; i < number_of_solutions; ++i)
                {
                    s_solutions[index][i].writeToStream(stream);

                    Q_INT32 pushes = s_pushes[index][i];
                    stream << pushes;

                    Q_INT32 linear_pushes = s_linear_pushes[index][i];
                    stream << linear_pushes;

                    Q_INT32 gem_changes = s_gem_changes[index][i];
                    stream << gem_changes;

                    Q_INT32 moves = s_moves[index][i];
                    stream << moves;

                    stream << s_dates[index][i];

                    stream << s_infos[index][i];
                }

                s_last_attempts[index].writeToStream(stream);

                Q_INT32 move_pointer = s_last_attempt_positions[index];
                stream << move_pointer;
            }
        }
    }

    s_was_modified = false;
}


int SolutionHolder::getIndexForMap(CompressedMap const & map)
{
    std::map<CompressedMap, int>::const_iterator const it = s_map_indices.find(map);

    if (it == s_map_indices.end())
    {
        return -1;
    }
    else
    {
        return it->second;
    }
}


bool SolutionHolder::hasSolution(CompressedMap const & map)
{
    int const index = getIndexForMap(map);

    return (index != -1) && hasSolution(index);
}


bool SolutionHolder::hasSolution(int index)
{
    return numberOfSolutions(index) > 0;
}


int SolutionHolder::numberOfSolutions(CompressedMap const & map)
{
    assert(hasSolution(map));

    return numberOfSolutions(getIndexForMap(map));
}


int SolutionHolder::numberOfSolutions(int index)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));

    return static_cast<int> (s_solutions[index].size());
}


int SolutionHolder::pushBestSolution(CompressedMap const & map)
{
    assert(hasSolution(map));

    return pushBestSolution(getIndexForMap(map));
}


int SolutionHolder::pushBestSolution(int index)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(!s_solutions.empty());

    int const nr_of_solutions = numberOfSolutions(index);

    int min_pushes = s_pushes[index][0];
    int min_moves = s_moves[index][0];
    int min_index = 0;

    for (int i = 1; i < nr_of_solutions; ++i)
    {
        int const act_pushes = s_pushes[index][i];
        int const act_moves = s_moves[index][i];

        if ((act_pushes < min_pushes) || ((act_pushes == min_pushes) && (act_moves < min_moves)))
        {
            min_pushes = act_pushes;
            min_moves = act_moves;
            min_index = i;
        }
    }

    return min_index;
}


int SolutionHolder::linearPushBestSolution(CompressedMap const & map)
{
    assert(hasSolution(map));

    return linearPushBestSolution(getIndexForMap(map));
}


int SolutionHolder::linearPushBestSolution(int index)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(!s_solutions.empty());

    int const nr_of_solutions = numberOfSolutions(index);

    int min_linear_pushes = s_linear_pushes[index][0];
    int min_moves = s_moves[index][0];
    int min_index = 0;

    for (int i = 1; i < nr_of_solutions; ++i)
    {
        int const act_linear_pushes = s_linear_pushes[index][i];
        int const act_moves = s_moves[index][i];

        if ((act_linear_pushes < min_linear_pushes) ||
            ((act_linear_pushes == min_linear_pushes) && (act_moves < min_moves)))
        {
            min_linear_pushes = act_linear_pushes;
            min_moves = act_moves;
            min_index = i;
        }
    }

    return min_index;
}


int SolutionHolder::gemChangeBestSolution(CompressedMap const & map)
{
    assert(hasSolution(map));

    return gemChangeBestSolution(getIndexForMap(map));

}


int SolutionHolder::gemChangeBestSolution(int index)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(!s_solutions.empty());

    int const nr_of_solutions = numberOfSolutions(index);

    int min_gem_changes = s_gem_changes[index][0];
    int min_moves = s_moves[index][0];
    int min_index = 0;

    for (int i = 1; i < nr_of_solutions; ++i)
    {
        int const act_gem_changes = s_gem_changes[index][i];
        int const act_moves = s_moves[index][i];

        if ((act_gem_changes < min_gem_changes) ||
            ((act_gem_changes == min_gem_changes) && (act_moves < min_moves)))
        {
            min_gem_changes = act_gem_changes;
            min_moves = act_moves;
            min_index = i;
        }
    }

    return min_index;
}


int SolutionHolder::moveBestSolution(CompressedMap const & map)
{
    assert(hasSolution(map));

    return moveBestSolution(getIndexForMap(map));
}


int SolutionHolder::moveBestSolution(int index)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(!s_solutions.empty());

    int const nr_of_solutions = numberOfSolutions(index);

    int min_pushes = s_pushes[index][0];
    int min_moves = s_moves[index][0];
    int min_index = 0;

    for (int i = 1; i < nr_of_solutions; ++i)
    {
        int const act_pushes = s_pushes[index][i];
        int const act_moves = s_moves[index][i];

        if ((act_moves < min_moves) || ((act_moves == min_moves) && (act_pushes < min_pushes)))
        {
            min_pushes = act_pushes;
            min_moves = act_moves;
            min_index = i;
        }
    }

    return min_index;
}


Movements SolutionHolder::movements(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    return Movements(s_solutions[getIndexForMap(map)][solution]);
}


Movements SolutionHolder::movements(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    return Movements(s_solutions[index][solution]);
}


int SolutionHolder::pushesInSolution(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    return pushesInSolution(getIndexForMap(map), solution);
}


int SolutionHolder::pushesInSolution(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    return s_pushes[index][solution];
}


int SolutionHolder::linearPushesInSolution(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    return linearPushesInSolution(getIndexForMap(map), solution);
}


int SolutionHolder::linearPushesInSolution(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    return s_linear_pushes[index][solution];
}


int SolutionHolder::gemChangesInSolution(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    return gemChangesInSolution(getIndexForMap(map), solution);
}


int SolutionHolder::gemChangesInSolution(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    return s_gem_changes[index][solution];
}


int SolutionHolder::movesInSolution(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    return movesInSolution(getIndexForMap(map), solution);
}


int SolutionHolder::movesInSolution(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    return s_moves[index][solution];
}


QDateTime const & SolutionHolder::dateOfSolution(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    return dateOfSolution(getIndexForMap(map), solution);
}


QDateTime const & SolutionHolder::dateOfSolution(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    return s_dates[index][solution];
}


QString const & SolutionHolder::infoOfSolution(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    return infoOfSolution(getIndexForMap(map), solution);
}


QString const & SolutionHolder::infoOfSolution(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    return s_infos[index][solution];
}


int SolutionHolder::addSolution(CompressedMap const & map, Movements const & moves, int number_of_pushes,
                                int number_of_linear_pushes, int number_of_gem_changes,
                                int number_of_moves, QString const & info)
{
    return addSolution(map, moves, number_of_pushes, number_of_linear_pushes, number_of_gem_changes,
                       number_of_moves, info, QDateTime::currentDateTime());
}


int SolutionHolder::addSolution(int index, Movements const & moves, int number_of_pushes,
                                int number_of_linear_pushes, int number_of_gem_changes,
                                int number_of_moves, QString const & info)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));

    return addSolution(index, moves, number_of_pushes, number_of_linear_pushes, number_of_gem_changes,
                       number_of_moves, info, QDateTime::currentDateTime());
}


int SolutionHolder::addSolution(CompressedMap const & map, Movements const & moves, int number_of_pushes,
                                int number_of_linear_pushes, int number_of_gem_changes, int number_of_moves,
                                QString const & info, QDateTime const & date_time)
{
    s_was_modified = true;

    int index = getIndexForMap(map);

    if (index == -1)
    {
        index = static_cast<int> (s_solutions.size());

        s_solutions.push_back(std::vector<CompressedMovements>());
        s_last_attempts.push_back(CompressedMovements(Movements()));
        s_last_attempt_positions.push_back(0);
        s_pushes.push_back(std::vector<int>());
        s_moves.push_back(std::vector<int>());
        s_linear_pushes.push_back(std::vector<int>());
        s_gem_changes.push_back(std::vector<int>());
        s_dates.push_back(std::vector<QDateTime>());
        s_infos.push_back(QStringList());
        s_map_indices.insert(std::make_pair(map, index));
    }

    return addSolution(index, moves, number_of_pushes, number_of_linear_pushes,
                       number_of_gem_changes, number_of_moves, info, date_time);
}


int SolutionHolder::addSolution(int index, Movements const & moves, int number_of_pushes,
                                int number_of_linear_pushes, int number_of_gem_changes,
                                int number_of_moves, QString const & info, QDateTime const & date_time)
{
    s_was_modified = true;

    int const number_of_solutions = numberOfSolutions(index);

    for (int i = 0; i <= number_of_solutions; ++i)
    {
        bool const last_one = (i == number_of_solutions);
        bool const better_pushes = last_one  || (s_pushes[index][i] > number_of_pushes);
        bool const equal_pushes = last_one || (s_pushes[index][i] == number_of_pushes);
        bool const better_moves = last_one || (equal_pushes && (s_moves[index][i] > number_of_moves));

        if (better_pushes || better_moves)
        {
            s_solutions[index].insert(s_solutions[index].begin() + i, CompressedMovements(moves));
            s_pushes[index].insert(s_pushes[index].begin() + i, number_of_pushes);
            s_linear_pushes[index].insert(s_linear_pushes[index].begin() + i, number_of_linear_pushes);
            s_gem_changes[index].insert(s_gem_changes[index].begin() + i, number_of_gem_changes);
            s_moves[index].insert(s_moves[index].begin() + i, number_of_moves);
            s_dates[index].insert(s_dates[index].begin() + i, date_time);
            s_infos[index].insert(s_infos[index].at(i), info);

            return i;
        }
    }

    assert(false);

    return -1;
}


void SolutionHolder::changeSolutionInfo(CompressedMap const & map, int solution, QString const & info)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    s_was_modified = true;

    changeSolutionInfo(getIndexForMap(map), solution, info);
}


void SolutionHolder::changeSolutionInfo(int index, int solution, QString const & info)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    s_infos[index][solution] = info;
}


void SolutionHolder::deleteSolution(CompressedMap const & map, int solution)
{
    assert(hasSolution(map));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(map));

    s_was_modified = true;

    deleteSolution(getIndexForMap(map), solution);
}


void SolutionHolder::deleteSolution(int index, int solution)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_solutions.size()));
    assert(hasSolution(index));
    assert(solution >= 0);
    assert(solution < numberOfSolutions(index));

    s_was_modified = true;

    s_solutions[index].erase(s_solutions[index].begin() + solution);
    s_pushes[index].erase(s_pushes[index].begin() + solution);
    s_moves[index].erase(s_moves[index].begin() + solution);
    s_linear_pushes[index].erase(s_linear_pushes[index].begin() + solution);
    s_gem_changes[index].erase(s_gem_changes[index].begin() + solution);
    s_dates[index].erase(s_dates[index].begin() + solution);
    s_infos[index].remove(s_infos[index].at(solution));
}


bool SolutionHolder::hasLastAttempt(CompressedMap const & map)
{
    int const index = getIndexForMap(map);

    if (index < 0)
    {
        return false;
    }
    else if (index >= static_cast<int> (s_last_attempts.size()))
    {
        return false;
    }

    return true;
}


Movements SolutionHolder::lastAttempt(CompressedMap const & map)
{
    assert(hasLastAttempt(map));

    int const index = getIndexForMap(map);

    Movements movements(s_last_attempts[index]);

    if (s_last_attempt_positions[index] < movements.numberOfMoves())
    {
        movements.setMovePointer(s_last_attempt_positions[index]);
    }

    return movements;
}


Movements SolutionHolder::lastAttempt(int index)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_last_attempts.size()));

    Movements movements(s_last_attempts[index]);
    movements.setMovePointer(s_last_attempt_positions[index]);

    return movements;
}


void SolutionHolder::setLastAttempt(CompressedMap const & map, Movements const & movements)
{
    s_was_modified = true;

    int const index = getIndexForMap(map);

    if (index == -1)
    {
        const int new_index = static_cast<int> (s_solutions.size());

        s_solutions.push_back(std::vector<CompressedMovements>());
        s_last_attempts.push_back(movements);
        s_last_attempt_positions.push_back(movements.movePointer());
        s_pushes.push_back(std::vector<int>());
        s_linear_pushes.push_back(std::vector<int>());
        s_gem_changes.push_back(std::vector<int>());
        s_moves.push_back(std::vector<int>());
        s_dates.push_back(std::vector<QDateTime>());
        s_infos.push_back(QStringList());
        s_map_indices.insert(std::make_pair(map, new_index));
    }
    else
    {
        s_last_attempts[index] = movements;
        s_last_attempt_positions[index] = movements.movePointer();
    }
}


void SolutionHolder::setLastAttempt(int index, Movements const & movements)
{
    assert(index >= 0);
    assert(index < static_cast<int> (s_last_attempts.size()));

    s_was_modified = true;
    s_last_attempts[index] = movements;
    s_last_attempt_positions[index] = movements.movePointer();
}

