/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_SOLUTION_HOLDER_INC_GUARD_H
#define EASYSOK_SOLUTION_HOLDER_INC_GUARD_H


#include <map>
#include <vector>

#include <qdatetime.h>
#include <qstring.h>
#include <qstringlist.h>

#include "compressed_map.h"
#include "compressed_movements.h"
#include "movements.h"




/**
 * This class is used to load/store solutions and retrieve them.
 *
 * Note that all methods are static.
 *
 * If there are more solutions for a map, the first ist the best and so on (determined
 * by the number of pushes).
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */


class SolutionHolder
{

public:

    /**
     * Reads in all previos stored levels.
     */

    static void load();


    /**
     * Stores all modifications.
     */

    static void save();


    /**
     * Returns an index for a given map.
     *
     * If there is not info for the map, -1 will be returned.
     *
     * If you have the index, the lookups will be much faster.
     */

    static int getIndexForMap(CompressedMap const & map);


    /**
     * Returns true, if a solution for the given map exists.
     *
     * @param map The compressed map to test.
     */

    static bool hasSolution(CompressedMap const & map);


    /**
     * @overload
     */

    static bool hasSolution(int index);


    /**
     * Returns the number of solutions of a give map.
     *
     * @param map The compressed map to test.
     */

    static int numberOfSolutions(CompressedMap const & map);


    /**
     * @overload
     */

    static int numberOfSolutions(int index);


    /**
     * Returns the index of the best solution in terms of the number of pushes.
     *
     * @param map The compressed map to use.
     */

    static int pushBestSolution(CompressedMap const & map);


    /**
     * @overload
     */

    static int pushBestSolution(int index);


    /**
     * Returns the index of the best solution in terms of the number of linear pushes.
     *
     * @param map The compressed map to use.
     */

    static int linearPushBestSolution(CompressedMap const & map);


    /**
     * @overload
     */

    static int linearPushBestSolution(int index);


    /**
     * Returns the index of the best solution in terms of the number of gem changes.
     *
     * @param map The compressed map to use.
     */

    static int gemChangeBestSolution(CompressedMap const & map);


    /**
     * @overload
     */

    static int gemChangeBestSolution(int index);


    /**
     * Returns the index of the best solution in terms of the number of moves.
     *
     * @param map The compressed map to use.
     */

    static int moveBestSolution(CompressedMap const & map);


    /**
     * @overload
     */

    static int moveBestSolution(int index);


    /**
     * Returns the movements for a given solutions of a map.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static Movements movements(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static Movements movements(int index, int solution);


    /**
     * Returns the number of pushes in the solution.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static int pushesInSolution(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static int pushesInSolution(int index, int solution);


    /**
     * Returns the number of linear pushes in the solution.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static int linearPushesInSolution(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static int linearPushesInSolution(int index, int solution);


    /**
     * Returns the number of gem changes in the solution.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static int gemChangesInSolution(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static int gemChangesInSolution(int index, int solution);


    /**
     * Returns the number of moves in the solution.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static int movesInSolution(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static int movesInSolution(int index, int solution);


    /**
     * Returns the date, when the solution was made.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static QDateTime const & dateOfSolution(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static QDateTime const & dateOfSolution(int index, int solution);


    /**
     * Returns additional info for a solution.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static QString const & infoOfSolution(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static QString const & infoOfSolution(int index, int solution);


    /**
     * Adds a new solution of a given map.
     *
     * Returns the position, at which the solution was added.
     *
     * The date of the solution is the current date.
     *
     * @param map The compressed map.
     * @param moves The movements.
     * @param number_of_pushes The number of pushes.
     * @param number_of_moves The number of moves.
     * @param info Additional info.
     */

    static int addSolution(CompressedMap const & map, Movements const & moves,
                           int number_of_pushes, int number_of_linear_pushes,
                           int number_of_gem_changes, int number_of_moves, QString const & info);


    /**
     * @overload
     */

    static int addSolution(int index, Movements const & moves,
                           int number_of_pushes, int number_of_linear_pushes,
                           int number_of_gem_changes, int number_of_moves, QString const & info);


    /**
     * Adds a new solution of a given map.
     *
     * Returns the position, at which the solution was added.
     *
     * @param map The compressed map.
     * @param moves The movements.
     * @param number_of_pushes The number of pushes.
     * @param number_of_moves The number of moves.
     * @param info Additional info.
     * @param date_time The date and time of the solution.
     */

    static int addSolution(CompressedMap const & map, Movements const & moves, int number_of_pushes,
                           int number_of_linear_pushes, int number_of_gem_changes,
                           int number_of_moves, QString const & info, QDateTime const & date_time);


    /**
     * @overload
     */

    static int addSolution(int index, Movements const & moves, int number_of_pushes,
                           int number_of_linear_pushes, int number_of_gem_changes,
                           int number_of_moves, QString const & info, QDateTime const & date_time);


    /**
     * Changes the info of a given solution.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     * @param info Additional info.
     */

    static void changeSolutionInfo(CompressedMap const & map, int solution, QString const & info);


    /**
     * @overload
     */

    static void changeSolutionInfo(int index, int solution, QString const & info);


    /**
     * Deletes a given solution of a map.
     *
     * @param map The compressed map.
     * @param solution The index of the solution.
     */

    static void deleteSolution(CompressedMap const & map, int solution);


    /**
     * @overload
     */

    static void deleteSolution(int index, int solution);


    /**
     * Returns true, if a last attempt for the given map exists.
     *
     * @param map The compressed map.
     */

    static bool hasLastAttempt(CompressedMap const & map);


    /**
     * Returns the movements of the last attempt.
     *
     * @param map The compressed map.
     */

    static Movements lastAttempt(CompressedMap const & map);


    /**
     * @overload
     */

    static Movements lastAttempt(int index);


    /**
     * Sets a new last attempt of a given map.
     *
     * @param map The compressed map.
     * @param movements The movements.
     */

    static void setLastAttempt(CompressedMap const & map, Movements const & movements);


    /**
     * @overload
     */

    static void setLastAttempt(int index, Movements const & movements);


private:

    /**
     * Here we store the id strings of all maps and their associated number.
     */

    static std::map<CompressedMap, int> s_map_indices;


    /**
     * Here we store the the solution movements.
     */

    static std::vector<std::vector<CompressedMovements> > s_solutions;


    /**
     * Here we store the last attempts.
     */

    static std::vector<CompressedMovements> s_last_attempts;


    /**
     * Here we store the positions of the last attempts.
     */

    static std::vector<int> s_last_attempt_positions;


    /**
     * Here we store the number of pushes in the solution.
     */

    static std::vector<std::vector<int> > s_pushes;


    /**
     * Here we store the number of moves in the solution.
     */

    static std::vector<std::vector<int> > s_moves;


    /**
     * Here we store the number of linear pushes in the solution.
     */

    static std::vector<std::vector<int> > s_linear_pushes;


    /**
     * Here we store the number of gem changes in the solution.
     */

    static std::vector<std::vector<int> > s_gem_changes;


    /**
     * Here we store the date.
     */

    static std::vector<std::vector<QDateTime> > s_dates;


    /**
     * Here we store the info.
     */

    static std::vector<QStringList > s_infos;


    /**
     * If true, the solution was modified.
     */

    static bool s_was_modified;
};


#endif
