/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



#include "solution_list_view.h"

#include <cassert>

#include <kglobal.h>
#include <klineeditdlg.h>
#include <klistview.h>
#include <klocale.h>

#include "compressed_map.h"
#include "solution_holder.h"




SolutionListView::SolutionListView(CompressedMap const & map, QWidget * parent,
                                   const char * name) :
    KListView(parent, name)
{
    setup(SolutionHolder::getIndexForMap(map));
}


SolutionListView::SolutionListView(int index, QWidget * parent, const char * name) :
    KListView(parent, name)
{
    setup(index);
}


int SolutionListView::numberOfSolutions() const
{
    return static_cast<int> (m_items.size());
}


bool SolutionListView::isHidden(int index) const
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    return m_hidden[index];
}


void SolutionListView::setHidden(int index, bool hide)
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    if (m_hidden[index])
    {
        if (!hide)
        {
            m_hidden[index] = 0;
            insertItem(m_items[index]);
        }
    }
    else
    {
        if (hide)
        {
            m_hidden[index] = 1;
            takeItem(m_items[index]);
        }
    }
}


int SolutionListView::pushes(int index) const
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    return m_items[index]->text(2).toInt();
}


void SolutionListView::setPushes(int index, int pushes)
{
    assert(index >= 0);
    assert(index < numberOfSolutions());
    assert(pushes > 0);

    m_items[index]->setText(2, QString::number(pushes));
}


int SolutionListView::linearPushes(int index) const
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    return m_items[index]->text(3).toInt();
}


void SolutionListView::setLinearPushes(int index, int pushes)
{
    assert(index >= 0);
    assert(index < numberOfSolutions());
    assert(pushes > 0);

    m_items[index]->setText(3, QString::number(pushes));
}


int SolutionListView::gemChanges(int index) const
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    return m_items[index]->text(4).toInt();
}


void SolutionListView::setGemChanges(int index, int gem_changes)
{
    assert(index >= 0);
    assert(index < numberOfSolutions());
    assert(gem_changes > 0);

    m_items[index]->setText(4, QString::number(gem_changes));
}


int SolutionListView::moves(int index) const
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    return m_items[index]->text(5).toInt();
}


void SolutionListView::setMoves(int index, int moves)
{
    assert(index >= 0);
    assert(index < numberOfSolutions());
    assert(moves > 0);

    m_items[index]->setText(6, QString::number(moves));
}


QString SolutionListView::annotation(int index) const
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    return m_items[index]->text(6);
}


void SolutionListView::setAnnotation(int index, QString const & annotation)
{
    assert(index >= 0);
    assert(index < numberOfSolutions());

    m_items[index]->setText(6, annotation);
}


std::vector<int> SolutionListView::selectedSolutions() const
{
    std::vector<int> result;

    int const number_of_solutions = numberOfSolutions();

    for (int i = 0; i < number_of_solutions; ++i)
    {
        if (m_hidden[i] == 0)
        {
            if (isSelected(m_items[i]))
            {
                result.push_back(i);
            }
        }
    }

    return result;
}


void SolutionListView::solutionSelected(QListViewItem * item)
{
    emit clickedSolution(item->text(0).toInt() - 1);
}


void SolutionListView::setup(int index)
{
    setShowSortIndicator(true);

    connect(this, SIGNAL(executed(QListViewItem *)), this, SLOT(solutionSelected(QListViewItem *)));

    addColumn(i18n("Nr."));
    setColumnAlignment(0, AlignRight);
    addColumn(i18n("When solved"));
    setColumnAlignment(1, AlignLeft);
    addColumn(i18n("Pushes"));
    setColumnAlignment(2, AlignRight);
    addColumn(i18n("Linear Pushes"));
    setColumnAlignment(3, AlignRight);
    addColumn(i18n("Gem changes"));
    setColumnAlignment(4, AlignRight);
    addColumn(i18n("Moves"));
    setColumnAlignment(5, AlignRight);
    addColumn(i18n("Annotation"));
    setColumnAlignment(6, AlignLeft);

    if (index == -1)
    {
        return;
    }

    int const number_of_solutions = SolutionHolder::numberOfSolutions(index);

    m_items.reserve(number_of_solutions);
    m_hidden.reserve(number_of_solutions);

    for (int i = 0; i < number_of_solutions; ++i)
    {
        QString const number = QString::number(i + 1).rightJustify(3, '0', true) + "  ";
        QString const date = KGlobal::locale()->formatDateTime(SolutionHolder::dateOfSolution(index, i), true);
        QString const pushes = QString::number(SolutionHolder::pushesInSolution(index, i));
        QString const linear_pushes = QString::number(SolutionHolder::linearPushesInSolution(index, i));
        QString const gem_changes = QString::number(SolutionHolder::gemChangesInSolution(index, i));
        QString const moves = QString::number(SolutionHolder::movesInSolution(index, i));
        QString const info = SolutionHolder::infoOfSolution(index, i);

        QListViewItem * item = new QListViewItem(this, number, date, pushes, linear_pushes,
                                                 gem_changes, moves, info);
        insertItem(item);

        if (i == 0)
        {
            setSelected(item, true);
        }

        m_items.push_back(item);
        m_hidden.push_back(0);
    }
}
