/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_SOLUTION_LIST_VIEW_INC_GUARD_H
#define EASYSOK_SOLUTION_LIST_VIEW_INC_GUARD_H


#include <vector>

#include <klistview.h>
#include <qstring.h>


class CompressedMap;
class QListViewItem;
class QString;




/**
 * Display the solutions of a map.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class SolutionListView : public KListView
{
    Q_OBJECT

public:

    /**
     * Creates the list view.
     *
     * @param compressed_map The compressed map of the level.
     * @param parent The parent widget.
     * @param name The internal name.
     */

    SolutionListView(CompressedMap const & map, QWidget * parent = 0, const char * name = 0);


    /**
     * @overload
     */

    SolutionListView(int index, QWidget * parent = 0, const char * name = 0);


    /**
     * Returns the number of solutions.
     */

    int numberOfSolutions() const;


    /**
     * Returns true, if the solution is hidden.
     *
     * @param index The index of the solution.
     */

    bool isHidden(int index) const;


    /**
     * Hides the solution with the given index.
     *
     * @param index The index of the solution to hide.
     * @param hide If true, the solution is hidden and unhidden otherwise.
     */

    void setHidden(int index, bool hide);


    /**
     * Returns the number of pushes in the solution.
     *
     * @param index The index of the solution.
     */

    int pushes(int index) const;


    /**
     * Sets the number of pushes in the solution.
     *
     * @param index The index of the solution.
     * @param pushes The new number of pushes in the solutions.
     */

    void setPushes(int index, int pushes);


    /**
     * Returns the number of linear pushes in the solution.
     *
     * @param index The index of the solution.
     */

    int linearPushes(int index) const;


    /**
     * Sets the number of linear pushes in the solution.
     *
     * @param index The index of the solution.
     * @param pushes The new number of linear pushes in the solutions.
     */

    void setLinearPushes(int index, int pushes);


    /**
     * Returns the number of gem changes in the solution.
     *
     * @param index The index of the solution.
     */

    int gemChanges(int index) const;


    /**
     * Sets the number of gem changes in the solution.
     *
     * @param index The index of the solution.
     * @param pushes The new number of gem changes in the solutions.
     */

    void setGemChanges(int index, int gem_changes);


    /**
     * Returns the number of moves in the solution.
     *
     * @param index The index of the solution.
     */

    int moves(int index) const;


    /**
     * Sets the number of moves in the solution.
     *
     * @param index The index of the solution.
     * @param moves The new number of moves in the solutions.
     */

    void setMoves(int index, int moves);


    /**
     * Returns the annotation of a solution.
     *
     * @index The index of the solution.
     */

    QString annotation(int index) const;


    /**
     * Sets a new annotation of a solution.
     *
     * @param index The index of the solution.
     * @param annotation The new annotation.
     */

    void setAnnotation(int index, QString const & annotation);


    /**
     * Returns a vector holding the indices of all selected solutions.
     */

    std::vector<int> selectedSolutions() const;


signals:

    /**
     * Emitted, when the user clicks a solution.
     *
     * @param index The index of the solution.
     */

    void clickedSolution(int index);


private slots:

    /**
     * Called, when the user selects a solution.
     */

    void solutionSelected(QListViewItem * item);


private:

    /**
     * Sets up the widget.
     *
     * @param index The index of the compressed map in the SolutionHolder.
     */

    void setup(int index);


    /**
     * The list view items.
     */

    std::vector<QListViewItem *> m_items;


    /**
     * Here we store 1 for every hidden item.
     */

    std::vector<int> m_hidden;
};


#endif
