/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "solution_optimize_dialog.h"

#include <cassert>

#include <qdatetime.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qvbox.h>

#include <klocale.h>

#include "compressed_map.h"
#include "map.h"
#include "move_optimizer.h"
#include "push_optimizer.h"
#include "solution_holder.h"
#include "solution_list_view.h"




SolutionOptimizeDialog::SolutionOptimizeDialog(CompressedMap const & map, bool opt_pushes,
                                               QWidget * parent, const char * name) :
    KDialogBase(parent, name, true, i18n("Optimize solutions"), Ok | Cancel | User1 | Help,
                Ok, true, i18n("Optmize")),
    m_map(map),
    m_opt_pushes(opt_pushes)
{
    m_index = SolutionHolder::getIndexForMap(map);
    assert(m_index != -1);

    m_list_view = new SolutionListView(m_index, makeVBoxMainWidget());
    m_list_view->setSelectionMode(QListView::Multi);

    int const number_of_solutions = SolutionHolder::numberOfSolutions(m_index);

    m_moves.resize(number_of_solutions, Movements());
    m_optimized.resize(number_of_solutions, 0);

    setHelp("solution-optimize-dialog");
}


void SolutionOptimizeDialog::slotUser1()
{
    std::vector<int> selected = m_list_view->selectedSolutions();
    int const number_of_selections = static_cast<int> (selected.size());
    Map map(m_map);

    for (int i = 0; i < number_of_selections; ++i)
    {
        int const solution = selected[i];

        if (m_optimized[solution] == false)
        {
            Movements moves = SolutionHolder::movements(m_index, solution);
            int number_of_pushes;
            int number_of_linear_pushes;
            int number_of_gem_changes;
            int number_of_moves;

            if (m_opt_pushes)
            {
                PushOptimizer optimizer(map, moves);
                moves = optimizer.moves();
                number_of_pushes = optimizer.numberOfPushes();
                number_of_linear_pushes = moves.linearPushes();
                number_of_gem_changes = moves.gemChanges();
                number_of_moves = optimizer.numberOfMoves();
            }
            else
            {
                MoveOptimizer optimizer(map, moves);
                moves = optimizer.moves();
                number_of_pushes = optimizer.numberOfPushes();
                number_of_linear_pushes = moves.linearPushes();
                number_of_gem_changes = moves.gemChanges();
                number_of_moves = optimizer.numberOfMoves();
            }

            m_list_view->setPushes(solution, number_of_pushes);
            m_list_view->setLinearPushes(solution, number_of_linear_pushes);
            m_list_view->setGemChanges(solution, number_of_gem_changes);
            m_list_view->setMoves(solution, number_of_moves);
            m_moves[solution] = moves;
            m_optimized[solution] = true;
        }
    }
}


void SolutionOptimizeDialog::slotOk()
{
    std::vector<Movements> moves;
    std::vector<int> number_of_pushes;
    std::vector<int> number_of_linear_pushes;
    std::vector<int> number_of_gem_changes;
    std::vector<int> number_of_moves;
    QStringList infos;
    std::vector<QDateTime> date_times;

    int const number_of_solutions = SolutionHolder::numberOfSolutions(m_index);

    for (int i = 0; i < number_of_solutions; ++i)
    {
        date_times.push_back(SolutionHolder::dateOfSolution(m_index, i));
        infos.append(SolutionHolder::infoOfSolution(m_index, i));

        if (m_optimized[i])
        {
            number_of_pushes.push_back(m_list_view->pushes(i));
            number_of_linear_pushes.push_back(m_list_view->linearPushes(i));
            number_of_gem_changes.push_back(m_list_view->gemChanges(i));
            number_of_moves.push_back(m_list_view->moves(i));
            moves.push_back(m_moves[i]);
        }
        else
        {
            number_of_pushes.push_back(SolutionHolder::pushesInSolution(m_index, i));
            number_of_linear_pushes.push_back(SolutionHolder::linearPushesInSolution(m_index, i));
            number_of_gem_changes.push_back(SolutionHolder::gemChangesInSolution(m_index, i));
            number_of_moves.push_back(SolutionHolder::movesInSolution(m_index, i));
            moves.push_back(SolutionHolder::movements(m_index, i));
        }
    }

    for (int i = number_of_solutions - 1; i >= 0; --i)
    {
        SolutionHolder::deleteSolution(m_index, i);
    }

    for (int i = 0; i < number_of_solutions; ++i)
    {
        SolutionHolder::addSolution(m_index, moves[i], number_of_pushes[i], number_of_linear_pushes[i],
                                    number_of_gem_changes[i], number_of_moves[i], infos[i], date_times[i]);
    }

    accept();
}
