/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "solver.h"


#include <algorithm>
#include <cassert>
#include <utility>

#include <qfile.h>
#include <qstring.h>
#include <qstringlist.h>
#include <qtextstream.h>

#include <kglobal.h>
#include <kstddirs.h>




int const Solver::s_unsolvable = 16383;


Solver::Solver(Map const & map, int cache_size) :
    m_map(map),
    m_start_map(map),
    m_empty_map(map),
    m_best_effort_max_depth(0),
    m_best_effort_lower_bound(s_unsolvable),
    m_depth(0),
    m_max_depth(1),
    m_width(map.width()),
    m_height(map.height()),
    m_size(m_width * m_height),
    m_max_cache_size(cache_size),
    m_cache_size(0),
    m_depth_counts(1, 0),
    m_max_depth_counts(0),
    m_reachable_map_gem(0),
    m_solve_finished(false)
{
    assert(map.isValid());
    assert(cache_size > 0);

    m_xy_offsets[0] = -1;
    m_xy_offsets[1] = 1;
    m_xy_offsets[2] = -m_width;
    m_xy_offsets[3] = m_width;

    setupDeadlockPatterns();

    for (int i = 0; i < m_size; ++i)
    {
        if (m_map.containsGem(i) && m_map.containsGoal(i) && isDeadlock(i, true))
        {
            m_map.setPiece(i, Map::WALL);
        }
    }

    m_map = Map(m_map.width(), m_map.height(), m_map.pieces());
    m_start_map = m_map;

    m_map.calcDeadlocks();

    for (int i = 0; i < m_size; ++i)
    {
        if ((m_map.containsGem(i) || m_map.canDropGem(i)) && !m_map.isDeadlock(i))
        {
            m_possible_gem_positions.push_back(i);
        }

        if (m_map.containsGoal(i))
        {
            m_goal_positions.push_back(i);
        }
    }

    m_gem_positions = getGemPositions(m_start_map);

    m_goals = static_cast<int> (m_goal_positions.size());

    std::vector<int> pieces = m_map.pieces();

    for (int k = 0; k < m_goals; ++k)
    {
        --pieces[m_goal_positions[k]];
    }

///   Why was this needed?
//    pieces[m_map.getIndex(m_map.keeper())] += 4;

    for (int k = 0; k < m_size; ++k)
    {
        pieces[k] &= 7;

        if (m_map.containsGem(k))
        {
            pieces[k] += 2;
        }
    }

    m_empty_map = Map(m_map.width(), m_map.height(), pieces);
    m_empty_map.calcDeadlocks();

    setupDistanceMap();

    m_reachable_map.resize(m_size * m_size / 8 + 1, 0);
}


bool Solver::prepare()
{
    if (m_reachable_map_gem == m_size)
    {
        return true;
    }

    int const gem_pos = m_reachable_map_gem;

    if (m_empty_map.canDropGem(gem_pos) && !m_map.isDeadlock(gem_pos))
    {
        m_empty_map.setPiece(gem_pos, Map::GEM);

        for (int i = 0; i < m_size; ++i)
        {
            if (m_empty_map.canDropKeeper(i))
            {
                m_empty_map.calcReachable(i);

                for (int j = 0; j < 4; ++j)
                {
                    if (m_empty_map.isReachable(gem_pos + m_xy_offsets[j]))
                    {
                        int const index = gem_pos * m_size + i;

                        m_reachable_map[index >> 3] |= (1 << (j + 4 * (index & 7)));
                    }
                }
            }
        }

        m_empty_map.setPiece(gem_pos, Map::EMPTY);
    }

    ++m_reachable_map_gem;

    return false;
}


bool Solver::solve(int steps)
{
    assert(steps >= 0);

    if (m_solve_finished)
    {
        return true;
    }

    m_act_min_depth = m_max_depth;
    m_act_max_depth = 0;

    for (int i = 0; i < steps; ++i)
    {
        if (doSingleStep())
        {
            m_solve_finished = true;

            return true;
        }
    }

    return false;
}


int Solver::depth() const
{
    return m_depth;
}


int Solver::actMinDepth() const
{
    return m_act_min_depth;
}


int Solver::actMaxDepth() const
{
    return m_act_max_depth;
}


int Solver::maxDepthReached() const
{
    return m_max_depth_reached;
}


int Solver::maxDepth() const
{
    return m_max_depth;
}


int Solver::nextMaxDepth() const
{
    return m_max_depth + m_min_add_depth;
}


double Solver::percentageCompleted() const
{
    double percentage = 0.0;
    double factor = 1.0;

    for (int i = 0; i < m_depth; ++i)
    {
        percentage += factor * static_cast<double> (m_pos[i]) / m_move_lengths[i];
        factor /= m_move_lengths[i];
    }

    return 100.0 * percentage;
}


Movements Solver::moves() const
{
    return m_solution_moves;
}


Movements Solver::bestEffort() const
{
    return m_best_effort;
}


bool Solver::doSingleStep()
{
    if (m_depth == 0)
    {
        return startSearch();
    }
    else if (m_pos.back() == m_move_lengths.back())
    {
        return collapse();
    }
    else if (static_cast<int> (m_hashs.size()) < m_max_depth)
    {
        expand();

        return false;
    }
    else
    {
        return processEndNodes();
    }
}


bool Solver::startSearch()
{
    assert(Hash(m_start_map) == Hash(m_map));

    m_min_add_depth = s_unsolvable;
    setupCache();

    m_pos.push_back(0);
    m_move_offsets.push_back(0);
    m_moves = validMoves();
    m_move_lengths.push_back(static_cast<int> (m_moves.size()));
    m_min_solve_moves.push_back(s_unsolvable);
    m_hashs.push_back(Hash(m_map, m_possible_gem_positions));
    ++m_depth;

    m_act_min_depth = std::min(m_act_min_depth, m_depth);
    m_act_max_depth = std::max(m_act_max_depth, m_depth);

    m_max_depth_reached = m_depth;

    if (m_moves.empty())
    {
        return true;
    }

    int const lower_bound = lowerBound(m_hashs.front(), m_gem_positions.front(), m_depth);

    if (lower_bound >= s_unsolvable)
    {
        return true;
    }

    if (lower_bound > m_max_depth)
    {
        m_max_depth = lower_bound;
    }

    return false;
}


bool Solver::collapse()
{
    int const solve_moves = std::min(s_unsolvable, m_min_solve_moves.back() + 1);
    updateCache(m_hashs.back(), solve_moves, m_depth);

    m_pos.pop_back();
    m_moves.erase(m_moves.end() - m_move_lengths.back(), m_moves.end());
    m_move_lengths.pop_back();
    m_move_offsets.pop_back();
    m_min_solve_moves.pop_back();
    m_hashs.pop_back();
    --m_depth;

    m_act_min_depth = std::min(m_act_min_depth, m_depth);
    m_act_max_depth = std::max(m_act_max_depth, m_depth);

    if (m_depth == 0)
    {
        if (m_min_add_depth > 0)
        {
            m_max_depth += m_min_add_depth;

            return false;
        }
        else
        {
            return true;
        }
    }
    else
    {
        if (m_min_solve_moves.back() > solve_moves)
        {
            m_min_solve_moves.back() = solve_moves;
        }

        int const raw_move = m_moves[m_move_offsets.back() + m_pos.back()];
        int const move_direction = raw_move & 3;
        int const moved_gem = raw_move >> 2;
        int const old_pos = m_gem_positions[moved_gem];
        int const new_pos = old_pos - m_xy_offsets[move_direction];

        m_gem_positions[moved_gem] = new_pos;
        m_map.setKeeper(new_pos - m_xy_offsets[move_direction]);
        m_map.moveGem(old_pos, new_pos);

        ++m_pos.back();
    }

    return false;
}


void Solver::expand()
{
    int const raw_move = m_moves[m_move_offsets.back() + m_pos.back()];
    int const move_direction = raw_move & 3;
    int const moved_gem = raw_move >> 2;
    int const old_pos = m_gem_positions[moved_gem];
    int const new_pos = old_pos + m_xy_offsets[move_direction];

    m_gem_positions[moved_gem] = new_pos;
    m_map.setKeeper(old_pos - m_xy_offsets[move_direction]);
    m_map.moveGem(old_pos, new_pos);
    m_map.setKeeper(old_pos);

    Hash const hash(m_map, m_possible_gem_positions);
    int const lower_bound = lowerBound(hash, new_pos, m_depth + 1);

    bool undo_move = false;

    if (lower_bound >= s_unsolvable)
    {
        undo_move = true;
    }
    else
    {
        if (m_depth > m_best_effort_max_depth)
        {
            m_best_effort = getFullMoves();
            m_best_effort_max_depth = m_depth;
        }
        else if ( (m_depth == m_best_effort_max_depth) &&
                  (lower_bound < m_best_effort_lower_bound) )
        {
            m_best_effort = getFullMoves();
            m_best_effort_lower_bound = lower_bound;
        }

        if (lower_bound + m_depth >= m_max_depth + m_min_add_depth)
        {
            if (m_min_solve_moves.back() > lower_bound)
            {
                m_min_solve_moves.back() = lower_bound;
            }

            undo_move = true;
        }
        else if (lower_bound + m_depth > m_max_depth)
        {
            if (lower_bound + m_depth < m_max_depth + m_min_add_depth)
            {
                m_min_add_depth = lower_bound + m_depth - m_max_depth;
            }

            if (m_min_solve_moves.back() > lower_bound)
            {
                m_min_solve_moves.back() = lower_bound;
            }

            undo_move = true;
        }
        else
        {
            std::vector<int> moves = validMoves();

            if (!moves.empty())
            {
                m_move_lengths.push_back(static_cast<int> (moves.size()));
                m_pos.push_back(0);
                m_move_offsets.push_back(static_cast<int> (m_moves.size()));
                m_moves.insert(m_moves.end(), moves.begin(), moves.end());
                m_min_solve_moves.push_back(s_unsolvable);
                m_hashs.push_back(hash);
                ++m_depth;

                m_act_min_depth = std::min(m_act_min_depth, m_depth);
                m_act_max_depth = std::max(m_act_max_depth, m_depth);

                m_max_depth_reached = std::max(m_max_depth_reached, m_depth);
            }
            else
            {
                updateCache(hash, s_unsolvable, m_depth + 1);

                undo_move = true;
            }
        }
    }

    if (undo_move)
    {
        m_gem_positions[moved_gem] = old_pos;
        m_map.setKeeper(old_pos - m_xy_offsets[move_direction]);
        m_map.moveGem(new_pos, old_pos);

        ++m_pos.back();
    }
}


bool Solver::processEndNodes()
{
    int const raw_move = m_moves[m_move_offsets.back() + m_pos.back()];
    int const move_direction = raw_move & 3;
    int const moved_gem = raw_move >> 2;
    int const old_pos = m_gem_positions[moved_gem];
    int const new_pos = old_pos + m_xy_offsets[move_direction];

    m_gem_positions[moved_gem] = new_pos;
    m_map.setKeeper(old_pos - m_xy_offsets[move_direction]);
    m_map.moveGem(old_pos, new_pos);
    m_map.setKeeper(old_pos);

    if (m_map.isSolved())
    {
        m_solution_moves = getFullMoves();

        return true;
    }
    else
    {
        Hash const hash(m_map, m_possible_gem_positions);

        int const lower_bound = lowerBound(hash, new_pos, m_depth + 1);

        if (m_min_add_depth > lower_bound)
        {
            m_min_add_depth = lower_bound;
        }

        if (m_min_solve_moves.back() > lower_bound)
        {
            m_min_solve_moves.back() = lower_bound;
        }

        if (lower_bound < s_unsolvable)
        {
            if (m_depth > m_best_effort_max_depth)
            {
                m_best_effort = getFullMoves();
                m_best_effort_max_depth = m_depth;
            }
            else if ( (m_depth == m_best_effort_max_depth) &&
                      (lower_bound < m_best_effort_lower_bound) )
            {
                m_best_effort = getFullMoves();
                m_best_effort_lower_bound = lower_bound;
            }
        }
    }

    m_map.setKeeper(old_pos - m_xy_offsets[move_direction]);
    m_map.moveGem(new_pos, old_pos);
    m_gem_positions[moved_gem] = old_pos;

    ++m_pos.back();

    return false;
}


std::vector<int> const & Solver::validMoves() const
{
    static std::vector<int> result;
    result.resize(m_goals * 4);

    m_map.calcReachable();

    int pos = 0;

    for (int i = 0; i < m_goals; ++i)
    {
        int gem_pos = m_gem_positions[i];

        for (int j = 0; j < 4; ++j)
        {
            int const gem_nb_index = gem_pos + m_xy_offsets[j];

            if (!m_map.isDeadlock(gem_nb_index) && m_map.canDropGem(gem_nb_index))
            {
                int const keeper_nb_index = gem_pos - m_xy_offsets[j];

                if (m_map.isReachable(keeper_nb_index))
                {
                    result[pos] = (i << 2) + j;
                    ++pos;
                }
            }
        }
    }

    result.resize(pos);

    return result;
}


int Solver::movesForGem(int keeper_pos, int gem_pos, int goal_pos) const
{
    assert(gem_pos >= 0);
    assert(gem_pos < m_size);
    assert(goal_pos >= 0);
    assert(goal_pos < m_goals);

    int min_moves = s_unsolvable;

    for (int i = 0; i < 4; ++i)
    {
        int const index = gem_pos * m_size + keeper_pos;

        if (m_reachable_map[index >> 3] & (1 << (i + 4 * (index & 7))))
        {
            min_moves = std::min(m_distance_map[i + 4 * (gem_pos + goal_pos * m_size)], min_moves);
        }
    }

    return min_moves;
}


bool Solver::isDeadlock(int position, bool ignore_on_goal) const
{
    assert(m_map.containsGem(position));

    if (!ignore_on_goal)
    {
        if (m_map.containsGoal(position))
        {
            return false;
        }
    }

    int const deadlocks = static_cast<int> (m_deadlock_pattern_lengths.size());
    int offset = 0;

    for (int i = 0; i < deadlocks; ++i)
    {
        int const size = m_deadlock_pattern_lengths[i];
        bool is_deadlock = true;

        for (int j = 0; j < size; ++j)
        {
            int const act_offset = j + offset;
            int const pos = position + m_deadlock_pattern_positions[act_offset];

            if ( (pos < 0) || (pos >= m_size) )
            {
                is_deadlock = false;

                break;
            }

            int const piece = m_map.getPiece(pos) | (m_map.isDeadlock(pos) << 3);

            if (!(m_deadlock_patterns[act_offset] & (1 << piece)))
            {
                is_deadlock = false;

                break;
            }
        }

        if (is_deadlock)
        {
            return true;
        }

        offset += size;
    }

    return false;
}


int Solver::minMovesForSolution(int moved_gem_pos) const
{
    if (isDeadlock(moved_gem_pos))
    {
        return s_unsolvable;
    }

    static std::vector<int> benefit_matrix;
    benefit_matrix.resize(m_goals * m_goals);

    int const keeper_pos = m_map.getIndex(m_map.keeper());

    for (int i = 0; i < m_goals; ++i)
    {
        for (int j = 0; j < m_goals; ++j)
        {
            int const benefit = s_unsolvable - movesForGem(keeper_pos, m_gem_positions[i], j);
            assert(benefit >= 0);
            assert(benefit <= s_unsolvable);
            benefit_matrix[j + i * m_goals] = benefit;
        }
    }

    return assignmentSolver(benefit_matrix, m_goals);
}


// This solver for the linear assignment problem uses an auction algorithm

int Solver::assignmentSolver(std::vector<int> & matrix, int _dim) const
{
    int dim = _dim;

    // First check, if there is any unreachable goal
    for (int i = 0; i < dim; ++i)
    {
        bool is_reachable = false;

        for (int j = i + dim * dim - dim; j >= 0; j -= dim)
        {
            if (matrix[j] != 0)
            {
                is_reachable = true;

                break;
            }
        }

        if (!is_reachable)
        {
            return s_unsolvable;
        }
    }

    int cost = 0;

    // Now we check, if there are any unassignable gems.
    // Further we remove gems which must go to a specific goal.

    bool removed_gem = false;

    do
    {
        removed_gem = false;

        for (int i = 0; i < dim; ++i)
        {
            int usable_goals = 0;
            int usable_goal = -1;
            int const offset = i * dim;

            for (int j = 0; j < dim; ++j)
            {
                if (matrix[offset + j] != 0)
                {
                    ++usable_goals;
                    usable_goal = j;

                    if (usable_goals > 1)
                    {
                        break;
                    }
                }
            }

            if (usable_goals == 0)
            {
                return s_unsolvable;
            }
            // If there is only one usable goal, we can remove the gem from the matrix
            // and build a smaller one
            else if (usable_goals == 1)
            {
                assert(usable_goal != -1);

                cost += s_unsolvable - matrix[usable_goal + i * dim];

                int const old_dim = dim;
                --dim;

                // Now remove the gem and goal from the matrix.
                std::vector<int> new_matrix(dim * dim);

                for (int j = 0; j < i; ++j)
                {
                    for (int k = 0; k < usable_goal; ++k)
                    {
                        new_matrix[k + j * dim] = matrix[k + j * old_dim];
                    }

                    for (int k = usable_goal + 1; k < old_dim; ++k)
                    {
                        new_matrix[k - 1 + j * dim] = matrix[k + j * old_dim];
                    }
                }

                for (int j = i + 1; j < old_dim; ++j)
                {
                    for (int k = 0; k < usable_goal; ++k)
                    {
                        new_matrix[k + (j - 1) * dim] = matrix[k + j * old_dim];
                    }

                    for (int k = usable_goal + 1; k < old_dim; ++k)
                    {
                        new_matrix[k - 1 + (j - 1) * dim] = matrix[k + j * old_dim];
                    }
                }

                new_matrix.swap(matrix);
                removed_gem = true;

                break;
            }
        }
    } while (removed_gem);

    std::vector<int> goal_assignments(dim, -1);
    std::vector<int> gem_assignments(dim, -1);

    if (!matrix.empty())
    {
        std::vector<double> prices(dim, -s_unsolvable * 10.0);

        for (int step = 0; step < 2; ++step)
        {
            if (step > 0)
            {
                std::fill(goal_assignments.begin(), goal_assignments.end(), -1);
                std::fill(gem_assignments.begin(), gem_assignments.end(), -1);
            }

            std::vector<int> unassigned(dim);

            for (int i = 0; i < dim; ++i)
            {
                unassigned[i] = i;
            }

            double epsilon;

            // In first step,  do only a rough calculation
            if (step == 0)
            {
                epsilon = 1.0;
            }
            else
            {
                // The minimum price increase. 1 / dim is sufficient, but just to be sure ...
                epsilon = 1.0 / (dim + 1);
            }

            static std::vector<int> new_unassigned;

            do
            {
                new_unassigned.resize(0);

                int const nr_unassigned = static_cast<int> (unassigned.size());

                for (int i = 0; i < nr_unassigned; ++i)
                {
                    int const gem = unassigned[i];
                    int const offset = gem * dim;

                    int best = -1;
                    double best_val = -1e10;
                    double second_val = -1e10;

                    for (int j = 0; j < dim; ++j)
                    {
                        double const benefit = matrix[offset + j] - prices[j];

                        if (benefit > best_val)
                        {
                            second_val = best_val;
                            best_val = benefit;
                            best = j;
                        }
                        else if (benefit > second_val)
                        {
                            second_val = benefit;
                        }
                    }

                    assert(best != -1);
                    assert(second_val > -1e10);

                    if (goal_assignments[best] != -1)
                    {
                        new_unassigned.push_back(goal_assignments[best]);
                    }

                    goal_assignments[best] = gem;
                    gem_assignments[gem] = best;
                    prices[best] += best_val - second_val + epsilon;
                }

                new_unassigned.swap(unassigned);

            } while (!unassigned.empty());
        }
    }

    for (int i = 0; i < dim; ++i)
    {
        int const benefit = matrix[gem_assignments[i] + i * dim];

        if (benefit != 0)
        {
            cost += s_unsolvable - benefit;
        }
        else
        {
            return s_unsolvable;
        }
    }

    return cost;
}


int Solver::lowerBound(Hash const & hash, int moved_gem_pos, int depth)
{
    assert(depth > 0);
    assert(Hash(m_map) == hash);

    std::map<Hash, CacheEntry>::iterator it = m_cache.find(hash);

    if (it != m_cache.end())
    {
        int const old_depth = it->second.depth();

        if (!it->second.wasTouched())
        {
            if (old_depth >= depth)
            {
                ++m_depth_counts[depth];
                --m_depth_counts[old_depth];

                it->second.setDepth(depth);
                it->second.touch();

                return it->second.movesToSolve();
            }
            else
            {
                return s_unsolvable;
            }
        }
        else
        {
            if (old_depth > depth)
            {
                ++m_depth_counts[depth];
                --m_depth_counts[old_depth];

                it->second.setDepth(depth);

                return it->second.movesToSolve();
            }
            else
            {
                return s_unsolvable;
            }
        }
    }
    else
    {
        int const lower_bound = minMovesForSolution(moved_gem_pos);
        insertInCache(hash, lower_bound, depth, true);

        return lower_bound;
    }
}


void Solver::updateCache(Hash const & hash, int moves_to_solve, int depth)
{
    assert(moves_to_solve >= 0);
    assert(depth > 0);

    std::map<Hash, CacheEntry>::iterator it = m_cache.find(hash);

    if (it != m_cache.end())
    {
        assert(it->second.wasTouched());
        assert(it->second.depth() >= depth);

        if (it->second.movesToSolve() < moves_to_solve)
        {
            it->second.setMovesToSolve(moves_to_solve);
        }
    }
    else
    {
        insertInCache(hash, moves_to_solve, depth, false);
    }
}


void Solver::insertInCache(Hash const & hash, int moves_to_solve, int depth, bool calculated)
{
    assert(depth > 0);

    if (m_cache_size >= m_max_cache_size)
    {
        int entries_to_delete = static_cast<int> (m_cache_size - m_max_cache_size * 0.8);
        int delete_depth = m_max_depth_counts;
        int deleted_entries = 0;

        while (deleted_entries < entries_to_delete)
        {
            deleted_entries += m_depth_counts[delete_depth];
            --delete_depth;
        }

        deleted_entries = 0;
        std::map<Hash, CacheEntry>::iterator it = m_cache.begin();
        std::map<Hash, CacheEntry>::const_iterator const end_it = m_cache.end();

        while ( (it != end_it) && (deleted_entries < entries_to_delete) )
        {
            int const depth = it->second.depth();

            if (!it->second.wasCalculated() && (depth >= delete_depth) )
            {
                m_cache.erase(it++);
                --m_depth_counts[depth];
                ++deleted_entries;
            }
            else
            {
                ++it;
            }
        }

        m_cache_size -= deleted_entries;
        entries_to_delete -= deleted_entries;

        delete_depth = m_max_depth_counts;
        deleted_entries = 0;

        while (deleted_entries < entries_to_delete)
        {
            deleted_entries += m_depth_counts[delete_depth];
            --delete_depth;
        }

        deleted_entries = 0;
        it = m_cache.begin();

        while (deleted_entries < entries_to_delete)
        {
            int const depth = it->second.depth();

            if (depth >= delete_depth)
            {
                m_cache.erase(it++);
                --m_depth_counts[depth];
                ++deleted_entries;
            }
            else
            {
                ++it;
            }
        }

        m_cache_size -= deleted_entries;
    }

    m_cache.insert(std::make_pair(hash, CacheEntry(moves_to_solve, depth, calculated)));

    if (depth > m_max_depth_counts)
    {
        m_max_depth_counts = depth;
        m_depth_counts.resize(depth + 1, 0);
    }

    ++m_depth_counts[depth];

    ++m_cache_size;
}


void Solver::setupCache()
{
    std::map<Hash, CacheEntry>::iterator it = m_cache.begin();
    std::map<Hash, CacheEntry>::const_iterator end_it = m_cache.end();
    int deleted = 0;

    while (it != end_it)
    {
        if (!it->second.wasTouched())
        {
            --m_depth_counts[it->second.depth()];
            m_cache.erase(it++);
            ++deleted;
        }
        else
        {
            it->second.untouch();
            ++it;
        }
    }

    m_cache_size -= deleted;
}


std::vector<int> Solver::getGemPositions(Map const & map) const
{
    std::vector<int> result;

    for (int i = 0; i < m_size; ++i)
    {
        if (map.containsGem(i))
        {
            result.push_back(i);
        }
    }

    return result;
}


void Solver::setupDistanceMap()
{
    // This is a BFS

    m_distance_map.reserve(4 * m_goals * m_size);

    for (int i = 0; i < m_goals; ++i)
    {
        std::vector<int> new_distance_map = m_empty_map.getDistanceMap(m_goal_positions[i], s_unsolvable);
        m_distance_map.insert(m_distance_map.end(), new_distance_map.begin(), new_distance_map.end());
    }
}


void Solver::setupDeadlockPatterns()
{
    QString filename = KGlobal::dirs()->findResource("data", "easysok/deadlocks/deadlocks.dat");

    if (!QFile::exists(filename))
    {
        return;
    }

    QFile file(filename);

    if (!file.open(IO_ReadOnly))
    {
        return;
    }

    QTextStream stream(&file);

    QStringList lines;

    while (!stream.atEnd())
    {
        lines.append(stream.readLine());
    }

    QStringList raw_patterns;

    int const deadlocked = 0xff00;
    int const non_keeper_deadlocked = (WALL | OUTSIDE | GEM | GEM_ON_GOAL | EMPTY) << 8;

    while (!lines.isEmpty())
    {
        QStringList pattern_lines;

        int height = 0;
        int width = -1;

        while (!lines.isEmpty() && !lines.first().isEmpty())
        {
            pattern_lines.append(lines.first());
            width = std::max(static_cast<int> (pattern_lines.last().length()), width);
            lines.remove(lines.begin());
            ++height;
        }

        while (!lines.isEmpty() && lines.first().isEmpty())
        {
            lines.remove(lines.begin());
        }

        for (int v = 0; v < 16; ++v)
        {
            // Rotate

            QStringList new_pattern_lines;

            for (int i = 0; i < width; ++i)
            {
                new_pattern_lines.append(QString());

                for (int j = 0; j < height; ++j)
                {
                    new_pattern_lines.last().append(pattern_lines[height - j - 1][i]);
                }
            }

            pattern_lines = new_pattern_lines;
            std::swap(width, height);

            if ((v & 3) == 0)
            {
                // Mirror horizontally

                new_pattern_lines = QStringList();

                for (int i = 0; i < height; ++i)
                {
                    new_pattern_lines.append(pattern_lines[height - i - 1]);
                }

                pattern_lines = new_pattern_lines;
            }

            if ((v & 7) == 0)
            {
                // MIrror vertically

                for (int i = 0; i < height; ++i)
                {
                    QString new_line;

                    for (int j = 0; j < width; ++j)
                    {
                        new_line.append(pattern_lines[i][width - j - 1]);
                    }

                    pattern_lines[i] = new_line;
                }
            }

            QString raw_pattern;
            int ref_x = -1;
            int ref_y = -1;
            std::vector<int> pieces;
            std::vector<int> positions;

            for (int i = 0; i < height; ++i)
            {
                QString const line = pattern_lines[i];
                raw_pattern += line + "\n";

                for (int j = 0; j < width; ++j)
                {
                    if (line[j] == '$')
                    {
                        ref_x = j;
                        ref_y = i;
                    }
                    else
                    {
                        if (line[j] == ' ')
                        {
                            pieces.push_back(EMPTY | (EMPTY << 8));
                            positions.push_back(j + m_width * i);
                        }
                        else if (line[j] == '#')
                        {
                            pieces.push_back(WALL | OUTSIDE);
                            positions.push_back(j + m_width * i);
                        }
                        else if (line[j] == '%')
                        {
                            pieces.push_back(WALL | OUTSIDE | GEM | GEM_ON_GOAL );
                            positions.push_back(j + m_width * i);
                        }
                        else if (line[j] == '*')
                        {
                            pieces.push_back(WALL | OUTSIDE | deadlocked);
                            positions.push_back(j + m_width * i);
                        }
                        else if (line[j] == '+')
                        {
                            pieces.push_back(WALL | OUTSIDE | GEM | GEM_ON_GOAL | deadlocked);
                            positions.push_back(j + m_width * i);
                        }
                        else if (line[j] == 'X')
                        {
                            pieces.push_back(WALL | OUTSIDE | GEM_ON_GOAL | GEM |
                                             EMPTY | non_keeper_deadlocked);
                            positions.push_back(j + m_width * i);
                        }
                        else
                        {
                            assert(line[j] == '?');
                        }
                    }

                }
            }

            if (raw_patterns.contains(raw_pattern) != 0)
            {
                continue;
            }

            raw_patterns.append(raw_pattern);

            int const offset = ref_x + m_width * ref_y;
            int const length = static_cast<int> (pieces.size());
            m_deadlock_pattern_lengths.push_back(length);

            for (int i = 0; i < length; ++i)
            {
                m_deadlock_pattern_positions.push_back(positions[i] - offset);
                m_deadlock_patterns.push_back(pieces[i]);
            }
        }
    }
}


Movements Solver::getFullMoves() const
{
    Movements result;
    int keeper = m_start_map.getIndex(m_start_map.keeper());
    std::vector<int> gem_positions = getGemPositions(m_start_map);
    int position = 0;

    for (int i = 0; i < m_depth; ++i)
    {
        position += m_pos[i];
        int const raw_move = m_moves[position];
        int const moved_gem = raw_move >> 2;
        int const direction = raw_move & 3;
        int const old_pos = gem_positions[moved_gem];
        int const new_pos = old_pos + m_xy_offsets[direction];
        int const from_keeper = old_pos - m_xy_offsets[direction];

        if (keeper != from_keeper)
        {
            result.addMove(Move(m_start_map.getPoint(keeper), m_start_map.getPoint(from_keeper), false));
        }

        keeper = old_pos;
        result.addMove(Move(m_start_map.getPoint(from_keeper), m_start_map.getPoint(keeper), true));
        gem_positions[moved_gem] = new_pos;

        position += m_move_lengths[i] - m_pos[i];
    }

    result = m_start_map.expandMoves(result, false);
    result.setToFirstPosition();

    return result;
}


Solver::CacheEntry::CacheEntry(int moves_to_solve, int depth, bool calculated) :
    m_data(moves_to_solve + (depth << 14) + (calculated << 28) + (1 << 29))
{
    assert(moves_to_solve >= 0);
    assert(moves_to_solve < 0x4000);
    assert(depth >= 0);
    assert(depth < 0x4000);
}


int Solver::CacheEntry::movesToSolve() const
{
    return m_data & 0x00003fff;
}


bool Solver::CacheEntry::unsolvable() const
{
    return movesToSolve() >= s_unsolvable;
}


void Solver::CacheEntry::setMovesToSolve(int moves_to_solve)
{
    assert(moves_to_solve >= 0);
    assert(moves_to_solve < 0x4000);

    m_data &= 0xefffc000;      // Erase calculation flag.
    m_data |= moves_to_solve;
}


int Solver::CacheEntry::depth() const
{
    return (m_data & 0x0fffc000) >> 14;
}


void Solver::CacheEntry::setDepth(int new_depth)
{
    assert(new_depth >= 0);
    assert(new_depth < 0x4000);

    m_data &= 0xf0003fff;
    m_data |= (new_depth << 14);
}


bool Solver::CacheEntry::wasCalculated() const
{
    return m_data & 0x10000000;
}


bool Solver::CacheEntry::wasTouched() const
{
    return m_data & 0x20000000;
}


void Solver::CacheEntry::touch()
{
    m_data |= 0x20000000;
}


void Solver::CacheEntry::untouch()
{
    m_data &= 0xdfffffff;
}
