/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_SOLVER_INC_GUARD_H
#define EASYSOK_SOLVER_INC_GUARD_H


#include <map>
#include <utility>
#include <vector>

#include "hash.h"
#include "map.h"
#include "move.h"
#include "movements.h"




/**
 * Used for solving a map.
 *
 * Uses IDA* and some simple enhancements.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class Solver
{

public:

    /**
     * Constructs the solver for the given map.
     *
     * @param map The map to solve.
     * @param cache_size The number of map hashs to put in the cache.
     */

    Solver(Map const & map, int cache_size);


    /**
     * Sets up the solver.
     *
     * You need to call this function as long as it does not returns true.
     */

    bool prepare();


    /**
     * Tries to solve the map.
     *
     * You need to call this function as long as it does not returns true.
     *
     * @param steps The maximum number of steps to try.
     * @return true If a solution or the proof of unsolvability could be found in the number
     *              of steps.
     */

    bool solve(int steps);


    /**
     * Returns the actual search depth.
     */

    int depth() const;


    /**
     * Returns the actual min depth for the last call to solve().
     */

    int actMinDepth() const;


    /**
     * Returns the actual max depth for the last call to solve().
     */

    int actMaxDepth() const;


    /**
     * Returns the max search depth reached.
     */

    int maxDepthReached() const;


    /**
     * Returns the actual maximum search depth.
     */

    int maxDepth() const;


    /**
     * Returns the next max depth for the next iteration.
     */

    int nextMaxDepth() const;


    /**
     * Returns the percentage (0 to 100) of the work in the actual iteration completed.
     */

    double percentageCompleted() const;


    /**
     * Returns the moves of the solution or and empty move list if it is not solvable.
     */

    Movements moves() const;


    /**
     * Returns the moves of the best current effort.
     */

    Movements bestEffort() const;


private:

    /**
     * Makes a single solve step.
     *
     * @return True, if the map is solved (the movements are then stored in m_solution_moves)
     *         or is proved to be unsolvable.
     */

    bool doSingleStep();


    /**
     * Starts the search with a new limit.
     *
     * @return True, if no solution could be found.
     */

    bool startSearch();


    /**
     * Callapses to an old depth.
     */

    bool collapse();


    /**
     * Expands to a new depth.
     */

    void expand();


    /**
     * Searche the end nodes.
     *
     * @return True, if an solution was found.
     */

    bool processEndNodes();


    /**
     * Returns the valid moves (menaing movements of gems) in compressed format .
     */

    std::vector<int> const & validMoves() const;


    /**
     * Returns the minimum number of moves to move a gem from pos1 to pos2.
     *
     * If there is no way from pos1 to pos2 the return -1.
     *
     * @param keeper_pos The position of the keeper.
     * @param gem_pos The start position of the gem as an index in m_gem_positions.
     * @param goal_pos The end position of the goal as an index in m_goal_positions.
     */

    int movesForGem(int keeper_pos, int gem_pos, int goal_pos) const;


    /**
     * Returns true, if the gem at the given position is deadlocked.
     *
     * @param map The map to use.
     * @param position The position of the gem.
     * @param ignore_on_goal If true, we count the gem as deadlocked even if it is on a goal.
     */

    bool isDeadlock(int position, bool ignore_on_goal = false) const;


    /**
     * Returns the minimum number of atomic gem moves to solve a map.
     *
     * If there are no possible moves, we return -1.
     *
     * @param map The map to use.
     * @param moved_gem_pos The end position of the last moved gem.
     */

    int minMovesForSolution(int moved_gem_pos) const;


    /**
     * Utility function used to find the minimum number of moves.
     */

    int assignmentSolver(std::vector<int> & matrix, int dim) const;


    /**
     * Returns the lower bound for the map.
     *
     * Uses the cache to get better results.
     *
     * @param map The map to use.
     * @param hash The hash of the map.
     * @param moved_gem_pos The end position of the last moved gem.
     * @param depth The depth of the search tree where we are.
     */

    int lowerBound(Hash const & hash, int moved_gem_pos, int depth);


    /**
     * Updates the cache entry for the given map.
     *
     * @param hash The hash of the map.
     * @param moves_to_solve The minimum number of moves to solve. This entry will be updated,
     *                       if it is greater than the previous value.
     * @param depth The depth of the search tree where we are.
     */

    void updateCache(Hash const & hash, int moves_to_solve, int depth);


    /**
     * Inserts the given entry in the cache and removes som elements of the cache if it is full.
     *
     * @param hash The hash value of the map.
     * @param moves_to_solve The minimum number of moves to solve. If the value is negative, the
     *                       entries is considered to be calculatable.
     * @param depth The depth of the search tree where we are.
     * @param calculated If true, the number of moves is calculated.
     */

    void insertInCache(Hash const & hash, int moves_to_solve, int depth, bool calculated);


    /**
     * Updates the cache after starting a new search.
     */

    void setupCache();


    /**
     * Returns a vector holding the positions (sorted) of all gems.
     *
     * @param map The map to use.
     */

    std::vector<int> getGemPositions(Map const & map) const;


    /**
     * Sets up the distance map.
     */

    void setupDistanceMap();


    /**
     * Sets up the deadlock patterns.
     */

    void setupDeadlockPatterns();


    /**
     * Returns the movements (including keeper moves) to get to the current map.
     */

    Movements getFullMoves() const;


    /**
     * Used to define entries in the cache.
     */

    class CacheEntry
    {

    public:

        /**
         * Constructs the cache entry.
         *
         * The new entrie is considered touched.
         *
         * @param moves_to_solve The number of moves to solve.
         * @param depth The depth of the corresponding map.
         * @param calculated If true, the entrie is directly calculated.
         */

        CacheEntry(int moves_to_solve, int depth, bool calculated);


        /**
         * Returns the number of moves to solve.
         */

        int movesToSolve() const;


        /**
         * Returns true, if the entry is unsolvable.
         */

        bool unsolvable() const;


        /**
         * Sets a new number of moves to solve.
         *
         * The entries is now considered to be not calculated.
         *
         * @param moves_to_solve The number of moves to solve.
         */

        void setMovesToSolve(int moves_to_solve);


        /**
         * Returns the depth.
         */

        int depth() const;


        /**
         * Sets a new depth.
         *
         * @param new_depth The depth of the corresponding map.
         */

        void setDepth(int new_depth);


        /**
         * Returns true, if the entry is calculated.
         */

        bool wasCalculated() const;


        /**
         * Returns true, if the entry was touched.
         */

        bool wasTouched() const;


        /**
         * Touches the entry.
         */

        void touch();


        /**
         * Untouches the entry.
         */

        void untouch();


    private:

        /**
         * Here we store the data.
         */

        Q_UINT32 m_data;
    };


    /**
     * The act map.
     */

    Map m_map;


    /**
     * The start map.
     */

    Map m_start_map;


    /**
     * The start map without gems, goals and keeper.
     */

    Map m_empty_map;


    /**
     * The move positions.
     */

    std::vector<int> m_pos;


    /**
     * The movement lengths.
     */

    std::vector<int> m_move_lengths;


    /**
     * The move offsets.
     */

    std::vector<int> m_move_offsets;


    /**
     * The minimum moves to solve the map.
     */

    std::vector<int> m_min_solve_moves;


    /**
     * The valid moves in a compressed format.
     */

    std::vector<int> m_moves;


    /**
     * The actual positions of the gems.
     */

    std::vector<int> m_gem_positions;


    /**
     * The solution moves.
     */

    Movements m_solution_moves;


    /**
     * The best effort moves.
     */

    Movements m_best_effort;


    /**
     * The max depth for the best effort.
     */

    int m_best_effort_max_depth;


    /**
     * The min lower bound for the best effort.
     */

    int m_best_effort_lower_bound;


    /**
     * Here we store the indices of positions where a moveable gem might reside.
     */

    std::vector<int> m_possible_gem_positions;


    /**
     * Here we store the indices of positions with a goal.
     */

    std::vector<int> m_goal_positions;


    /**
     * Here we store the hashs for the maps.
     */

    std::vector<Hash> m_hashs;


    /**
     * The number of goals.
     */

    int m_goals;


    /**
     * The actual depth.
     */

    int m_depth;


    /**
     * The actual min depth in the last call to solve.
     */

    int m_act_min_depth;


    /**
     * The actual max depth in the last call to solve.
     */

    int m_act_max_depth;


    /**
     * The maximum depth reached.
     */

    int m_max_depth_reached;


    /**
     * The actual max depth.
     */

    int m_max_depth;


    /**
     * The width of the map.
     */

    int const m_width;


    /**
     * The height of the map.
     */

    int const m_height;


    /**
     * m_width * m_height.
     */

    int const m_size;


    /**
     * The offsets for atomic moves.
     */

    int m_xy_offsets[4];


    /**
     * The current mimimum distance to a solution at the current depth.
     */

    int m_min_add_depth;


   /**
    * True, if at least one move seems to lead to a solution.
    */

    bool m_solution_possible;


    /**
     * The map used to find cycles in the move graph.
     */

    std::map<Hash, CacheEntry> m_cache;


    /**
     * The maximum cache size.
     */

    int m_max_cache_size;


    /**
     * The actual cache size.
     */

    int m_cache_size;


    /**
     * Here we store the number of entries of a given depth.
     */

    std::vector<int> m_depth_counts;


    /**
     * Here we store the max depth for m_depth_counts.
     */

    int m_max_depth_counts;


    /**
     * Here we store the distance map.
     */

    std::vector<int> m_distance_map;


    /**
     * Here we store the the reachable fields for gems.
     */

    std::vector<Q_UINT32> m_reachable_map;


    /**
     * Here we store the current index of the gem position for which we calc the reachable map.
     */

    int m_reachable_map_gem;


    /**
     * If true, no further calls ot solve are needed.
     */

    bool m_solve_finished;


    /**
     * Used to define the patterns.
     */

    enum PatternType
    {
        KEEPER = 1 << 0,
        KEEPER_ON_GOAL = 1 << 1,
        GEM = 1 << 2,
        GEM_ON_GOAL = 1 << 3,
        EMPTY = 1 << 4,
        GOAL = 1 << 5,
        WALL = 1 << 6,
        OUTSIDE = 1 << 7
    };


    /**
     * Here we store the deadlock patterns.
     */

    std::vector<int> m_deadlock_patterns;


    /**
     * Here we store the deadlock pattern positions.
     */

    std::vector<int> m_deadlock_pattern_positions;


    /**
     * Here we store the length of the patterns.
     */

    std::vector<int> m_deadlock_pattern_lengths;


    /**
     * The number of pushes for unsolvable positions.
     */

    static int const s_unsolvable;
};


#endif
