/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "solver_dialog.h"

#include <qtimer.h>

#include <kapp.h>
#include <kconfig.h>
#include <klocale.h>

#include "solver.h"




SolverDialog::SolverDialog(Map const & map, QWidget * parent, char const * name) :
    QMessageBox(i18n("Solver"), i18n("Preparing solver"), Information, Cancel | Default, NoButton, NoButton,
                parent, name, true),
    m_steps_rest(0),
    m_steps_million(0)
{
    KConfig * config = KApplication::kApplication()->config();
    config->setGroup("");

    m_positions_per_step = config->readNumEntry("Solver steps per call", 1000);
    int cache_size = config->readNumEntry("Solver cache size", 100000);
    cache_size = std::min(std::max(cache_size, 1000), 10000000);

    m_solver = new Solver(map, cache_size);

    m_timer = new QTimer(this);
    connect(m_timer, SIGNAL(timeout()), this, SLOT(prepareSolver()));

    m_timer->start(100, true); // 0 may prevent the dialog from appearing correctly
}


SolverDialog::~SolverDialog()
{
    delete m_solver;
}


Solver const * SolverDialog::solver() const
{
    return m_solver;
}


QString SolverDialog::positionsExaminedAsText() const
{
    QString text;

    if (m_steps_million < 10)
    {
        int const thousands = m_steps_million * 1000 + m_steps_rest / 1000;
        text = i18n("%1 thousand").arg(thousands);
    }
    else
    {
        text = i18n("%1 million").arg(m_steps_million);
    }

    return text;
}


void SolverDialog::prepareSolver()
{
    if (m_solver->prepare())
    {
        disconnect(m_timer, SIGNAL(timeout()), this, SLOT(prepareSolver()));
        connect(m_timer, SIGNAL(timeout()), this, SLOT(performSolve()));
    }

    m_timer->start(0, true);
}


void SolverDialog::performSolve()
{
    m_steps_rest += m_positions_per_step;

    while (m_steps_rest >= 1000000)
    {
        m_steps_rest -= 1000000;
        ++m_steps_million;
    }


    if (m_solver->solve(m_positions_per_step))
    {
        accept();
    }
    else
    {
        QString text = i18n("Actual maximum search depth: %1\n").arg(m_solver->maxDepth());
        text += i18n("Actual minimum search depth: %1\n").arg(m_solver->actMinDepth());
        text += i18n("Actual maximum search depth: %1\n").arg(m_solver->actMaxDepth());
        text += i18n("Maximum search depth reached in iteration: %1\n").arg(m_solver->maxDepthReached());
        text += "Examined positions: " + positionsExaminedAsText();

        setText(text);

        m_timer->start(0, true);

        adjustSize();
    }
}
