/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "theme.h"

#include <algorithm>
#include <cassert>
#include <cstdlib>

#include <qdom.h>
#include <qpoint.h>
#include <qstringlist.h>

#include <klocale.h>
#include <kmessagebox.h>

#include "dom_helper.h"
#include "map.h"




char const * Theme::s_elements[33] =
{
    "LeftKeeperOnFloor",
    "RightKeeperOnFloor",
    "UpKeeperOnFloor",
    "DownKeeperOnFloor",
    "LeftKeeperOnDeadFloor",
    "RightKeeperOnDeadFloor",
    "UpKeeperOnDeadFloor",
    "DownKeeperOnDeadFloor",
    "LeftKeeperOnGoal",
    "RightKeeperOnGoal",
    "UpKeeperOnGoal",
    "DownKeeperOnGoal",
    "LeftMovingKeeper",
    "RightMovingKeeper",
    "UpMovingKeeper",
    "DownMovingKeeper",
    "LeftVirtualKeeper",
    "RightVirtualKeeper",
    "UpVirtualKeeper",
    "DownVirtualKeeper",
    "GemOnFloor",
    "GemOnDeadFloor",
    "GemOnGoal",
    "MovingGem",
    "Floor",
    "DeadFloor",
    "Goal",
    "LeftArrow",
    "RightArrow",
    "UpArrow",
    "DownArrow",
    "Wall",
    "Outside"
};


char const * Theme::s_short_elements[33] =
{
    "KeeperOnFloor",
    "KeeperOnFloor",
    "KeeperOnFloor",
    "KeeperOnFloor",
    "KeeperOnDeadFloor",
    "KeeperOnDeadFloor",
    "KeeperOnDeadFloor",
    "KeeperOnDeadFloor",
    "KeeperOnGoal",
    "KeeperOnGoal",
    "KeeperOnGoal",
    "KeeperOnGoal",
    "MovingKeeper",
    "MovingKeeper",
    "MovingKeeper",
    "MovingKeeper",
    "VirtualKeeper",
    "VirtualKeeper",
    "VirtualKeeper",
    "VirtualKeeper",
    "GemOnFloor",
    "GemOnDeadFloor",
    "GemOnGoal",
    "MovingGem",
    "Floor",
    "DeadFloor",
    "Goal",
    "LeftArrow",
    "RightArrow",
    "UpArrow",
    "DownArrow",
    "Wall",
    "Outside"
};

int const Theme::s_types_for_hidden_gems[33] =
{
    LEFT_KEEPER_ON_FLOOR,
    RIGHT_KEEPER_ON_FLOOR,
    UP_KEEPER_ON_FLOOR,
    DOWN_KEEPER_ON_FLOOR,
    LEFT_KEPPER_ON_DEAD_FLOOR,
    RIGHT_KEPPER_ON_DEAD_FLOOR,
    UP_KEPPER_ON_DEAD_FLOOR,
    DOWN_KEPPER_ON_DEAD_FLOOR,
    LEFT_KEEPER_ON_GOAL,
    RIGHT_KEEPER_ON_GOAL,
    UP_KEEPER_ON_GOAL,
    DOWN_KEEPER_ON_GOAL,
    LEFT_MOVING_KEEPER,
    RIGHT_MOVING_KEEPER,
    UP_MOVING_KEEPER,
    DOWN_MOVING_KEEPER,
    LEFT_VIRTUAL_KEEPER,
    RIGHT_VIRTUAL_KEEPER,
    UP_VIRTUAL_KEEPER,
    DOWN_VIRTUAL_KEEPER,
    EMPTY_FLOOR,
    EMPTY_DEAD_FLOOR,
    EMPTY_GOAL,
    MOVING_GEM,
    EMPTY_FLOOR,
    EMPTY_DEAD_FLOOR,
    EMPTY_GOAL,
    LEFT_ARROW,
    RIGHT_ARROW,
    UP_ARROW,
    DOWN_ARROW,
    WALL,
    OUTSIDE
};

int const Theme::s_types_for_hidden_goals[33] =
{
    LEFT_KEEPER_ON_FLOOR,
    RIGHT_KEEPER_ON_FLOOR,
    UP_KEEPER_ON_FLOOR,
    DOWN_KEEPER_ON_FLOOR,
    LEFT_KEPPER_ON_DEAD_FLOOR,
    RIGHT_KEPPER_ON_DEAD_FLOOR,
    UP_KEPPER_ON_DEAD_FLOOR,
    DOWN_KEPPER_ON_DEAD_FLOOR,
    LEFT_KEEPER_ON_FLOOR,
    RIGHT_KEEPER_ON_FLOOR,
    UP_KEEPER_ON_FLOOR,
    DOWN_KEEPER_ON_FLOOR,
    LEFT_MOVING_KEEPER,
    RIGHT_MOVING_KEEPER,
    UP_MOVING_KEEPER,
    DOWN_MOVING_KEEPER,
    LEFT_VIRTUAL_KEEPER,
    RIGHT_VIRTUAL_KEEPER,
    UP_VIRTUAL_KEEPER,
    DOWN_VIRTUAL_KEEPER,
    GEM_ON_FLOOR,
    GEM_ON_DEAD_FLOOR,
    GEM_ON_FLOOR,
    MOVING_GEM,
    EMPTY_FLOOR,
    EMPTY_DEAD_FLOOR,
    EMPTY_FLOOR,
    LEFT_ARROW,
    RIGHT_ARROW,
    UP_ARROW,
    DOWN_ARROW,
    WALL,
    OUTSIDE
};


Theme::Theme(QDomElement const & dom_element) :
    m_hide_gems(false),
    m_hide_goals(false),
    m_outside_as_wall(false)
{
    assert(dom_element.tagName() == "EasySokTheme");

    m_left_border = DomHelper::getDouble(dom_element, "leftborder", 0.0);
    m_right_border = DomHelper::getDouble(dom_element, "rightborder", 0.0);
    m_upper_border = DomHelper::getDouble(dom_element, "upperborder", 0.0);
    m_lower_border = DomHelper::getDouble(dom_element, "lowerborder", 0.0);

    m_name = dom_element.attribute("name", i18n("unknown"));
    m_background_image = dom_element.attribute("image", "");
    m_background_color = DomHelper::getColor(dom_element);

    QDomNodeList childs = dom_element.childNodes();
    int const nr_of_childs = static_cast<int> (childs.count());
    int index = 0;

    for (int i = 0; i < nr_of_childs; ++i)
    {
        QDomNode node = childs.item(i);

        if (node.isElement())
        {
            QDomElement element = node.toElement();

            if (element.tagName() == s_elements[index])
            {
                addAlternates(element);
                ++index;
            }
            else if (element.tagName() == s_short_elements[index])
            {
                for (int j = 0; j < 4; ++j)
                {
                    addAlternates(element);
                }

                index += 4;
            }
            else
            {
                assert(false);
            }
        }
        else
        {
            assert(false);
        }
    }

    assert(index == 33);
}


int Theme::suggestFieldSize() const
{
    int max_size = 0;
    int const nr_of_piece_images = nrOfPieceImages();

    for (int i = 0; i < nr_of_piece_images; ++i)
    {
        max_size = std::max(max_size, pieceImage(i).size());
    }

    return max_size;
}


bool Theme::hideGems() const
{
    return m_hide_gems;
}


void Theme::setHideGems(bool hide)
{
    m_hide_gems = hide;
}


bool Theme::hideGoals() const
{
    return m_hide_goals;
}


void Theme::setHideGoals(bool hide)
{
    m_hide_goals = hide;
}


bool Theme::outsideAsWall() const
{
    return m_outside_as_wall;
}


void Theme::setOutsideAsWall(bool outside_as_wall)
{
    m_outside_as_wall = outside_as_wall;
}


QString const & Theme::name() const
{
    return m_name;
}


QString const & Theme::backgroundImage() const
{
    return m_background_image;
}


QColor Theme::backgroundColor() const
{
    return m_background_color;
}


std::vector<int> const & Theme::imageIndices(ImageType type, QPoint const & position, Map const & map) const
{
    static std::vector<int> indices;
    indices.resize(0);

    int real_type = type;

    if (m_hide_goals)
    {
        real_type = s_types_for_hidden_goals[real_type];
    }

    if (m_hide_gems)
    {
        real_type = s_types_for_hidden_gems[real_type];
    }

    if (m_outside_as_wall && (real_type == OUTSIDE))
    {
        real_type = WALL;
    }

    int pattern;

    if ( (real_type != WALL) && (real_type != OUTSIDE) )
    {
        pattern = getNonWallPattern(position, map);
    }
    else
    {
        pattern = getWallPattern(position, map);
    }

    int const nr_of_alternates = m_alternate_sizes[real_type];
    int const alternate_offset = m_alternate_starts[real_type];

    for (int i = 0; i < nr_of_alternates; ++i)
    {
        if ((m_patterns[i + alternate_offset] & pattern) == pattern)
        {
            int nr_of_piece_images = m_alternative_piece_image_nr[i + alternate_offset];
            int piece_offset = m_alternative_piece_image_starts[i + alternate_offset];
            int raw_index = ((position.x() + position.y()) % nr_of_piece_images) + piece_offset;

            indices.push_back(m_piece_image_indices[raw_index]);
        }
    }

    return indices;
}


std::vector<int> const & Theme::imageIndicesFromPosition(int direction, QPoint const & position,
                                                         Map const & map) const
{
    static std::vector<int> empty_indices;

    int const piece = map.getPiece(position);
    int const is_crossed = map.isCrossed(position);
    int const piece_and_crossed = piece + (is_crossed << 3);

    switch (piece_and_crossed)
    {
    case Map::KEEPER:
        return imageIndices(static_cast<ImageType> (Theme::LEFT_KEEPER_ON_FLOOR + direction), position, map);
        break;

    case Map::KEEPER + 8:
        return imageIndices(static_cast<ImageType> (Theme::LEFT_KEPPER_ON_DEAD_FLOOR + direction),
                            position, map);
        break;

    case Map::KEEPER_ON_GOAL:
        return imageIndices(static_cast<ImageType> (Theme::LEFT_KEEPER_ON_GOAL + direction), position, map);
        break;

    case Map::GEM:
        return imageIndices(Theme::GEM_ON_FLOOR, position, map);
        break;

    case Map::GEM + 8:
        return imageIndices(Theme::GEM_ON_DEAD_FLOOR, position, map);
        break;

    case Map::GEM_ON_GOAL:
        return imageIndices(Theme::GEM_ON_GOAL, position, map);
        break;

    case Map::EMPTY:
        return imageIndices(Theme::EMPTY_FLOOR, position, map);
        break;

    case Map::EMPTY + 8:
        return imageIndices(Theme::EMPTY_DEAD_FLOOR, position, map);
        break;

    case Map::GOAL:
        return imageIndices(Theme::EMPTY_GOAL, position, map);
        break;

    case Map::WALL:
        return imageIndices(Theme::WALL, position, map);
        break;

    case Map::OUTSIDE:
        return imageIndices(Theme::OUTSIDE, position, map);
        break;

    default:
        return empty_indices;
        break;
    }
}


PieceImage const & Theme::pieceImage(int index) const
{
    assert(index >= 0);
    assert(index < static_cast<int> (m_piece_images.size()));

    return m_piece_images[index];
}


int Theme::nrOfPieceImages() const
{
    return static_cast<int> (m_piece_images.size());
}


double Theme::upperBorder() const
{
    return m_upper_border;
}


double Theme::lowerBorder() const
{
    return m_lower_border;
}


double Theme::leftBorder() const
{
    return m_left_border;
}


double Theme::rightBorder() const
{
    return m_right_border;
}


void Theme::addAlternates(QDomElement const & dom_element)
{
    QDomNodeList childs = dom_element.childNodes();
    int const nr_of_childs = static_cast<int> (childs.count());

    m_alternate_starts.push_back(static_cast<int> (m_patterns.size()));
    m_alternate_sizes.push_back(nr_of_childs);

    for (int i = 0; i < nr_of_childs; ++i)
    {
        QDomNode node = childs.item(i);

        if (node.isElement())
        {
            QDomElement element = node.toElement();

            if (element.tagName() == "PieceImageAlternates")
            {
                addPieceImageAlternates(element);
            }
            else if (element.tagName() == "WallPieceAlternates")
            {
                addWallPieceAlternates(element);
            }
            else
            {
                assert(false);
            }
        }
        else
        {
            assert(false);
        }
    }
}


void Theme::addPieceImageAlternates(QDomElement const & dom_element)
{
    assert(dom_element.tagName() == "PieceImageAlternates");

    int pattern = patternForPieceImage(dom_element, "upperleft");
    pattern <<= 3;
    pattern += patternForPieceImage(dom_element, "upper");
    pattern <<= 3;
    pattern += patternForPieceImage(dom_element, "upperright");
    pattern <<= 3;
    pattern += patternForPieceImage(dom_element, "left");
    pattern <<= 3;
    pattern += patternForPieceImage(dom_element, "right");
    pattern <<= 3;
    pattern += patternForPieceImage(dom_element, "lowerleft");
    pattern <<= 3;
    pattern += patternForPieceImage(dom_element, "lower");
    pattern <<= 3;
    pattern += patternForPieceImage(dom_element, "lowerright");

    m_patterns.push_back(pattern);

    addAlternatesImages(dom_element);
}


void Theme::addWallPieceAlternates(QDomElement const & dom_element)
{
    assert(dom_element.tagName() == "WallPieceAlternates");

    int pattern = patternForWallPiece(dom_element, "upperleft");
    pattern <<= 3;
    pattern += patternForWallPiece(dom_element, "upper");
    pattern <<= 3;
    pattern += patternForWallPiece(dom_element, "upperright");
    pattern <<= 3;
    pattern += patternForWallPiece(dom_element, "left");
    pattern <<= 3;
    pattern += patternForWallPiece(dom_element, "right");
    pattern <<= 3;
    pattern += patternForWallPiece(dom_element, "lowerleft");
    pattern <<= 3;
    pattern += patternForWallPiece(dom_element, "lower");
    pattern <<= 3;
    pattern += patternForWallPiece(dom_element, "lowerright");

    m_patterns.push_back(pattern);

    addAlternatesImages(dom_element);
}


void Theme::addAlternatesImages(QDomElement const & dom_element)
{
    QDomNodeList childs = dom_element.childNodes();
    int const nr_of_childs = static_cast<int> (childs.count());

    assert(nr_of_childs > 0);

    m_alternative_piece_image_starts.push_back(static_cast<int> (m_piece_image_indices.size()));
    m_alternative_piece_image_nr.push_back(nr_of_childs);

    for (int i = 0; i < nr_of_childs; ++i)
    {
        QDomNode node = childs.item(i);

        if (node.isElement())
        {
            QDomElement element = node.toElement();

            if (element.tagName() == "PieceImage")
            {
                PieceImage piece_image = PieceImage(element);
                int const nr_of_piece_images = static_cast<int> (m_piece_images.size());
                bool already_there = false;

                for (int j = 0; j < nr_of_piece_images; ++j)
                {
                    if (m_piece_images[j] == piece_image)
                    {
                        m_piece_image_indices.push_back(j);
                        already_there = true;

                        break;
                    }
                }

                if (!already_there)
                {
                    m_piece_image_indices.push_back(static_cast<int> (m_piece_images.size()));
                    m_piece_images.push_back(piece_image);
                }
            }
            else
            {
                assert(false);
            }
        }
        else
        {
            assert(false);
        }
    }
}


int Theme::patternForPieceImage(QDomElement const & dom_element, QString const & name)
{
    QString val = dom_element.attribute(name, "");

    int pattern = 7;

    if (val == "wall")
    {
        pattern = 1;
    }
    else if (val == "empty")
    {
        pattern = 2;
    }
    else if (val == "goal")
    {
        pattern = 4;
    }
    else if (val == "nonwall")
    {
        pattern = 6;
    }
    else if (val == "nonempty")
    {
        pattern = 5;
    }
    else if (val == "nongoal")
    {
        pattern = 3;
    }

    return pattern;
}


int Theme::patternForWallPiece(QDomElement const & dom_element, QString const & name)
{
    QString val = dom_element.attribute(name, "");

    int pattern = 7;

    if (val == "wall")
    {
        pattern = 1;
    }
    else if (val == "inside")
    {
        pattern = 2;
    }
    else if (val == "outside")
    {
        pattern = 4;
    }
    else if (val == "nonwall")
    {
        pattern = 6;
    }
    else if (val == "noninside")
    {
        pattern = 5;
    }
    else if (val == "nonoutside")
    {
        pattern = 3;
    }
    else if ( (val != "") && (val != "all") )
    {
        assert(false);
    }

    return pattern;
}


int Theme::getNonWallPattern(QPoint const & position, Map const & map) const
{
    int const with = map.width();
    int const height = map.height();

    int pattern = 0;

    for (int y = -1; y <= 1; ++y)
    {
        int const ry = position.y() + y;

        for (int x = -1; x <= 1; ++x)
        {
            if ( (x == 0) && (y == 0) )
            {
                continue;
            }

            pattern <<= 3;

            int const rx = position.x() + x;

            if ( (rx >= 0) && (ry >= 0) && (rx < with) && (ry < height) )
            {
                int const piece = map.getPiece(QPoint(rx, ry));

                if (m_hide_goals)
                {
                    switch(piece)
                    {
                    case Map::KEEPER:
                    case Map::GEM:
                    case Map::EMPTY:
                    case Map::KEEPER_ON_GOAL:
                    case Map::GEM_ON_GOAL:
                    case Map::GOAL:
                        pattern += 2;
                        break;

                    default:
                        pattern += 1;
                        break;
                    }
                }
                else
                {
                    switch(piece)
                    {
                    case Map::KEEPER:
                    case Map::GEM:
                    case Map::EMPTY:
                        pattern += 2;
                        break;

                    case Map::KEEPER_ON_GOAL:
                    case Map::GEM_ON_GOAL:
                    case Map::GOAL:
                        pattern += 4;
                        break;

                    default:
                        pattern += 1;
                        break;
                    }
                }
            }
            else
            {
                pattern += 1;
            }
        }
    }

    return pattern;
}


int Theme::getWallPattern(QPoint const & position, Map const & map) const
{
    int const with = map.width();
    int const height = map.height();

    int pattern = 0;

    for (int y = -1; y <= 1; ++y)
    {
        int const ry = position.y() + y;

        for (int x = -1; x <= 1; ++x)
        {
            if ( (x == 0) && (y == 0) )
            {
                continue;
            }

            pattern <<= 3;

            int const rx = position.x() + x;

            if ( (rx >= 0) && (ry >= 0) && (rx < with) && (ry < height) )
            {
                int const piece = map.getPiece(QPoint(rx, ry));

                if (m_outside_as_wall)
                {
                    switch(piece)
                    {
                    case Map::WALL:
                    case Map::OUTSIDE:
                        pattern += 1;
                        break;

                    default:
                        pattern += 2;
                        break;
                    }
                }
                else
                {
                    switch(piece)
                    {
                    case Map::WALL:
                        pattern += 1;
                        break;

                    case Map::OUTSIDE:
                        pattern += 4;
                        break;

                    default:
                        pattern += 2;
                        break;
                    }
                }
            }
            else
            {
                pattern += 4;
            }
        }
    }

    return pattern;
}
