/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#ifndef EASYSOK_THEME_INC_GUARD_H
#define EASYSOK_THEME_INC_GUARD_H


#include <vector>

#include <qcolor.h>
#include <qstring.h>

#include "piece_image.h"


class Map;
class QDomElement;
class QPoint;




/**
 * This class represents a theme.
 *
 * @author Ralf Schmelter (ralf.schmelter@t-online.de).
 * @version 0.2
 */

class Theme
{

public:

    /**
     * Creates a theme from an dom element.
     *
     * @param dom_element The dom element to use.
     */

    Theme(QDomElement const & dom_element);


    /**
     * Suggest a size of a field.
     */

    int suggestFieldSize() const;


    /**
     * Returns true, if we hide the gems.
     */

    bool hideGems() const;


    /**
     * Hides all gems.
     *
     * @param hide If true, all gems are hidden.
     */

    void setHideGems(bool hide);


    /**
     * Returns true, if we hide the goals.
     */

    bool hideGoals() const;


    /**
     * Hides all goals.
     *
     * @param hide If true, all goals are hidden.
     */

    void setHideGoals(bool hide);


    /**
     * Returns true if outside fields are treated as wall.
     */

    bool outsideAsWall() const;


    /**
     * Sets if outside fields are treated as wall.
     *
     * @param outside_as_wall If true, outside fields are treated as wall.
     */

    void setOutsideAsWall(bool outside_as_wall);


    /**
     * Returns the name of the theme.
     */

    QString const & name() const;


    /**
     * Returns the name background image or an empty string if we dont use one.
     */

    QString const & backgroundImage() const;


    /**
     * Returns the background color, which will be used if we don't have a background
     * image or it could not be loaded.
     */

    QColor backgroundColor() const;


    /**
     * Defines the image types.
     */

    enum ImageType
    {
        /**
         * A keeper looking left on a non dead floor.
         */

        LEFT_KEEPER_ON_FLOOR,


        /**
         * A keeper looking right on a non dead floor.
         */

        RIGHT_KEEPER_ON_FLOOR,


        /**
         * A keeper looking up on a non dead floor.
         */

        UP_KEEPER_ON_FLOOR,


        /**
         * A keeper looking down on a non dead floor.
         */

        DOWN_KEEPER_ON_FLOOR,


        /**
         * A keeper looking left on a dead floor.
         */

        LEFT_KEPPER_ON_DEAD_FLOOR,


        /**
         * A keeper looking right on a dead floor.
         */

        RIGHT_KEPPER_ON_DEAD_FLOOR,


        /**
         * A keeper looking up on a dead floor.
         */

        UP_KEPPER_ON_DEAD_FLOOR,


        /**
         * A keeper looking down on a dead floor.
         */

        DOWN_KEPPER_ON_DEAD_FLOOR,


        /**
         * A keeper looking left on a goal field.
         */

        LEFT_KEEPER_ON_GOAL,


        /**
         * A keeper looking right on a goal field.
         */

        RIGHT_KEEPER_ON_GOAL,


        /**
         * A keeper looking up on a goal field.
         */

        UP_KEEPER_ON_GOAL,


        /**
         * A keeper looking down on a goal field.
         */

        DOWN_KEEPER_ON_GOAL,


        /**
         * A moving keeper looking left.
         */

        LEFT_MOVING_KEEPER,


        /**
         * A moving keeper looking right.
         */

        RIGHT_MOVING_KEEPER,


        /**
         * A moving keeper looking down.
         */

        DOWN_MOVING_KEEPER,


        /**
         * A moving keeper looking up.
         */

        UP_MOVING_KEEPER,


        /**
         * The virtual keeper looking left.
         */

        LEFT_VIRTUAL_KEEPER,


        /**
         * The virtual keeper looking right.
         */

        RIGHT_VIRTUAL_KEEPER,


        /**
         * The virtual keeper up.
         */

        UP_VIRTUAL_KEEPER,


        /**
         * The virtual keeper down.
         */

        DOWN_VIRTUAL_KEEPER,


        /**
         * A gem on a non dead floor.
         */

        GEM_ON_FLOOR,


        /**
         * A gem on a dead floor.
         */

        GEM_ON_DEAD_FLOOR,


        /**
         * A gem on a goal.
         */

        GEM_ON_GOAL,


        /**
         * A moving gem.
         */

        MOVING_GEM,


        /**
         * The empty floor.
         */

        EMPTY_FLOOR,


        /**
         * The empty dead floor.
         */

        EMPTY_DEAD_FLOOR,


        /**
         * An empty goal field.
         */

        EMPTY_GOAL,


        /**
         * A left arrow.
         */

        LEFT_ARROW,


        /**
         * A right arrow.
         */

        RIGHT_ARROW,


        /**
         * An up arrow.
         */

        UP_ARROW,


        /**
         * A down arrow.
         */

        DOWN_ARROW,


        /**
         * A wall field.
         */

        WALL,


        /**
         * An outside field.
         */

        OUTSIDE
    };

    /**
     * Returns the indices of the images requested.
     *
     * @param type The type of the images.
     * @param position The position of the image (is ignored for moving images).
     * @param map The map to use.
     */

    std::vector<int> const & imageIndices(ImageType type, QPoint const & position, Map const & map) const;


    /**
     * Returns the indices of the images of the piece at the given position.
     *
     * @param direction The direction of the keeper.
     * @param position The position of the field.
     * @param map The map to use.
     */

    std::vector<int> const & imageIndicesFromPosition(int direction, QPoint const & position,
                                                      Map const & map) const;


    /**
     * Returns the piece image with the given index.
     *
     * @param index The index of the piece image.
     */

    PieceImage const & pieceImage(int index) const;


    /**
     * Returns the number of piece images.
     */

    int nrOfPieceImages() const;


    /**
     * Returns the upper border given relative to the field size.
     */

    double upperBorder() const;


    /**
     * Returns the lower border given relative to the field size.
     */

    double lowerBorder() const;


    /**
     * Return the left border given relative to the field size.
     */

    double leftBorder() const;


    /**
     * Returns the right border given relative to the field size.
     */

    double rightBorder() const;


private:

    /**
     * Adds the alternates of an image type.
     *
     * @param dom_element The dom element to use.
     */

    void addAlternates(QDomElement const & dom_element);


    /**
     * Adds piece image alternates.
     *
     * @param dom_element The dom element to use.
     */

    void addPieceImageAlternates(QDomElement const & dom_element);


    /**
     * Adds wall piece alternates.
     *
     * @param dom_element The dom element to use.
     */

    void addWallPieceAlternates(QDomElement const & dom_element);


    /**
     * Addes the alternates piece images of the dom element
     *
     * @param dom_element The dom element to use.
     */

    void addAlternatesImages(QDomElement const & dom_element);


    /**
     * Returns the pattern value for the given attribute of the dom element for piece images.
     *
     * @param dom_element The dom element to use.
     * @param name The name of the attribute to use.
     */

    int patternForPieceImage(QDomElement const & dom_element, QString const & name);


    /**
     * Returns the pattern value for the given attribute of the dom element for wall pieces.
     *
     * @param dom_element The dom element to use.
     * @param name The name of the attribute to use.
     */

    int patternForWallPiece(QDomElement const & dom_element, QString const & name);


    /**
     * Returns the pattern for a non wall piece at the given position in the map.
     *
     * @pattern position The position of the image.
     * @pattern map The map to use.
     */

    int getNonWallPattern(QPoint const & position, Map const & map) const;


    /**
     * Returns the pattern for a wall piece at the given position in the map.
     *
     * @pattern position The position of the image.
     * @pattern map The map to use.
     */

    int getWallPattern(QPoint const & position, Map const & map) const;


    /**
     * The name of the theme.
     */

    QString m_name;


    /**
     * The upper border.
     */

    double m_upper_border;


    /**
     * The lower border.
     */

    double m_lower_border;


    /**
     * The left border.
     */

    double m_left_border;


    /**
     * The right border.
     */

    double m_right_border;


    /**
     * The background color to use.
     */

    QColor m_background_color;


    /**
     * The background image to use.
     */

    QString m_background_image;


    /**
     * The start index of the image type in the alternates.
     */

    std::vector<int> m_alternate_starts;


    /**
     * The number of alternates for the image types.
     */

    std::vector<int> m_alternate_sizes;


    /**
     * The start index of alternative piece images for the alternates.
     */

    std::vector<int> m_alternative_piece_image_starts;


    /**
     * The number of alternative piece images for the alternates.
     */

    std::vector<int> m_alternative_piece_image_nr;


    /**
     * The alternates patterns.
     */

    std::vector<int> m_patterns;


    /**
     * Points to the piece indices.
     */

    std::vector<int> m_piece_image_indices;


    /**
     * The piece images.
     */

    std::vector<PieceImage> m_piece_images;


    /**
     * If true, all gems are hidden.
     */

    bool m_hide_gems;


    /**
     * If true, all goals are hidden.
     */

    bool m_hide_goals;


    /**
     * If true, outside fields are treated as wall.
     */

    bool m_outside_as_wall;


    /**
     * Here we store the names of the elements of the root element of the theme file.
     */

    static char const * s_elements[33];


    /**
     * Here we store the alternative short names of the elements of the root element of the theme file.
     */

    static char const * s_short_elements[33];


    /**
     * The translation table if we don't show gems.
     */

    static int const s_types_for_hidden_gems[33];


    /**
     * The translation table if we don't show goals.
     */

    static int const s_types_for_hidden_goals[33];
};


#endif
