/*
 *   EasySok --- A(nother) sokoban game for KDE.
 *
 *   Copyright (C) 2001 by Ralf Schmelter (ralf.schmelter@t-online.de).
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 2 as
 *   published by the Free Software Foundation.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "theme_file.h"

#include <cassert>
#include <cstdlib>

#include <qstring.h>
#include <qstringlist.h>

#include <klocale.h>
#include <kmessagebox.h>




ThemeFile::ThemeFile(QString const & filename) :
    m_filename(filename),
    m_line_nr(0),
    m_file(filename),
    m_stream(&m_file)
{
    if (!m_file.open(IO_ReadOnly))
    {
        KMessageBox::error(0, i18n("Could not open theme file '%1'!").arg(m_filename));

        std::exit(1);
    }
}


bool ThemeFile::isValid() const
{
    return m_file.isOpen();
}


QRgb ThemeFile::readColor()
{
    QString line = readString();
    line = line.simplifyWhiteSpace();

    QStringList parts = QStringList::split(' ', line);

    if (parts.count() != 4)
    {
        QString msg = i18n("Invalid color specification in theme file '%1' on line '%2'");
        msg = msg.arg(m_filename).arg(m_line_nr);

        KMessageBox::error(0, msg);

        std::exit(1);
    }

    return qRgba(parts[0].toInt(), parts[1].toInt(), parts[2].toInt(), parts[3].toInt());
}


QString ThemeFile::readString()
{
    QString line = m_stream.readLine();
    line = line.stripWhiteSpace();
    ++m_line_nr;

    return line;
}


QString ThemeFile::readString(QString const & expected)
{
    QString line = m_stream.readLine();
    line = line.stripWhiteSpace();
    ++m_line_nr;

    if ( (expected.length() != line.length()) || (line.find(expected, 0, false) != 0) )
    {
        QString msg = i18n("Error in theme-file '%1', line %2.\nExpected '%3',\n but got '%4'");
        msg = msg.arg(m_filename).arg(m_line_nr).arg(expected).arg(line);

        KMessageBox::error(0, msg);

        std::exit(1);
    }

    return line;
}


QString ThemeFile::readString(QString const & expected1, QString const & expected2)
{
    QString line = m_stream.readLine();
    line = line.stripWhiteSpace();
    ++m_line_nr;

    bool const fits1 = (expected1.length() == line.length()) && (line.find(expected1, 0, false) == 0);
    bool const fits2 = (expected2.length() == line.length()) && (line.find(expected2, 0, false) == 0);

    if (!fits1 && !fits2)
    {
        QString msg = i18n("Error in theme-file '%1', line %2.\nExpected '%3' or '%4',\n but got '%5'");
        msg = msg.arg(m_filename).arg(m_line_nr).arg(expected1).arg(expected2).arg(line);

        KMessageBox::error(0, msg);

        std::exit(1);
    }

    return line;
}


QString ThemeFile::readString(QString const & expected1, QString const & expected2, QString const & expected3)
{
    QString line = m_stream.readLine();
    line = line.stripWhiteSpace();
    ++m_line_nr;

    bool const fits1 = (expected1.length() == line.length()) && (line.find(expected1, 0, false) == 0);
    bool const fits2 = (expected2.length() == line.length()) && (line.find(expected2, 0, false) == 0);
    bool const fits3 = (expected3.length() == line.length()) && (line.find(expected3, 0, false) == 0);

    if (!fits1 && !fits2 && !fits3)
    {
        QString msg = i18n("Error in theme-file '%1', line %2.\nExpected '%3', '%4' or '%5',\n but got '%6'");
        msg = msg.arg(m_filename).arg(m_line_nr).arg(expected1).arg(expected2).arg(expected3).arg(line);

        KMessageBox::error(0, msg);

        std::exit(1);
    }

    return line;
}


int ThemeFile::readInt()
{
    QString line = readString();
    line = line.simplifyWhiteSpace();

    return line.toInt();
}


double ThemeFile::readDouble()
{
    QString line = readString();
    line = line.simplifyWhiteSpace();

    return line.toDouble();
}


int ThemeFile::lineNr() const
{
    return m_line_nr;
}


QString const & ThemeFile::filename() const
{
    return m_filename;
}
